-- everything for espio in battle mod
local espio = chaotix.espio

-- command to switch how invis controls
COM_AddCommand(espio.InvisModeCommand, function(p, mode)
	p.espio_invismode = $ or espio.InvisModeDefault	-- make sure this exists
	local modes = {"Hold", "Toggle"}
	
	if not (mode)
		CONS_Printf(p, "\""..espio.InvisModeCommand.."\" is \""..modes[p.espio_invismode].."\" default is \""..modes[espio.InvisModeDefault].."\"")
		return
	else
		if (tonumber(mode))	-- number handling
			mode = tonumber($)
			if not (mode >= 1 and mode <= #modes)	-- only 2 possible values so I can use this
				CONS_Printf(p, "\""..mode.."\" is not a possible value for \""..espio.InvisModeCommand.."\"")
				return
			end
			
			p.espio_invismode = mode
			chaotix.WriteConfig(p)	-- update config
		else	-- string handling
			for index = 1, #modes
				-- not this mode?
				if (mode:lower() != modes[index]:lower())
					continue
				end
						
				p.espio_invismode = index
				chaotix.WriteConfig(p)	-- update config
				return
			end
					
			CONS_Printf(p, "\""..mode.."\" is not a possible value for \""..espio.InvisModeCommand.."\"")
			return
		end
	end
end)

-- loading espio's stuff
local loaded = false
espio.BattleSetup = function()
	if not (CBW_Battle) then return end	-- battle isnt loaded
	if (loaded) then return end	-- already setup for battle? dont run
	local B = CBW_Battle
	local S = B.SkinVars
	
	-- espio's priority
	local function espio_battlePriority(p)
		local mo = p.mo
		if (mo and mo.valid)
		and (mo.espio_teleport_homing)
			local atk, def = 1, 1
			local text = "focus attack"
			B.SetPriority(p, atk, def, nil, atk, def, text)
		elseif (p.actionstate == 2)
			-- espio should have no priority, the tornado IS the priority
			local atk, def = 0, 1
			B.SetPriority(p, atk, def, nil, atk, def, "leaf swirl tornado")
		else
			local spinjump = ((p.pflags & PF_JUMPED) and not (p.pflags & PF_NOJUMPDAMAGE))
			local spinning = (p.pflags & PF_SPINNING)
			
			if (spinjump or spinning)
				-- make espio have a drill spin instead of a spin attack
				local atk, def = 1, 1
				local text = (spinning) and "drill-spin attack" or "jumping drill-spin attack"
				B.SetPriority(p, atk, def, nil, atk, def, text)
			end
		end
	end
	
	-- separate function for leaf swirl functionality, just so I can return on the end conditions
	local function espio_leafSwirl(p)
		local mo = p.mo
		
		-- end conditions
		local frametics = espio.LeafSwirlFrameTics
		local minduration = espio.LeafSwirlMinDuration
		local maxduration = espio.LeafSwirlMaxDuration
		local swirltics = p.actiontime - (frametics[1] + frametics[2] + frametics[3])
		
		-- show the exhaust!
		p.exhaustmeter = max(0, min(FRACUNIT, FRACUNIT - ((swirltics * FRACUNIT) / maxduration)))
		
		if ((swirltics >= minduration) and not (p.cmd.buttons & p.battleconfig_special))	-- we're not holding down the button anymore (for some reason i cant use doaction here)
		or (swirltics >= maxduration)	-- can't hold it down anymore
		or (P_PlayerInPain(p))	-- we're in pain
			if not (P_PlayerInPain(p))	-- we wouldnt want to set any state here
				mo.state = (P_IsObjectOnGround(mo)) and S_PLAY_STND or S_PLAY_FALL
			end
			
			p.actionstate = 0
			p.exhaustmeter = FRACUNIT
			mo.espio_leafswirl_parryhack = 3	-- need this here too
			B.ApplyCooldown(p, espio.LeafSwirlCooldown)
			
			local tornado = mo.espio_leafswirl
			if (tornado and tornado.valid)
				tornado.tics = espio.LeafSwirlLingerTics + 3	-- + 3 for how the leaf swirl despawns
				mo.espio_leafswirl = nil
				
				-- give it our momentum! (will trigger code in the mobjthinker)
				tornado.momx = mo.momx
				tornado.momy = mo.momy
				tornado.momz = mo.momz
			end
			return
		end
		
		-- set actiontext + actionrings
		p.actiontext = "Leaf Swirl"
		p.actionrings = espio.LeafSwirlRingCost
		
		-- set action2text to the exhaust
		if (p.exhaustmeter < FRACUNIT)
			p.action2text = "Exhaust "..100*p.exhaustmeter/FRACUNIT.."%"
		end
		
		-- set state
		if (mo.state != S_ESPIO_LEAFSWIRL) then mo.state = S_ESPIO_LEAFSWIRL end
				
		local frame
		local tics = 0
		for i = 0, #frametics-1
			tics = $+frametics[i+1]
			if (p.actiontime <= tics)
				-- reduce momentum
				mo.momx = FixedMul(espio.LeafSwirlXYMultiplier, $)
				mo.momy = FixedMul(espio.LeafSwirlXYMultiplier, $)
				mo.momz = FixedMul(espio.LeafSwirlZMultiplier, $)
				
				-- set frame
				frame = i
				break
			end
		end
		if (frame == nil) then frame = 6 end	-- assume we're on the final frame
		chaotix.SetFrame(mo, frame)
		
		-- never be "grounded"
		if (P_IsObjectOnGround(mo)) then P_SetOrigin(mo, mo.x, mo.y, mo.z + P_MobjFlip(mo)) end	-- make sure we're never "grounded" (affected by friction)
		
		-- final frame is attack time
		if (frame == 6)
			-- set momentum
			p.pflags = ($|PF_FULLSTASIS)
			mo.momx = FixedMul(espio.LeafSwirlDriftMultiplier, $)
			mo.momy = FixedMul(espio.LeafSwirlDriftMultiplier, $)
			mo.momz = 0	-- no z momentum!
			
			-- we'll need this to get parried!
			mo.espio_leafswirl_parryhack = 3	
			
			-- tornado spawning
			if (p.actiontime == (tics + 1))
				local tornado = P_SpawnMobjFromMobj(mo, 0, 0, espio.LeafSwirlTornadoSpawnOffset, MT_ESPIO_LEAFSWIRL_TORNADO)
				tornado.target = mo	-- we'll want to save this to know who spawned it
				tornado.name = "leaf swirl tornado"	-- so we have an actual hit message!
				mo.espio_leafswirl = tornado	-- make sure the tornado is saved in us!
				
				-- make sure it's at least always on the floor
				local checkz = tornado.z
				if (mo.eflags & MFE_VERTICALFLIP)
					checkz = tornado.z + tornado.height
				end
				
				if ((mo.eflags & MFE_VERTICALFLIP) and checkz > P_CeilingzAtPos(mo.x, mo.y, mo.z, mo.height))
				or (not (mo.eflags & MFE_VERTICALFLIP) and checkz < P_FloorzAtPos(mo.x, mo.y, mo.z, mo.height))
					local newz = P_FloorzAtPos(mo.x, mo.y, mo.z, mo.height)
					if (mo.eflags & MFE_VERTICALFLIP)
						newz = P_CeilingzAtPos(mo.x, mo.y, mo.z, mo.height)
					end
					
					P_SetOrigin(tornado, tornado.x, tornado.y, newz)
				end
			end
		end
	end
	
	local function espio_battleSpecial(mo, doaction)
		local p = mo.player
		local grounded = P_IsObjectOnGround(mo)
		
		-- ability names
		p.actiontext = ((grounded or p.climbing) and not mo.espio_teleport) and "Invisibility" or "Leaf Swirl"
		p.actionrings = ((grounded or p.climbing) and not mo.espio_teleport) and espio.InvisRingCost or espio.LeafSwirlRingCost
		p.action2text = nil	-- this'll get changed later if we're invisible
		
		-- make sure lingering saved tornados are properly reduced in time
		if (p.actionstate != 2)
			local tornado = mo.espio_leafswirl
			if (tornado and tornado.valid)
				tornado.tics = espio.LeafSwirlLingerTics + 3	-- + 3 for how the leaf swirl despawns
				mo.espio_leafswirl = nil
				B.ApplyCooldown(p, espio.LeafSwirlCooldown)	-- properly apply cooldowns
			end
		end
		
		-- main actionstate block
		if (p.actionstate == 0)	-- starting the move?
			if (doaction == 1)	-- just pressed it
				B.PayRings(p)
				p.actionstate = ((grounded or p.climbing) and not mo.espio_teleport) and 1 or 2	-- invisibility is 1, leaf swirl is 2
				p.actiontime = 0	-- timer for abilities
				p.canguard = false
					
				-- init functions on each state
				if (p.actionstate == 1)	-- invisibility
					mo.espio_ringdraintics = espio.InvisRingDrainTicBounds[2]
					if (p.rings) then mo.flags2 = $|(MF2_FRET) end	-- prevent targeting
						
					-- set opacity!
					mo.espio_battleopacity = (p.rings == 0) and (espio.InvisNoRingFade * 10) or 0
					mo.espio_battleopacity_sound = nil	-- make sure this is reset
						
					-- visuals
					local dusts = 42
					for i = 1, dusts
						local radius, height = (mo.radius*9/4)>>FRACBITS, (mo.height)>>FRACBITS
						local x = P_RandomRange(-(radius), radius)*FRACUNIT
						local y = P_RandomRange(-(radius), radius)*FRACUNIT
						local z = P_RandomRange(0, height)*FRACUNIT
							
						local dust = P_SpawnMobjFromMobj(mo, x, y, z, MT_SPINDUST)
						local savescale = dust.scale
						if (mo.eflags & MFE_UNDERWATER)
							dust.state = S_WPLD1 
							dust.scale = $/2
						else
							dust.tics = $/2
						end
						dust.momz = FixedMul(savescale, P_RandomRange(2, 5)*FRACUNIT) * P_MobjFlip(mo)
					end
					chaotix.SpawnSoundMobj(mo, sfx_s3k82)	-- just so the sound doesnt follow us!
				else	-- leaf swirl
					p.pflags = ($|PF_FULLSTASIS) & ~(PF_JUMPED|PF_SHIELDABILITY)
					p.secondjump = 0
					p.climbing = 0	-- get off the wall!
					mo.state = S_ESPIO_LEAFSWIRL
				end
			end
		else	-- ability thinker
			p.actiontime = $+1
			p.canguard = false
			
			if (p.actionstate == 1)	-- invisibility
				-- we were hit, decloak!
				if (P_PlayerInPain(p)) then p.actionstate = 0 return end
				
				-- set actiontext + actionrings
				p.actiontext = "Invisibility"
				p.actionrings = espio.InvisRingCost
				
				-- instantly go to waiting while invisible
				if (mo.state == S_PLAY_STND) then mo.state = S_PLAY_WAIT end
				
				-- drain rings while we're invisible based on velocity
				local speedthreshold = FixedDiv(FixedMul(p.normalspeed, espio.InvisWalkThreshold), mo.scale)
				local speed = FixedDiv(FixedHypot(mo.momx, mo.momy), mo.scale)
				if (p.climbing)
					-- make this based off of momz instead, using wallrun speed
					local normalspeed = FixedMul(p.normalspeed, espio.WallRunSpeedModifier)
					speedthreshold = FixedDiv(FixedMul(normalspeed, espio.InvisWalkThreshold), mo.scale)
					speed = FixedDiv(mo.momz, mo.scale)
				end
				
				local ticbounds = espio.InvisRingDrainTicBounds
				local ticrange = ticbounds[2] - ticbounds[1]
				local mintics = max(1, ticbounds[2] - min(ticrange, (speed * ticrange) / speedthreshold))
				
				mo.espio_ringdraintics = min(mintics, $) - 1
				if not (mo.espio_ringdraintics)
					mo.espio_ringdraintics = mintics	-- reset timer
					P_GivePlayerRings(p, -1)
				end
				
				-- end conditions
				p.espio_invismode = $ or espio.InvisModeDefault	-- make sure this exists
				
				if ((p.espio_invismode == 1) and not doaction)	-- HOLD to keep invis on
				or ((p.espio_invismode == 2) and (doaction == 1))	-- TOGGLE to disable invis
					p.actionstate = 0
					mo.flags2 = $ & ~(MF2_FRET)	-- let us be targeted again
					B.ApplyCooldown(p, espio.InvisCooldown)
				end
			else	-- leaf swirl
				espio_leafSwirl(p)	-- easier to just have it in a function
			end
		end
		
		-- handle the leaf swirl hack
		if (mo.espio_leafswirl_parryhack)
			mo.espio_leafswirl_parryhack = $-1
		end
		
		-- get how opaque we should be based on momentum
		if (p.actionstate == 1)	-- we're in the invisibility state, adjust based on momentum!
			-- speed values
			local speedthreshold = FixedDiv(FixedMul(p.normalspeed, espio.InvisWalkThreshold), mo.scale)
			local speed = FixedDiv(FixedHypot(mo.momx, mo.momy), mo.scale)
			if (p.climbing)
				-- make this based off of momz instead, using wallrun speed
				local normalspeed = FixedMul(p.normalspeed, espio.WallRunSpeedModifier)
				speedthreshold = FixedDiv(FixedMul(normalspeed, espio.InvisWalkThreshold), mo.scale)
				speed = FixedDiv(abs(mo.momz), mo.scale)
			end
			
			-- set how close we are to being visible as our action2text
			if (p.rings == 0)	-- we're visible when we have no rings regardless
				p.action2text = "Visible!"
				p.action2textflags = (leveltime&4 <= 2) and 2 or 0	-- TF_YELLOW = 2 (these arent exposed)
				mo.flags2 = $ & ~(MF2_FRET)	-- let us be targeted again
			else
				local speedvalue = min(100, speed*100/speedthreshold)
				p.action2text = "Speed "..speedvalue.."%"
				p.action2textflags = (speedvalue == 100 and (leveltime&4 <= 2)) and 2 or 0	-- TF_YELLOW = 2 (these arent exposed)
				mo.flags2 = $|(MF2_FRET)	-- prevent targeting
			end
			
			-- now get the target opacity
			local targetopacity = ((p.rings == 0) and espio.InvisNoRingFade or 0) * 10
			if (mo.espio_battleopacity < targetopacity)
				mo.espio_battleopacity = min(targetopacity, $ + espio.InvisTransparencyAdjust)
			elseif (mo.espio_battleopacity > targetopacity)
				mo.espio_battleopacity = max(targetopacity, $ - espio.InvisTransparencyAdjust)
			end
		else
			-- bring espio's transparency back to normal
			if (mo.espio_battleopacity != nil)
				mo.flags2 = $ & ~(MF2_FRET)	-- let us be targeted again
				mo.espio_battleopacity = min(100, $ + espio.InvisTransparencyAdjust)
				
				-- play sound!
				if not (mo.espio_battleopacity_sound)
					mo.espio_battleopacity_sound = true
					S_StartSound(mo, sfx_prloop)
				end
				
				-- fully opaque, reset time!
				if (mo.espio_battleopacity == 100)
					mo.espio_battleopacity = nil
					mo.espio_battleopacity_sound = nil
					
					-- reset this while we're at it
					mo.frame = ($ & ~FF_TRANSMASK)
					mo.flags2 = $ & ~(MF2_DONTDRAW)
				end
			end
		end
		
		-- set espio's transparency
		if (mo.espio_battleopacity != nil)
			-- now set our transparency!
			local trans = 10 - (mo.espio_battleopacity/10)
			//print("Transparency: "..trans)
			if (trans == 10)	-- fully transparent, use MF2_DONTDRAW
				mo.frame = ($ & ~FF_TRANSMASK) 	-- remove transparency, we wont need it here
				mo.flags2 = $|(MF2_DONTDRAW)
			else
				mo.frame = ($ & ~FF_TRANSMASK)
				mo.flags2 = $ & ~(MF2_DONTDRAW)
				if (trans >= 1 and trans <= 9)
					mo.frame = $|(trans << FF_TRANSSHIFT)
				end
			end
		end
	end
	
	-- espio's collision 
	local function espio_battlePreCollide(s, t, p, mo, atk, def, weight, hurt, pain, ground, angle, thrust, thrust2, collisiontype)
		-- make sure we can stay in invis?
		if (p[s].actionstate == 1)
			mo[s].espio_battleinvis = true
		end
	end
	
	local function espio_battlePostCollide(s, t, p, mo, atk, def, weight, hurt, pain, ground, angle, thrust, thrust2, collisiontype)
		if not (p[s] and mo[s]) then return end
		if not (p[t] and mo[t]) then return end
		
		-- laziness check for hurt value (so we don't have to use 2 different checks...)
		local newhurt = hurt
		if (abs(newhurt) == 1)
		and (s == 2)	-- s is actually t
			newhurt = $ * -1
		end
		
		-- invisibility interactions!
		if (mo[s].espio_battleinvis)
			if not (newhurt)	-- blink!
				local targetopacity = espio.InvisBumpFade * 10
				p[s].actionstate = 1	-- make sure we stay in invis
				mo[s].espio_battleopacity = max(targetopacity, $)
				S_StartSound(mo[s], sfx_prloop)
			else	-- someone got hurt, knock em out of invis
				p[s].actionstate = 0
				mo[s].flags2 = $ & ~(MF2_FRET)	-- let us be targeted again
				B.ApplyCooldown(p[s], espio.InvisHitCooldown)
			end
			
			-- reset this
			mo[s].espio_battleinvis = nil
		end
		
		-- fix momentum after a teleport
		if (mo[s].espio_teleport_battleinfo)
		and (newhurt == 1)
		and not (p[t].guard)	-- we don't wanna go flying away after a parry!
			-- fix up player state
			p[s].pflags = $|(PF_JUMPED)
			mo[s].state = S_PLAY_JUMP
			
			-- fix up momentum
			local battleinfo = mo[s].espio_teleport_battleinfo
			mo[s].momx = battleinfo["momx"]
			mo[s].momy = battleinfo["momy"]
			mo[s].momz = battleinfo["momz"]
			espio.ResetTeleport(mo[s])
		end
	end
	
	-- espio exhaust
	local espio_battleExhaust = function(p)
		if not (espio.AbilityCheck(p)) then return end
		local mo = p.mo
			
		-- conditions to lower it
		if (mo.espio_ceilingCling)
		or (mo.espio_chargejump)
			local maxtime = (mo.espio_ceilingCling) and espio.CeilingExhaustTime or espio.ChargeThrustExhaustTime
			p.exhaustmeter = max(0, $ - (FRACUNIT/maxtime))
		end
		
		if (mo.espio_wallrun)
			-- get the amount of tics to reduce
			local momz = mo.momz
			if (mo.espio_wallrun_lastmomz != nil) then momz = mo.espio_wallrun_lastmomz end
			momz = $ * P_MobjFlip(mo)
			
			local removetics = 0
			if (momz > 0)	-- this check exists to prevent a nasty overflow
				local tics = (mo.espio_wallrun_timer*100)/FRACUNIT
				local maxspeed = FixedMul(p.normalspeed, espio.WallRunSpeedModifier)
				maxspeed = FixedMul(mo.scale, $)
				
				-- adjust maxspeed depending on situations
				local speedup = (p.powers[pw_sneakers] or P_PlayerFullbright(p))
				if (speedup)
					maxspeed = $ * 5 / 3	-- same as source's maxspeed
				end
				//if (mo.eflags & MFE_UNDERWATER)
					//maxspeed = $ >> 1
				//end
			
				local speed = momz >> FRACBITS
				maxspeed = $ >> FRACBITS
				
				removetics = max(0, (speed * espio.WallExhaustMaxRemove) / maxspeed)
			end
			
			p.exhaustmeter = max(0, $ - removetics)
		end
		
		-- refill here
		if ((P_IsObjectOnGround(mo) and not mo.espio_ceilingCling) or P_PlayerInPain(p))
		and not (p.actionstate == 2)	-- not in leaf swirl!
			p.exhaustmeter = FRACUNIT
		end
	
		return true
	end
	
	-- setup espio's skinvars
	S[espio.SKIN] = {
		weight = 105,
		special = espio_battleSpecial,
		func_priority_ext = espio_battlePriority,
		func_precollide = espio_battlePreCollide,
		func_postcollide = espio_battlePostCollide,
		func_exhaust = espio_battleExhaust,	-- CE based
		guard_frame = 6,
	}
	
	-- slight edit to B.DummyLockOn to fix homing invisible espio players!
	B.DummyLockOn = function(mo,target)
		mo.height = target.height
		mo.scale = target.scale
		mo.flags = $|MF_NOGRAVITY
		if target.player.playerstate == PST_LIVE then
			B.DummyTeleport(mo) //Reposition
		end
		
		//Get aimable
		if target.player.playerstate == PST_LIVE and not target.player.intangible
		and not (target.flags & MF_NOCLIPTHING)
		and not (target.flags2 & MF2_FRET)	-- espio fix
			mo.flags2 = $|MF2_INVERTAIMABLE //Allow lock-on
		else
			//Target is not alive. Disallow lock-on.
			mo.flags2 = $&~MF2_INVERTAIMABLE
		end
	end
	
	loaded = true
end