-- defining everything for espio
chaotix.espio = {}
local espio = chaotix.espio
espio.SKIN = "espio"
chaotix.AddCharacter(espio)

rawset(_G, "CA_ESPIOREAL", 95)

-- espio's sounds
local hijack_sounds = {
	{sound = "sfx_cdfm50", caption = "Charged"},
	{sound = "sfx_kc5b", caption = "Launch"},
	{sound = "sfx_kc45", caption = "Cling"},
	{sound = "sfx_s251", caption = "Dust"},
}
for k, v in pairs(hijack_sounds)
	-- these are now espio's sounds
	sfxinfo[_G[v.sound]].caption = v.caption
end

local new_sounds = {
	{sound = "sfx_eplock", caption = "Locked on"},
}
for k, v in pairs(new_sounds)
	chaotix.SafeFreeslot(v.sound)
	sfxinfo[_G[v.sound]].caption = v.caption
end

-- debug
espio.Debug = false

-- colour shift constants
espio.ColourShiftCommand = "espio_shiftcolors"
espio.ColourShiftTics = 20	-- how long it takes to go from colour to colour
espio.ColourShiftDefault = {SKINCOLOR_SAPPHIRE, SKINCOLOR_APPLE, SKINCOLOR_YELLOW, SKINCOLOR_SALMON}

-- focus jump constants
espio.FocusJumpBase = 8*FRACUNIT	-- jump height with no charge
espio.FocusJumpAdd = 9*FRACUNIT/4	-- how much height we get per charge level

-- charge jump constants
espio.ChargeJumpXYMultiplier = FRACUNIT*955/1000	-- how much we slow down while charging (horizontally)
espio.ChargeJumpZMultiplier = FRACUNIT*9/10	-- ditto, but vertically
espio.ChargeJumpSlowfallThreshold = -FRACUNIT*15/10
espio.ChargeJumpMin = FRACUNIT*9	-- min power
espio.ChargeJumpMax = FRACUNIT*20	-- max power
espio.ChargeJumpSuperMultiplier = FRACUNIT*12/10	-- increased strength for super as a bonus
espio.ChargeJumpDuration = 9/5*TICRATE	-- how long it takes to get from min to max
espio.ChargeJumpVisualTics = 19

-- charge thrust constants (battlemod)
espio.ChargeThrustRingCost = 5
espio.ChargeThrustMinCooldown = TICRATE/2
espio.ChargeThrustMaxCooldown = TICRATE
espio.ChargeThrustMin = FRACUNIT*28	-- min power
espio.ChargeThrustMax = FRACUNIT*49	-- max power

-- teleport constants
espio.TeleportStartupTics = TICRATE/5	-- the pause before deploying the tracer (also doubles for homing timing)
espio.TeleportXYMultiplier = FRACUNIT*80/100
espio.TeleportTracerSpeed = FRACUNIT*48
espio.TeleportTracerSpeedMultiplier = FRACUNIT	-- multiplies our momentum we add on top of espio.TeleportTracerSpeed
espio.TeleportMinTracingTics = TICRATE/4	-- how long we NEED to hold it
espio.TeleportMaxTracingTics = TICRATE*2/3	-- how long we CAN hold it (also doubles as the time to reach max distance
espio.TeleportZTracingMinModifier = FRACUNIT*2/3	-- how much the tracer's XY momentum is affected (lowest)
espio.TeleportZTracingMaxModifier = FRACUNIT/4	-- how much the tracer's XY momentum is affected (highest)
espio.TeleportZTracingShift = FRACUNIT/2	-- how much it vertically shifts
espio.TeleportTracingFrameTics = 4	-- how long it takes to go to the next frame
espio.TeleportHomingScalar = 1896	-- for every <x> FRACUNITs, TICRATE to travel
espio.TeleportJumpSpeed = FRACUNIT*36
espio.TeleportJumpHeight = FRACUNIT*11
espio.TeleportJumpSpinModifier = FRACUNIT*2/3
espio.TeleportJumpXYModifier = FRACUNIT/3
espio.TeleportJumpZModifier = FRACUNIT*4/3

-- wall run constants
espio.WallRunTics = 5*TICRATE
espio.WallRunSpeedModifier = FRACUNIT/2
espio.WallRunSpeedSlide = FRACUNIT/5
espio.WallRunTiredThreshold = FRACUNIT*3/8
espio.WallRunJumpSpeed = 7*FRACUNIT
espio.WallRunJumpMinHeight = 10*FRACUNIT
espio.WallRunDelay = TICRATE/3
espio.WallRunFastThreshold = 16*FRACUNIT
espio.WallRunFriction = 64000
espio.WallRunAccelerationThreshold = 10*FRACUNIT
espio.WallRunSidewaysModifier = FRACUNIT*2/5
espio.WallRunDustTicBounds = {3, 12}
espio.WallRunDustMomBounds = {FRACUNIT*3/4, 11*FRACUNIT}
espio.WallRunMaxFrameTics = TICRATE/3
espio.WallRunMaxFrameSpeed = FRACUNIT*12

-- ceiling cling constants
espio.CeilingClingRange = FRACUNIT*20

-- exhaust constants (battle mod)
espio.WallExhaustMaxRemove = FRACUNIT/112	-- scales up to this depending on maxspeed
espio.WallExhaustLoss = FRACUNIT/5
espio.CeilingExhaustTime = 4*TICRATE
espio.ChargeThrustExhaustTime = 3*TICRATE

-- teleport constants (battle mod)
espio.TeleportBattleMultiplier = FRACUNIT*4/9	-- reduces launch speed

-- invisibility constants (battle mod)
espio.InvisModeCommand = "espio_invismode"
espio.InvisModeDefault = 2
espio.InvisRingCost = 5
espio.InvisCooldown = TICRATE/2
espio.InvisWalkMultiplier = FRACUNIT*3/4
espio.InvisWalkThreshold = FRACUNIT*3/5
espio.InvisRingDrainTicBounds = {TICRATE*2/3, TICRATE}	-- {fastest drain, slowest drain}
espio.InvisNoRingFade = 5
espio.InvisTransparencyAdjust = 6
espio.InvisBumpFade = 8
espio.InvisHitCooldown = TICRATE*5/2

-- leaf swirl constants (battle mod)
espio.LeafSwirlRingCost = 10
espio.LeafSwirlCooldown = TICRATE*9/4
espio.LeafSwirlXYMultiplier = FRACUNIT*93/100
espio.LeafSwirlZMultiplier = FRACUNIT*9/10	-- (this is only applied during the startup)
espio.LeafSwirlFrameTics = {5, 2, 2}	-- the tornado will come out at: sum of these + 1
espio.LeafSwirlTornadoSpawnOffset = -56*FRACUNIT
espio.LeafSwirlDamageRange = 56*FRACUNIT
espio.LeafSwirlLaunchHeight = 16*FRACUNIT
espio.LeafSwirlLeafOffset = 24	-- we dont want this in fracunits
espio.LeafSwirlProtectTics = 2
espio.LeafSwirlMinDuration = TICRATE*2/3	-- how long we HAVE to do the attack
espio.LeafSwirlMaxDuration = TICRATE*4/3	-- how long we CAN do the attack
espio.LeafSwirlLingerTics = TICRATE/3	-- how long the tornado lingers after stopping
espio.LeafSwirlDriftMultiplier = FRACUNIT*97/100

-- I/O constants
espio.ConfigName = "client/chaotixchars/espioconfig.cfg"

-- make sure we have the right ability
espio.AbilityCheck = function(p)
	local mo = p.mo
	
	-- man
	local abilitycheck = p.charability != CA_ESPIOREAL
	if (srb2p) and (mo and mo.valid) then abilitycheck = (mo.skin != espio.SKIN) end
	
	if (abilitycheck and not p.espio_wallrun)	-- we use CA_GLIDEANDCLIMB hacks for wall runs, so we need a buffer
		return false
	end
	
	-- we have espio's ability if we're here
	return true
end

-- player animation setter
function A_SetEspioPanim(mo, var1, var2)
	if mo.player
		mo.player.panim = var2
	end
end

-- espio roll state
chaotix.SafeFreeslot("S_ESPIO_ROLL")
states[S_ESPIO_ROLL] = {SPR_PLAY, A|SPR2_JUMP, 2, A_SetEspioPanim, 0, PA_ROLL, S_ESPIO_ROLL}

-- espio chargejump state
chaotix.SafeFreeslot("S_ESPIO_CHARGEJUMP")
states[S_ESPIO_CHARGEJUMP] = {SPR_PLAY, A|SPR2_ROLL, 4, A_SetEspioPanim, 0, PA_ABILITY, S_ESPIO_CHARGEJUMP}

-- espio's teleport ability
chaotix.SafeFreeslot("S_ESPIO_TELEPORT_STARTUP", "S_ESPIO_TELEPORT_TRACING", "S_ESPIO_TELEPORT_HOMING")
states[S_ESPIO_TELEPORT_STARTUP] = {SPR_PLAY, A|SPR2_FIRE, -1, A_SetEspioPanim, 0, PA_DASH, S_ESPIO_TELEPORT_STARTUP}
states[S_ESPIO_TELEPORT_TRACING] = {SPR_PLAY, A|SPR2_FIRE, -1, A_SetEspioPanim, 0, PA_DASH, S_ESPIO_TELEPORT_TRACING}
states[S_ESPIO_TELEPORT_HOMING] = {SPR_PLAY, A|SPR2_ROLL, 1, A_SetEspioPanim, 0, PA_DASH, S_ESPIO_TELEPORT_HOMING}

chaotix.SafeFreeslot("S_ESPIO_TELEPORT_TRACER", "MT_ESPIO_TELEPORT_TRACER")
chaotix.SafeFreeslot("SPR_INVS")	-- because SPR_NULL cant cast shadows for some reason
states[S_ESPIO_TELEPORT_TRACER] = {SPR_THOK, A, -1, nil, 0, 0, S_ESPIO_TELEPORT_TRACER}
mobjinfo[MT_ESPIO_TELEPORT_TRACER] = {
	spawnstate = S_ESPIO_TELEPORT_TRACER,
	deathstate = S_NULL,
	radius = mobjinfo[MT_PLAYER].radius,	-- will be auto-corrected to the player's radius
	height = FRACUNIT*55, -- espio's default height
	flags = MF_NOGRAVITY,
}

-- target enemies!
local nonenemiesdisregard = MF_SPRING
espio.TeleportTargetEnemies = function(mo, mo2)
	if (mo.tracer) then return end	-- already got a target!
	if (mo.target.espio_teleport_state != 2) then return end
				
	-- height checks
	local topz = mo.z + mo.height
	local bottomz = mo.z
	if not ((topz >= mo2.z) and (bottomz <= mo2.z + mo2.height)) then return end	-- z checks
	
	-- make sure they're a valid target!
	if (mo2 == mo.target) then return end	-- easiest check
	
	if not (mo2.player)
	and (!((mo2.flags & (MF_ENEMY|MF_BOSS|MF_MONITOR)
	and (mo2.flags & MF_SHOOTABLE)) or (mo2.flags & MF_SPRING)) == !(mo2.flags2 & MF2_INVERTAIMABLE)) -- allows if it has the flags desired XOR it has the invert aimable flag
		return -- not a valid target
	end
	
	//CTF monitor 
	if mo2.type == MT_RING_REDBOX and not(G_GametypeHasTeams() and mo.target.player.ctfteam == 1) then return end
	if mo2.type == MT_RING_BLUEBOX and not(G_GametypeHasTeams() and mo.target.player.ctfteam == 2) then return end
	
	if (mo2.flags2 & MF2_FRET) return end
	if (mo2.flags & nonenemiesdisregard) return end
-- 	if (mo2.type == MT_PLAYER) return end -- Don't chase after other players!
	
	// Player stuff!
	local B = CBW_Battle
	if (B and mo2.type == MT_TARGETDUMMY) then return end	-- MT_TARGETDUMMY only exists in battle
	
	if (mo2.player)
		if not (B) then return end	-- only target in battle!
		
		if (B.MyTeam(mo.target.player, mo2.player) or mo2.player.spectator) then return end //Disallow targeting teammates
		if (mo2.player.intangible) then return end//Disallow targeting air dodge
	end
				
	-- set target then stop the search
	mo.tracer = mo2	-- so other searches are halted
	mo.espio_enemy = mo2
	//S_StartSound(mo2, sfx_cdfm39)
	S_StartSound(mo2, sfx_eplock)
end
chaotix.AddHook("MobjCollide", espio.TeleportTargetEnemies, MT_ESPIO_TELEPORT_TRACER)
chaotix.AddHook("MobjMoveCollide", espio.TeleportTargetEnemies, MT_ESPIO_TELEPORT_TRACER)

-- get a line to latch onto, or go through a spin gap or bustable wall
espio.TeleportSpinGapTries = 8
chaotix.AddHook("MobjMoveBlocked", function(mo, thing, line)
	if (mo.target.espio_teleport_state != 2) then return end
	
	if line and not mo.espio_line and not (line.flags & ML_NOCLIMB)
	and not mo.tracer
		if mo.espio_spingap != nil
			mo.espio_line = line
			return
		end
		
		-- make an object to go check for a position we can teleport to
		local p = mo.target and mo.target.player
		local thok = P_SpawnMobjFromMobj(mo, 0,0,0, MT_GHOST)
		thok.height = p and P_GetPlayerSpinHeight(p) or 32*mo.scale
		thok.radius = mo.radius
		thok.target = mo
		A_CapeChase(thok)
		
		P_InstaThrust(thok, mo.angle, mo.radius*4)
		local mom = {x = thok.momx, y = thok.momy, z = 0}
		
		-- get sectors we could move through
		local bustablesectors = {}
		for i = 1, espio.TeleportSpinGapTries
			P_RailThinker(thok)
			if not (thok and thok.valid) then mo.espio_line = line return end
			local sector = thok.subsector and thok.subsector.sector
			if sector and not bustablesectors[#sector]
				bustablesectors[#sector] = sector
			end
		end
		
		-- get bustable fofs we could move through and disable their collision
		local bustablefofs = {}
		for i, sector in pairs(bustablesectors)
			for fof in sector.ffloors()
				if fof and fof.valid
				and fof.flags & (FF_EXISTS|FF_SOLID|FF_BUSTUP)
				== (FF_EXISTS|FF_SOLID|FF_BUSTUP)
					table.insert(bustablefofs, fof)
					fof.flags = $ &~ FF_SOLID
				end
			end
		end
		
		-- move the thok forward, see if it can fit through a spin gap
		A_CapeChase(thok)
		thok.flags = MF_NOBLOCKMAP|MF_NOGRAVITY
		for i = 1, espio.TeleportSpinGapTries
			if not (thok and thok.valid) then break end
			thok.momx = mom.x
			thok.momy = mom.y
			thok.momz = mom.z
			mo.espio_spingap = not P_RailThinker(thok)
			
			-- uh oh! we're touching a bustable fof that we disabled
			local sector = thok.subsector and thok.subsector.sector
			if sector
				for fof in sector.ffloors()
					if fof.flags & (FF_EXISTS|FF_SOLID|FF_BUSTUP)
					== (FF_EXISTS|FF_BUSTUP)
					and P_InsideANonSolidFFloor(thok, fof)
						mo.espio_spingap = false
						break
					end
				end
			end
			
			if mo.espio_spingap and thok and thok.valid
			and P_CheckPosition(thok, thok.x, thok.y)
			and (thok.ceilingz - thok.floorz) >= mo.height -- make sure we can fit
				break
			else
				mo.espio_spingap = false
			end
		end
		
		-- re-enable the collision all fofs we disabled
		for i, fof in ipairs(bustablefofs)
			fof.flags = $ | FF_SOLID
		end
		
		-- thok wasn't deleted?
		if thok and thok.valid
			-- thok fit through gap?
			if mo.espio_spingap
				-- teleport through gap
				local pos = {x = mo.x, y = mo.y, z = mo.z}
				mo.tracer = thok
				A_CapeChase(mo, 1)
				P_MoveOrigin(mo, mo.x, mo.y, mo.z)
				
				-- uh oh, seems like we're gonna get crushed...
				local crushed = not (P_CheckPosition(mo, mo.x, mo.y)
				and (mo.ceilingz - mo.floorz) >= mo.height)
				mo.tracer = nil
				P_RemoveMobj(thok)
				
				if crushed
					P_MoveOrigin(mo, pos.x, pos.y, pos.z)
				else
					return true
				end
			else
				P_RemoveMobj(thok)
			end
		else
			mo.espio_spingap = false
		end
		mo.espio_line = line
	end
end, MT_ESPIO_TELEPORT_TRACER)

-- battle mod support
chaotix.SafeFreeslot("S_ESPIO_LEAFSWIRL")
states[S_ESPIO_LEAFSWIRL] = {SPR_PLAY, A|SPR2_TRNS, 3, nil, 0, 0, S_ESPIO_LEAFSWIRL}

local defaultlinger = espio.LeafSwirlMaxDuration + espio.LeafSwirlLingerTics
chaotix.SafeFreeslot("S_ESPIO_LEAFSWIRL_TORNADO", "MT_ESPIO_LEAFSWIRL_TORNADO")
states[S_ESPIO_LEAFSWIRL_TORNADO] = {SPR_NULL, 0, defaultlinger + 2, nil, 0, 0, S_NULL}
mobjinfo[MT_ESPIO_LEAFSWIRL_TORNADO] = {
	spawnstate = S_ESPIO_LEAFSWIRL_TORNADO,
	deathstate = S_NULL,
	deathsound = sfx_s3k4b,
	radius = espio.LeafSwirlDamageRange,
	height = espio.LeafSwirlLaunchHeight,
	flags = MF_NOBLOCKMAP|MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT,
}

-- instead of just using the ACZ dust layers, we'll have our own since for some reason they have 256*FRACUNIT HEIGHT... WHAT?!
for i = 1, 5 do chaotix.SafeFreeslot("S_ESPIO_LEAFSWIRL_DUSTLAYER"..i) end
chaotix.SafeFreeslot("MT_ESPIO_LEAFSWIRL_DUSTLAYER", "SPR_ESDL")
states[S_ESPIO_LEAFSWIRL_DUSTLAYER1] = {SPR_ESDL, 0|FF_PAPERSPRITE|FF_TRANS70, 2, nil, 0, 0, S_ESPIO_LEAFSWIRL_DUSTLAYER2}
states[S_ESPIO_LEAFSWIRL_DUSTLAYER2] = {SPR_ESDL, 0|FF_PAPERSPRITE|FF_TRANS70, 4, nil, 0, 0, S_ESPIO_LEAFSWIRL_DUSTLAYER3}
states[S_ESPIO_LEAFSWIRL_DUSTLAYER3] = {SPR_ESDL, 0|FF_PAPERSPRITE|FF_TRANS80, 4, nil, 0, 0, S_ESPIO_LEAFSWIRL_DUSTLAYER4}
states[S_ESPIO_LEAFSWIRL_DUSTLAYER4] = {SPR_ESDL, 0|FF_PAPERSPRITE|FF_TRANS80, 4, nil, 0, 0, S_ESPIO_LEAFSWIRL_DUSTLAYER5}
states[S_ESPIO_LEAFSWIRL_DUSTLAYER5] = {SPR_ESDL, 0|FF_PAPERSPRITE|FF_TRANS90, 4, nil, 0, 0, S_NULL}
mobjinfo[MT_ESPIO_LEAFSWIRL_DUSTLAYER] = {
	spawnstate = S_ESPIO_LEAFSWIRL_DUSTLAYER1,
	deathstate = S_NULL,
	height = 22*FRACUNIT,
	flags = MF_NOGRAVITY|MF_NOBLOCKMAP|MF_NOCLIP|MF_SCENERY,
}
for i = 1, 5 do chaotix.SafeFreeslot("S_ESPIO_LEAFSWIRL_LEAF"..i) end
states[S_ESPIO_LEAFSWIRL_LEAF1] = {SPR_ESDL, B|FF_PAPERSPRITE, 4, nil, 0, 0, S_ESPIO_LEAFSWIRL_LEAF2}
states[S_ESPIO_LEAFSWIRL_LEAF2] = {SPR_ESDL, B|FF_PAPERSPRITE, 4, nil, 0, 0, S_ESPIO_LEAFSWIRL_LEAF3}
states[S_ESPIO_LEAFSWIRL_LEAF3] = {SPR_ESDL, B|FF_PAPERSPRITE|FF_TRANS30, 4, nil, 0, 0, S_ESPIO_LEAFSWIRL_LEAF4}
states[S_ESPIO_LEAFSWIRL_LEAF4] = {SPR_ESDL, B|FF_PAPERSPRITE|FF_TRANS60, 4, nil, 0, 0, S_ESPIO_LEAFSWIRL_LEAF5}
states[S_ESPIO_LEAFSWIRL_LEAF5] = {SPR_ESDL, B|FF_PAPERSPRITE|FF_TRANS90, 4, nil, 0, 0, S_NULL}

-- helper functions
local function isHigher(ref, zpos, zpos2)
	if (ref.eflags & MFE_VERTICALFLIP)
		return (zpos < zpos2)
	else
		return (zpos > zpos2)
	end
end
local function isLower(ref, zpos, zpos2)
	return not isHigher(ref, zpos, zpos2)
end

espio.LeafSwirlProtectionAdd = function(mo)
	if not (server and server.valid) then return end
	if not (server.espio_leafswirlprotectlist) then server.espio_leafswirlprotectlist = {} end

	mo.espio_leafswirlprotect = espio.LeafSwirlProtectTics
	table.insert(server.espio_leafswirlprotectlist, mo)
end

chaotix.AddHook("MobjThinker", function(mo)
	-- conditions to destroy the leaf swirl tornado!
	if (mo.tics > 3)	
		-- parry hack
		local target_parrycheck
		
		if ((mo.target and mo.target.valid) and mo.target.health)
		and (mo.target.player.tumble and mo.target.player.tumble > 0)
		and (mo.target.espio_leafswirl_parryhack)
			target_parrycheck = true
		end
		
		-- run conditions
		if not (mo.target and mo.target.valid)	-- target is gone??
		or not (mo.target.health)	-- target is death
		or (target_parrycheck)	-- target got parried
			mo.tics = 3
		end
	end
	
	-- follow target, otherwise drift
	local target = mo.target
	if ((target and target.valid) and target.health)
	and (target.espio_leafswirl == mo)	-- little hacky way to check if they're casting
		local zpos = target.z + FixedMul(target.scale, espio.LeafSwirlTornadoSpawnOffset)
		P_MoveOrigin(mo, target.x, target.y, zpos)
					
		-- make sure it's at least always on the floor
		local checkz = mo.z
		if (target.eflags & MFE_VERTICALFLIP)
			checkz = mo.z + mo.height
		end
					
		if ((target.eflags & MFE_VERTICALFLIP) and checkz > P_CeilingzAtPos(target.x, target.y, target.z, target.height))
		or (not (target.eflags & MFE_VERTICALFLIP) and checkz < P_FloorzAtPos(target.x, target.y, target.z, target.height))
			local newz = P_FloorzAtPos(target.x, target.y, target.z, target.height)
			if (target.eflags & MFE_VERTICALFLIP)
				newz = P_CeilingzAtPos(target.x, target.y, target.z, target.height)
			end
						
			P_MoveOrigin(mo, mo.x, mo.y, newz)
		end
	else
		mo.momx = FixedMul(espio.LeafSwirlDriftMultiplier, $)
		mo.momy = FixedMul(espio.LeafSwirlDriftMultiplier, $)
	end
	
	-- FUCK flipped gravity
	local flipoffset = -52*FRACUNIT	-- FUCK FLIPPED GRAVITY x2
	
	-- for setting the hitbox later on
	local bestbottom, besttop
	
	-- rotate spawned layers with a chain system
	local offset = espio.LeafSwirlLeafOffset
	local layer = mo.tracer
	while (layer and layer.valid)
		local leaf = layer.espio_leaf	
		if (layer.espio_leaf)
			-- leaves spawn further out
			layer.espio_leaf = $+2
			leaf = layer.espio_leaf
		end
		
		-- rotate around the center (the tornado)
		local xy_offset = (leaf) and (offset + leaf) or 0
		
		layer.angle = $ + (ANG1*(leaf and 9 or 14))	-- rotate it a tiny bit
		local angle = layer.angle - ANGLE_90	-- the layers are all offset forward by ANGLE_90
		local x = mo.x + (((mo.radius/mo.scale) + xy_offset) * FixedMul(mo.scale, cos(angle)))
		local y = mo.y + (((mo.radius/mo.scale) + xy_offset) * FixedMul(mo.scale, sin(angle)))
		P_MoveOrigin(layer, x, y, layer.z)
		if (leaf) then layer.rollangle = $ + ANG1 end
		
		//P_TeleportMove(layer, layer.x + (15 * FixedMul(mo.scale, cos(layer.angle))), layer.y + (15 * FixedMul(mo.scale, sin(layer.angle))), layer.z)
		//layer.scale = mo.scale
		//layer.angle = $ + (ANG1*14)
		layer.momx = mo.momx
		layer.momy = mo.momy
		
		-- make them stop at the ceiling
		if not (leaf)	-- unless it's a leaf
			if (layer.momz)
				local checkz = layer.z + layer.height
				local checkheight = P_CeilingzAtPos(mo.x, mo.y, mo.z, 1)
				if (mo.eflags & MFE_VERTICALFLIP)
					checkz = layer.z
					checkheight = P_FloorzAtPos(mo.x, mo.y, mo.z, 1)
				end
				//print("CheckZ: "..checkz)
				//print("CheckHeight: "..checkheight)
				
				if ((mo.eflags & MFE_VERTICALFLIP) and checkz <= checkheight)
				or (not (mo.eflags & MFE_VERTICALFLIP) and checkz >= checkheight)
					layer.momz = 0
					
					local snapz = (mo.eflags & MFE_VERTICALFLIP) and checkheight or (checkheight - layer.height)
					P_MoveOrigin(layer, layer.x, layer.y, snapz)
					//print("stop")
				end
			end
			
			-- set hitbox
			local layertop = layer.z + layer.height
			local layerbottom = layer.z
			if (layer.eflags & MFE_VERTICALFLIP)
				layertop = layer.z + FixedMul(mo.scale, flipoffset)
				layerbottom = layer.z + layer.height + FixedMul(mo.scale, flipoffset)
			end
			
			if (mo.tornadohitboxes == nil)	-- we dont have hitboxes set
				bestbottom, besttop = layerbottom, layertop
			else	-- we already got one, check heights to reset it
				-- bottom hitbox
				if (bestbottom == nil)
				or (isLower(mo, layerbottom, bestbottom))	-- can we get much lower???
					bestbottom = layerbottom
				end
				
				-- top hitbox
				if (besttop == nil)
				or (isHigher(mo, layertop, besttop))	-- can we get much higher???
					besttop = layertop
				end
			end
		end
			
		-- go to the next layer in the chain (if there is one)
		layer = layer.tracer
	end
	
	-- spawn layers
	if not (mo.extravalue1%6)
	and (mo.tics > 2)
		for i = 1, 4
			local angle = mo.angle + ANGLE_90*i
			local x = (mo.radius/mo.scale) * cos(angle)
			local y = (mo.radius/mo.scale) * sin(angle)
		
			local layer = P_SpawnMobjFromMobj(mo, x, y, 0, MT_ESPIO_LEAFSWIRL_DUSTLAYER)
			layer.angle = angle + ANGLE_90
			layer.color = (G_GametypeHasTeams()) and target.color or SKINCOLOR_WHITE
			layer.flags = $|(MF_NOCLIPHEIGHT)
			P_SetObjectMomZ(layer, 12*FRACUNIT)
			
			-- set hitboxes here too
			local layertop = layer.z + layer.height
			local layerbottom = layer.z
			if (layer.eflags & MFE_VERTICALFLIP)
				layertop = layer.z + FixedMul(mo.scale, flipoffset)
				layerbottom = layer.z + layer.height + FixedMul(mo.scale, flipoffset)
			end
			
			if (mo.tornadohitboxes == nil)	-- we dont have hitboxes set
				bestbottom, besttop = layerbottom, layertop
			else	-- we already got one, check heights to reset it
				-- bottom hitbox
				if (bestbottom == nil)
				or (isLower(mo, layerbottom, bestbottom))	-- can we get much lower???
					bestbottom = layerbottom
				end
				
				-- top hitbox
				if (besttop == nil)
				or (isHigher(mo, layertop, besttop))	-- can we get much higher???
					besttop = layertop
				end
			end
			
			-- chain?
			layer.tracer = mo.tracer
			mo.tracer = layer
		end
		
		-- spawn leaves
		local leaves = 6
		for i = 1, leaves
			local angle = mo.angle + ((ANG1*(360/leaves))*i)
			local x = ((mo.radius/mo.scale) + offset) * cos(angle)
			local y = ((mo.radius/mo.scale) + offset) * sin(angle)
			
			local layer = P_SpawnMobjFromMobj(mo, x, y, FRACUNIT*24, MT_ESPIO_LEAFSWIRL_DUSTLAYER)
			layer.state = S_ESPIO_LEAFSWIRL_LEAF1
			layer.angle = angle + ANGLE_90
			layer.flags = $|(MF_NOCLIPHEIGHT)
			P_SetObjectMomZ(layer, 7*FRACUNIT)
			
			if target and target.player
				P_ColorTeamMissile(layer, target.player)
				if layer.color then layer.colorized = true end
			end
			
			-- chain?
			layer.espio_leaf = 1	-- just so we can distinguish these to make them spaced more outwards
			layer.tracer = mo.tracer
			mo.tracer = layer
		end
	end
	mo.extravalue1 = $+1
	
	-- set hitboxes
	if ((bestbottom != nil) and (besttop != nil))	-- these need to exist
		if (mo.tornadohitboxes == nil)	-- hasnt been defined yet
			mo.tornadohitboxes = {bestbottom, besttop}
		else
			-- set bottom hitbox
			if (isLower(mo, mo.tornadohitboxes[1], bestbottom))
				mo.tornadohitboxes[1] = bestbottom
			end
			
			-- set top hitbox
			if (isLower(mo, mo.tornadohitboxes[2], besttop))
				mo.tornadohitboxes[2] = besttop
			end
		end
	end
	
	-- make sure the bottom hitbox is always the bottom of the tornado while still spawning layers
	if (mo.tics > 2)
		local bottomz = mo.z
		if (mo.eflags & MFE_VERTICALFLIP) then bottomz = mo.z + mo.height + FixedMul(mo.scale, flipoffset) end
				
		mo.tornadohitboxes[1] = bottomz
	end
	
	-- damage stuff around us
	local function espio_tornadoBlockmap(ref, found)
		if not ((found and found.valid) and found.health) then return end
		if (ref == found) then return end	-- brew
		
		-- easy checks
		if not ((mo.tornadohitboxes[2] >= found.z) and (mo.tornadohitboxes[1] <= found.z + found.height)) then return end	-- z checks
		if not (P_CheckSight(ref, found)) then return end	-- self explanatory
		
		-- battle team checks
		local B = CBW_Battle
		if (found.player)	-- we gotta do our own checks here
			if not (B.PlayerCanBeDamaged(found.player)) then return end	-- cant be damaged
			
			if (found == ref.target)	-- it's ourself!
			or (B.MyTeam(found.player, ref.target.player)) -- same team
				espio.LeafSwirlProtectionAdd(found)	-- offer some tics of protection from projectiles!
				return
			end
		else
			if (found == ref.target)
				espio.LeafSwirlProtectionAdd(found)
				return
			end
			
			if (found.flags & MF_MISSILE)	-- destroy projectiles!
				-- try not to destroy projectiles on your own team
				local source = found.target
				if not (source and source.valid) then return end
				if not (source.player) then return end
				
				-- same team? don't destroy
				if (B.MyTeam(source.player, ref.target.player)) then return end
			else
				-- damage enemies and monitors!
				if not (found.flags & (MF_SHOOTABLE|MF_ENEMY|MF_MONITOR)) then return end
			end	
		end
		
		-- ACTUAL distance check, so it's ciruclar and not square
		local dist = mo.radius + FixedMul(mo.scale, 16*FRACUNIT)
		if not (R_PointToDist2(found.x, found.y, ref.x, ref.y) <= dist) then return end
		
		-- if they're in the leaf swirl actionstate, make them the inflictor so they can be parried out!
		local inflictor = (ref.target.player.actionstate == 2) and ref.target or ref
		P_DamageMobj(found, inflictor, ref.target)
		
		if ((found and found.valid) and found.health)	-- doublecheck this, we only want to launch them if they're still alive!
		and not (found.player and not P_PlayerInPain(found.player))	-- assume a parry happened!
			if (found.flags & MF_MISSILE)	-- explode the missile
				P_ExplodeMissile(found)
			else
				-- make sure they're launched AWAY from the tornado
				found.angle = R_PointToAngle2(found.x, found.y, ref.x, ref.y)
				P_InstaThrust(found, found.angle, -FixedHypot(found.momx, found.momy))
				
				-- launch them!
				if not (found.flags & MF_BOSS)
					P_SetObjectMomZ(found, ref.height)
				end
			end
		end
	end
	
	if (mo.tornadohitboxes != nil)
		searchBlockmap("objects", espio_tornadoBlockmap, mo, mo.x - mo.radius, mo.x + mo.radius, mo.y - mo.radius, mo.y + mo.radius)
		
		-- hitbox check
		if (espio.Debug)
			if not (mo.debugthoks)
				mo.debugthoks = {}
				
				for i = 1, 2
					local thok = P_SpawnMobj(mo.x, mo.y, mo.tornadohitboxes[i], MT_THOK)
					thok.frame = ($|FF_FULLBRIGHT) & ~FF_TRANSMASK	-- give fullbright, remove transparency
					thok.color = (i == 2) and SKINCOLOR_WHITE or SKINCOLOR_BLACK
					thok.flags = $ & ~(MF_SHOOTABLE|MF_ENEMY|MF_MONITOR)	-- we dont want our debug thoks to be killed
					thok.flags = $|(MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT)
					thok.tics = -1
					
					thok.debugthoknum = i
					table.insert(mo.debugthoks, thok)
				end
			else
				for k, v in pairs(mo.debugthoks)
					P_TeleportMove(v, mo.x, mo.y, mo.tornadohitboxes[v.debugthoknum])
				end
			end
		end
	else
		if (espio.Debug)
			if (mo.debugthoks)
				for k, v in pairs(mo.debugthoks)
					P_RemoveMobj(v)
				end
				mo.debugthoks = nil
			end
		end
	end
	
	-- play sounds
	if not (S_SoundPlaying(mo, sfx_s3kcel)) then S_StartSound(mo, sfx_s3kcel) end	-- active sound
	if (mo.tics == 3)	-- death sound
		-- doing this so it doesnt get cut off
		local soundmo = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
		soundmo.state = S_INVISIBLE
		soundmo.fuse = 3*TICRATE
		S_StartSound(soundmo, mo.info.deathsound) 
	end
	
	-- dont despawn if we still have dustlayers!
	if (mo.tics <= 2)	-- cant check 1 because it despawns at 0
	and (mo.tracer and mo.tracer.valid)	-- mo.tracer should always be the last spawn layer
		mo.tics = 2
	end
end, MT_ESPIO_LEAFSWIRL_TORNADO)

-- protect us from projectiles inside a leaf swirl!
espio.LeafSwirlProtectionTics = function(mo)
	if not (mo.health) then return false end
	if not (mo.espio_leafswirlprotect) then return false end

	mo.espio_leafswirlprotect = $-1
	return true
end

espio.LeafSwirlProtect = function(mo, inflictor)
	if not ((mo and mo.valid) and mo.health) then return end
	if not (inflictor and inflictor.valid) then return end
	
	if (mo.espio_leafswirlprotect)
	and (inflictor.flags & MF_MISSILE)
		return false
	end
end

espio.LeafSwirlProtectionIterate = function()
	if not (server and server.valid) then return end
	if not (server.espio_leafswirlprotectlist) then server.espio_leafswirlprotectlist = {} end

	for key,mo in pairs(server.espio_leafswirlprotectlist) do
		if not (mo and mo.valid and espio.LeafSwirlProtectionTics(mo))
			table.remove(server.espio_leafswirlprotectlist, key)
			continue
		end
	end
end
	

-- MF_NOBLOCKMAP removal for leaf swirl projectile killing
local infoindex = 1
espio.NoBlockmapRemoval = function()
	while (mobjinfo[infoindex].spawnstate or mobjinfo[infoindex].health or mobjinfo[infoindex].radius or mobjinfo[infoindex].height) and (infoindex < #mobjinfo)
		-- check info and remove MF_NOBLOCKMAP from projectiles
		local info = mobjinfo[infoindex]
		if ((info.flags & (MF_MISSILE|MF_NOBLOCKMAP)) == (MF_MISSILE|MF_NOBLOCKMAP))
			info.flags = $ & ~(MF_NOBLOCKMAP)
		end
		
		-- increase index
		infoindex = $+1
	end
end

-- prune and phlox
chaotix.SafeFreeslot("SKINCOLOR_PRUNE")
skincolors[SKINCOLOR_PRUNE] = {
	name = "Prune",
	ramp = {176, 202, 205, 206, 207, 185, 185, 186, 168, 169, 169, 253, 253, 254, 254, 254},
	invcolor = SKINCOLOR_MINT,
	invshade = 7,
	chatcolor = V_PURPLEMAP,
	accessible = true
}
M_MoveColorAfter(SKINCOLOR_PRUNE, SKINCOLOR_NEON)

chaotix.SafeFreeslot("SKINCOLOR_PHLOX")
skincolors[SKINCOLOR_PHLOX] = {
	name = "Phlox",
	ramp = {144, 160, 161, 162, 163, 163, 164, 165, 166, 167, 167, 168, 168, 169, 253, 254},
	invcolor = SKINCOLOR_FOREST,
	invshade = 9,
	chatcolor = V_PURPLEMAP,
	accessible = true
}
M_MoveColorAfter(SKINCOLOR_PHLOX, SKINCOLOR_PASTEL)

-- super raspberry
chaotix.SafeFreeslot("SKINCOLOR_SUPERRASPBERRY1")
skincolors[SKINCOLOR_SUPERRASPBERRY1] = {
	name = "Super Raspberry 1",
	ramp = {0, 0, 0, 0, 200, 200, 200, 202, 202, 180, 180, 182, 182, 184, 254, 254},
	accessible = false
}
chaotix.SafeFreeslot("SKINCOLOR_SUPERRASPBERRY2")
skincolors[SKINCOLOR_SUPERRASPBERRY2] = {
	name = "Super Raspberry 2",
	ramp = {200, 200, 200, 200, 201, 201, 201, 202, 202, 182, 182, 184, 184, 184, 254, 254},
	accessible = false
}
chaotix.SafeFreeslot("SKINCOLOR_SUPERRASPBERRY3")
skincolors[SKINCOLOR_SUPERRASPBERRY3] = {
	name = "Super Raspberry 3",
	ramp = {200, 200, 201, 201, 202, 202, 202, 203, 203, 183, 183, 185, 185, 185, 254, 254},
	accessible = false
}
chaotix.SafeFreeslot("SKINCOLOR_SUPERRASPBERRY4")
skincolors[SKINCOLOR_SUPERRASPBERRY4] = {
	name = "Super Raspberry 4",
	ramp = {200, 200, 201, 201, 203, 203, 204, 204, 204, 184, 184, 185, 185, 185, 254, 254},
	accessible = false
}
chaotix.SafeFreeslot("SKINCOLOR_SUPERRASPBERRY5")
skincolors[SKINCOLOR_SUPERRASPBERRY5] = {
	name = "Super Raspberry 5",
	ramp = {200, 200, 202, 202, 204, 204, 205, 205, 205, 185, 185, 186, 186, 186, 254, 254},
	accessible = false
}