-- handles drilling, visuals AND certain functionality
local espio = chaotix.espio

local function espio_spinTrail(p)
	local mo = p.mo
	
	-- determine spin state
	espio.PlayerHeight(p)
	
	-- let us use espio's special roll again after we stop rolling normally
	if p.espio_normalroll and mo.state != S_PLAY_ROLL
		p.espio_normalroll = false
		p.charflags = $ &~ SF_NOJUMPSPIN
	end
	
	-- use a normal spin trail in this scenario
	if not (mo.state == S_PLAY_JUMP
	or mo.state == S_PLAY_SPINDASH
	or mo.state == S_ESPIO_ROLL)
		if mo.espio_drillspin
			p.spinitem = skins[p.skin].spinitem
			if p.spinitem == -1 then p.spinitem = MT_THOK end
			mo.espio_drillspin = false
		end
		return
	end
	p.spinitem = MT_NULL	-- reset this just incase
	mo.espio_drillspin = true
	
	if ((p.pflags & PF_SPINNING) and (P_AproxDistance(p.speed, mo.momz) > FixedMul(15<<FRACBITS, mo.scale)) and not (p.pflags & PF_JUMPED))
	and (P_IsObjectOnGround(mo) and (leveltime%2) and not (mo.eflags & MFE_GOOWATER))
		for i = 1, 2
			local dust = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_SPINDUST)
			
			-- set state under the right conditions
			if (mo.eflags & (MFE_TOUCHWATER|MFE_UNDERWATER)) 
				dust.state = S_SPINDUST_BUBBLE1
			elseif (p.powers[pw_shield] == SH_ELEMENTAL)
				dust.state = S_SPINDUST_FIRE1
			end
			
			-- set target and scale
			dust.target = mo
			dust.scale = mo.scale*2/3
			
			-- momentum
			local momz = 2*FRACUNIT + (P_RandomFixed()<<2)
			local angle = R_PointToAngle2(0, 0, mo.momx, mo.momy) + (ANG1*P_RandomRange(-50, 50))
			local speed = FRACUNIT + (P_RandomFixed()<<3)
			
			P_SetObjectMomZ(dust, momz)
			P_InstaThrust(dust, angle, -(speed))
			P_TryMove(dust, dust.x + dust.momx, dust.y + dust.momy, true)
		end
	end
end

-- when rolling, use the drill roll if it wouldn't crush us
espio.PlayerHeight = function(p)
	local mo = p.mo
	if (not mo) or mo.skin != espio.SKIN or p.espio_normalroll then return end
	
	-- force normal roll when bouncing with the bubble shield
	if mo.state == S_PLAY_FALL and p.secondjump == UINT8_MAX
	and p.pflags & PF_JUMPED and p.pflags & PF_THOKKED
	and not (p.pflags & (PF_NOJUMPDAMAGE|PF_SPINNING|PF_STARTJUMP))
		mo.state = S_PLAY_ROLL
		p.espio_normalroll = true
		p.charflags = $ | SF_NOJUMPSPIN
		return
	end
	
	-- always allow a normal roll under certain conditions
	if mo.state == S_PLAY_ROLL
	and (p.powers[pw_carry] or p.pflags & PF_SHIELDABILITY
	or mo.eflags & MFE_SPRUNG -- spin boosters
	or P_PlayerTouchingSectorSpecial(p, 4, 7)) -- force spin
		p.espio_normalroll = true
		p.charflags = $ | SF_NOJUMPSPIN
		return
	end
	
	-- in a roll...
	if not ((mo.state == S_PLAY_ROLL or mo.state == S_PLAY_JUMP
	or mo.state == S_PLAY_SPINDASH or mo.state == S_ESPIO_ROLL)
	and P_PlayerShouldUseSpinHeight(p)) then return end
	
	local height = P_GetPlayerHeight(p)
	
	-- no crushing, use drill roll
	if (mo.ceilingz - mo.floorz >= height)
		if mo.state == S_PLAY_ROLL
			mo.state = S_ESPIO_ROLL
		end
		return height
		
	-- uh oh! emergency rolling!
	else
		if mo.state != S_PLAY_ROLL
			mo.state = S_PLAY_ROLL
		end
		p.espio_normalroll = true
		p.charflags = $ | SF_NOJUMPSPIN
	end
end

local function espio_resetBuster(mo, dashspeed)
	if (dashspeed)
		mo.espio_fakedashspeed = nil
	end
	
	if (mo.espio_debrisobject)
		mo.espio_debrisobject = nil
	end
end

-- handles the spindash drilling into bustable floors
local function espio_floorBuster(p)
	local mo = p.mo
	if not (p.pflags & PF_STARTDASH and mo.state == S_PLAY_SPINDASH)
		espio_resetBuster(mo, true)
		return
	end
	
	-- get the rover we'll be checking
	local rover = mo.floorrover
	if (mo.eflags & MFE_VERTICALFLIP) then rover = mo.ceilingrover end
	
	if not (rover and rover.valid) then espio_resetBuster(mo) return end
	if not (rover.flags & FF_EXISTS) then espio_resetBuster(mo) return end
	if not ((rover.flags & FF_STRONGBUST) or (rover.flags & FF_BUSTUP)) then espio_resetBuster(mo) return end
	-- ^ after all these checks we should be good
		
	-- init var
	mo.espio_fakedashspeed = $ or 0
		
	local soundcalculation = (6*(mo.espio_fakedashspeed - p.mindash))/(p.maxdash - p.mindash)
	mo.espio_fakedashspeed = max(p.mindash, min(p.maxdash, $+FRACUNIT))
			
	-- redefine calculation then play spindash sound
	local chargecalculation = (6*(mo.espio_fakedashspeed - p.mindash))/(p.maxdash - p.mindash)
	if (soundcalculation != chargecalculation)
		-- try to get the object?
		if not (mo.espio_debrisobject)	-- this is actually a really laggy method so dont do this too much
		end
		
		-- spawn dirt from player
		for i = 1, 5
			local rock = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_PLAYER)
			rock.scale = mo.scale/3
			rock.sprite = SPR_ROIA
			rock.flags = MF_SCENERY
			rock.fuse = TICRATE*3/4
			P_InstaThrust(rock, ANG1*P_RandomRange(1, 360), P_RandomRange(2, 5)*rock.scale)
			P_SetObjectMomZ(rock, P_RandomRange(6, 12)*FRACUNIT)
		end
		
		S_StartSound(mo, sfx_crumbl)	-- time to revv the engine!
	end
	
	-- use normal roll frames
	if (mo.skin == espio.SKIN)
	and (mo.sprite2 == SPR2_SPIN)
		mo.sprite2 = SPR2_JUMP
		
		local step = (p.maxdash - p.mindash)/4
		local speed = (p.dashspeed - p.mindash)
		if (speed > 3*step)
			chaotix.SetFrame(mo, leveltime%6)
		elseif (speed > step)
			chaotix.SetFrame(mo, (leveltime%12)/2)
		else
			chaotix.SetFrame(mo, (leveltime%18)/3)
		end
	end
	
	-- break the bustable floor on max charge!
	if (mo.espio_fakedashspeed != p.maxdash) then return end
	
	-- bust the floor
	EV_CrumbleChain(rover)
	if (rover.master.flags & ML_EFFECT5)
		P_LinedefExecute(P_AproxDistance(rover.master.dx, rover.master.dy)>>FRACBITS, mo, mo.subsector.sector)
	end
	
	-- set player state
	p.pflags = $|(PF_JUMPED|PF_THOKKED|PF_SPINNING)
	mo.state = S_ESPIO_ROLL
	P_SetObjectMomZ(mo, chaotix.WaterMom(mo, 5*FRACUNIT))
	S_StartSound(mo, sfx_crumbl)	-- time to revv the engine!
end

local function espio_rollAnimating(p)
	local mo = p.mo
	if (mo.state != S_ESPIO_ROLL) then return end
	
	-- recreate how S_PLAY_ROLL animates
	local speed = FixedDiv(p.speed, FixedMul(mo.scale, mo.movefactor))	-- for some reason zoomtubes animate slower if they're slow as well
	if speed > 16<<FRACBITS
		mo.tics = min($, 1)
	end
	p.panim = PA_ROLL
end

espio.DrillSpin = function(p)
	local mo = p.mo
	if not ((mo and mo.valid) and mo.health and mo.skin == espio.SKIN)
		-- Make sure we reset normalroll
		if p.espio_normalroll
			p.espio_normalroll = false
			if skins[p.skin].name == espio.SKIN
				p.charflags = $ &~ SF_NOJUMPSPIN
			end
		end
		return
	end
	
	espio_spinTrail(p)
	espio_floorBuster(p)
	espio_rollAnimating(p)
end