-- theoretical espio ability: focus shift
local espio = chaotix.espio

-- start the teleport
local st_startup = 1
local st_tracing = 2
local st_homing = 3

espio.AbilitySpecial = function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	if not (espio.AbilityCheck(p)) then return end
	
	if (p.pflags & PF_THOKKED) then return end	-- thokked state
	if (p.gotcrystal) then return end	-- battlemod support
	if (CBW_Battle and p.exhaustmeter == 0) then return end	-- we're exhausted (battlemod)
	
	p.pflags = ($|PF_THOKKED) & ~(PF_JUMPED|PF_STARTJUMP|PF_SPINNING)
	mo.state = S_ESPIO_TELEPORT_STARTUP	-- set state
	mo.espio_teleport = true -- we'll need this
	mo.espio_teleport_state = st_startup
end

espio.TeleportLeaves = function(mo)
	local p = mo.player
	local angle = mo.angle
	if p then angle = p.cmd.angleturn<<16 end
	angle = $ + ANGLE_90
	
	for i = 1, 12
		local vertangle = i*(360/12*ANG1)
		local mobj = P_SpawnMobjFromMobj(mo, 0,0,
		FixedDiv(mo.height/2, mo.scale), MT_THOK)
		
		mobj.flags = $ &~ MF_NOGRAVITY
		mobj.rollangle = vertangle
		mobj.frame = $ &~ FF_PAPERSPRITE
		
		mobj.height = FRACUNIT
		mobj.scale = $/2
		mobj.spritexscale = $*2
		mobj.spriteyscale = mobj.spritexscale
		
		//State
		mobj.state = S_ESPIO_LEAFSWIRL_LEAF1
		mobj.tics = -1
		if P_RandomChance(FRACUNIT/4)
			mobj.frame = $ + P_RandomKey(G)
		end
		mobj.fuse = TICRATE
		
		//Thrust
		local speed = 3*P_RandomFixed() + FRACUNIT
		vertangle = $ + P_SignedRandom()*ANG1
		
		//Horizontal
		mobj.angle = angle
		local thrust = FixedMul(speed, sin(vertangle))
		P_InstaThrust(mobj, angle, FixedMul(thrust, mo.scale))
		
		//Vertical
		thrust = FixedMul(speed, cos(vertangle))
		mobj.momz = FixedMul(thrust, mo.scale)*P_MobjFlip(mo)
		mobj.momz = $ + 2*mo.scale*P_MobjFlip(mo)
		
		//3D!
		P_Thrust(mobj, angle - ANGLE_90, P_SignedRandom()*mo.scale*4/INT8_MAX)
		
		//Move a bit
		for i = 1, 8
			if mobj and mobj.valid
				P_RailThinker(mobj)
			end
		end
	end
end

espio.ResetTeleport = function(mo)
	local p = mo.player
	
	mo.espio_teleport = nil
	mo.espio_teleport_speed = nil
	mo.espio_teleport_tics = nil
	mo.espio_teleport_zshift = nil
	mo.espio_teleport_travelpos = nil
	mo.espio_teleport_traveltics = nil
	mo.espio_teleport_target = nil
	mo.espio_teleport_angle = nil
	mo.espio_teleport_homing = nil
	mo.espio_teleport_battleinfo = nil	-- reset this here, too
	
	-- spawn dust with state
	if (mo.espio_teleport_state == st_homing)
		for i = 1, 16
			local d = P_SpawnMobj(mo.x, mo.y, mo.z+mo.height/2, MT_SPINDUST)
			d.momx = P_RandomRange(-16, 16)*mo.scale
			d.momy = P_RandomRange(-16, 16)*mo.scale
			d.momz = P_RandomRange(-16, 16)*mo.scale
		end
		espio.TeleportLeaves(mo)
		S_StartSound(mo, sfx_s3k82)
	end
	mo.espio_teleport_state = nil
	
	-- fix flags & analog
	if (p and p.valid)
	and (p.espio_teleport_fixflags)	-- needs to be player for map swaps
		p.pflags = $ & ~PF_NOCLIP
		if (mo.health)	-- death has its own flags
			mo.flags = $ & ~(MF_NOCLIP|MF_NOCLIPTHING|MF_NOCLIPHEIGHT)
		end
		mo.flags2 = $ & ~MF2_DONTDRAW
		
		p.espio_teleport_fixflags = nil
	end
	
	-- revert state
	if mo.state >= S_ESPIO_TELEPORT_STARTUP and mo.state <= S_ESPIO_TELEPORT_HOMING
		mo.state = S_PLAY_FALL
	end
	
	-- remove tracer
	local tracer = mo.espio_teleport_tracer
	if (tracer and tracer.valid)
		P_RemoveMobj(tracer)
	end
end

--Get angle for teleporting
espio.TeleportAngle = function(p, start)
	local mo = p.mo
	if not p.mo then return end
	local angle = mo.angle
	
	-- influence our teleport angle with movement in simple
	-- but not backwards! that would be confusing with the rising back input
	if p.pflags & PF_ANALOGMODE
	and not (mo.flags2 & MF2_TWOD or twodlevel)
	and (start or p.cmd.forwardmove or p.cmd.sidemove)
		angle = p.cmd.angleturn<<16
		if p.cmd.forwardmove or p.cmd.sidemove
			angle = $ + R_PointToAngle2(0,0,
			abs(p.cmd.forwardmove)*FRACUNIT, -p.cmd.sidemove*FRACUNIT)
		end
	end
	return angle
end

-- Set the tracer's frame depending on displayplayer
espio.SetTracerVisibility = function(tracer, p)
	if (p == displayplayer)
	or (p == secondarydisplayplayer)
		-- show the tracer
		local mo = p.mo
		tracer.skin = mo.skin
		tracer.color = mo.color
		tracer.sprite = SPR_PLAY
		tracer.sprite2 = SPR2_FALL
		tracer.frame = A|FF_MODULATE
	else
		-- hide!
		tracer.sprite = SPR_INVS
		tracer.frame = A
	end
end

espio.Teleport = function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	if not (espio.AbilityCheck(p)) then espio.ResetTeleport(mo) return end
	
	-- the rest of the function is for the move itself
	if not (mo.espio_teleport)
		-- reset necessary shit
		espio.ResetTeleport(mo)
		
		-- dont do anything else
		return
	end
	
	-- ending the move early
	if not (mo.health)	-- we're dead
	or not (mo.state >= S_ESPIO_TELEPORT_STARTUP and mo.state <= S_ESPIO_TELEPORT_HOMING)	-- we got knocked out of it
	or (p.gotflag or p.gotcrystal)	-- we got the flag (or the crystal in battlemod)
	or (CBW_Battle and p.exhaustmeter == 0)	-- we're exhausted (battlemod)
		-- reset vars (and state if needed)
		if (mo.state >= S_ESPIO_TELEPORT_STARTUP and mo.state <= S_ESPIO_TELEPORT_HOMING) then mo.state = S_PLAY_FALL end
		espio.ResetTeleport(mo)
		return
	end
	
	mo.espio_teleport_tics = $ and $+1 or 1	-- increase tics for the ability
	local tics = mo.espio_teleport_tics
	
	local frames = 6	-- this'll be useful to expose
	if (mo.espio_teleport_state == st_startup)
		-- face our angle and frame
		p.drawangle = espio.TeleportAngle(p, true)
		chaotix.SetFrame(mo, (tics * (frames - 2)) / espio.TeleportStartupTics)
		
		-- decide what state to go to
		if (tics >= espio.TeleportStartupTics)
			-- transfer speed (affected by angle so we dont get fast 180 turns)
			local angle = R_PointToAngle2(0, 0, mo.momx, mo.momy) - mo.angle
			local anglefactor = abs(cos((chaotix.AngToInt(angle)/2) * ANG1))
			local anglespeed = FixedMul(anglefactor, FixedHypot(mo.momx, mo.momy))
			//mo.espio_teleport_speed = max(FixedMul(mo.scale, chaotix.WaterMom(mo, espio.TeleportJumpSpeed)), anglespeed)
			mo.espio_teleport_speed = max(FixedMul(mo.scale, chaotix.WaterMom(mo, p.normalspeed)), anglespeed)
			
			-- spawn the tracer
			local tracer = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_ESPIO_TELEPORT_TRACER)
			tracer.target = mo
			tracer.angle = p.drawangle
			tracer.radius = mo.radius	-- trust me this is a good thing
			tracer.height = mo.height
			tracer.shadowscale = FRACUNIT
			tracer.espio_speed = FixedMul(mo.scale, espio.TeleportTracerSpeed) + FixedMul(espio.TeleportTracerSpeedMultiplier, max(0, anglespeed))
			P_InstaThrust(tracer, tracer.angle, tracer.espio_speed)
			mo.espio_teleport_tracer = tracer	-- link it to espio
			espio.SetTracerVisibility(tracer, p)	-- hide it if we're not a display player
			
			-- spawn the arrows
			for i = 1, 2
				if i > 1 and not (p.pflags & PF_ANALOGMODE) then break end
				
				local arrow = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_THOK)
				arrow.state = S_LOCKON2
				arrow.frame = $ | FF_TRANS70 //| FF_REVERSESUBTRACT
				arrow.color = mo.color
				arrow.colorized = true
				
				arrow.frame = $ | FF_PAPERSPRITE
				arrow.rollangle = $ + ANGLE_90
				
				if p != displayplayer and p != secondarydisplayplayer
					arrow.flags2 = $ | MF2_DONTDRAW
				end
				
				local vertical = (p.height/3) &~ UINT16_MAX
				if i == 1
					arrow.target = mo
					local angle = p.drawangle
					p.drawangle = R_PointToAngle2(mo.x, mo.y, tracer.x, tracer.y)
					A_CapeChase(arrow, vertical, 0)
					p.drawangle = angle
				else
					arrow.target = tracer
					A_CapeChase(arrow, vertical, 0)
				end
				
				-- link the arrow to espio
				mo['espio_teleport_arrow'+i] = arrow
			end
				
			-- set state
			mo.state = S_ESPIO_TELEPORT_TRACING
			mo.espio_teleport_state = st_tracing
			mo.espio_teleport_tics = 0	-- reset tics
			mo.espio_teleport_zshift = 0	-- keep track of how much we shift it
			S_StartSound(mo, sfx_s3k42)
			chaotix.SetFrame(mo, frames - 2)
			return
		end
	elseif (mo.espio_teleport_state == st_tracing)
		-- end state early
		local tracer = mo.espio_teleport_tracer
		if not (tracer and tracer.valid)
			mo.state = S_PLAY_FALL
			espio.ResetTeleport(mo)
			return
		end
		
		-- set the tracer's momentum
		local countdown = espio.TeleportMaxTracingTics - tics
		local speed = (countdown * tracer.espio_speed) / espio.TeleportMaxTracingTics
		tracer.momx = FixedMul(speed, cos(tracer.angle))
		tracer.momy = FixedMul(speed, sin(tracer.angle))
		
		-- move the arrows if they exist
		for i = 1, 2
			local arrow = mo['espio_teleport_arrow'+i]
			if arrow and arrow.valid 
				arrow.tics = 2
				if not arrow.threshold
					arrow.frame = ($ &~ FF_TRANSMASK) | FF_TRANS30
					arrow.threshold = 1
				end
				
				arrow.movefactor = $ + speed/(6/i)
				local movement = (FixedDiv(arrow.movefactor, mo.scale)) &~ UINT16_MAX
				local vertical = i == 1 and (((tracer.z - mo.z)/2) &~ UINT16_MAX) or 0
				vertical = $ + ((p.height/3) &~ UINT16_MAX)
				if arrow.target == mo
					p.drawangle = R_PointToAngle2(mo.x, mo.y, tracer.x, tracer.y)
					A_CapeChase(arrow, vertical, movement)
				elseif arrow.target
					if (tracer.espio_enemy and tracer.espio_enemy.valid)
					and (arrow.target != tracer.espio_enemy)
						arrow.target = tracer.espio_enemy
					end
					
					local target = arrow.target
					local angle = target.angle
					target.angle = espio.TeleportAngle(p)
					A_CapeChase(arrow, vertical, movement)
					target.angle = angle
				end
			end
		end
		
		-- adjust the tracer's z momentum
		if (p.cmd.forwardmove < 0)
			-- increase tracer's z momentum to gain height
			local forwardmove_modifier = (abs(p.cmd.forwardmove) * FRACUNIT) / 50
			tracer.momz = $ + (FixedMul(forwardmove_modifier, FixedMul(mo.scale, espio.TeleportZTracingShift)) * P_MobjFlip(mo))
			
			-- track how much we've added
			local zshift_add = FRACUNIT / espio.TeleportMaxTracingTics
			mo.espio_teleport_zshift = $ + FixedMul(forwardmove_modifier, zshift_add)
			
			-- now adjust momentum based on our zshift (TODO: potentially rework this?)
			local modifier_range = espio.TeleportZTracingMinModifier - espio.TeleportZTracingMaxModifier
			local modifier = espio.TeleportZTracingMinModifier - abs(FixedMul(mo.espio_teleport_zshift, modifier_range))
			tracer.momx = FixedMul(modifier, $)
			tracer.momy = FixedMul(modifier, $)
		end
		
		-- visuals
		p.drawangle = espio.TeleportAngle(p)
		
		local visible = true
		if mo.frame & FF_TRANSMASK
		or mo.flags2 & (MF2_DONTDRAW | MF2_SHADOW)
			visible = false
		end
		
		if (tics%2)
		and (visible)
			local g = P_SpawnGhostMobj(mo)
			g.tics = 3
		end
		chaotix.SetFrame(mo, frames - (tics > espio.TeleportTracingFrameTics and 1 or 2))
		
		-- cut momentum
		local multiplier = espio.TeleportXYMultiplier
		mo.momx = FixedMul(multiplier, $)
		mo.momy = FixedMul(multiplier, $)
		mo.momz = 0	-- honestly just kill z momentum
		
		-- start the teleport!
		if (tics >= espio.TeleportMinTracingTics and not (p.cmd.buttons & BT_JUMP))
		or (tics >= espio.TeleportMaxTracingTics)
		or (tracer.espio_line) or (tracer.espio_spingap) or (tracer.espio_enemy)
		or (((not (tracer.eflags & MFE_VERTICALFLIP) and tracer.z + tracer.height >= tracer.ceilingz)
		or (tracer.eflags & MFE_VERTICALFLIP and tracer.z <= tracer.floorz)) and p.cmd.forwardmove < 0)
			local target = (tracer.espio_enemy and tracer.espio_enemy.valid) and tracer.espio_enemy or tracer
			
			-- handle the tracer
			if (target != tracer)
				-- not needed anymore!
				mo.espio_teleport_homing = true	-- lol
				P_RemoveMobj(tracer)
			else
				-- halt its momentum
				tracer.momx, tracer.momy, tracer.momz = 0, 0, 0
			end
			
			-- set state
			mo.state = S_ESPIO_TELEPORT_HOMING
			mo.espio_teleport_state = st_homing
			mo.espio_teleport_tics = 0	-- reset tics
			
			-- get how long it'll take to reach our target
			local travel_totaldist = FixedHypot(FixedHypot(target.x - mo.x, target.y - mo.y), target.z - mo.z) / FRACUNIT	-- okay fixed point was too much
			local travel_tics = (FixedDiv(travel_totaldist, mo.scale) * TICRATE) / espio.TeleportHomingScalar
			mo.espio_teleport_traveltics = max(1, travel_tics)
			mo.espio_teleport_travelpos = {x = mo.x, y = mo.y, z = mo.z}	-- also store our current position
			mo.espio_teleport_target = target	-- set target
			mo.espio_teleport_angle = R_PointToAngle2(mo.x, mo.y, target.x, target.y)	-- cache this for the release
			S_StartSound(mo, sfx_s3k82)
			
			-- set flags
			p.pflags = $|PF_NOCLIP
			mo.flags = $|(MF_NOCLIP|MF_NOCLIPTHING|MF_NOCLIPHEIGHT)
			mo.flags2 = $|MF2_DONTDRAW
			p.espio_teleport_fixflags = true	-- important
			
			-- dust
			for i = 1, 16
				local d = P_SpawnMobj(mo.x, mo.y, mo.z+mo.height/2, MT_SPINDUST)
				d.momx = P_RandomRange(-16, 16)*mo.scale
				d.momy = P_RandomRange(-16, 16)*mo.scale
				d.momz = P_RandomRange(-16, 16)*mo.scale
			end
			espio.TeleportLeaves(mo)
			
			return
		end
	elseif (mo.espio_teleport_state == st_homing)
		local validTarget = false
		
		-- track target
		local target = mo.espio_teleport_target
		
		if (target and target.valid)
		and (target.health)
			-- home in on our target
			local pos = mo.espio_teleport_travelpos
			local angle = R_PointToAngle2(pos.x, pos.y, target.x, target.y)
			local distance = R_PointToDist2(pos.x, pos.y, target.x, target.y)
			local height = pos.z - target.z
			
			local travel_dist = ease.outquad((tics * FRACUNIT) / mo.espio_teleport_traveltics, distance)
			local travel_height = ease.outquad((tics * FRACUNIT) / mo.espio_teleport_traveltics, height)
		
			local xpos = pos.x + FixedMul(travel_dist, cos(angle))
			local ypos = pos.y + FixedMul(travel_dist, sin(angle))
			local zpos = pos.z - travel_height
			
			mo.momx, mo.momy, mo.momz = 0, 0, 0
			mo.espio_teleport_angle = angle
			P_MoveOrigin(mo, xpos, ypos, zpos)
			mo.flags2 = $|MF2_DONTDRAW	-- all sneaky-like
			
			-- got a valid target still
			validTarget = true
		end
		
		-- end the teleport
		if (tics >= mo.espio_teleport_traveltics)
		or not (validTarget)
			-- re-adjust for anglespeed (again)
			local angle = mo.espio_teleport_angle - mo.angle
			local anglefactor = abs(cos((chaotix.AngToInt(angle)/2) * ANG1))
			local anglespeed = FixedMul(anglefactor, mo.espio_teleport_speed)
			
			local speed = max(FixedMul(mo.scale, chaotix.WaterMom(mo, p.normalspeed)), anglespeed)
			local height = chaotix.WaterMom(mo, espio.TeleportJumpHeight)
			
			-- adjust based on our z momentum
			if (mo.espio_teleport_zshift)
				-- modify speed/height dependant on the zshift
				local xy_modifier_range = espio.TeleportJumpXYModifier - FRACUNIT
				local z_modifier_range = espio.TeleportJumpZModifier - FRACUNIT
				speed = FixedMul(FRACUNIT + FixedMul(mo.espio_teleport_zshift, xy_modifier_range), $)
				height = FixedMul(FRACUNIT + FixedMul(mo.espio_teleport_zshift, z_modifier_range), $)
			end
			
			-- holding spin should reduce the height we gain
			if (p.cmd.buttons & BT_SPIN) then height = FixedMul(espio.TeleportJumpSpinModifier, $) end
			
			-- set launch angle
			local launchangle = espio.TeleportAngle(p)
			
			if (target and target.valid) 
			and (target.espio_line)	-- we got a line!
			and (chaotix.CheckLine(p, target.espio_line))	-- it's a valid wall to cling to
			and (p.cmd.buttons & BT_JUMP)	-- gotta be holding jump to cling, anyways!
				-- face the wall
				local line = target.espio_line
						
				local whichside = P_PointOnLineSide(mo.x, mo.y, line)
				launchangle = R_PointToAngle2(line.v1.x, line.v1.y, line.v2.x, line.v2.y)
				launchangle = $ + (ANGLE_90 * (whichside and -1 or 1))
						
				-- conditions to not get any height
				if not (p.cmd.forwardmove > 0)	-- not holding forward
				or not (mo.espio_wallrun_timer)	-- no wallrun stamina
					height = 0
				end
			end
			
			-- nerf speed a little if in battlemod and the target is a player
			if (target.player)
			and (CBW_Battle)
				speed = FixedMul(espio.TeleportBattleMultiplier, $)
				//print("reduce speed")
			end
			
			-- launch out
			local homing = (mo.espio_teleport_homing) and true or false
			p.pflags = $|(PF_JUMPED|PF_THOKKED)
			mo.state = S_PLAY_JUMP
			P_InstaThrust(mo, launchangle, speed)
			P_SetObjectMomZ(mo, height)
			S_StartSound(mo, sfx_kc5b)
			
			-- dust (now done in espio.ResetTeleport)
			/*for i = 1, 16
				local d = P_SpawnMobj(mo.x, mo.y, mo.z+mo.height/2, MT_SPINDUST)
				d.momx = P_RandomRange(-16, 16)*mo.scale
				d.momy = P_RandomRange(-16, 16)*mo.scale
				d.momz = P_RandomRange(-16, 16)*mo.scale
			end*/
			
			mo.espio_spikebreaker = true
			espio.ResetTeleport(mo)
			mo.z = $ + P_MobjFlip(mo)
			mo.espio_spikebreaker = false
			
			-- homing stuff
			if (homing)
				p.pflags = $ & ~(PF_THOKKED)	-- reset thokked state
				mo.espio_teleport_battleinfo = {
					momx = mo.momx,
					momy = mo.momy,
					momz = mo.momz,
				}
				mo.espio_teleport_homing = true
			end
			
			return
		end
	end
end

espio.TeleportShouldDamage = function(mo, inflictor, source, damage, damagetype)
	if not (mo and mo.valid)	-- not valid!
	or not (mo.espio_teleport and mo.espio_teleport_state == st_homing)	-- not focus shifting!
		return 
	end
	
	if ((inflictor and inflictor.valid) and (inflictor.flags & MF_BOXICON))	-- dont get protected from eggman boxes!
	or (damagetype & DMG_DEATHMASK)	-- not an instakill sector
		return
	end
	
	-- get protected!
	return false
end

-- prevent jump/spin moves during teleport
chaotix.AddHook("JumpSpecial", function(p)
	if p and p.mo and p.mo.espio_teleport_state then return true end
end, MT_PLAYER)
chaotix.AddHook("SpinSpecial", function(p)
	if p and p.mo and p.mo.espio_teleport_state then return true end
end, MT_PLAYER)

-- Destroy spikes on touch
chaotix.AddHook("MobjMoveCollide", function(mo, other)
	if not(mo.valid and mo.espio_spikebreaker)
	or not(mo.z <= other.z + other.height and other.z <= mo.z + mo.height)
	or not (other.valid and (other.type == MT_SPIKE or other.type == MT_WALLSPIKE))
		return
	end
	if (other.type == MT_SPIKE or other.type == MT_WALLSPIKE)
		P_KillMobj(other, mo, mo)
	end
end, MT_PLAYER)
