-- Everything related to Espio's Wallrun
local espio = chaotix.espio

espio.resetWallRun = function(p)
	local mo = p.mo
	
	p.espio_wallrun, mo.espio_wallrun = nil, nil
	mo.espio_wallrun_lastmomz = nil
	mo.espio_wallrun_slidemomz = nil
	mo.espio_wallrun_leveltime = nil
	mo.espio_wallrun_frame = nil
	mo.espio_wallrun_frametics = nil
end

-- handle the delay before we can cling onto a wall
espio.clingDelay = function(p)
	local mo = p.mo
	
	if (P_IsObjectOnGround(mo))
		-- we only wanna reset this on the ground
		mo.espio_wallrun_clingdelay = espio.WallRunDelay
	else
		mo.espio_wallrun_clingdelay = $ and $-1
	end
end

espio.wallCling = function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	if not (espio.AbilityCheck(p)) then return end
	if (mo.espio_wallrun) then return end	-- we're already running up a wall!
	
	if (p.lastlinehit != -1)	-- we hit a line?
		local l, s = lines[p.lastlinehit], sides[p.lastsidehit]
		
		if ((p.pflags & PF_JUMPED and not mo.espio_wallrun_clingdelay)
		or (mo.state == S_ESPIO_TELEPORT_STARTUP and not mo.espio_teleport_tracer))	-- we need to be jumping
		and (p.cmd.buttons & BT_JUMP)	-- we also need to be holding jump
		and not (CBW_Battle and p.exhaustmeter == 0)	-- we're exhausted (battle mod)
		and not (p.gotflag or p.gotcrystal)	-- we have the flag (or the crystal in battlemod)
		and (chaotix.CheckLine(p, l))	-- make sure everything is good here
			-- face the wall
			local whichside = P_PointOnLineSide(mo.x, mo.y, l)
			local angle = R_PointToAngle2(l.v1.x, l.v1.y, l.v2.x, l.v2.y)
			angle = $ + (ANGLE_90 * (whichside and -1 or 1))
			p.drawangle = angle
			
			-- little hack to face the wall without moving the camera (SMS Alfredo)
			local thrustx = P_ReturnThrustX(mo, angle, p.height)
			local thrusty = P_ReturnThrustY(mo, angle, p.height)
			local thok = P_SpawnMobjFromMobj(mo, thrustx, thrusty, 0, MT_THOK)
				
			local oldtarget = mo.target
			mo.target = thok
			A_FaceTarget(mo)
			
			mo.target = oldtarget
			thok.state = S_NULL
				
			-- make us start running up it (this is a dirty climbing hack)
			p.charability = CA_GLIDEANDCLIMB
			p.climbing = 5
			p.espio_wallrun, mo.espio_wallrun = true, true	-- trust me there's a reason we have two saved
			mo.espio_wallrun_lastmomz = mo.momz
			mo.espio_wallrun_leveltime = leveltime
			mo.espio_wallrun_forwardmove = max(p.cmd.forwardmove, 0)
			S_StartSound(mo, sfx_kc45)
				
			-- reset stuff on cling
			p.pflags = $ & ~(PF_JUMPED|PF_STARTJUMP|PF_SHIELDABILITY)
			p.secondjump = 0
			
			-- visual stuff
			mo.espio_wallrun_dusttics = espio.WallRunDustTicBounds[2]
			mo.espio_wallrun_frame = 0
			mo.espio_wallrun_frametics = espio.WallRunMaxFrameTics
			
			-- set speed based on momz
			mo.espio_wallrun_speed = mo.momz
			if (mo.espio_wallrun_lastmomz != nil)
				mo.espio_wallrun_speed = mo.espio_wallrun_lastmomz
			end
			
			-- set horizontal momentum (stolen from mighty)
			if p.cmd.sidemove and not (twodlevel or (mo.flags2 & MF2_TWOD))
				local momx, momy = mo.espio_oldmomx or mo.momx, mo.espio_oldmomy or mo.momy
				local moangle = R_PointToAngle2(0, 0, momx, momy)
				local hitangle = (angle - ANGLE_90) - moangle
				local speedfactor = abs(cos(hitangle))
				local speed = FixedMul(speedfactor, FixedHypot(momx, momy))
				
				local reverseangle = (angle + ANGLE_180) - moangle
				if (chaotix.AngToInt(reverseangle) > 180)
					speed = $ * -1
				end
				mo.espio_wallrun_shift = FixedMul(chaotix.mighty.WallKickMomStabilizer, speed)
				
				-- influence how much we keep horizontal momentum via sidemove
				if (mo.espio_wallrun_shift > 0) == (p.cmd.sidemove < 0) -- same direction
					mo.espio_wallrun_shift = $*abs(p.cmd.sidemove)/50
				else
					mo.espio_wallrun_shift = 0
				end
			else
				mo.espio_wallrun_shift = 0
			end
			
			-- lose exhaust on cling (battlemod)
			if (CBW_Battle)
				local warningtic = FRACUNIT/3
				local loss = espio.WallExhaustLoss
				
				if p.exhaustmeter and (p.exhaustmeter > warningtic) then
					p.exhaustmeter = max(warningtic, $ - loss)
				else
					p.exhaustmeter = max(0, $ - loss)
				end
			end
			
			return
		end
		
		-- reset these to check for wall collision again
		p.lastlinehit, p.lastsidehit = -1, -1	
	end
	mo.espio_oldmomx, mo.espio_oldmomy = mo.momx, mo.momy	-- i hate that I have to store this...
end

espio.WallRun = function(p)
	local mo = p.mo
	if not ((mo and mo.valid) and mo.health) then return end
	if not (espio.AbilityCheck(p)) then espio.resetWallRun(p) return end
	
	-- handle stamina
	if (mo.espio_wallrun_timer == nil) then mo.espio_wallrun_timer = FRACUNIT end	-- init var to fix certain edge cases
	
	if (P_IsObjectOnGround(mo))	-- get stamina back when grounded
	or (p.powers[pw_carry])	-- or when being carried
		mo.espio_wallrun_timer = FRACUNIT 
	end
	
	espio.clingDelay(p)	-- the delay before we CAN cling
	espio.wallCling(p)	-- the actual cling stuff
	-- ^ cling to walls
	
	if (p.climbing)
		-- lower stamina
		-- forward check
		local forward
		if (twodlevel or (mo.flags2 & MF2_TWOD))	-- fuck
			forward = (p.cmd.sidemove > 0 and mo.angle == 0 or p.cmd.sidemove < 0 and mo.angle == ANGLE_180)
		else
			forward = p.cmd.forwardmove > 0
		end
		
		-- only lose tics if we are running up!
		if (forward)
			local ticsremove = (FRACUNIT/espio.WallRunTics)
			mo.espio_wallrun_timer = max(0, $ - ticsremove)
		end
		
		-- if we're not holding jump, get off the wall!
		if not (p.cmd.buttons & BT_JUMP)
			-- jump off
			p.charability = CA_ESPIOREAL
			p.climbing = 0
			P_DoJump(p, true)
			p.pflags = $ & ~(PF_STARTJUMP|PF_THOKKED)
			S_StartSound(mo, sfx_jump)
			
			local angle = mo.angle + ANGLE_180
			mo.angle = angle
			P_InstaThrust(mo, angle, FixedMul(p.jumpfactor, FixedMul(mo.scale, espio.WallRunJumpSpeed)))
			
			-- momz is a lot more annoying than I thought
			mo.momz = FixedMul(p.jumpfactor, mo.espio_wallrun_lastmomz)
			
			-- minimum height
			local minheight = FixedMul(p.jumpfactor, FixedMul(mo.scale, chaotix.WaterMom(mo, espio.WallRunJumpMinHeight)))
			
			if ((mo.eflags & MFE_VERTICALFLIP) and mo.momz > -(minheight))
			or (not (mo.eflags & MFE_VERTICALFLIP) and mo.momz < minheight)
				mo.momz = (minheight)*P_MobjFlip(mo)
			end
			espio.resetWallRun(p)
			
			-- reset these to prevent reclings
			p.lastlinehit = -1
			p.lastsidehit = -1
			return
		end
	else	-- make sure our ability is reset if we're not wall running anymore
		if (p.espio_wallrun)
			if (mo.espio_wallrun)	-- THIS is why we had both saved
				if not (P_IsObjectOnGround(mo))
				and not (P_PlayerInPain(p))
					p.pflags = $|(PF_STARTJUMP) & ~(PF_THOKKED)
					mo.momz = mo.espio_wallrun_lastmomz*3/4
					
					if (p.cmd.buttons & BT_SPIN)
						-- minimum height
						if ((mo.eflags & MFE_VERTICALFLIP) and mo.momz > FixedMul(mo.scale, -3*FRACUNIT))
						or (not (mo.eflags & MFE_VERTICALFLIP) and mo.momz < FixedMul(mo.scale, 3*FRACUNIT))
							mo.momz = FixedMul(mo.scale, FRACUNIT*3) * P_MobjFlip(mo)
						end
					else
						mo.espio_wallrun_clingdelay = espio.WallRunDelay	-- this helps more than you'd think
					end
				end
				
				S_StartSound(mo, sfx_kc45)
			end
			
			-- this needs to always run, otherwise we can keep the wrong ability through levels
			p.charability = CA_ESPIOREAL
			espio.resetWallRun(p)
			
			-- reset these to prevent reclings
			p.lastlinehit = -1
			p.lastsidehit = -1
			return
		else
			-- store this for when we cling
			mo.espio_wallrun_lastmomz = mo.momz
		end
	end
end

chaotix.AddHook("PreThinkFrame", do
	for p in players.iterate
		local mo = p.mo
		
		if not (espio.AbilityCheck(p)) then continue end
		if not (p.climbing) then continue end
		
		-- sidemove stuff
		mo.espio_wallrun_sidemove = p.cmd.sidemove
		p.cmd.sidemove = 0
		
		-- slight hack to make horizontal momentum capped
		-- setting forward move to be -1/1 max so we can still have wall scrolling momentum without too much power
		local forward
		if (twodlevel or (mo.flags2 & MF2_TWOD))	-- fuck
			local right = abs(mo.angle) < ANGLE_90 and mo.angle != ANGLE_180
			forward = (mo.espio_wallrun_sidemove > 0 and right
			or mo.espio_wallrun_sidemove < 0 and not right)
			p.cmd.forwardmove = abs(mo.espio_wallrun_sidemove)
		else
			forward = p.cmd.forwardmove > 0
		end
		if (forward)
			mo.espio_wallrun_forwardmove = p.cmd.forwardmove	-- store this
			p.cmd.forwardmove = min(1, $)
			continue 
		end
		
		-- wall sliding cmd stuff stuff
		p.cmd.forwardmove = -1
	end
end)

-- playerthink allows us to edit momentum while climbing
chaotix.AddHook("PlayerThink", function(p)
	local mo = p.mo
	if not ((mo and mo.valid) and mo.health) then return end
	
	-- wall running checks
	if not (espio.AbilityCheck(p)) then return end
	if not (p.climbing) then return end
	
	local wallruntimer = leveltime - mo.espio_wallrun_leveltime
	-- ^ add some difference between multiple espios
		
	-- forward check
	local forward
	if (twodlevel or (mo.flags2 & MF2_TWOD))	-- fuck
		local right = abs(mo.angle) < ANGLE_90 and mo.angle != ANGLE_180
		forward = (mo.espio_wallrun_sidemove > 0 and right
		or mo.espio_wallrun_sidemove < 0 and not right)
	else
		forward = p.cmd.forwardmove > 0
	end
	
	-- give a speed boost when you're super or have super sneakers
	local speedup = (p.powers[pw_sneakers] or P_PlayerFullbright(p))
	local modifier = espio.WallRunSpeedModifier
	local accelThreshold = FixedMul(mo.scale, espio.WallRunAccelerationThreshold)
	
	-- sideways shifting
	if mo.espio_wallrun_sidemove and not (twodlevel or (mo.flags2 & MF2_TWOD))
		-- get stats for shifting around the wall
		local speed = abs(mo.espio_wallrun_shift)
		local thrustfactor = p.thrustfactor
		local acceleration = p.accelstart + (FixedDiv(speed, mo.scale)>>FRACBITS) * p.acceleration
		
		-- adjust the stats depending on situations
		local maxshift_modifier = espio.WallRunSidewaysModifier
		local maxshift = FixedMul(FixedMul(maxshift_modifier, p.normalspeed), modifier)
		maxshift = FixedMul(mo.scale, $)
		
		if (speedup)
			maxshift = $ * 5 / 3	-- same as source's maxspeed
			thrustfactor = $ * 2
			acceleration = p.accelstart/2 + (FixedDiv(speed, mo.scale)>>FRACBITS) * p.acceleration/2
		end
		if (mo.eflags & MFE_UNDERWATER)
			maxshift = $ >> 1
			acceleration = $ * 2 / 3
			accelThreshold = $ / 2
		end
		if (mo.espio_wallrun_speed < accelThreshold)
			acceleration = $ * 5 / 3
		end
		
		-- increase sideways momentum
		local oldShift = abs(mo.espio_wallrun_shift)
		local movepushsideways = FixedMul(modifier, -(mo.espio_wallrun_sidemove or 0) * (thrustfactor * acceleration))
		mo.espio_wallrun_shift = $ + FixedMul(movepushsideways, mo.scale)
		
		-- normalize shift
		local newShift = abs(mo.espio_wallrun_shift)
		if (newShift > maxshift)
			if (oldShift > maxshift)
				if (newShift > oldShift)
					mo.espio_wallrun_shift = FixedMul(FixedDiv($, newShift), oldShift)
				end
			else
				mo.espio_wallrun_shift = FixedMul(FixedDiv($, newShift), maxshift)
			end
		end
		
		P_InstaThrust(mo, mo.angle + ANGLE_90, mo.espio_wallrun_shift)
		mo.espio_wallrun_shift = FixedMul($, espio.WallRunFriction)	-- arbitrary friction number
	else
		if mo.espio_wallrun_shift
			local removemomz = FixedMul(mo.scale*2, espio.WallRunSpeedSlide)
			if mo.espio_wallrun_shift > 0
				mo.espio_wallrun_shift = $ - removemomz
				mo.espio_wallrun_shift = max($, 0)
			else
				mo.espio_wallrun_shift = $ + removemomz
				mo.espio_wallrun_shift = min($, 0)
			end
		end
		
		P_InstaThrust(mo, mo.angle + ANGLE_90, mo.espio_wallrun_shift)
	end
	
	-- detect if we're moving forward
	if (forward)
	and (mo.espio_wallrun_timer)	-- we need stamina too
		
		-- get stats for running up the wall
		local speed = abs(mo.espio_wallrun_speed)
		local thrustfactor = p.thrustfactor
		local acceleration = p.accelstart + (FixedDiv(speed, mo.scale)>>FRACBITS) * p.acceleration
		
		-- set maxspeed based on tics left
		local tics = (mo.espio_wallrun_timer*100)/FRACUNIT
		local maxspeed = FixedMul(p.normalspeed, modifier) >> FRACBITS
		maxspeed = (((tics * maxspeed) / 100) * FRACUNIT)	-- adjust the maxspeed based on the timer
		maxspeed = FixedMul(mo.scale, $)
		
		-- adjust the stats depending on situations
		if (speedup)
			maxspeed = $ * 5 / 3	-- same as source's maxspeed
			thrustfactor = $ * 2
			acceleration = p.accelstart/2 + (FixedDiv(speed, mo.scale)>>FRACBITS) * p.acceleration/2
		end
		if (mo.eflags & MFE_UNDERWATER)
			maxspeed = $ >> 1
			acceleration = $ * 2 / 3
			accelThreshold = $ / 2
		end
		if (mo.espio_wallrun_speed < accelThreshold)
			acceleration = $ * 5 / 3
		end
		
		-- increase the wall momentum
		local oldSpeed = mo.espio_wallrun_speed
		local movepushupwards = FixedMul(modifier, (mo.espio_wallrun_forwardmove or 0) * (thrustfactor * acceleration))
		mo.espio_wallrun_speed = $ + FixedMul(movepushupwards, mo.scale)
		
		-- normalize speed
		local newSpeed = mo.espio_wallrun_speed
		if (newSpeed > maxspeed)
			if (oldSpeed > maxspeed)
				if (newSpeed > oldSpeed)
					mo.espio_wallrun_speed = FixedMul(FixedDiv($, newSpeed), oldSpeed)
				end
			else
				mo.espio_wallrun_speed = FixedMul(FixedDiv($, newSpeed), maxspeed)
			end
		end
		
		-- scale the wall!
		mo.momz = $ + (mo.espio_wallrun_speed * P_MobjFlip(mo))
		mo.espio_wallrun_lastmomz = mo.momz
		mo.espio_wallrun_slidemomz = nil

		-- set frames
		local faster = false
		
		if (mo.espio_wallrun_speed > FixedMul(mo.scale, espio.WallRunFastThreshold))	-- above running frame fasttics speed
		or (mo.espio_wallrun_speed < accelThreshold)	-- regaining speed
		or (mo.espio_wallrun_timer < espio.WallRunTiredThreshold)	-- getting tired
			faster = true
		end
	
		if (mo.state != S_PLAY_CLIMB) then mo.state = S_PLAY_CLIMB end
		local frame = (wallruntimer%8)/2
		if (faster) then frame = wallruntimer%4 end
		chaotix.SetFrame(mo, frame, SPR2_CLMB)
		
		-- extra effects
		if (mo.espio_wallrun_timer < espio.WallRunTiredThreshold)	-- getting tired
		or (mo.espio_wallrun_speed < accelThreshold)	-- regaining speed
			if ((wallruntimer%4) == 1 or (wallruntimer%4) == 3)
				for i = 1, 2
					-- distance to the wall
					local distance = 13
					local xoff = distance*cos(mo.angle)
					local yoff = distance*sin(mo.angle)
					
					-- make the dust come from espio's shoes
					local angle = mo.angle - (((wallruntimer%4) == 1) and ANGLE_270 or ANGLE_90)
					local distance = 14 + P_RandomRange(-8, 8)
					xoff = $ + distance*cos(angle)
					yoff = $ + distance*sin(angle)
					
					local dust = P_SpawnMobjFromMobj(mo, xoff, yoff, P_RandomRange(-2, 2)*FRACUNIT, MT_SPINDUST)
					dust.scale = $*2/3
					if (mo.eflags & MFE_UNDERWATER)
						dust.state = S_SPINDUST_BUBBLE1
						S_StartSoundAtVolume(mo, _G["sfx_bubbl"..P_RandomRange(1, 5)], 128)
					else
						S_StartSoundAtVolume(mo, sfx_s251, 96)
					end
					
					-- make the dust burst out a bit from us
					local momz = 2*FRACUNIT + (P_RandomFixed()<<2)
					local angle = mo.angle + (ANG1*P_RandomRange(-35, 35))
					local speed = (FRACUNIT/4) + (P_RandomFixed()<<3)
					
					P_SetObjectMomZ(dust, momz)
					P_InstaThrust(dust, angle, -(speed))
					P_TryMove(dust, dust.x + dust.momx, dust.y + dust.momy, true)
				end
			end
		end
		
		-- apply friction
		mo.espio_wallrun_speed = FixedMul($, espio.WallRunFriction)	-- arbitrary friction number
		
		-- make sure frame stuff is reset here
		mo.espio_wallrun_frame = 0
		mo.espio_wallrun_frametics = espio.WallRunMaxFrameTics
	else	-- slide down
		if (mo.espio_wallrun_slidemomz == nil)
			-- assume this is when we start sliding
			mo.espio_wallrun_slidemomz = mo.espio_wallrun_lastmomz/2
		end
		
		-- modify slide momz
		local removemomz = FixedMul(mo.scale, espio.WallRunSpeedSlide)*P_MobjFlip(mo)
		-- flip this again in goop
		if (mo.eflags & MFE_GOOWATER)
			removemomz = $ * -1
		end
		
		mo.espio_wallrun_slidemomz = $ - (chaotix.WaterMom(mo, removemomz))
		mo.espio_wallrun_speed = mo.espio_wallrun_slidemomz	-- update this
		
		if (mo.state != S_PLAY_CLING) then mo.state = S_PLAY_CLING end
		chaotix.SetFrame(mo, 0, SPR2_CLNG)
		mo.momz = $ + mo.espio_wallrun_slidemomz
		mo.espio_wallrun_lastmomz = mo.momz
		
		-- set frame and spawn dust!
		local speed = FixedHypot(mo.momz, mo.espio_wallrun_shift)	-- take both momxy and momz into account
	
		if (speed)
			-- frame animation
			local frame_speed = FixedMul(mo.scale, espio.WallRunMaxFrameSpeed)
			local frame_tics = espio.WallRunMaxFrameTics
			local maxtics = max(1, frame_tics - ((speed * frame_tics) / frame_speed))
			mo.espio_wallrun_frametics = $ and min(maxtics, $) or maxtics
			
			local numframes = 2
			mo.espio_wallrun_frametics = $-1
			if not (mo.espio_wallrun_frametics)
				mo.espio_wallrun_frame = $+1
					
				-- reset frame after max frame is reached
				if (mo.espio_wallrun_frame > (numframes - 1))
					mo.espio_wallrun_frame = 0
				end
			end
			chaotix.SetFrame(mo, mo.espio_wallrun_frame, SPR2_CLNG)
		
			-- dust particles on the wall
			local mombounds = espio.WallRunDustMomBounds
			if (speed > mombounds[1])
				speed = $ - mombounds[1]
				
				-- duration is based on how fast we're going
				local momrange = FixedMul(mo.scale, mombounds[2] - mombounds[1])
				local ticbounds = espio.WallRunDustTicBounds
				local ticrange = ticbounds[2] - ticbounds[1]
				
				local mintics = max(ticbounds[1], ticbounds[2] - ((speed * ticrange) / momrange))
				mo.espio_wallrun_dusttics = min(mintics, $) - 1
				if not (mo.espio_wallrun_dusttics)
					for i = 1, 2
						-- distance to the wall
						local distance = 13
						local xoff = distance*cos(mo.angle)
						local yoff = distance*sin(mo.angle)
						
						-- distance to espio's hand (with some randomization)
						local distance = 15 + P_RandomRange(-8, 8)
						xoff = $ + distance*cos(mo.angle - ANGLE_90)
						yoff = $ + distance*sin(mo.angle - ANGLE_90)
						local zoff = FRACUNIT*42 + (P_RandomRange(-8, 8)*FRACUNIT)
						
						local dust = P_SpawnMobjFromMobj(mo, xoff, yoff, zoff, MT_SPINDUST)
						dust.state = (mo.eflags & MFE_UNDERWATER) and S_SPINDUST_BUBBLE1 or $
						dust.scale = $*2/3
						S_StartSound(mo, sfx_s3k7e)
					end
					
					-- reset tics
					mo.espio_wallrun_dusttics = mintics
				end
			end
		else
			-- reset frame info
			mo.espio_wallrun_frame = 0
			mo.espio_wallrun_frametics = espio.WallRunMaxFrameTics
		end
	end
		
	-- visual ghost (copypaste from source)
	if (speedup) and (abs(mo.momz) > FixedMul(mo.scale, FRACUNIT*10))
		local g = P_SpawnGhostMobj(mo)
		g.fuse = 2
		if (g.tracer) then g.tracer.fuse = 2 end
			
		if (leveltime%2)
			g.frame = $ & ~(FF_TRANSMASK)
			g.frame = $|FF_TRANS70
			if (g.tracer)
				g.tracer.frame = $ & ~(FF_TRANSMASK)
				g.tracer.frame = $|(FF_TRANS70)
			end
		end
	end
end)