-- everything for mighty in battle mod
local mighty = chaotix.mighty

-- loading mighty's stuff
local loaded = false
mighty.BattleSetup = function()
	if not (CBW_Battle) then return end	-- battle isnt loaded
	if (loaded) then return end	-- already setup for battle? dont run
	local B = CBW_Battle
	local S = B.SkinVars
	
	-- mighty's battle special states (defined above for use as constants)
	local st_charging = 1	-- charging the punch
	local st_thrusting = 2	-- released, thrusting forward
	local st_punch = 3	-- punched someone, recoil state
	
	-- mighty's priority
	local function mighty_battleShell(mo, other)
		local xyangle = B.GetCollideRelativeAngle
		return (abs(xyangle(mo, other, true)) >= ANGLE_135)
	end
	B.AddPriorityFunction("mighty_shell", mighty_battleShell)

	local function mighty_battlePriority(p)
		local mo = p.mo
		
		-- punch? (megaton punch interactions are special, where defence IS the attack!)
		if (p.actionstate == st_thrusting)
			B.SetPriority(p, 0, 1, "knuckles_glide", 0, 3, "megaton punch")
			return
		end
		
		-- hammer dropping? (needs to be beyond the startup)
		if (mo.mighty_hammerdrop and mo.mighty_hammerdrop_timer) 
		and (mo.mighty_hammerdrop_timer >= mighty.HammerDropStartupTics)
			B.SetPriority(p, 2, 2, "fang_tailbounce", 3, 3, "hammer drop")
			return
		end
		
		-- let's assume no other priority function has been run, this'll be what these will be set to
		if not (p.battle_atk or p.battle_def)
		and (p.battle_sfunc == "can_damage")
		and not (p.tumble and p.tumble > 0)	-- not in tumble
		and not (p.guard == 1)	-- not guarding
		and not (P_PlayerInPain(p))	-- not in pain
			B.SetPriority(p, 0, 0, "mighty_shell", 0, 1, "shell")
			return
		end
	end
	
	-- mighty's battle special
	local function angtoint(a)
		if not a return 0 end
		a = $/ANG1
		if a < 0
			a = $+360
		end
		return a
	end
	
	local function mighty_battleSpecial(mo, doaction)
		local p = mo.player
		
		-- go back to actionstate 0 if hit
		if (P_PlayerInPain(p)) then p.actionstate = 0 end
		
		p.actiontext = "Megaton Punch"	-- name pending
		p.actionrings = mighty.PunchRingCost
		p.action2text = nil	-- this'll eventually be how much the punch is charged!
		
		-- try to start the visual flash
		if (p.actionstate == st_charging)
		or (p.actionstate == st_thrusting)
			local charge = p.actiontime
			if (p.actionstate == st_thrusting)
				-- use the stored value instead
				charge = mo.mighty_battle_charge
			end
			charge = min(mighty.PunchChargeTics, $)	-- okay
			mo.mighty_battle_flashcharge = charge	-- expose this
			
			-- trigger the visual flash
			if (charge)
				local maxcharge = mighty.PunchChargeTics
				local frequency_range = mighty.PunchFlashMaxFrequency - mighty.PunchFlashMinFrequency
				local frequency = mighty.PunchFlashMinFrequency + ((charge * frequency_range) / maxcharge)
				
				mo.mighty_battle_flashtics = $ and min(frequency, $) - 1 or 0
				if not (mo.mighty_battle_flashtics)
					-- set flash and reset
					mo.mighty_battle_flash = mighty.PunchFlashTics
					mo.mighty_battle_flashtics = frequency
				end
			else
				-- reset this
				mo.mighty_battle_flashtics = 0
			end
		end
		
		-- show the visual flash
		if (mo.mighty_battle_flash)
			local charge, maxcharge = mo.mighty_battle_flashcharge, mighty.PunchChargeTics
			local colour_index = max(1, (charge * #mighty.PunchFlashColours) / maxcharge)
			local colour = mighty.PunchFlashColours[colour_index]
			
			local maxtrans = mighty.PunchFlashMaxTrans
			local trans = 10 - ((mo.mighty_battle_flash * maxtrans) / mighty.PunchFlashTics)
				
			if (trans >= 0 and trans <= 9)
				local gmo = mo
				while (gmo and gmo.valid)
					local g = chaotix.SpawnGhostMobj(mo)
					g.frame = ($ & ~FF_TRANSMASK) | (trans << FF_TRANSSHIFT)
					g.color, g.colorized = colour, true
					g.blendmode = AST_ADD
					g.tics = 1
						
					-- attempt to spawn for a followmobj as well?
					gmo = g.tracer
				end
			end
				
			mo.mighty_battle_flash = $-1
		end
		
		if not (p.actionstate)
			if (doaction == 1)	-- just pressed it
				-- go to the charging state!
				p.actionstate = st_charging
				S_StartSound(mo, sfx_mcharg)
				p.actiontime = 0
				p.canguard = false
				B.PayRings(p)
				
				-- reset player stuff
				p.pflags = $ & ~(PF_JUMPED|PF_SPINNING|PF_STARTDASH)
				
				-- set state + frame
				mo.state = S_MIGHTY_PUNCH_CHARGE
				local frame = (not P_IsObjectOnGround(mo)) and 1 or 0
				chaotix.SetFrame(mo, frame)
				
				-- lower normalspeed
				p.normalspeed = FixedMul(mighty.PunchChargeWalkMultiplier, skins[mo.skin].normalspeed)
				p.mighty_battle_fixnormalspeed = true	-- we'll need this
			end
		elseif (p.actionstate == st_charging)			
			-- set state + frame
			if (mo.state != S_MIGHTY_PUNCH_CHARGE) then mo.state = S_MIGHTY_PUNCH_CHARGE end
			local frame = (not P_IsObjectOnGround(mo)) and 1 or 0
			chaotix.SetFrame(mo, frame)
		
			-- increase charge
			p.actiontime = $ + 1
			local chargetext = min(100, (p.actiontime * 100) / mighty.PunchChargeTics)
			p.action2text = "Charge "..chargetext.."%"
			p.action2textflags = (chargetext == 100 and (leveltime&4 <= 2)) and 2 or 0	-- TF_YELLOW = 2 (these arent exposed)
			
			-- handle stasis
			p.normalspeed = FixedMul(mighty.PunchChargeWalkMultiplier, skins[mo.skin].normalspeed)
			p.mighty_battle_fixnormalspeed = true	-- we'll need this
			if (P_IsObjectOnGround(mo))
				p.pflags = $|(PF_STASIS)
				
				-- make sure drawangle is preserved
				local diff = mo.angle - p.drawangle
				local factor = 8
				if (diff)
					diff = (angtoint(diff) > 180) and InvAngle(InvAngle(diff)/factor) or $/factor
					p.drawangle = $ + diff
				end
			end
			
			-- offset the sprites
			local offset_tics = min(mighty.PunchChargeTics, p.actiontime)
			local offset = FixedMul((offset_tics * FRACUNIT) / mighty.PunchChargeTics, mighty.PunchMaxOffset)
			mo.spritexoffset = offset * (((p.actiontime%4) >= 2) and -1 or 1)
			mo.mighty_battle_offsetted = true	-- just for resetting it
			
			-- go to the thrusting state?
			if not (doaction)	-- not holding down anymore
			or (p.actiontime >= mighty.PunchChargeReleaseTics)	-- CANT hold down anymore
				local charge = min(mighty.PunchChargeTics, p.actiontime)
				mo.mighty_battle_charge = charge	-- save this
				
				-- calculate thrust speed
				local mombounds = mighty.PunchThrustMomentumBounds
				local momrange = mombounds[2] - mombounds[1]
				local speed = FixedMul(mo.scale, min(mombounds[2], mombounds[1] + ((charge * momrange) / mighty.PunchChargeTics)))
				
				-- if we're already going faster than this, use our current speed!
				if (FixedHypot(mo.momx, mo.momy) > speed)
					speed = FixedHypot(mo.momx, mo.momy)
				end
				mo.mighty_battle_thrustspeed = speed
				mo.mighty_battle_thrustangle = mo.angle
				
				-- calculate tic bound
				local ticbounds = mighty.PunchThrustTicBounds
				local ticrange = ticbounds[2] - ticbounds[1]
				mo.mighty_battle_thrusttics = min(ticbounds[2], ticbounds[1] + ((charge * ticrange) / mighty.PunchChargeTics))
				
				-- fix normalspeed
				B.ResetPlayerProperties(p)
				p.mighty_battle_fixnormalspeed = nil
				
				-- set new state
				p.actionstate = st_thrusting
				p.actiontime = 0	-- reset this for the thrust
				p.drawangle = mo.angle
				mo.state = S_MIGHTY_PUNCH_THRUST
				P_InstaThrust(mo, mo.angle, speed)
				S_StartSound(mo, sfx_plunch)
				S_StopSoundByID(mo,sfx_mcharg)
			end
		elseif (p.actionstate == st_thrusting)
			-- set state
			if (mo.state != S_MIGHTY_PUNCH_THRUST) then mo.state = S_MIGHTY_PUNCH_THRUST end
			
			-- cancel using jump input to jump!
			if (p.cmd.buttons & BT_JUMP)
			and not (p.lastbuttons & BT_JUMP)
				-- account for analog
				local maxangle = mighty.PunchCancelMaxAngle
				local cameraangle = (p.cmd.angleturn * FRACUNIT)	-- camera-based
				local anglediff = mo.mighty_battle_thrustangle - cameraangle
				
				if (p.pflags & PF_ANALOGMODE)
					local inputangle = R_PointToAngle2(0, 0, p.cmd.forwardmove*FRACUNIT, -p.cmd.sidemove*FRACUNIT)
					anglediff = $ - inputangle
					if abs(anglediff) > ANGLE_90 or anglediff == ANGLE_180
						anglediff = InvAngle($) + ANGLE_180
					end
				end
				
				anglediff = max(-(maxangle), min(maxangle, $ / ANG1)) * ANG1	-- min/max it
				local angle = mo.mighty_battle_thrustangle - anglediff
				
				mo.state = S_PLAY_JUMP
				mo.mighty_hammerdrop_ghosttics = mighty.HammerDropGhostTics	-- reuse this
				P_InstaThrust(mo, angle, FixedHypot(mo.momx, mo.momy))
				P_SetObjectMomZ(mo, mighty.PunchCancelHeight)
				S_StartSound(mo, sfx_jump)
				
				-- reset actionstate
				p.actionstate = 0
				B.ResetPlayerProperties(p, true, true)	-- i WOULDVE liked to keep thokked status, but stasis forces a hammer drop when you hold jump...
				B.ApplyCooldown(p, mighty.PunchCancelCooldown)
				return
			end
			
			-- increase thrust time
			p.actiontime = $ + 1
			
			-- thrust towards angle
			local angle = mo.mighty_battle_thrustangle
			local speed = mo.mighty_battle_thrustspeed
			
			-- ramp up the speed!
			local range = mighty.PunchThrustMaxRamp - FRACUNIT
			local modifier = FRACUNIT + ((p.actiontime * range) / mo.mighty_battle_thrusttics)
			
			p.pflags = $|(PF_FULLSTASIS)
			p.drawangle = angle	-- only set drawangle so we can look around
			p.cangaurd = false
			P_InstaThrust(mo, angle, FixedMul(modifier, speed))
			
			-- cancel Z momentum!
			mo.momz = FixedMul(mighty.PunchThrustZMultiplier, $)
			
			-- visuals
			if not (p.actiontime%2)
				local charge, maxcharge = mo.mighty_battle_charge, mighty.PunchChargeTics
				local colour_index = max(1, (charge * #mighty.PunchFlashColours) / maxcharge)
				local colour = mighty.PunchFlashColours[colour_index]
				
				local g = P_SpawnGhostMobj(mo)
				g.color, g.colorized = colour, true
				//g.blendmode = AST_ADD
				g.tics = 4
			end
			
			-- reset stuff when thrust time is up
			if (p.actiontime >= mo.mighty_battle_thrusttics)
				-- reset action state
				local resetflags = not P_IsObjectOnGround(mo)
				p.actionstate = 0
				B.ResetPlayerProperties(p, resetflags, resetflags)	-- i WOULDVE liked to keep thokked status, but stasis forces a hammer drop when you hold jump...
				B.ApplyCooldown(p, mighty.PunchCooldown)
				if (P_IsObjectOnGround(mo)) then mo.state = S_PLAY_RUN end
			end
		elseif (p.actionstate == st_punch)
			-- set state + frame
			local tics = 0
			local frame = 0
			for i = 1, #mighty.PunchHitSmearTics
				local ticsadd = mighty.PunchHitSmearTics[i]
				tics = $ + ticsadd
				if (p.actiontime < tics) then break end
				
				frame = $+1
			end
			
			if (mo.state != S_MIGHTY_PUNCH_HIT) then mo.state = S_MIGHTY_PUNCH_HIT end
			chaotix.SetFrame(mo, frame)
			
			-- raise timer
			p.actiontime = $ + 1
			
			-- stall momentum for extra effect
			if (p.actiontime <= mighty.PunchHitStallTics)
				p.pflags = $|(PF_STASIS)	-- no moving!
				mo.momx, mo.momy, mo.momz = 0, 0, 0
			end
			p.drawangle = mo.mighty_battle_thrustangle	-- set this too
			p.pflags = $|(PF_JUMPSTASIS)	-- no jumping!
			
			-- end the punch animation after a set time
			if (p.actiontime >= mighty.PunchHitTics)
				-- reset action state
				p.actionstate = 0
				B.ResetPlayerProperties(p)
				B.ApplyCooldown(p, mighty.PunchCooldown)
				if (P_IsObjectOnGround(mo)) then mo.state = S_PLAY_RUN end
			end
		end
		
		-- fix normalspeed again, just incase
		if not (p.actionstate)
		and (p.mighty_battle_fixnormalspeed)
			B.ResetPlayerProperties(p)
			p.mighty_battle_fixnormalspeed = nil
		end
		
		-- fix sprite offset
		if (p.actionstate != st_charging)
		and (mo.mighty_battle_offsetted)
			mo.spritexoffset = 0
			mo.mighty_battle_offsetted = nil
		end
	end
	
	-- mighty's collision functions (for the hammer drop!)
	local function mighty_battlePreCollide(s, t, p, mo, atk, def, weight, hurt, pain, ground, angle, thrust, thrust2, collisiontype)
		if not (p[s] and mo[s]) then return end
		
		-- save vars under certain condition (before the collision stuff happens!)
		if (p[s].actionstate == st_thrusting)
			mo[s].mighty_battle_thrustsave = FixedHypot(mo[s].momx, mo[s].momy)
		end
		
		if (mo[s].mighty_hammerdrop and mo[s].mighty_hammerdrop_timer) 
		and (mo[s].mighty_hammerdrop_timer >= mighty.HammerDropStartupTics)
			mo[s].mighty_hammerdrop_battlemomz = mo[s].momz	-- we'll be transferring this all to our victim!
		end
	end
	
	local function mighty_battlePostCollide(s, t, p, mo, atk, def, weight, hurt, pain, ground, angle, thrust, thrust2, collisiontype)
		if not (p[s] and mo[s]) then return end
		
		-- remove these saved vars
		if (mo[s].mighty_battle_thrustsave != nil)
			mo[s].mighty_battle_thrustsave = nil
		end
		if (mo[s].mighty_hammerdrop_battlemomz != nil)
			mo[s].mighty_hammerdrop_battlemomz = nil
		end
		
		-- remove recoil if we got tumbled
		if (mo[s].mighty_battle_removerecoil)
			mo[s].recoilangle = nil
			mo[s].recoilthrust = nil
		end
	end
	
	local function mighty_battleCollide(s, t, p, mo, atk, def, weight, hurt, pain, ground, angle, thrust, thrust2, collisiontype)
		if not (p[s] and mo[s]) then return end
		if not (p[t] and mo[t]) then return end
		
		-- laziness check for hurt value (so we don't have to use 2 different checks...)
		local newhurt = hurt
		if (abs(newhurt) == 1)
		and (s == 2)	-- s is actually t
			newhurt = $ * -1
		end
		
		-- megaton punch
		if (p[s].actionstate == st_thrusting)
			-- get parried!
			if (p[t].guard == 1)	-- they're guarding
				S_StartSound(mo[t], sfx_cdpcm9)
				S_StartSound(mo[t], sfx_s259)
				p[t].guard = 2
				p[t].guardtics = 9
				B.ControlThrust(mo[t], FRACUNIT/2)
				
				-- graphical effects
				local sh = P_SpawnMobjFromMobj(mo[t], 0, 0, 0, MT_BATTLESHIELD)
				sh.target = mo[t]
				for n = 0, 16
					local dust = P_SpawnMobj(mo[t].x, mo[t].y, mo[t].z, MT_DUST)
					if (dust and dust.valid)
						P_InstaThrust(dust, mo[t].angle + ANGLE_22h * n, mo[t].scale*36)
					end
				end
				P_SpawnMobjFromMobj(mo[s], 0, 0, 0, MT_EXPLODE)
				
				-- affect source
				if (mo[s].health and p[s].powers[pw_flashing])
					p[s].powers[pw_flashing] = 0
					local nega = P_SpawnMobjFromMobj(mo[s], 0, 0, 0, MT_NEGASHIELD)
					nega.target = mo[s]
				end
				
				-- affect attacker
				if (p[s].powers[pw_invulnerability])
					p[s].powers[pw_invulnerability] = 0
					P_RestoreMusic(p[s])
				end
				
				local angle = R_PointToAngle2(mo[t].x - mo[t].momx, mo[t].y - mo[t].momy, mo[s].x - mo[s].momx, mo[s].y - mo[s].momy)
				local thrust = FRACUNIT*10
				if twodlevel then thrust = B.TwoDFactor($) end
				P_SetObjectMomZ(mo[s], thrust)
				B.DoPlayerTumble(p[s], 45, angle, mo[s].scale * 3, true)
				return true
			end
			
			-- ...but let's see if we can give em a punch!
			-- megaton punch is special, uses its defence as the "attacking" power!
			if (def[s] > atk[t])
			and (def[s] > def[t])
				-- set punching state
				p[s].actionstate = st_punch
				p[s].actiontime = 0
				mo[s].state = S_MIGHTY_PUNCH_HIT
				mo[s].momx, mo[s].momy, mo[s].momz = 0, 0, 0
				
				-- visual effect!
				local shake = false
				if (p[s] == consoleplayer)
				or (p[t] == consoleplayer)
					shake = true
				end
	
				local scale = 4
				S_StartSoundAtVolume(mo[s], sfx_s3k49, 200)
				S_StartSound(mo[t], sfx_s3k61)
				if (shake) then P_StartQuake(FRACUNIT*12, 3) end
				
				local vfx = P_SpawnMobjFromMobj(mo[t], 0, 0, mo[t].height/2, MT_SPINDUST)
				if (vfx and vfx.valid)
					vfx.scale = mo[t].scale * scale/5
					vfx.destscale = vfx.scale * 3
					vfx.colorized = true
					vfx.color = SKINCOLOR_WHITE
					vfx.state = S_BCEBOOM
				end
				
				-- we'll do this in the post collide, so momentum won't be overidden
				local angle = mo[s].mighty_battle_thrustangle
				if (angle == nil) then angle = R_PointToAngle2(mo[s].x, mo[s].y, mo[t].x, mo[t].y) end	-- fallback
				local thrust = FixedMul(mighty.PunchHitKnockbackModifier, mo[s].mighty_battle_thrustsave)
				
				P_SetObjectMomZ(mo[t], chaotix.WaterMom(mo[t], mighty.PunchHitKnockbackHeight))
				B.DoPlayerTumble(p[t], mighty.PunchHitTumbleTics, angle, thrust, true)
				mo[t].mighty_battle_tumbled = true	-- we'll need this for wall detection
				mo[t].mighty_battle_tumbler = mo[s]	-- this too
				return true
			else
				-- reset player state properly
				p[s].actionstate = 0
				local state = (P_IsObjectOnGround(mo[s])) and S_PLAY_SKID or S_PLAY_FALL
				mo[s].state = state
				return
			end
		end
		
		-- hammer drop
		if (mo[s].mighty_hammerdrop_battlemomz != nil)	-- we got momentum to give em
		and ((newhurt >= 1) or pain[t])	-- we gotta hit/trade, or they need to be in pain state (combo them!)
		and (atk[s] > 1)	-- we def got the spike hitbox!
			-- now let's try to spike them!
			if not (ground[t])	-- they need to be airborne
				-- send em packin'
				local angle = R_PointToAngle2(mo[s].x, mo[s].y, mo[t].x, mo[t].y)
				P_SetObjectMomZ(mo[t], mo[s].mighty_hammerdrop_battlemomz)
				mo[t].mighty_tumble = true	-- we'll do some more with this
				S_StartSound(mo[t], sfx_hamspk)
				
				-- cap our momentum
				local modifier = mighty.HammerDropSpikeModifier
				mo[t].momx = FixedMul(modifier, $)
				mo[t].momy = FixedMul(modifier, $)
				
				-- tumble start stuff (rest will be handled below
				p[t].panim = PA_PAIN
				p[t].pflags = $&~(PF_GLIDING|PF_JUMPED|PF_BOUNCING|PF_SPINNING|PF_THOKKED|PF_SHIELDABILITY)
				
				p[t].tumble = 2
				p[t].tumble_nostunbreak = true
				p[t].airdodge_spin = 0
				if not (p[t].actionsuper) then p[t].actionstate = 0 end
				
				S_StartSound(mo[t], sfx_s3k98)
				//S_StartSoundAtVolume(mo[t], sfx_kc38, 70)	-- this gets prevented, replay below
				return
			end
		end
		
		-- shell deflections should tumble mighty!
		if ((def[s] == 1) and (p[s].battle_sfunc == "mighty_shell"))	-- we got the shell armour
		and (atk[t] > 0)	-- they need an attacking power!
		and (newhurt == 0 or newhurt == 1)	-- basic clash (or they somehow got hurt by the shell????????
			-- first check for conditions where we wouldnt of taken damage from the collision
			if not (p[s].powers[pw_super] or p[s].powers[pw_invulnerability])	-- not invincible
			and not (p[s].powers[pw_flashing])	-- not in flashtics
			and not (p[s].tumble and p[s].tumble > 0)	-- we've already been tumbled!
				-- tumble the player
				B.DoPlayerTumble(p[s], mighty.ShellTumbleTics, angle[t], FixedMul(mo[t].scale, mighty.ShellTumbleKnockback), true)
				P_SetObjectMomZ(mo[s], chaotix.WaterMom(mo[s], mighty.ShellTumbleZKnockback))
				mo[s].mighty_battle_removerecoil = true
				
				-- play shell sfx
				S_StartSound(mo[s], sfx_cdfm46)
				return
			end
		end
	end
	
	-- setup mighty's skinvars
	S[mighty.SKIN] = {
		weight = 120,
		guard_frame = 0,
		special = mighty_battleSpecial,
		func_priority_ext = mighty_battlePriority,
		func_precollide = mighty_battlePreCollide,
		func_collide = mighty_battleCollide,
		func_postcollide = mighty_battlePostCollide,
	}
	
	loaded = true
end

-- mighty's tumble interactions
mighty.HammerDropBattleTumble = function()
	if not (CBW_Battle) then return end	-- battle isnt loaded
	local B = CBW_Battle
	
	for p in players.iterate
		local mo = p.mo
		if not ((mo and mo.valid) and mo.health) then continue end
		
		-- reset these if we're not tumbled
		if (mo.mighty_battle_tumbled)
			if not (p.tumble)
				mo.mighty_battle_tumbled = nil
				mo.mighty_battle_tumbler = nil
			else
				-- a bit of a hack to update credit history!
				local tumbler = mo.mighty_battle_tumbler
				if (tumbler and tumbler.valid) then B.PlayerCreditPusher(p, tumbler) end
			end
		end
		
		-- rest of the iteration is mighty's custom hammer drop tumble
		if not (mo.mighty_tumble) then continue end
		
		-- pretty nasty hack, but tumble isnt usually consistently this low
		if (p.tumble and p.tumble > 6)
			-- assume a new tumble has been established
			mo.mighty_tumble = nil
			mo.mighty_tumble_sound = nil
			mo.mighty_tumble_bounces = nil
			mo.mighty_tumble_flashingfreeze = nil
			continue
		end
		
		-- register the bounces on the ground
		if (P_IsObjectOnGround(mo))
			mo.mighty_tumble_bounces = $ and $+1 or 1
		end
		
		-- end the bouncing (doing the check like this to conquer load order)
		local bounces = mo.mighty_tumble_bounces or 0
		
		if ((mo.momz * P_MobjFlip(mo)) > 0)
		and (bounces >= mighty.HammerDropSpikeBounces)
			mo.mighty_tumble = nil
			mo.mighty_tumble_sound = nil
			mo.mighty_tumble_bounces = nil
			
			-- we'll have to reset it ourself
			if (mo.mighty_tumble_flashingfreeze)
				p.powers[pw_flashing] = (3*TICRATE) - 1
				mo.mighty_tumble_flashingfreeze = nil
			end
				
			-- might as well end the tumble ourself
			p.tumble = nil
			p.lockmove = false
			p.drawangle = mo.angle
			S_StopSoundByID(mo, sfx_kc38)
			B.ResetPlayerProperties(p, false, false)
			
			-- go into falling animation
			p.panim = PA_FALL
			mo.state = S_PLAY_FALL
				
			-- don't do anything else
			continue
		end
		
		-- start the sound up again
		if not (mo.mighty_tumble_sound)
			mo.mighty_tumble_sound = true
			S_StartSoundAtVolume(mo, sfx_kc38, 70)
		end
		
		-- perma-tumble (while keeping the effects working)
		if (p.tumble == nil)
		or (p.tumble <= 2)
			p.tumble = 6	-- reset to 6 so the tumble afterimage is still spawned
			p.tumble_nostunbreak = true
		end
		
		-- nasty hack so invuln period is correctly applied
		if (P_PlayerInPain(p) and (p.powers[pw_flashing] == 3*TICRATE))
			mo.mighty_tumble_flashingfreeze = true
		end
		if (mo.mighty_tumble_flashingfreeze) then p.powers[pw_flashing] = (3*TICRATE) - 1 end	-- gotta keep it one below so tumble doesnt get ended early
	end
end

-- for mighty's megaton punch damage on collision!
chaotix.AddHook("MobjMoveBlocked", function(mo, object, line)
	if not (mo.mighty_battle_tumbled) then return end
	local p = mo.player
	local tumbler = mo.mighty_battle_tumbler
	local B = CBW_Battle
	
	-- visual effect!
	local scale = 6
	S_StartSoundAtVolume(mo, sfx_s3k49, 200)
	S_StartSound(mo, sfx_s3k9b)
	if (p == consoleplayer) then P_StartQuake(FRACUNIT*14, 5) end
				
	local vfx = P_SpawnMobjFromMobj(mo, 0, 0, mo.height/2, MT_SPINDUST)
	if (vfx and vfx.valid)
		vfx.scale = mo.scale * scale/5
		vfx.destscale = vfx.scale * 3
		vfx.colorized = true
		vfx.color = SKINCOLOR_WHITE
		vfx.state = S_BCEBOOM
	end
				
	-- "bounce" off the wall
	P_BounceMove(mo)
	local bouncemomx, bouncemomy = mo.momx, mo.momy
	
	-- set hurttxt for the P_DamageMobj
	local oldhurttxt
	if (tumbler and tumbler.valid)
		oldhurttxt = tumbler.player.battle_hurttxt
		tumbler.player.battle_hurttxt = "megaton punch impact"	-- maybe this'll work for the hurt message!
	end
	
	-- hurt em good
	P_DamageMobj(mo, tumbler, tumbler)
	mo.mighty_battle_tumbled = nil
	mo.mighty_battle_tumbler = nil
	
	-- restore old hurttxt (who knows, might be needed?)
	if (tumbler and tumbler.valid)
		tumbler.player.battle_hurttxt = oldhurttxt
	end
	
	-- NOW we bounce off the wall
	if (mo.health)
		mo.momx, mo.momy = bouncemomx, bouncemomy
	end
	
	-- end tumble
	p.tumble = nil
	p.lockmove = false
	S_StopSoundByID(mo, sfx_kc38)
	if (mo.health) then B.ResetPlayerProperties(p, false, false) end
end, MT_PLAYER)