-- everything related to mighty's hammerdrop
local mighty = chaotix.mighty

-- reset stuff related to the hammer drop
mighty.ResetHammerDrop = function(mo)
	mo.mighty_hammerdrop = nil
	mo.mighty_hammerdrop_timer = nil
	mo.mighty_hammerdrop_lastmom = nil
	
	-- stretch
	if (mo.mighty_hammerdrop_stretch)
		mo.spritexscale = FRACUNIT
		mo.spriteyscale = FRACUNIT
		mo.mighty_hammerdrop_stretch = nil
	end
	
	-- chaos mode support (?)
	mo.chaos_enemypierce = nil
end

-- start the hammerdrop here
mighty.AbilitySpecial = function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	if not (mighty.AbilityCheck(p)) then return end
	
	if (p.pflags & PF_THOKKED) then return end	-- thokked state
	if (p.gotcrystal) then return end	-- battlemod support
	if (p.actionstate) then return end	-- battlemod support (2)
	
	p.pflags = ($|PF_THOKKED|PF_JUMPDOWN) & ~(PF_STARTJUMP|PF_SPINNING)
	mo.state = S_MIGHTY_HAMMERDROP_STARTUP	-- set state
	mo.mighty_hammerdrop = true
end

-- so we can damage enemies CLOSE to the ground
local function isObjectCloseToGround(mo)
	local leeway = FixedMul(mo.scale, mighty.HammerDropFloorDistanceLeeway)
	
	if (mo.eflags & MFE_VERTICALFLIP)
		if ((mo.z + mo.height) >= (mo.ceilingz - leeway))
			return true
		end
	else
		if (mo.z <= (mo.floorz + leeway))
			return true
		end
	end
	
	return false
end

mighty.doHammerDropSearch = function(ref, found, totaldist, heavy)
	if not ((found and found.valid) or found.health) then return end
	if (ref == found) then return end
	if (P_MobjFlip(ref) != P_MobjFlip(found)) then return end
				
	-- quickly just do spike stuff here
	if (found.type == MT_SPIKE or found.type == MT_WALLSPIKE)
		local dist = P_AproxDistance(P_AproxDistance(found.x - ref.x, found.y - ref.y), found.z - ref.z)
		local spikedist = FixedMul(mighty.HammerDropSpikeDistMultiplier, totaldist)
					
		-- break them if they're nearby
		if ((dist <= spikedist) and P_IsObjectOnGround(found))
			P_KillMobj(found, ref, ref)
		end
					
		return
	end
	
	if (found.type == MT_METALSONIC_BATTLE)
	and (found.state == S_METALSONIC_FLOAT)
		found.momx, found.momy = 0, 0
		found.mighty_stunned = true	-- we'll stop him in a thinker
		found.mighty_stuntimer = mighty.HammerDropMetalStunTime	-- how long he'll be stunned for
		found.mighty_stunangle = found.angle	-- the angle he was stunned at
		return
	end

	if not ((found.flags & (MF_SHOOTABLE|MF_ENEMY|MF_MONITOR)) or found.player) then return end
	if not (P_CheckSight(ref, found)) then return end
				
	-- launch
	local grounded = P_IsObjectOnGround(found)	-- need to save this
	
	if ((found and found.valid) and found.health) -- gotta be alive
	and not (found.player or (found.flags & MF_BOSS))	-- don't launch players or bosses
	and (grounded)	-- might as well use this since we've stored it
		P_SetObjectMomZ(found, chaotix.WaterMom(found, 6*FRACUNIT))
	end

	-- damage nearby objects!
	local dist = FixedHypot(FixedHypot(found.x - ref.x, found.y - ref.y), found.z - ref.z)
	local damagedist = heavy and totaldist or FixedMul(mighty.HammerDropDamageDistMultiplier, totaldist)
				
	local groundcheck = (found.player) and (grounded) or (isObjectCloseToGround(found))
	if ((dist <= damagedist) and groundcheck)
		if (found.player)
			local B = CBW_Battle
			if not (B) then return end	-- no battle
			
			-- can they even be tumbled?
			if (B.MyTeam(ref.player, found.player)) then return end -- same team
			if (found.player.tumble) then return end	-- don't tumble-lock them!
			
			-- tumble them for a combo!
			local oldmomx, oldmomy = found.momx, found.momy
			
			local angle = R_PointToAngle2(ref.x, ref.y, found.x, found.y)
			P_SetObjectMomZ(found, chaotix.WaterMom(found, mighty.HammerDropTumbleHeight))
			B.DoPlayerTumble(found.player, heavy and chaotix.Heavy.tumbleTics
			or mighty.HammerDropTumbleTics, angle, 0, true)
			
			-- restore old momentum, but heavily gimped
			local multiplier = mighty.HammerDropMomXYMultiplier
			found.momx = FixedMul(multiplier, oldmomx)
			found.momy = FixedMul(multiplier, oldmomy)
		else
			P_DamageMobj(found, ref, ref, 1, 0)
		end
	end
end

mighty.doHammerDrop = function(p)
	local mo = p.mo
	
	local lastmomz = abs(mo.mighty_hammerdrop_lastmom[3] or mo.momz)
	local modifier = (P_PlayerFullbright(p)) and mighty.HammerDropTotalDistSuperModifier or FRACUNIT
	local totaldist = FixedMul(modifier, FixedMul(mighty.HammerDropTotalDistModifier, lastmomz))
	
	-- search for things to damage
	searchBlockmap("objects", function(ref, found)
		return mighty.doHammerDropSearch(ref, found, totaldist)
	end, mo, mo.x - totaldist, mo.x + totaldist, mo.y - totaldist, mo.y + totaldist)
				
	-- spawn some dust!
	local dusts = 15
	for i = 1, dusts
		local angle = (360/dusts*i)*ANG1

		local dust = P_SpawnMobj(mo.x + 32*cos(angle), mo.y + 32*sin(angle), mo.z, MT_THOK)
		dust.state = (mo.eflags & MFE_UNDERWATER) and S_WPLD1 or S_XPLD1	-- water's nice
		dust.scale = (mo.eflags & MFE_UNDERWATER) and mo.scale*4/5 or mo.scale
		//dust.momx, dust.momy = mo.momx, mo.momy
		P_Thrust(dust, angle, FixedMul(mighty.HammerDropDustDistMultiplier, lastmomz))
	end
	
	-- local quake
	if (p == displayplayer)
	or (p == secondarydisplayplayer)
		P_StartQuake(mighty.HammerDropQuakeIntensity, mighty.HammerDropQuakeTics)
	end
	
	-- sounds
	S_StartSound(mo, sfx_hamlnd)
	if (mo.eflags & MFE_UNDERWATER) then S_StartSound(mo, sfx_splash) end
end

-- hammer drop function
mighty.HammerDrop = function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	
	-- checks for the hammer drop conditions
	if not (mighty.AbilityCheck(p))
	or not (mo.health)
		mo.mighty_hammerdrop_ghosttics = nil
		mighty.ResetHammerDrop(mo) 
		return
	end
	
	-- spawn ghosts
	if (mo.mighty_hammerdrop_ghosttics)
		if not (mo.mighty_hammerdrop_ghosttics%2)
			local g = P_SpawnGhostMobj(mo)
			if (g.tracer and g.tracer.valid)
				local fmo = p.followmobj
			
				if (fmo and fmo.valid)
				and (fmo.type == MT_MIGHTY_HAMMERDROP)
					P_RemoveMobj(g.tracer)
				end
			end
		end
		mo.mighty_hammerdrop_ghosttics = $-1
	end
	
	-- rotate after hitting a spring with the hammer drop
	if (mo.state == S_MIGHTY_HAMMERDROP_SPRING)
		-- make sure this exists
		if (mo.mighty_hammerdrop_twirlangle == nil)
			-- start at the player's current drawangle
			mo.mighty_hammerdrop_twirlangle = p.drawangle
		end
		
		-- rotate the player based on momz
		local angle_add = FixedAngle(FixedMul(mighty.HammerDropSpringTwirlMultiplier, (mo.momz/mo.scale)*FRACUNIT))
		mo.mighty_hammerdrop_twirlangle = $ + angle_add
		p.drawangle = mo.mighty_hammerdrop_twirlangle
		
		-- when our momz declines completely, go to falling
		-- (just like S_PLAY_SPRING)
		if ((mo.momz * P_MobjFlip(mo)) <= 0)
			mo.state = S_PLAY_FALL
			mo.mighty_hammerdrop_twirlangle = nil
		end
	else
		mo.mighty_hammerdrop_twirlangle = nil
	end
	
	-- dont continue any further if we arent IN a hammer drop
	if not (mo.mighty_hammerdrop) then return end
	
	-- stop conditions for the hammerdrop
	if not (mo.state >= S_MIGHTY_HAMMERDROP_STARTUP and mo.state <= S_MIGHTY_HAMMERDROP)	-- we got knocked out of it
	or (p.gotflag or p.gotcrystal)	-- we got the flag (or the crystal in battlemod)
		if (mo.state >= S_MIGHTY_HAMMERDROP_STARTUP and mo.state <= S_MIGHTY_HAMMERDROP)
			mo.state = S_PLAY_ROLL
			mo.mighty_hammerdrop_ghosttics = mighty.HammerDropGhostTics
		end
		
		-- ground hit, damage stuff around us!
		if (mo.eflags & MFE_JUSTHITFLOOR)
		and not (mo.mighty_hammerdrop_preventstomp)
		and (mo.mighty_hammerdrop_timer and mo.mighty_hammerdrop_timer >= mighty.HammerDropStartupTics)
		and not (P_PlayerInPain(p))	-- not hurt
		and not (p.powers[pw_carry])	-- not being carried
			mo.state = S_PLAY_ROLL
			
			if (p.cmd.buttons & BT_SPIN)	-- this calls for a spindash!
				-- get last momz
				local lastmomz = abs(mo.mighty_hammerdrop_lastmom[3] or mo.momz)
				//print("lastmomz: "..tostring(FixedDiv(lastmomz, mo.scale) / FRACUNIT))
				
				-- get the speed + function
				local speed = min(FixedMul(mighty.HammerDropDashMultiplier, lastmomz), FixedMul(mo.scale, p.maxdash))
				//print("speed: "..tostring(FixedDiv(speed, mo.scale) / FRACUNIT))
				local func = P_InstaThrust
				if (mo.standingslope)	-- extremely ghetto but this should work
					local lastmomx = (mo.mighty_hammerdrop_lastmom[1] or mo.momx)
					local lastmomy = (mo.mighty_hammerdrop_lastmom[2] or mo.momy)
					speed = $ - FixedHypot(lastmomx, lastmomy)
					func = P_Thrust	
				end
				
				-- launch off!
				p.pflags = $|PF_SPINNING
				func(mo, mo.angle, speed)
				S_StartSound(mo, sfx_zoom)
				
				-- spawn dust
				for i = 1, 16
					-- spawn the dust (alongside override states)
					local dust = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_SPINDUST)
					if (mo.eflags & (MFE_TOUCHWATER|MFE_UNDERWATER))
						dust.state = S_SPINDUST_BUBBLE1
					elseif (p.powers[pw_shield] == SH_ELEMENTAL)
						dust.state = S_SPINDUST_FIRE1
					end
					
					-- set vars
					dust.target = mo
					dust.destscale = (2*mo.scale)/3
					dust.scale = dust.destscale
					if (mo.eflags & MFE_VERTICALFLIP)
						-- honestly this might be irrelevant but i'll copy the source regardless
						P_SetOrigin(dust, dust.x, dust.y, (mo.z + mo.height) - dust.height)
					end
					
					-- set momentum
					local frac_speed = FixedMul(FRACUNIT*10, FixedDiv(mo.scale, speed))
					local prandom = {P_RandomFixed()<<2, P_RandomRange(-30, 30), P_RandomFixed()<<3}
					P_SetObjectMomZ(dust, frac_speed/50 + prandom[1])
					P_InstaThrust(dust, mo.angle + (prandom[2]*ANG1), -FixedMul(frac_speed/12 + FRACUNIT + prandom[3], mo.scale))
					P_TryMove(dust, dust.x + dust.momx, dust.y + dust.momy, true)
				end
			else
				-- different recoil state depending on battlemod
				local B = CBW_Battle
				if (B and B.BattleGametype())
					p.pflags = ($|PF_THOKKED) & ~(PF_JUMPED)
					mo.state = S_PLAY_FALL	-- so we have a bigger window to punish him!
				else
					p.pflags = $|(PF_JUMPED|PF_THOKKED)
				end
				
				-- do the actual recoil
				p.pflags = $|PF_JUMPDOWN
				mo.mighty_hammerdrop_ghosttics = mighty.HammerDropGhostTics
				P_SetObjectMomZ(mo, chaotix.WaterMom(mo, mighty.HammerDropRecoilMomentum))
				mighty.doHammerDrop(p)
			end
		end
		
		S_StopSoundByID(mo, sfx_hamdrp)
		mighty.ResetHammerDrop(mo)
		return
	end
	
	-- hammerdrop timer for the ability
	mo.mighty_hammerdrop_timer = $ and $+1 or 1
	local timer = mo.mighty_hammerdrop_timer
	
	-- reset this here (from mighty_protection.lua)
	mo.mighty_hammerdrop_preventstomp = nil	
	
	p.pflags = $|PF_JUMPDOWN
	if (timer == mighty.HammerDropStartupTics)	-- start the hammer drop!
		-- set momz
		local modifier = (P_PlayerFullbright(p)) and mighty.HammerDropSuperStartModifier or FRACUNIT
		P_SetObjectMomZ(mo, FixedMul(modifier, chaotix.WaterMom(mo, mighty.HammerDropStartMomentum)))
		mo.mighty_hammerdrop_lastmom = {mo.momx, mo.momy, mo.momz}	-- save this
		
		-- kill momentum a tiny bit
		mo.momx = FixedMul(mighty.HammerDropStartupModifier, $)
		mo.momy = FixedMul(mighty.HammerDropStartupModifier, $)
		
		-- start sounds
		S_StartSound(mo, sfx_zoom)
		S_StartSound(mo, sfx_hamdrp)
		
		-- followmobj shenanigans
		local fmo = p.followmobj
		
		if (fmo and fmo.valid)
		and (fmo.type == MT_MIGHTY_HAMMERDROP)
			fmo.reactiontime = 10
			fmo.reactiondelay = 2
			mighty.UpdateFollowMobj(p, fmo, fmo.reactiontime)
		end
	elseif (timer > mighty.HammerDropStartupTics)
		-- stop in goop
		if (mo.eflags & MFE_GOOWATER)	
		or ((mo.momz*P_MobjFlip(mo)) > 0)
			mo.state = S_PLAY_ROLL
			mo.momz = $ - ($/3) -- reduce by a third
			mo.mighty_hammerdrop_ghosttics = mighty.HammerDropGhostTics
			mighty.ResetHammerDrop(mo)
			return
		end
		
		-- set stuff for this state
		timer = $ - mighty.HammerDropStartupTics	-- hehehehe
		if (timer == 1)	then mo.state = S_MIGHTY_HAMMERDROP end	-- JUST started, set state
		
		-- buildup momentum
		local modifier = (P_PlayerFullbright(p)) and mighty.HammerDropSuperBuildupModifier or FRACUNIT
		P_SetObjectMomZ(mo, FixedMul(modifier, chaotix.WaterMom(mo, mighty.HammerDropBuildupMomentum)), true)
		
		-- ghosts!
		local g = P_SpawnGhostMobj(mo)
		if (g.tracer and g.tracer.valid)
			local fmo = p.followmobj
		
			if (fmo and fmo.valid)
			and (fmo.type == MT_MIGHTY_HAMMERDROP)
				P_RemoveMobj(g.tracer)
			end
		end
		
		-- stretch
		local stretch_speed = (mo.momz * P_MobjFlip(mo)) - mighty.HammerDropStartMomentum
		local stretch_speed_modifier = min(0, stretch_speed)
		mo.spritexscale = FRACUNIT - FixedMul(stretch_speed_modifier, mighty.HammerDropXStretch)
		mo.spriteyscale = FRACUNIT - FixedMul(stretch_speed_modifier, mighty.HammerDropYStretch)
		mo.mighty_hammerdrop_stretch = true
		
		-- save this
		mo.mighty_hammerdrop_lastmom = {mo.momx, mo.momy, mo.momz}
		
		-- break floors! (we can return here since it's the last thing in the block!)
		local rover = mo.floorrover
		
		-- rover checks
		if not (rover and rover.valid) then return end
		if not (rover.flags & FF_EXISTS) then return end
		if not ((rover.flags & FF_STRONGBUST) or (rover.flags & FF_BUSTUP)) then return end
		
		-- height checks
		local topheight = rover.topheight - mo.momz
		local bottomheight = rover.bottomheight - mo.momz
			
		if (mo.z >= topheight) then return end
		if (mo.z + mo.height < bottomheight) then return end
		
		-- break it if we're here!
		EV_CrumbleChain(rover)
		if (rover.master.flags & ML_EFFECT5)
			P_LinedefExecute(P_AproxDistance(rover.master.dx, rover.master.dy)>>FRACBITS, mo, mo.subsector.sector)
		end
	end
end

-- additional stuff for metal sonic
local function mighty_resetMetal(metal)
	-- reset vars
	metal.flags2 = $ & ~(MF2_DONTDRAW)
	metal.mighty_stunned = nil
	metal.mighty_stuntimer = nil
	metal.mighty_stunangle = nil
	S_StopSoundByID(metal, sfx_s250)
end
	
chaotix.AddHook("MobjThinker", function(metal)
	if not ((metal and metal.valid) and metal.health) then return end
	if not (metal.mighty_stunned) then return end	-- shouldnt be stunned
	
	-- shouldnt be stunned anymore
	if (metal.state != S_METALSONIC_FLOAT)
		mighty_resetMetal(metal)
		return
	end
	
	-- halt metal for an easier hit!
	metal.momx, metal.momy = 0, 0
	metal.angle = metal.mighty_stunangle
	metal.frame = O
	
	-- play the malfunction sound
	if not (S_SoundPlaying(metal, sfx_s250))
		-- malfunction
		S_StartSound(metal, sfx_s250)
	end
	
	-- shake around our frozen position!
	local shift = mighty.HammerDropMetalStunShift
	local xoff = P_RandomRange(-(shift), shift) * FRACUNIT
	local yoff = P_RandomRange(-(shift), shift) * FRACUNIT
	
	local faker = P_SpawnMobjFromMobj(metal, xoff, yoff, 0, MT_THOK)
	faker.sprite, faker.frame = metal.sprite, metal.frame
	faker.angle = metal.angle
	faker.color = metal.color
	faker.tics = 2
	metal.flags2 = $|MF2_DONTDRAW	-- making my life easier
	
	-- visuals!!
	local chance = mighty.HammerDropMetalStunSparkChance
	if not (metal.mighty_stuntimer % 2)
		local rad = (metal.radius)>>FRACBITS
		local xoff = P_RandomRange(-(rad), rad)<<FRACBITS
		local yoff = P_RandomRange(-(rad), rad)<<FRACBITS
		local zoff = P_RandomRange(0, (metal.height)>>FRACBITS)<<FRACBITS
						
		local spark = P_SpawnMobjFromMobj(faker, xoff, yoff, zoff, MT_WATERZAP)
	end
	
	local g = P_SpawnGhostMobj(faker)
	g.tics = 3
					
	-- reduce timer
	metal.mighty_stuntimer = $ and $-1
	if not (metal.mighty_stuntimer)	-- dont be stunned anymore
		metal.frame = E	-- reset frame
		mighty_resetMetal(metal)
		return
	end
end, MT_METALSONIC_BATTLE)

-- make sure we can damage
chaotix.AddHook("PlayerCanDamage", function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	if not (mo.mighty_hammerdrop) then return end
	
	return true
end)

-- make the hammerdrop go through enemies
chaotix.AddHook("TouchSpecial", function(special, mo)
	local p = mo.player
	if not ((mo and mo.valid) and mo.health) then return end
	if not (P_PlayerCanDamage(p, special)) then return end
	
	-- hammerdrop checks
	if not (mo.mighty_hammerdrop) then return end
	if not (mo.mighty_hammerdrop_timer) then return end
	if not (mo.mighty_hammerdrop_timer >= mighty.HammerDropStartupTics) then return end
	
	-- exclude players from this!
	if (special.player) then return end
	
	-- handle these a little differently
	if (special.type == MT_BIGMINE) 
		mo.state = S_PLAY_ROLL
		mo.mighty_hammerdrop_ghosttics = mighty.HammerDropGhostTics
		mighty.ResetHammerDrop(mo)
		return 
	end
	
	-- keep the pterabyte check from source
	if (special.type == MT_PTERABYTE and special.target == mo and special.extravalue1 == 1)
		return
	end
	
	-- make sure we still damage them!
	P_DamageMobj(special, mo, mo, 1, 0)
	return true
end)

-- hammer drop collisions
mighty.HammerDropCollide = function(mo, mo2)
	-- object checks
	if not ((mo and mo.valid) and mo.health) then return end
	if not ((mo2 and mo2.valid) and mo2.health) then return end
	
	-- hammerdrop check
	if not (mo.mighty_hammerdrop) then return end
	
	-- springshell collision (it's up here because we use a special height check)
	if (mo2.type == MT_SPRINGSHELL)
		local tmz = (mo2.eflags & MFE_VERTICALFLIP and -((mo.z + mo.height) or 1) or mo.z)
		local tmznext = (mo2.eflags & MFE_VERTICALFLIP and -mo.momz or mo.momz) + tmz
		local thzh = (mo2.eflags & MFE_VERTICALFLIP and -(mo2.z or 1) or mo2.z + mo2.height)
		local sprarea = FixedMul(8*FRACUNIT, mo2.scale) * P_MobjFlip(mo2)

		if (((tmznext <= thzh) and (tmz > thzh)) or ((tmznext > thzh - sprarea) and (tmznext < thzh))) then
			if (mo.mighty_hammerdrop_timer)
			and (mo.mighty_hammerdrop_timer >= mighty.HammerDropStartupTics)	-- only while we're dropping
				mo.mighty_hammerdrop_springinfo = {vertical = mobjinfo[mo2.type].mass}
			end
			
			mighty.ResetHammerDrop(mo)
			return
		end
	end
	
	-- break spikes with the hammerdrop
	if (mo2.type == MT_SPIKE or mo2.type == MT_WALLSPIKE)
	and (mo.mighty_hammerdrop_timer)
	and (mo.mighty_hammerdrop_timer >= mighty.HammerDropStartupTics)
		-- adjust this based on momz
		local topheight = (mo2.z + mo2.height) - mo.momz
		local bottomheight = mo2.z - mo.momz
		
		if ((mo.z <= topheight)
		and (bottomheight <= mo.z + mo.height))
			P_KillMobj(mo2, mo, mo)
			return false	-- they're getting destroyed, we WOULDNT collide
		end
	end
	
	-- the rest of these use regular height checks
	if not ((mo.z <= mo2.z + mo2.height)
	and (mo2.z <= mo.z + mo.height))
		return
	end
	
	-- monitor collision
	if (mo2.flags & MF_MONITOR)
	and (mo.mighty_hammerdrop_timer)
	and (mo.mighty_hammerdrop_timer >= mighty.HammerDropStartupTics)
		-- pierce the monitor
		P_DamageMobj(mo2, mo, mo, 1, 0)
		if (mo2.flags & MF_GRENADEBOUNCE)	-- land directly on golden monitors
			mo.player.pflags = $|(PF_JUMPED|PF_THOKKED)
			mo.state = S_PLAY_ROLL
			mo.mighty_hammerdrop_ghosttics = mighty.HammerDropGhostTics
			P_SetObjectMomZ(mo, chaotix.WaterMom(mo, mighty.HammerDropRecoilMomentum))
			mighty.doHammerDrop(mo.player)
			mighty.ResetHammerDrop(mo)
		end
		
		return true
	end
	
	-- spring collision
	if (mo2.flags & MF_SPRING)
		-- amplify momentum
		mo.mighty_hammerdrop_springinfo = {}
		local springinfo = mo.mighty_hammerdrop_springinfo
		
		if (mo.mighty_hammerdrop_timer)
		and (mo.mighty_hammerdrop_timer >= mighty.HammerDropStartupTics)	-- only while we're dropping
			if (mobjinfo[mo2.type].mass)	-- vertical
				springinfo.vertical = mobjinfo[mo2.type].mass
			end
			if (mobjinfo[mo2.type].damage)	-- horizontal
				springinfo.horizontal = mobjinfo[mo2.type].damage
			end
		end
		
		-- set state correctly
		if not (mobjinfo[mo2.type].mass)	-- vertical
		and (mobjinfo[mo2.type].damage)	-- horizontal
			mo.state = S_PLAY_ROLL
		end
		
		mighty.ResetHammerDrop(mo)
		return
	end
end
chaotix.AddHook("MobjCollide", mighty.HammerDropCollide, MT_PLAYER)
chaotix.AddHook("MobjMoveCollide", mighty.HammerDropCollide, MT_PLAYER)

-- multiply momentum on springs
mighty.HammerDropSprings = function(p)
	local mo = p.mo
	if not ((mo and mo.valid) and mo.health) then return end
	local springinfo = mo.mighty_hammerdrop_springinfo
	
	-- check springinfo
	if not (springinfo) then return end
	if not (springinfo["vertical"] or springinfo["horizontal"]) then
		mo.mighty_hammerdrop_springinfo = nil
		return
	end

	if (springinfo["vertical"])	-- multiply vertical momentum
		mo.momz = FixedMul(mighty.HammerDropSpringMultiplier, $)
		
		-- spring twirl! (like cd/mania)
		if not (springinfo["horizontal"])
			mo.state = S_MIGHTY_HAMMERDROP_SPRING
			mo.mighty_hammerdrop_twirlangle = p.drawangle
		end
	end
	if (springinfo["horizontal"])	-- multiply horizontal momentum
		mo.momx = FixedMul(mighty.HammerDropSpringMultiplier, $)
		mo.momy = FixedMul(mighty.HammerDropSpringMultiplier, $)
	end
	mo.mighty_hammerdrop_ghosttics = mighty.HammerDropGhostTics
	mo.mighty_hammerdrop_springinfo = nil
	S_StartSound(mo, sfx_sprong)
end