-- funny extras for mighty
local mighty = chaotix.mighty

-- mighty's shell can pop off when you fall in a pit!
mighty.ShellDeath = function(mo, inflictor, source, damagetype)
	if (mo.skin != mighty.SKIN) then return end
	if (damagetype != DMG_DEATHPIT) then return end
	
	-- random chance to have our shell pop off
	if (P_RandomChance(mighty.ShellDeathChance))
		mo.mighty_shelldeath = 1
	end
end
mighty.ShellDeathThinker = function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	
	-- check if should have our shell pop off!
	if not (mo.mighty_shelldeath) then return end
	if (mo.health)	-- we're alive???
	or (mo.skin != mighty.SKIN)	-- we are somehow not mighty anymore
	or (p.isjettysyn)	-- NO
		mo.mighty_shelldeath = nil
		return
	end
	
	-- pop the shell off!
	local threshold = FixedMul(mo.scale, mighty.ShellDeathThreshold)
	if ((mo.momz * P_MobjFlip(mo)) < threshold)	-- we're falling fast enough
	and (mo.mighty_shelldeath != 2)	-- we haven't already done the pop
		-- change sprite
		mo.sprite2 = SPR2_TAL0
		
		-- spawn the shell
		local shell = P_SpawnMobjFromMobj(mo, 0, 0, 19*FRACUNIT, MT_GHOST)
		shell.target = mo
		shell.color = mo.color
		shell.flags = MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT
		P_SpawnMobjFromMobj(shell, 0, 0, 0, MT_EXPLODE)
		S_StartSound(shell, sfx_cdfm15)
		shell.state = S_MIGHTY_SHELL
		
		-- set this so we don't do this again
		mo.mighty_shelldeath = 2	
	end
end

-- mighity.
mighty.Mighity = function(p)
	if (p.mighity != nil) then return end	-- already set
	p.mighity = P_RandomChance(mighty.MighityChance)
end
mighty.MighityReset = function(gamemap)
	for p in players.iterate
		p.mighity = nil
	end
end
mighty.DrawHUDName = function(v, p, c)
	if ((maptol & TOL_NIGHTS) or G_IsSpecialStage(gamemap)) then return end
	if (modeattacking) then return end
	if not (hud.enabled("lives")) then return end
	if (skins[p.skin].name != mighty.SKIN) then return end
	
	-- set flags
	local flags = V_YELLOWMAP
	if (p.spectator)
		flags = V_GRAYMAP
	elseif (gametyperules & GTR_TAG)
	and (p.pflags & PF_TAGIT)
		flags = V_ORANGEMAP
	elseif (G_GametypeHasTeams())
	and (p.ctfteam)	-- we're on a team!
		flags = (p.ctfteam == 1) and V_REDMAP or V_BLUEMAP
	end
	flags = $|(V_HUDTRANS|hudinfo[HUD_LIVES].f|V_PERPLAYER)
	
	-- draw "name" (if only we could edit skins[mighty.SKIN].hudname...)
	local name_appendix = (consoleplayer.mighity) and "ity" or "ty"	-- make sure this is consoleplayer for extra madness
	v.drawString(hudinfo[HUD_LIVES].x + 18, hudinfo[HUD_LIVES].y, "MIGH"..name_appendix, flags)
end

-- sexy
chaotix.AddHook("PlayerMsg", function(p, msgtype, target, msg)
	if (msgtype) then return end	-- needs to be a regular message
	
	-- mobj checks
	local mo = p.mo
	if not ((mo and mo.valid) and mo.health) then return end	-- gotta be alive
	if (mo.skin != mighty.SKIN) then return end	-- we gotta be mighty
	
	-- both state checks
	if (mo.state >= S_MIGHTY_SEXY and mo.state <= S_MIGHTY_SEXY_AIR)	-- already doing it
	or (p.actionstate)	-- actionstate
	or (p.isjettysyn)	-- NO
	or (p.powers[pw_nocontrol])	-- no control
		return
	end
	
	-- check message
	msg = $:lower()
	if (msg == "sexy")	-- sexy
		local panim, state
		
		if (P_IsObjectOnGround(mo))
			-- fail conditions
			if (p.speed) then return end	-- can't be moving!
			
			-- start the grounded sexy
			state = S_MIGHTY_SEXY
			panim = PA_IDLE
			p.pflags = $ & ~(PF_SPINNING|PF_STARTDASH)
		else
			-- fail conditions
			if (mo.mighty_hammerdrop)	-- specific check for the hammer drop
			or (P_PlayerInPain(p))	-- we got hurt
			or (p.powers[pw_carry])	-- we got carried
			or (p.panim == PA_PAIN)	-- pain pain pain pain pain
			or (p.airdodge) 	-- air dodging
			or (p.climbing)	-- climbing
				return 
			end
			
			-- start the aerial sexy
			state = S_MIGHTY_SEXY_AIR
			panim = PA_FALL
			p.pflags = $ & ~(PF_JUMPED|PF_STARTJUMP|PF_SPINNING)
		end
		
		-- we got sexy
		if not (panim and state) then return end
		
		if (mo.state != state) then mo.state = state end
		p.panim = panim
		mo.mighty_sexy = mighty.SexyTime
		S_StartSound(p.mo, sfx_vnboom)
			
		local g = P_SpawnGhostMobj(mo)
		g.destscale = FixedMul(FRACUNIT*5/4, mo.scale)
		g.scalespeed = FRACUNIT/24
		g.tics = TICRATE/4
		return true
	end
end)