-- where all the protection is done
local mighty = chaotix.mighty

-- check if a player is protected
mighty.IsPlayerProtected = function(p)
	local mo = p.mo
	if not ((mo and mo.valid) and mo.health) then return false end
	
	-- general protection
	if (p.pflags & PF_JUMPED)
	or (p.pflags & PF_SPINNING)
		return true
	end
	
	-- if we're here, we're not protected
	return false
end

-- this is purely just for spike sectors
mighty.ProtectionTics = function(p)
	local mo = p.mo
	if not ((mo and mo.valid) and mo.health) then return end
	if (mo.skin != mighty.SKIN)	-- mighty exclusive, his shell is pretty mighty
		mo.mighty_protection_reflecttics = nil
		mo.mighty_protection_reflectmo = nil
		
		mo.mighty_protection_spiketics = nil
		mo.mighty_protection_spikethokked = nil
		
		mo.mighty_protection_fliptics = nil
		return
	end
	
	-- handle timers
	mo.mighty_protection_reflecttics = $ and $-1 or 0
	mo.mighty_protection_spiketics = $ and $-1 or 0
	mo.mighty_protection_fliptics = $ and $-1 or 0
	mo.mighty_protection_sidetics = $ and $-1 or 0
	
	-- reset stuff based on timers
	if not (mo.mighty_protection_reflecttics)
		mo.mighty_protection_reflectmo = nil
	end
	if not (mo.mighty_protection_spiketics)
		mo.mighty_protection_spikethokked = nil	-- no need for this anymore
	end
	
	-- update tics if we're protected
	if not (mighty.IsPlayerProtected(p)) then return end
	mo.mighty_protection_spiketics = 2
	mo.mighty_protection_spikethokked = (p.pflags & PF_THOKKED)
end
local function mighty_onSpikeSec(mo)
	if (P_IsObjectOnGround(mo))
	and (GetSecSpecial(mo.subsector.sector.special, 1) == 5)
		return true
	end
	
	-- if we're here, we're not on a spike sector
	return false
end

mighty.ShouldDamage = function(mo, inflictor, source, damage, damagetype)
	if not ((mo and mo.valid) and mo.health) then return end
	if (mo.skin != mighty.SKIN) then return end	-- mighty exclusive, his shell is pretty mighty
	local p = mo.player
	
	-- we dont NEED to be protected right now
	if (p.powers[pw_flashing])	-- flashtics
	or (p.powers[pw_invulnerability])	-- invincible
	or (P_PlayerFullbright(p))	-- super
		return
	end
	
	-- conditions to protect mighty under
	if not (mighty.IsPlayerProtected(p))
	and not (mighty_onSpikeSec(mo) and mo.mighty_protection_spiketics)
		-- flipped spike protection
		if ((inflictor and inflictor.valid) and inflictor.type == MT_SPIKE)
		and (mo.mighty_protection_fliptics)
		and (P_MobjFlip(mo) != P_MobjFlip(inflictor))
			return false
		end
		
		return
	end
	
	-- spikes are a special case, handle separately
	if ((inflictor and inflictor.valid) and inflictor.type == MT_SPIKE)	-- normal spies
	or (not (inflictor and inflictor.valid) and mighty_onSpikeSec(mo))	-- spike sectors
		-- hammer dropping exclusive behaviour
		local hammerdrop_check = false
		if (mo.mighty_hammerdrop)
		and (mo.mighty_hammerdrop_timer)
		and (mo.mighty_hammerdrop_timer >= mighty.HammerDropStartupTics)
			local bounce_momz = chaotix.WaterMom(mo, FRACUNIT*11)
			if (inflictor and inflictor.valid)
				-- get protected from flipped spikes momentarily
				if (P_MobjFlip(mo) != P_MobjFlip(inflictor))
					bounce_momz = $ * -1
					mo.mighty_protection_fliptics = 5
				end
				
				-- also break the spike
				P_KillMobj(inflictor, mo, mo)
			end
			
			-- bounce up from the spike/surface
			p.pflags = ($|PF_JUMPED|PF_THOKKED) & ~(PF_SPINNING)
			mo.state = S_PLAY_JUMP
			mo.mighty_hammerdrop_ghosttics = 6
			mo.mighty_hammerdrop_preventstomp = true
			S_StartSound(mo, sfx_cdfm46)
			P_SetObjectMomZ(mo, bounce_momz)
			
			-- set this so we can end the hook afterwards
			hammerdrop_check = true
		end
		mighty.ResetHammerDrop(p)	-- reset this regardless of conditions
		if (hammerdrop_check) then return false end	-- end here if we did hammerdrop stuff
		
		local threshold = FixedMul(p.normalspeed, mighty.SpikeSpeedThreshold)
		if (FixedHypot(mo.momx, mo.momy) > FixedMul(mo.scale, threshold))
			local bounce_momz = chaotix.WaterMom(mo, FRACUNIT*4)
			if (inflictor and inflictor.valid)
			and (P_MobjFlip(mo) != P_MobjFlip(inflictor))
				bounce_momz = $ * -1
			end
			
			-- bounce up while staying in current state
			P_SetObjectMomZ(mo, bounce_momz)
			S_StartSound(mo, sfx_cdfm46)
			
			-- make sure we roll again with spike sectors
			if (not (inflictor and inflictor.valid) and mighty_onSpikeSec(mo))
			and not (mighty.IsPlayerProtected(p))
				local thokflag = mo.mighty_protection_spikethokked or 0
				p.pflags = $|(thokflag|PF_JUMPED)
				mo.state = S_PLAY_JUMP
			end
		else
			local bounce_momz = chaotix.WaterMom(mo, FRACUNIT*8)
			if (inflictor and inflictor.valid)
				-- get protected from flipped spikes momentarily
				if (P_MobjFlip(mo) != P_MobjFlip(inflictor))
					bounce_momz = $ * -1
					mo.mighty_protection_fliptics = 5
				end
			end
			
			-- not enough speed, bounce up in falling frames
			p.pflags = $ & ~(PF_JUMPED|PF_SPINNING)
			mo.state = S_PLAY_FALL
			S_StartSound(mo, sfx_spkprt)
			P_SetObjectMomZ(mo, bounce_momz)
		end
		
		return false
	end

	-- handle everything else with the protection defs
	if (inflictor and inflictor.valid)
		-- is the object blacklisted completely?
		if (mighty.GetProtectionDef(inflictor.type, "blacklisted"))
			return
		end
		
		-- is it from a player? dont make it unfair
		if (source and source.valid)
		and (source.player and not (CBW_Battle and CBW_Battle.BattleGametype))
			return
		end
		
		-- is it fire related?
		if (inflictor.flags & MF_FIRE) then return end
		
		-- prioritise defined functions
		if (mighty.GetProtectionDef(inflictor.type, "func"))
			local func = mighty.GetProtectionDef(inflictor.type, "func")
			return func(mo, inflictor, source)
		end
		
		-- missiles can be handled in a bunch
		if (inflictor.flags & MF_MISSILE)
			-- get pushed back by the missile
			local pushback_modifier = mighty.ProtectionDefaultPushbackModifier
			if (mighty.GetProtectionDef(inflictor.type, "pushback_modifier"))
				pushback_modifier = mighty.GetProtectionDef(inflictor.type, "pushback_modifier")
			end
			mo.momx = $ + (FixedMul(pushback_modifier, inflictor.momx))
			mo.momy = $ + (FixedMul(pushback_modifier, inflictor.momy))
			mo.momz = $ + (FixedMul(pushback_modifier, inflictor.momz))
			
			-- show them deflecting off?
			if not (mighty.GetProtectionDef(inflictor.type, "prevent_visual"))
				local projectile = P_SpawnMobjFromMobj(inflictor, 0, 0, 0, MT_MIGHTY_PROJECTILE)
				projectile.state = S_INVISIBLE
				projectile.tics = -1
				
				-- copy state
				projectile.angle = inflictor.angle
				projectile.sprite = inflictor.sprite
				projectile.frame = inflictor.frame
				
				-- copy info
				projectile.radius = inflictor.radius
				projectile.height = inflictor.height
				
				-- launch the projectile away
				local angle = R_PointToAngle2(mo.x, mo.y, inflictor.x, inflictor.y)
				local thrust = FixedMul(mighty.ProtectionThrustModifier, FixedHypot(inflictor.momx, inflictor.momy))
				P_InstaThrust(projectile, angle, thrust)
				projectile.momz = -(FixedMul(mighty.ProtectionThrustModifier, inflictor.momz))
			end
			
			-- destroy the missile
			P_KillMobj(inflictor)
			S_StartSound(mo, sfx_cdfm46)
			return false
		end
		
		-- bounce from these
		if (inflictor.flags & MF_PAIN)
			-- try not to reflect multiple times
			local rmo = mo.mighty_protection_reflectmo
			if ((rmo and rmo.valid) and rmo.health)	-- it still exists
			and (mo.mighty_protection_reflecttics)	-- still got tics left
			and (inflictor == rmo)	-- same mobj
				return false
			end
			
			local angle = R_PointToAngle2(mo.x, mo.y, inflictor.x, inflictor.y)
			P_InstaThrust(mo, angle, -(FixedHypot(mo.momx, mo.momy)))
			mo.momz = -(mo.momz)
			
			-- add their momentum if possible
			mo.momx = $ + (inflictor.momx)
			mo.momy = $ + (inflictor.momx)
			mo.momz = $ + (inflictor.momx)
			
			mo.mighty_protection_reflectmo = inflictor
			mo.mighty_protection_reflecttics = 5	-- dont get hit by the same inflictor till after this
			S_StartSound(mo, sfx_cdfm46)
			return false
		end
	end
end