-- mighty's wall kick
local mighty = chaotix.mighty

mighty.ResetWallKick = function(p, resetangle)
	local mo = p.mo
	
	p.mighty_wallkick, mo.mighty_wallkick = nil, nil
	mo.mighty_wallkick_momentum = nil
	mo.mighty_wallkick_dusttics = nil
	mo.mighty_wallkick_savedmomentum = nil
	mo.mighty_wallkick_oldinfo = nil
	mo.mighty_wallkick_frame = nil
	mo.mighty_wallkick_frametics = nil
end

local function mighty_wallCling(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	if not (mighty.AbilityCheck(p)) then return end
	if (mo.mighty_wallkick) then return end	-- we're already running up a wall!
	
	if (p.lastlinehit != -1)
		local l = lines[p.lastlinehit]
		
		if (p.pflags & PF_JUMPED)	-- we need to be jumping
		and (p.cmd.buttons & BT_JUMP)	-- we also need to be holding jump
		and (p.cmd.forwardmove or p.cmd.sidemove)	-- we need to actually be moving
		and not (mo.mighty_wallkick_delay)	-- so we can actually jump against walls
		and not (p.actionstate) -- battlemod support
		and not (CBW_Battle and p.exhaustmeter == 0)	-- we're exhausted (battle mod)
		and not (p.gotflag or p.gotcrystal)	-- we have the flag (or the crystal in battlemod)
		and (chaotix.CheckLine(p, l))	-- make sure everything is good here
			local momx, momy = mo.mighty_oldmomx or mo.momx, mo.mighty_oldmomy or mo.momy
			
			-- store momentum
			local wallangle = R_PointToAngle2(l.v1.x, l.v1.y, l.v2.x, l.v2.y)
			local moangle = R_PointToAngle2(0, 0, momx, momy)
			local hitangle = wallangle - moangle
			local speedfactor = abs(cos(hitangle))
			local speed = FixedMul(speedfactor, FixedHypot(momx, momy))
			
			local reverseangle = (wallangle - ANGLE_90) - moangle
			if (chaotix.AngToInt(reverseangle) > 180)
				speed = $ * -1
			end
			speed = FixedMul(mighty.WallKickMomStabilizer, speed)
	
			local momz = mo.momz
			if ((momz * P_MobjFlip(mo)) < 0)
				momz = FixedMul(mighty.WallKickMomZFallingStabilizer, momz)
			else
				momz = FixedMul(mighty.WallKickMomStabilizer, momz)
			end
			mo.mighty_wallkick_momentum = {speed, momz}
			
			-- face the wall
			local whichside = P_PointOnLineSide(mo.x, mo.y, l)
			local angle = R_PointToAngle2(l.v1.x, l.v1.y, l.v2.x, l.v2.y)
			angle = $ + (ANGLE_90 * (whichside and -1 or 1))
			p.drawangle = angle
			
			-- little hack to face the wall without moving the camera (SMS Alfredo)
			local thrustx = P_ReturnThrustX(mo, angle, p.height)
			local thrusty = P_ReturnThrustY(mo, angle, p.height)
			local thok = P_SpawnMobjFromMobj(mo, thrustx, thrusty, 0, MT_THOK)
				
			local oldtarget = mo.target
			mo.target = thok
			A_FaceTarget(mo)
				
			mo.target = oldtarget
			thok.state = S_NULL
			
			-- cling to the wall!
			p.charability = CA_GLIDEANDCLIMB
			p.climbing = 5
			p.mighty_wallkick, mo.mighty_wallkick = true, true	-- trust me there's a reason we have two saved
			mo.state = (speed or momz) and S_PLAY_CLIMB or S_PLAY_CLING
			mo.mighty_wallkick_dusttics = mighty.WallKickDustTicBounds[2]
			S_StartSound(mo, sfx_kc55)
			
			-- setup frames
			if (mo.state == S_PLAY_CLIMB)
				mo.mighty_wallkick_frame = 0
				mo.mighty_wallkick_frametics = mighty.WallKickMaxFrameTics
			end
				
			-- reset stuff on cling
			p.pflags = $ & ~(PF_JUMPED|PF_STARTJUMP|PF_SPINNING|PF_THOKKED|PF_SHIELDABILITY)
			p.secondjump = 0
			
			-- dust effect
			local speed = FRACUNIT*7/2
			local amount = 8
			local diff = 360/amount
			
			local horzangle = angle + ANGLE_90
			for i = 1, amount
				local vertangle = (i*diff)*ANG1
				
				-- spawn the dust
				local distance = FRACUNIT*13
				local xoff = FixedMul(cos(angle), distance)
				local yoff = FixedMul(sin(angle), distance)
				local zoff = FixedMul(FRACUNIT/2, mo.height)
						
				local mobj = P_SpawnMobjFromMobj(mo, xoff, yoff, zoff, MT_THOK)
				mobj.target = mo
				
				-- horizontal
				local horizontal_thrust = FixedMul(sin(vertangle), speed)
				P_InstaThrust(mobj, horzangle, FixedMul(mo.scale, horizontal_thrust))
				mobj.angle = horzangle
				
				-- vertical
				local vertical_thrust = FixedMul(cos(vertangle), speed)
				P_SetObjectMomZ(mobj, vertical_thrust)
				
				-- set state under the right conditions
				mobj.scale = FixedMul(FRACUNIT*2/3, $)
				mobj.state = S_SPINDUST2
				
				if (mo.eflags & (MFE_TOUCHWATER|MFE_UNDERWATER)) 
					mobj.state = S_SPINDUST_BUBBLE1
				elseif (p.powers[pw_shield] == SH_ELEMENTAL)
					mobj.state = S_SPINDUST_FIRE1
				end
			end
			
			return
		end
		
		-- reset these to check for wall collision again
		p.lastlinehit, p.lastsidehit = -1, -1
	end
	mo.mighty_oldmomx, mo.mighty_oldmomy = mo.momx, mo.momy	-- i hate that I have to store this...
end

-- prevent movement while clinged as mighty!
mighty.PreThinkFrame = function()
	for p in players.iterate
		local mo = p.mo
		if not ((mo and mo.valid) and mo.health) then continue end
		if not (mo.mighty_wallkick) then continue end
		
		-- prevent movement, but store forwardmove/sidemove values
		local cmd = p.cmd
		p.mighty_forwardmove = cmd.forwardmove
		p.mighty_sidemove = cmd.sidemove
		
		cmd.forwardmove = 0
		cmd.sidemove = 0
	end
end

mighty.WallKick = function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	if not (mighty.AbilityCheck(p)) then mighty.ResetWallKick(p) return end
	
	-- wallkick delay
	if (P_IsObjectOnGround(mo))
		mo.mighty_wallkick_delay = mighty.WallKickDelay
		mo.mighty_wallkick_tics = mighty.WallKickSlideTics
	else
		mo.mighty_wallkick_delay = $ and $-1
	end
	
	-- cling to walls
	mighty_wallCling(p)
	
	if (p.climbing)
		-- if we're not super, lose our grip!
		if not (P_PlayerFullbright(p))
			mo.mighty_wallkick_tics = $ and $-1
		else
			mo.mighty_wallkick_tics = mighty.WallKickSlideTics
		end
		
		-- store old position (trust me this'll be useful)
		if (mo.mighty_wallkick_tics)	-- gotta still be able to cling
		and (mo.mighty_wallkick_momentum[1] or mo.mighty_wallkick_momentum[2])	-- gotta have some momentum
			mo.mighty_wallkick_oldinfo = $ or {}
			
			local info = {
				pos = {mo.x, mo.y, mo.z},
				lastlinehit = p.lastlinehit,
			}
			table.insert(mo.mighty_wallkick_oldinfo, info)
			
			-- prevent the table from getting too big
			if (#mo.mighty_wallkick_oldinfo > mighty.WallKickMaxInfoSave)
				table.remove(mo.mighty_wallkick_oldinfo, 1)
			end
		end
		
		-- if we're not holding jump, jump off the wall!
		if not (p.cmd.buttons & BT_JUMP)
			-- reset character ability
			p.charability = CA_MIGHTY
			
			-- get jump angle (make sure it's restricted too)
			local maxangle = mighty.WallKickJumpMaxAngle	-- max walljump angle
			local cameraangle = (p.cmd.angleturn * FRACUNIT)
			local camerachange = true
			local anglediff = mo.angle - cameraangle
			
			-- account for wall kick modes
			if (p.pflags & PF_ANALOGMODE)	-- not fully camera based
				if p.mighty_forwardmove or p.mighty_sidemove
					local inputangle = R_PointToAngle2(0, 0, p.mighty_forwardmove*FRACUNIT, -p.mighty_sidemove*FRACUNIT)
					anglediff = $ - inputangle
					if abs(anglediff) > ANGLE_90 or anglediff == ANGLE_180
						anglediff = InvAngle($) + ANGLE_180
					end
					
					camerachange = false -- no camera change!
				else
					-- inputting nothing jumps off the wall directly with a camera change
					anglediff = 0
				end
			end
			
			anglediff = max(-(maxangle), min(maxangle, $ / ANG1)) * ANG1	-- min/max it
			local angle = mo.angle + ANGLE_180 + anglediff
			
			-- jump off the wall!
			local jumpspeed = FixedMul(p.jumpfactor, FixedMul(mo.scale, mighty.WallKickJumpSpeed))
			local jumpheight = FixedMul(p.jumpfactor, mighty.WallKickJumpHeight)
			if (P_PlayerFullbright(p))
				-- increase power for super forms!
				jumpspeed = FixedMul(mighty.WallKickSuperModifier, $)
				jumpheight = FixedMul(mighty.WallKickSuperModifier, $)
			end
			
			P_DoJump(p, true)
			p.pflags = $ & ~(PF_STARTJUMP)
			if camerachange
				mo.angle = angle
			else
				mo.angle = cameraangle
			end
			p.drawangle = angle
			
			P_InstaThrust(mo, angle, chaotix.WaterMom(mo, jumpspeed))
			P_SetObjectMomZ(mo, chaotix.WaterMom(mo, jumpheight))
			S_StartSound(mo, sfx_jump)
			
			-- reset wallkick vars
			mighty.ResetWallKick(p)
			
			-- reset these to prevent reclings
			p.lastlinehit = -1
			p.lastsidehit = -1
			return
		end
	else
		if (p.mighty_wallkick)
			-- this is why we have it saved in mo as well
			if (mo.mighty_wallkick)
				if not (P_IsObjectOnGround(mo))	
					if (p.cmd.buttons & BT_SPIN)
						-- minimum height
						if ((mo.eflags & MFE_VERTICALFLIP) and mo.momz > FixedMul(mo.scale, -3*FRACUNIT))
						or (not (mo.eflags & MFE_VERTICALFLIP) and mo.momz < FixedMul(mo.scale, 3*FRACUNIT))
							mo.momz = FixedMul(mo.scale, FRACUNIT*3) * P_MobjFlip(mo)
						end
					else
						/* now done in the PlayerThink hook to prevent camera jank
						-- try to stay clinged?
						if (mo.mighty_wallkick_oldinfo)
						and (#mo.mighty_wallkick_oldinfo >= mighty.WallKickMinInfoSave)	-- got enough to safely do this
						and (mo.mighty_wallkick_tics)	-- we're sliding off!
							-- go back to old position
							local info = mo.mighty_wallkick_oldinfo[1]	-- this'll be the oldest position
							P_SetOrigin(mo, info.pos[1], info.pos[2], info.pos[3])
							
							-- remove momentum
							mo.momx, mo.momy, mo.momz = 0, 0, 0
							mo.mighty_wallkick_momentum[1] = 0
							mo.mighty_wallkick_momentum[2] = 0
							
							-- little hack to face the wall without moving the camera (SMS Alfredo)
							local l = lines[info.lastlinehit]
							
							local whichside = P_PointOnLineSide(mo.x, mo.y, l)
							local angle = R_PointToAngle2(l.v1.x, l.v1.y, l.v2.x, l.v2.y)
							angle = $ + (ANGLE_90 * (whichside and -1 or 1))
							p.drawangle = angle
							
							-- little hack to face the wall without moving the camera (SMS Alfredo)
							local thrustx = P_ReturnThrustX(mo, angle, p.height)
							local thrusty = P_ReturnThrustY(mo, angle, p.height)
							local thok = P_SpawnMobjFromMobj(mo, thrustx, thrusty, 0, MT_THOK)
								
							local oldtarget = mo.target
							mo.target = thok
							A_FaceTarget(mo)
								
							mo.target = oldtarget
							thok.state = S_NULL
											
							-- stay clinged
							p.climbing = 5
							p.mighty_wallkick, mo.mighty_wallkick = true, true	-- trust me there's a reason we have two saved
							mo.state = S_PLAY_CLING
							mo.mighty_wallkick_dusttics = mighty.WallKickDustTicBounds[2]
							
							-- delete old info, not needed anymore
							mo.mighty_wallkick_oldinfo = nil
							return
						end*/
						
						-- set momentum
						if (mo.mighty_wallkick_savedmomentum)
							local moms = {"momx", "momy", "momz"}
							for i = 1, 3
								if not (mo.mighty_wallkick_savedmomentum[i]) then continue end
								local mom = moms[i]
								mo[mom] = mo.mighty_wallkick_savedmomentum[i]
							end
						end
						
						mo.mighty_wallkick_delay = mighty.WallKickDelay	-- this helps more than you'd think
					end
				end
				
				S_StartSound(mo, sfx_kc55)
			end
			
			-- this needs to always run, otherwise we can keep the wrong ability through levels
			p.charability = CA_MIGHTY
			mighty.ResetWallKick(p)
			
			-- reset these to prevent reclings
			p.lastlinehit = -1
			p.lastsidehit = -1
			return
		end
	end
end

-- playerthink lets us edit momentum while climbing
mighty.PlayerThink = function(p)
	local mo = p.mo
	if not ((mo and mo.valid) and mo.health) then return end
	
	-- wall kick checks
	if not (mighty.AbilityCheck(p)) then return end
	if not (p.climbing)
		-- gotta do this here, since we can INSTANTLY restick and prevent camera shit
		if not (p.cmd.buttons & BT_SPIN)
		and (mo.mighty_wallkick_oldinfo)
		and (#mo.mighty_wallkick_oldinfo >= mighty.WallKickMinInfoSave)	-- got enough to safely do this
		and (mo.mighty_wallkick_tics)	-- we're sliding off!
			-- go back to old position
			local info = mo.mighty_wallkick_oldinfo[1]	-- this'll be the oldest position
			P_SetOrigin(mo, info.pos[1], info.pos[2], info.pos[3])
							
			-- remove momentum
			mo.momx, mo.momy, mo.momz = 0, 0, 0
			mo.mighty_wallkick_momentum[1] = 0
			mo.mighty_wallkick_momentum[2] = 0
							
			-- face the wall
			local l = lines[info.lastlinehit]
							
			local whichside = P_PointOnLineSide(mo.x, mo.y, l)
			local angle = R_PointToAngle2(l.v1.x, l.v1.y, l.v2.x, l.v2.y)
			angle = $ + (ANGLE_90 * (whichside and -1 or 1))
			p.drawangle = angle
							
			-- little hack to face the wall without moving the camera (SMS Alfredo)
			local thrustx = P_ReturnThrustX(mo, angle, p.height)
			local thrusty = P_ReturnThrustY(mo, angle, p.height)
			local thok = P_SpawnMobjFromMobj(mo, thrustx, thrusty, 0, MT_THOK)
								
			local oldtarget = mo.target
			mo.target = thok
			A_FaceTarget(mo)
								
			mo.target = oldtarget
			thok.state = S_NULL
											
			-- stay clinged
			p.climbing = 5
			p.mighty_wallkick, mo.mighty_wallkick = true, true	-- trust me there's a reason we have two saved
			mo.state = S_PLAY_CLING
			mo.mighty_wallkick_dusttics = mighty.WallKickDustTicBounds[2]
							
			-- delete old info, not needed anymore
			mo.mighty_wallkick_oldinfo = nil
		else
			-- nvm, cease functions
			return
		end
	end
	
	-- remove horizontal momentum then set it
	local removemomxy = abs(mo.mighty_wallkick_momentum[1]) - FixedMul(mighty.WallKickSlideModifier, abs(mo.mighty_wallkick_momentum[1]))
	if (removemomxy < FixedMul(mo.scale, mighty.WallKickMinXYSlide))
		-- would just use max() but it'd be too long on one line and this is just about the same thing
		removemomxy = FixedMul(mo.scale, mighty.WallKickMinXYSlide)
	end
		
	local minmax_func = max
	if (mo.mighty_wallkick_momentum[1] < 0)
		removemomxy = $ * -1
		minmax_func = min
	end
	mo.mighty_wallkick_momentum[1] = minmax_func(0, $ - (removemomxy / ((mo.eflags & MFE_UNDERWATER) and 2 or 1)))
		
	local l = lines[p.lastlinehit]
	if (l and l.valid)	-- we need this
		local wallangle = R_PointToAngle2(l.v2.x, l.v2.y, l.v1.x, l.v1.y)
		P_InstaThrust(mo, wallangle, mo.mighty_wallkick_momentum[1])	-- this currently doesn't apply for scrollers, too bad!
	end
	
	-- remove vertical momentum then set it
	if (mo.mighty_wallkick_tics)	-- we can still cling to the wall without penalty!
		if (mo.mighty_wallkick_momentum[2])	-- we still have momentum! remove it quickly
			-- calculate how much to reduce momz by
			local removemomz = FixedMul(mo.scale, mighty.WallKickSlideClingSpeed)
			local minmax_func = max
			-- flip these if we need to increase to 0
			if (mo.mighty_wallkick_momentum[2] < 0)
				removemomz = $ * -1
				minmax_func = min
			end
			
			mo.mighty_wallkick_momentum[2] = minmax_func(0, $ - (removemomz / ((mo.eflags & MFE_UNDERWATER) and 2 or 1)))
		end
	else
		-- first calculate slide speed
		local slidespeed = mighty.WallKickSlideClingSpeed
		if (mo.eflags & MFE_GOOWATER)
			if ((mo.mighty_wallkick_momentum[2] * P_MobjFlip(mo)) > 0)
				slidespeed = mighty.WallKickSlideFallSpeed
			end
		else
			if ((mo.mighty_wallkick_momentum[2] * P_MobjFlip(mo)) < 0)
				slidespeed = mighty.WallKickSlideFallSpeed
			end
		end
		slidespeed = FixedMul(mo.scale, $)
		
		-- now remove it from our momz
		local removemomz = slidespeed * P_MobjFlip(mo)
		if (mo.eflags & MFE_GOOWATER)
			removemomz = $ * -1
		end
		mo.mighty_wallkick_momentum[2] = $ - (removemomz / ((mo.eflags & MFE_UNDERWATER) and 2 or 1))
	end
	mo.momz = $ + mo.mighty_wallkick_momentum[2]
	
	-- save these for launching off the wall...
	mo.mighty_wallkick_savedmomentum = {mo.momx, mo.momy, mo.momz}
	
	-- visuals!
	local speed = FixedHypot(mo.mighty_wallkick_momentum[2], mo.mighty_wallkick_momentum[1])	-- take both momxy and momz into account
	
	if (speed)
		-- frame animation
		local frame_speed = FixedMul(mo.scale, mighty.WallKickMaxFrameSpeed)
		local frame_tics = mighty.WallKickMaxFrameTics
		local maxtics = max(1, frame_tics - ((speed * frame_tics) / frame_speed))
		mo.mighty_wallkick_frametics = $ and min(maxtics, $) or maxtics
		
		local numframes = 2
		mo.mighty_wallkick_frametics = $-1
		if not (mo.mighty_wallkick_frametics)
			mo.mighty_wallkick_frame = $+1
				
			-- reset frame after max frame is reached
			if (mo.mighty_wallkick_frame > (numframes - 1))
				mo.mighty_wallkick_frame = 0
			end
		end
		
		mo.state = S_PLAY_CLIMB
		chaotix.SetFrame(mo, mo.mighty_wallkick_frame)
	
		-- dust particles on the wall
		local mombounds = mighty.WallKickDustMomBounds
		if (speed > mombounds[1])
			speed = $ - mombounds[1]
			
			-- duration is based on how fast we're going
			local momrange = FixedMul(mo.scale, mombounds[2] - mombounds[1])
			local ticbounds = mighty.WallKickDustTicBounds
			local ticrange = ticbounds[2] - ticbounds[1]
			
			local mintics = max(ticbounds[1], ticbounds[2] - ((speed * ticrange) / momrange))
			mo.mighty_wallkick_dusttics = min(mintics, $) - 1
			if not (mo.mighty_wallkick_dusttics)
				for i = 1, 2	-- for both shoes
					for j = 1, 2
						-- distance to the wall
						local distance = 13
						local xoff = distance*cos(mo.angle)
						local yoff = distance*sin(mo.angle)
								
						-- distance to espio's hand (with some randomization)
						local angle = mo.angle - ((i == 1) and ANGLE_270 or ANGLE_90)
						local distance = 12 + P_RandomRange(-8, 8)
						xoff = $ + distance*cos(angle)
						yoff = $ + distance*sin(angle)
						local zoff = FRACUNIT*((i == 1) and 6 or 3) + (P_RandomRange(-8, 8)*FRACUNIT)
								
						local dust = P_SpawnMobjFromMobj(mo, xoff, yoff, zoff, MT_SPINDUST)
						dust.state = (mo.eflags & MFE_UNDERWATER) and S_SPINDUST_BUBBLE1 or $
						dust.scale = $*2/3
						P_InstaThrust(dust, R_PointToAngle2(0, 0, mo.momx, mo.momy), FixedMul(FRACUNIT/4, -(FixedHypot(mo.momx, mo.momy))))
						dust.momz = FixedMul(FRACUNIT/4, mo.momz * -1)
						S_StartSound(mo, sfx_s3k7e)
					end
				end
				
				-- reset tics
				mo.mighty_wallkick_dusttics = mintics
			end
		end
	else
		-- reset frame info
		mo.mighty_wallkick_frame = 0
		mo.mighty_wallkick_frametics = mighty.WallKickMaxFrameTics
	end
end