-- everything for charmy in battle mod
local charmy = chaotix.charmy

local loaded = false
charmy.BattleSetup = function()
	if not (CBW_Battle) then return end	-- battle isnt loaded
	if (loaded) then return end	-- already setup for battle? dont run
	local B = CBW_Battle
	local S = B.SkinVars
	
	-- charmy's priority
	local function charmy_battleHoming(mo, other)
		-- :)
		return true
	end
	B.AddPriorityFunction("charmy_homing", charmy_battleHoming)
	
	local function charmy_battlePriority(p)
		local mo = p.mo
		
		-- fix up charmy's priority while charging, he should have none!
		if (mo.state == S_CHARMY_ASCENT_CHARGE)
		or (mo.state == S_CHARMY_STINGER_CHARGE)
			B.SetPriority(p, 0, 0, "can_damage", 1, 1, "attack")
			return
		end
		
		-- quick ascent!
		if (mo.charmy_ascent)
		and (mo.charmy_ascent_tics)
			B.SetPriority(p, 1, 1, "tails_fly", 1, 1, "quick ascent")
			return
		end
		
		-- stinger attacks!
		if (mo.charmy_stinger)
		and (mo.charmy_stinger_tics)
			local satk = (mo.charmy_stinger_maxed) and 2 or 1	-- fully charging will give us some priority!
			local sfunc = "knuckles_glide"
			local text = "stinger attack"
			
			-- change sfunc for dives (and the hit text!)
			if (mo.charmy_stinger_state >= 2)
				sfunc = "fang_tailbounce"
				text = "stinger dive attack"
			end
			
			-- append "homing" to homing attacks!
			if (mo.charmy_stinger_homing)
				sfunc = "charmy_homing"	-- this'll make every hit a sweetspot!
				text = "homing "..$
			end
			
			B.SetPriority(p, 1, 0, sfunc, satk, 1, text)
			return
		end
	end
	
	-- charmy's special
	local function atan(x) -- x is fixed_t
		return R_PointToAngle2(0,0,FRACUNIT,x)
	end

	local st_pollentoss = 1
	
	local function charmy_battleSpecial(mo, doaction)
		local p = mo.player
		local grounded = P_IsObjectOnGround(mo)
		
		-- ability names
		p.actiontext = "Pollen Mark"
		p.actionrings = charmy.PollenRingCost
		
		-- attempt to throw the pollen!
		if not (p.actionstate)
			if (doaction == 1)	-- just pressed it
				-- go to the charging state!
				p.actionstate = st_pollentoss
				p.actiontime = 0
				p.canguard = false
				B.PayRings(p)
				B.ApplyCooldown(p, charmy.PollenCooldown)
				
				-- reset player stuff
				p.pflags = $ & ~(PF_JUMPED|PF_SPINNING|PF_STARTDASH)
				
				-- set state + frame
				mo.state = S_CHARMY_POLLENTOSS
				p.drawangle = mo.angle
			end
		elseif (p.actionstate == st_pollentoss)
			-- give stasis while throwing
			p.pflags = $|(PF_FULLSTASIS)
			p.actiontime = $ + 1
			
			-- set state + frame
			local frametics = charmy.PollenTossFrameTics
			local tics = 1
			local frame = 0
			for i = 1, #frametics
				local ticsadd = frametics[i]
				tics = $ + ticsadd
				if (p.actiontime < tics) then break end
				
				frame = $+1
			end
			
			if (mo.state != S_CHARMY_POLLENTOSS) then mo.state = S_CHARMY_POLLENTOSS end
			chaotix.SetFrame(mo, frame)
			
			-- throw out the pollen at a certain point!
			local tosstics = 1 + frametics[1] + frametics[2]
			if (p.actiontime <= tosstics)
				local modifier = charmy.PollenTossModifier
				p.drawangle = mo.angle	-- during the startup, follow player angle
				mo.momx = FixedMul(modifier, $)
				mo.momy = FixedMul(modifier, $)
				mo.momz = FixedMul(modifier, $)
				
				-- toss it out!
				if (p.actiontime == tosstics)
					-- some homing calculation
					local closestmo = nil
		
					local twod = (twodlevel or mo.flags2 & MF2_TWOD)
					local ringdist, span
					if not(twod)
						ringdist = RING_DIST*2
						span = ANG30
					else
						ringdist = RING_DIST
						span = ANG20
					end

					local maxdist = FixedMul(ringdist, mo.scale)
					local closestdist = 0
					searchBlockmap("objects", function(ref, found)
						if (ref == found) then return end
						if (found.flags & MF_NOCLIPTHING) then return end
						if (found.health <= 0) then return end -- dead
							
						-- check if we can home in on a player
						if not (found.player) then return end	-- ONLY home in on players!
						if (found.flags2 & MF2_FRET) then return end
						if (mo.type == MT_TARGETDUMMY) then return end //We won't be relying on this for player detection
						if (found.player and (CBW_Battle.MyTeam(ref.player, found.player) or found.player.spectator)) then return end //Disallow targeting teammates
						if (found.player and found.player.intangible) then return end//Disallow targeting air dodge

						//Do angle/distance checks
						local zdist = (ref.z + ref.height/2) - (found.z + found.height/2)
						local dist = P_AproxDistance(ref.x - found.x, ref.y - found.y)
						//CBW: 	Made the angle checks their own locals, for readability purposes
						//		I also unsigned the angle checks, which appears to correct failed OutOfBounds checks above the player.
						local xyz_angle = abs(R_PointToAngle2(0, 0, dist, zdist))
						local xy_angle = abs(R_PointToAngle2(
								ref.x + P_ReturnThrustX(ref, ref.angle, ref.radius),
								ref.y + P_ReturnThrustY(ref, ref.angle, ref.radius),
								found.x, found.y
							) - ref.angle)
								
								
						dist = P_AproxDistance(dist, zdist)
						if (dist > maxdist)
							return -- out of range
						end
					-- 	print("\x82!!!VERT","# Span check: "..xyz_angle/ANG1,"# Limit: "..span*2/ANG1,"#Within Range: "..tostring(abs(xyz_angle) <= span))
						if (xyz_angle > span)
							return -- Don't home outside of desired angle!
						end


						if (twod
						and abs(ref.y - found.y) > ref.radius)
							return -- not in your 2d plane
						end

						if ((closestmo and closestmo.valid) and (dist > closestdist))
							return
						end
					-- 	print("\x82!!!HORZ","# Span check: "..xy_angle/ANG1,"# Limit: "..span*2/ANG1,"#Within Range: "..tostring(abs(xy_angle) <= span))
						if (xy_angle > span)
							return -- behind back
						end

						if not (P_CheckSight(ref, found))
							return -- out of sight
						end

						closestmo = found
						closestdist = dist
					end, mo, mo.x - maxdist, mo.x + maxdist, mo.y - maxdist, mo.y + maxdist)
					
					-- the projectile
					local pollen = P_SpawnMobjFromMobj(mo, 0, 0, (mo.height / 2), MT_CHARMY_POLLEN)
					pollen.target = mo
					pollen.color = (G_GametypeHasTeams()) and mo.color or SKINCOLOR_GOLD
					pollen.momx, pollen.momy = mo.momx, mo.momy
					P_Thrust(pollen, mo.angle, FixedMul(mo.scale, chaotix.WaterMom(mo, charmy.PollenTossSpeed)))
					S_StartSound(mo, sfx_vcclap)
					
					-- redirect momentum for homing!
					if (closestmo and closestmo.valid)
						local speed = FixedHypot(pollen.momz, FixedHypot(pollen.momx, pollen.momy))
						local dist = R_PointToDist2(mo.x, mo.y, closestmo.x, closestmo.y)
						local zdiff = closestmo.z - mo.z   
						
						local angle = R_PointToAngle2(pollen.x, pollen.y, closestmo.x, closestmo.y)
						local zangle = atan(FixedDiv(zdiff, dist))
							
						P_InstaThrust(pollen, angle, FixedMul(speed, cos(zangle)))
						P_SetObjectMomZ(pollen, FixedMul(speed, sin(zangle)))
					end
					
					-- recoil
					P_Thrust(mo, mo.angle, FixedMul(mo.scale, chaotix.WaterMom(mo, charmy.PollenTossXYRecoil)))
					P_SetObjectMomZ(mo, chaotix.WaterMom(mo, charmy.PollenTossZRecoil))
				end
			end
			
			-- end after a set duration
			if (p.actiontime == charmy.PollenTossTics)
				p.pflags = $ & ~(PF_FULLSTASIS)
				if (P_IsObjectOnGround(mo)) then mo.state = S_PLAY_RUN end
				
				-- reset action state
				local resetflags = not P_IsObjectOnGround(mo)
				p.actionstate = 0
				B.ResetPlayerProperties(p, false, false)
			end
		end
	end
	
	-- charmy's exhaust (VERY important)
	local charmy_battleExhaust = function(p)
		local mo = p.mo
		
		-- slowfall 
		if (mo.state == S_PLAY_FALL)	-- slowfall activates with falling frames
		and ((mo.momz * P_MobjFlip(mo)) < 0)	-- only when descending!
			local maxtime = charmy.SlowFallExhaustTime
			p.exhaustmeter = max(0, $-(FRACUNIT/maxtime))
		end
		
		-- exhaust while charging
		if (mo.state == S_CHARMY_ASCENT_CHARGE)	-- quick ascent charge
		or (mo.state == S_CHARMY_STINGER_CHARGE and mo.charmy_stinger_state >= 2)	-- stinger air charge
			local maxtime = charmy.AirChargeExhaustTime
			p.exhaustmeter = max(0, $-(FRACUNIT/maxtime))
		end
	end
	
	-- setup espio's skinvars
	S[charmy.SKIN] = {
		weight = 60,
		shields = 2,
		guard_frame = 1,
		special = charmy_battleSpecial,
		func_priority_ext = charmy_battlePriority,
		func_exhaust = charmy_battleExhaust,
	}
	
	loaded = true
end

-- handle marked players
charmy.MarkPlayer = function(p, source)
	local mo = p.mo
	if not ((mo and mo.valid) and mo.health) then return end
	
	mo.charmy_marked = charmy.PollenMarkTics	-- mark them for death, they will be BRUTALISED
	mo.charmy_markcolour = (G_GametypeHasTeams()) and source.color or SKINCOLOR_GOLD
	S_StartSound(mo, sfx_polm)
end

local function charmy_ghostCopy(ghost, mo)
	-- these need to be set directly
	ghost.target = mo
	ghost.tracer = mo
	ghost.fuse = 2
	ghost.scale = mo.scale
	ghost.destscale = mo.scale
	ghost.flags2 = (mo.flags2 & MF2_SPLAT) | MF2_LINKDRAW
	
	-- the rest can be copied pretty easily
	local fields = {
		"skin",
		"rollangle",
		"state", "tics",
		"sprite", "sprite2", "frame",
		"renderflags", "blendmode",
		"spritexscale", "spriteyscale", "spritexoffset", "spriteyoffset"
	}
	for _, field in ipairs(fields)
		ghost[field] = mo[field]
	end
	
	-- now move to our source object
	local health = mo.health
	mo.health = 1
	A_CapeChase(ghost)
	mo.health = health
end

charmy.PollenMark = function(p)
	for p in players.iterate
		local mo = p.mo
		if not (mo and mo.valid) then continue end
		
		-- the ghost we'll use for the mark overlay
		local ghost = mo.charmy_markghost
	
		if not (mo.charmy_marked)
			-- remove the ghost overlay
			if (ghost and ghost.valid) then P_RemoveMobj(ghost) end
			
			-- might as well just remove these
			mo.charmy_marked = nil
			mo.charmy_markghost = nil
			mo.charmy_markcolour = nil
			continue 
		end
		mo.charmy_marked = $ and $-1
		
		-- overlay the player to tint them
		if not (ghost and ghost.valid)
			ghost = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_GHOST)
			mo.charmy_markghost = ghost	-- store it
		end
		charmy_ghostCopy(ghost, mo)
		
		ghost.color = mo.charmy_markcolour
		ghost.colorized = true
		ghost.blendmode = AST_ADD
		
		local trans = 10 - min(charmy.MarkedOverlayTrans, (mo.charmy_marked * charmy.MarkedOverlayTrans) / charmy.MarkedOverlayFade)
		if (trans == charmy.MarkedOverlayTrans)
			-- pulsate when not fading
			local pulsate_tics = charmy.MarkedOverlayPulsateTics
			local pulsate_angle = (((mo.charmy_marked % pulsate_tics) * 360) / pulsate_tics) * ANG1
			local pulsate_trans = (charmy.MarkedOverlayPulsateAdd * cos(pulsate_angle)) / FRACUNIT
			trans = max(1, min(9, $ + pulsate_trans))
		end
		
		if (trans >= 1 and trans <= 9)
			ghost.frame = ($ & ~FF_TRANSMASK) | (trans << FF_TRANSSHIFT)
		else
			ghost.flags2 = $|(MF2_DONTDRAW)
		end
		
		-- particles
		if not (mo.charmy_marked % charmy.MarkedParticleFrequency)
			local radius = mo.radius >> FRACBITS
			local xoff = P_RandomRange(-(radius), radius) * FRACUNIT
			local yoff = P_RandomRange(-(radius), radius) * FRACUNIT
			local zoff = P_RandomRange(0, mo.height >> FRACBITS) * FRACUNIT
		
			local pollen = P_SpawnMobjFromMobj(mo, xoff, yoff, zoff, MT_CHARMY_POLLEN_PARTICLE)
			pollen.color = mo.charmy_markcolour
			pollen.fuse = TICRATE*3
			pollen.charmy_pollen_tics = 0
		end
	end
end