-- charmy's ability, quick ascent
local charmy = chaotix.charmy

charmy.AbilitySpecial = function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	if not (charmy.AbilityCheck(p, 1)) then return end
	
	-- charmy state checks
	if (mo.charmy_ascent_count >= charmy.QuickAscentCount) then return end	-- out of charges
	if (mo.charmy_ascent) then return end
	if (mo.state == S_CHARMY_STINGER_CHARGE) then return end
	
	-- other checks
	if (p.gotcrystal) then return end	-- battlemod support
	if (CBW_Battle and p.exhaustmeter == 0) then return end	-- we're exhausted (battlemod)
	
	-- start the charge!
	local charges = mo.charmy_ascent_count
	
	local moving = abs(p.cmd.forwardmove) > 35 or abs(p.cmd.sidemove) > 35
	p.pflags = $ & ~(PF_STARTJUMP)
	mo.state = S_CHARMY_ASCENT_CHARGE
	mo.sprite2 = moving and SPR2_GLID or SPR2_FLY_
	mo.charmy_ascent = true
	mo.charmy_ascent_direction = moving and 1 or 0	-- lock direction
	mo.charmy_ascent_charge = 1
	
	-- frame related stuff (shouldn't need to be reset)
	mo.charmy_ascent_vibration = charmy.QuickAscentChargeVibrationTics
	mo.charmy_ascent_frametics = 0
	
	-- increase count
	if not (p.powers[pw_super])
		mo.charmy_ascent_count = $ and $+1 or 1
	end
	mo.charmy_ascent_leveltime = leveltime	-- just for the sweat
	S_StartSound(mo, charmy.QuickAscentChargeSounds[max(1, mo.charmy_ascent_count)])
end

-- reset stuff related to quick ascent
charmy.ResetQuickAscent = function(mo)
	mo.charmy_ascent = nil
	mo.charmy_ascent_direction = nil
	mo.charmy_ascent_charge = nil
	mo.charmy_ascent_maxed = nil
	mo.charmy_ascent_angle = nil
	mo.charmy_ascent_tics = nil
	mo.charmy_ascent_ghosttics = nil
	mo.charmy_ascent_leveltime = nil
end

-- main quick ascent function
charmy.QuickAscent = function(p)
	local mo = p.mo
	if not ((mo and mo.valid) and mo.health) then return end
	
	-- ability check
	if not (charmy.AbilityCheck(p, 1))
		charmy.ResetQuickAscent(mo)
		return
	end
	
	-- if we're grounded, reset charges!
	if (mo.charmy_ascent_count == nil)
	or (P_IsObjectOnGround(mo))	-- we're considered landed
	or (p.powers[pw_super])	-- ALWAYS have max charges!
	or (p.powers[pw_carry] == 2) -- reset on tails carry
	or (p.powers[pw_carry] == 7) -- reset on rope hang
	or (p.powers[pw_carry] == 8) -- reset on mace swing
	or (p.powers[pw_carry] == 10) -- reset on rollout rock
		mo.charmy_ascent_count = 0
	end
	
	-- we're not even doing it
	if not (mo.charmy_ascent) then return end
	
	-- stop conditions for quick ascent
	if not (mo.state >= S_CHARMY_ASCENT_CHARGE and mo.state <= S_CHARMY_ASCENT)	-- we got knocked out of it
	or (p.gotflag or p.gotcrystal)	-- we got the flag (or the crystal in battlemod)
		-- reset to falling frames
		if (mo.state >= S_CHARMY_ASCENT_CHARGE and mo.state <= S_CHARMY_ASCENT)
			mo.state = S_PLAY_FALL
		end
		
		p.pflags = $ & ~(PF_STARTDASH|PF_SPINNING)
		charmy.ResetQuickAscent(mo)
		return
	end
	
	if (mo.state == S_CHARMY_ASCENT_CHARGE)
		-- release when we let go of jump!
		if not (p.cmd.buttons & BT_JUMP)
			local charge = mo.charmy_ascent_charge
			local bounds = charmy.QuickAscentThrustBounds
			local range = bounds[2] - bounds[1]
			local thrust = chaotix.WaterMom(mo, bounds[1] + ((charge * range) / charmy.QuickAscentChargeTics))
			
			-- set state
			local direction = mo.charmy_ascent_direction
			mo.state = S_CHARMY_ASCENT
			mo.sprite2 = (direction) and SPR2_GLID or SPR2_FLY_
			
			-- thrust the player depending on direction
			local charges = mo.charmy_ascent_count
			p.pflags = $ & ~(PF_STARTDASH|PF_SPINNING)
			S_StartSound(mo, charmy.QuickAscentThrustSounds[max(1, mo.charmy_ascent_count)])
			for i = 1, 3 do S_StopSoundByID(mo, charmy.QuickAscentChargeSounds[i]) end
			
			if (direction)	-- diagonal
				-- trig will make this a bit weaker, so give it a boost
				thrust = FixedMul(charmy.QuickAscentThrustDiagonalBoost, $)
				
				-- apply the thrust
				local angle = R_PointToAngle2(0, 0, mo.momx, mo.momy) - mo.angle
				local anglefactor = cos(angle)
				local anglespeed = FixedMul(anglefactor, FixedHypot(mo.momx, mo.momy))
				local xyspeed = FixedMul(mo.scale, FixedMul(thrust, cos(ANGLE_45)))
				
				P_InstaThrust(mo, mo.angle, max(anglespeed, xyspeed))	-- use current speed if we're going fast enough!
				P_SetObjectMomZ(mo, FixedMul(thrust, sin(ANGLE_45)))
				
				-- angle will be modified for diagonal ones
				p.drawangle = mo.angle
				mo.charmy_ascent_angle = mo.angle
			else	-- straight up
				mo.momx = FixedMul(charmy.QuickAscentThrustZMultiplier, $)
				mo.momy = FixedMul(charmy.QuickAscentThrustZMultiplier, $)
				P_SetObjectMomZ(mo, thrust)
			end
			
			-- set tics depending on z momentum
			local tics = charmy.QuickAscentTicConversion[1]
			local fracs = charmy.QuickAscentTicConversion[2]
			
			mo.charmy_ascent_tics = ((abs(FixedDiv(mo.momz, mo.scale))* tics) / fracs)
			mo.charmy_ascent_ghosttics = mo.charmy_ascent_tics/2
			if (mo.eflags & MFE_UNDERWATER) then mo.charmy_ascent_tics = $*2 end
			return
		end
		
		-- charge up!
		p.pflags = $|(PF_STARTDASH|PF_SPINNING)
		if not (CBW_Battle and p.exhaustmeter == 0)	-- we're exhausted (battlemod)
			mo.momx = FixedMul(charmy.QuickAscentChargeXYMultiplier, $)
			mo.momy = FixedMul(charmy.QuickAscentChargeXYMultiplier, $)
			mo.momz = FixedMul(charmy.QuickAscentChargeZMultiplier, $)
		end
		
		mo.charmy_ascent_charge = min(charmy.QuickAscentChargeTics, $+1)
		if (mo.charmy_ascent_charge == charmy.QuickAscentChargeTics)
		and not (mo.charmy_ascent_maxed)	-- havent already done it
			mo.charmy_ascent_maxed = true
			S_StartSound(mo, sfx_s3ka2)
		end
		
		-- frame handling
		if (mo.charmy_ascent_direction) then p.drawangle = mo.angle end
		
		local charge = mo.charmy_ascent_charge
		local bounds = charmy.QuickAscentChargeVibrationStartup
		local range = bounds[1] - bounds[2]
		local maxtics = bounds[1] - ((charge * range) / charmy.QuickAscentChargeTics)
		
		local sprite2 = (mo.charmy_ascent_direction) and SPR2_GLID or SPR2_FLY_
		if (mo.charmy_ascent_vibration)
			-- use the vibration frame!
			chaotix.SetFrame(mo, 1, sprite2)
			mo.charmy_ascent_vibration = $-1
			
			-- go back to the first frame
			if not (mo.charmy_ascent_vibration)
				chaotix.SetFrame(mo, 0, sprite2)
			end
		else
			chaotix.SetFrame(mo, 0, sprite2)
			mo.charmy_ascent_frametics = min(maxtics, $ or maxtics)	-- make sure this exists
			
			-- decrease frametics
			mo.charmy_ascent_frametics = $-1
			if not (mo.charmy_ascent_frametics)
				chaotix.SetFrame(mo, 1, sprite2)
				mo.charmy_ascent_vibration = charmy.QuickAscentChargeVibrationTics
			end
		end
		
		-- spawn sweat if it's our last dash!
		if (mo.charmy_ascent_count == charmy.QuickAscentCount)
			mo.charmy_ascent_leveltime = $ or leveltime
			local ascent_leveltime = leveltime - mo.charmy_ascent_leveltime
			
			if not (ascent_leveltime % charmy.QuickAscentSweatFrequency)
				local sweat = P_SpawnMobjFromMobj(mo, 0, 0, mo.height, MT_CHARMY_SWEAT)
				sweat.target = mo	-- so it can track us
				S_StartSound(mo, sfx_s3k6d)
			end
		end
	elseif (mo.state == S_CHARMY_ASCENT)	-- practically an endlag state
		-- adjust how we're drawn
		if (mo.charmy_ascent_direction)
			if (FixedHypot(mo.momx, mo.momy))
				mo.charmy_ascent_angle = R_PointToAngle2(0, 0, mo.momx, mo.momy)
			end
			
			-- drawangle should be the momentum angle!
			p.drawangle = mo.charmy_ascent_angle
		end
		chaotix.SetFrame(mo, 0, (mo.charmy_ascent_direction) and SPR2_GLID or SPR2_FLY_)
		
		-- ghosts for good effect
		if (mo.charmy_ascent_ghosttics)
		and (mo.charmy_ascent_ghosttics % 2)
			chaotix.SpawnSpeedTrail(mo, true, {scale = mo.scale/2})
			local g = P_SpawnGhostMobj(mo)
			g.tics = 3
		end
		mo.charmy_ascent_ghosttics = $ and $-1
				
		-- natural end
		mo.charmy_ascent_tics = $ and $-1
		if not (mo.charmy_ascent_tics)
			mo.state = S_PLAY_FALL
			charmy.ResetQuickAscent(mo)
		end
	end
end