-- functions to start the stinger charge
local charmy = chaotix.charmy

charmy.SpinSpecial = function(p)
	if (p.pflags & PF_SPINDOWN) then return end	-- make sure it's not continuous
	
	local mo = p.mo
	if not (mo and mo.valid) then return end
	if not (charmy.AbilityCheck(p, 2)) then return end
	if (p.exiting) then return end
	
	-- charmy state checks
	if (mo.charmy_stinger) then return end	-- easy state check
	if (mo.state == S_CHARMY_ASCENT_CHARGE) then return end
	
	-- insta-unstuck
	if (mo.charmy_stinger_stuck and mo.charmy_stinger_stuckline)
		mo.charmy_stinger_stuck = -1
		return
	end
	
	-- other checks
	if not (P_IsObjectOnGround(mo)) then return end

	-- start the charge!
	p.pflags = $ & ~(PF_STARTJUMP)
	mo.state = S_CHARMY_STINGER_CHARGE
	mo.charmy_stinger = true
	mo.charmy_stinger_state = 1
	mo.charmy_stinger_charge = 1
	S_StartSound(mo, sfx_s3k8e)
	
	-- frame related stuff (shouldn't need to be reset)
	mo.charmy_stinger_vibration = charmy.StingerChargeVibrationTics
	mo.charmy_stinger_frametics = 0
end
charmy.JumpSpinSpecial = function(p)
	if (p.pflags & PF_SPINDOWN) then return end	-- make sure it's not continuous
	
	local mo = p.mo
	if not (mo and mo.valid) then return end
	if not (charmy.AbilityCheck(p, 2)) then return end
	if (p.exiting) then return end
	
	-- charmy state checks
	if (mo.charmy_stinger) then return end
	if (mo.charmy_stinger_stuck) then return end
	if (mo.state == S_CHARMY_ASCENT_CHARGE) then return end
	
	-- start the charge!
	local moving = abs(p.cmd.forwardmove) > 35 or abs(p.cmd.sidemove) > 35
	p.pflags = $ & ~(PF_STARTJUMP)
	mo.state = S_CHARMY_STINGER_CHARGE
	mo.sprite2 = moving and SPR2_TWIN or SPR2_MLEL
	mo.charmy_stinger = true
	mo.charmy_stinger_state = moving and 3 or 2	-- lock direction
	mo.charmy_stinger_charge = 1
	S_StartSound(mo, sfx_s3k8e)
	
	-- frame related stuff (shouldn't need to be reset)
	mo.charmy_stinger_vibration = charmy.StingerChargeVibrationTics
	mo.charmy_stinger_frametics = 0
end

-- reset stuff related to the stinger
charmy.ResetStinger = function(mo)
	mo.charmy_stinger = nil
	mo.charmy_stinger_state = nil
	mo.charmy_stinger_charge = nil
	mo.charmy_stinger_maxed = nil
	mo.charmy_stinger_tics = nil
	mo.charmy_stinger_ghosttics = nil
	mo.charmy_stinger_angle = nil
	mo.charmy_stinger_target = nil
	mo.charmy_stinger_homing = nil
	mo.charmy_stinger_homingspeed = nil
end

-- main function for the stinger
local function atan(x) -- x is fixed_t
    return R_PointToAngle2(0,0,FRACUNIT,x)
end

charmy.Stinger = function(p)
	local mo = p.mo
	if not ((mo and mo.valid) and mo.health) then return end
	
	-- ability check
	if not (charmy.AbilityCheck(p, 2))
		charmy.ResetStinger(mo)
		return
	end
	
	-- special handling for getting stuck!
	if (mo.charmy_stinger_stuck)
		-- early end
		if (mo.state != S_CHARMY_STINGER_STUCK)
			mo.flags = $ & ~(MF_NOTHINK)
			mo.charmy_stinger_stuck = nil
			mo.charmy_stinger_stuckline = nil
			mo.charmy_stinger_stuckdust = nil
			S_StartSound(mo, sfx_s3k9a)
			return
		end
		
		-- stuck thinker
		if (mo.charmy_stinger_stuckline != nil)
			local line = mo.charmy_stinger_stuckline
			local whichside = P_PointOnLineSide(mo.x, mo.y, line)
			local stuckangle = R_PointToAngle2(line.v1.x, line.v1.y, line.v2.x, line.v2.y)
			stuckangle = $ + (ANGLE_90 * (whichside and -1 or 1))
			
			p.drawangle = stuckangle
			
			-- spawn dust particles (done here so it aligns with our momentum cutting...)
			if not (mo.charmy_stinger_stuckdust)
				local dusts = 6
				for i = 1, dusts
					local angle = (360/dusts*i)*ANG1
					local xythrust = FixedMul(mo.scale, FixedMul(cos(angle), FRACUNIT*2))
					local zthrust = FixedMul(sin(angle), FRACUNIT*2)
					local dustmobj = (mo.eflags & MFE_UNDERWATER) and MT_SMALLBUBBLE or MT_DUST
					
					local dust = P_SpawnMobj(mo.x, mo.y, mo.z + (mo.height/4), dustmobj)
					dust.angle = stuckangle
					P_Thrust(dust, dust.angle + ANGLE_90, xythrust)
					P_SetObjectMomZ(dust, zthrust)
				end
				
				S_StartSound(mo, sfx_s3k9e)
				mo.charmy_stinger_stuckdust = true
			end
		end
	
		p.pflags = $|(PF_JUMPDOWN)
		mo.flags = $|(MF_NOTHINK)
		mo.momx, mo.momy, mo.momz = 0, 0, 0
		
		-- decrease timer
		mo.charmy_stinger_stuck = $-1
		
		-- conditions to break out
		if (mo.charmy_stinger_stuck <= 0)	-- timer ran out
			-- get unstuck
			mo.flags = $ & ~(MF_NOTHINK)
			mo.state = S_PLAY_FALL
			S_StartSound(mo, sfx_s3k9a)
			
			if (mo.charmy_stinger_stuckline == nil)	-- grounded
				-- get some momentum, if we want it
				local angle = (p.cmd.angleturn<<16) + R_PointToAngle2(0, 0, p.cmd.forwardmove*FRACUNIT, -(p.cmd.sidemove*FRACUNIT))
				local strength = FixedHypot(p.cmd.forwardmove*FRACUNIT, -(p.cmd.sidemove*FRACUNIT)) / FRACUNIT	-- so it's 50 max
				local thrust = FixedMul(mo.scale, (strength * charmy.StingerStuckPopThrust) / 50)
				
				p.pflags = ($|PF_JUMPED|PF_NOJUMPDAMAGE) 
				P_Thrust(mo, angle, chaotix.WaterMom(mo, thrust))
				P_SetObjectMomZ(mo, chaotix.WaterMom(mo, charmy.StingerStuckPopUp))
			else
				if not (P_IsObjectOnGround(mo)) then p.pflags = $|(PF_JUMPED|PF_NOJUMPDAMAGE) end
				P_Thrust(mo, p.drawangle, chaotix.WaterMom(mo, -(charmy.StingerStuckPopThrust)))
			end
			
			-- reset stuck vars
			mo.charmy_stinger_stuck = nil
			mo.charmy_stinger_stuckline = nil
			mo.charmy_stinger_stuckdust = nil
		end
		
		-- don't do anything else
		return
	end
	
	-- make sure we're doing the dive
	if not (mo.charmy_stinger) then return end
	
	-- end conditions for the stinger
	if not (mo.state >= S_CHARMY_STINGER_CHARGE and mo.state <= S_CHARMY_STINGER)	-- we got knocked out of it
		-- reset to falling frames
		if (mo.state >= S_CHARMY_STINGER_CHARGE and mo.state <= S_CHARMY_STINGER)
			mo.state = S_PLAY_FALL
		end
		
		-- get stuck in the ground!
		if (mo.eflags & MFE_JUSTHITFLOOR)
		and (mo.charmy_stinger_tics)	-- we're thrusting
		and (mo.charmy_stinger_state > 1)	-- a downwards angle of some kind
		and not (P_PlayerInPain(p))	-- not hurt
		and not (p.powers[pw_carry])	-- not being carried
			p.pflags = ($|(PF_JUMPDOWN)) & ~(PF_SPINNING)
			mo.flags = $|(MF_NOTHINK)
			mo.state = S_CHARMY_STINGER_STUCK
			mo.momx, mo.momy, mo.momz = 0, 0, 0
			mo.charmy_stinger_stuck = charmy.StingerStuckTics
			S_StartSound(mo, sfx_s3k9e)
			
			-- spawn dust
			local dustmobj = (mo.eflags & MFE_UNDERWATER) and MT_SMALLBUBBLE or MT_DUST
			A_MultiShot(mo, (dustmobj << 16) + 6, -40)
		end
		
		charmy.ResetStinger(mo)
		return
	end
	
	if (mo.state == S_CHARMY_STINGER_CHARGE)
		local state = mo.charmy_stinger_state
		
		-- search for targets to home in on!
		local closestmo = nil
		
		if (state != 2)	-- downwards stingers have no homing!
		and (CBW_Battle)
			local twod = (twodlevel or mo.flags2 & MF2_TWOD)
			local ringdist, span
			if not(twod)
				ringdist = RING_DIST*2
				span = ANG30
			else
				ringdist = RING_DIST
				span = ANG20
			end

			local maxdist = FixedMul(ringdist, mo.scale)
			local closestdist = 0
			local nonenemiesdisregard = MF_SPRING
			searchBlockmap("objects", function(ref, found)
				if (ref == found) then return end
				if (found.flags & MF_NOCLIPTHING) then return end
				if (found.health <= 0) then return end -- dead
				
				-- check if we can home in on the player!
				if (found.flags2 & MF2_FRET) then return end
				if (mo.type == MT_TARGETDUMMY) then return end //We won't be relying on this for player detection
				if (found.player and (CBW_Battle.MyTeam(ref.player, found.player) or found.player.spectator)) then return end //Disallow targeting teammates
				if (found.player and found.player.intangible) then return end//Disallow targeting air dodge
				if not (found.charmy_marked) then return end	-- charmy needs to mark them first!

				//Do angle/distance checks
				local zdist = (ref.z + ref.height/2) - (found.z + found.height/2)
				local dist = P_AproxDistance(ref.x - found.x, ref.y - found.y)
				//CBW: 	Made the angle checks their own locals, for readability purposes
				//		I also unsigned the angle checks, which appears to correct failed OutOfBounds checks above the player.
				local xyz_angle = abs(R_PointToAngle2(0, 0, dist, zdist))
				local xy_angle = abs(R_PointToAngle2(
						ref.x + P_ReturnThrustX(ref, ref.angle, ref.radius),
						ref.y + P_ReturnThrustY(ref, ref.angle, ref.radius),
						found.x, found.y
					) - ref.angle)
					
					
				dist = P_AproxDistance(dist, zdist)
				if (dist > maxdist)
					return -- out of range
				end
		-- 		print("\x82!!!VERT","# Span check: "..xyz_angle/ANG1,"# Limit: "..span*2/ANG1,"#Within Range: "..tostring(abs(xyz_angle) <= span))
				if (xyz_angle > span)
					return -- Don't home outside of desired angle!
				end


				if (twod
				and abs(ref.y - found.y) > ref.radius)
					return -- not in your 2d plane
				end

				if ((closestmo and closestmo.valid) and (dist > closestdist))
					return
				end
		-- 		print("\x82!!!HORZ","# Span check: "..xy_angle/ANG1,"# Limit: "..span*2/ANG1,"#Within Range: "..tostring(abs(xy_angle) <= span))
				if (xy_angle > span)
					return -- behind back
				end

				if not (P_CheckSight(ref, found))
					return -- out of sight
				end

				closestmo = found
				closestdist = dist
			end, mo, mo.x - maxdist, mo.x + maxdist, mo.y - maxdist, mo.y + maxdist)
		end
		
		-- release the stinger!
		if not (p.cmd.buttons & BT_SPIN)
			-- unlike quick ascent, we'll use mindash/maxdash for values!
			local charge = mo.charmy_stinger_charge
			local range = p.maxdash - p.mindash
			local thrust = chaotix.WaterMom(mo, p.mindash + ((charge * range) / charmy.StingerChargeTics))
			
			-- set state
			mo.state = S_CHARMY_STINGER
			if (state == 1)
				mo.sprite2 = SPR2_MLEE
			else
				mo.sprite2 = (state == 2) and SPR2_MLEL or SPR2_TWIN 
			end
			
			-- thrust the player depending on direction
			p.pflags = ($|PF_SPINNING) & ~(PF_STARTDASH)
			S_StartSound(mo, sfx_cdfm35)
			S_StopSoundByID(mo, sfx_s3k8e)
			
			if (state == 1)	-- forwards thrust
				P_InstaThrust(mo, mo.angle, thrust)
				
				-- store angle
				p.drawangle = mo.angle
				mo.charmy_stinger_angle = mo.angle
			else
				if (state == 2)	-- straight down
					thrust = FixedMul(charmy.StingerDownThrustModifier, $)	-- make it less powerful..
					
					mo.momx = FixedMul(charmy.StingerThrustZMultiplier, $)
					mo.momy = FixedMul(charmy.StingerThrustZMultiplier, $)
					P_SetObjectMomZ(mo, -(thrust))
				else	-- diagonal
					local diagonalangle = charmy.StingerThrustDiagonalAngle
					
					-- trig, everyone's fav
					local angle = R_PointToAngle2(0, 0, mo.momx, mo.momy) - mo.angle
					local anglefactor = cos(angle)
					local anglespeed = FixedMul(anglefactor, FixedHypot(mo.momx, mo.momy))
					local xyspeed = FixedMul(mo.scale, FixedMul(thrust, cos(diagonalangle)))
					
					P_InstaThrust(mo, mo.angle, max(anglespeed, xyspeed))	-- use current speed if we're going fast enough!
					P_SetObjectMomZ(mo, FixedMul(-(thrust), sin(diagonalangle)))
					
					-- store angle
					p.drawangle = mo.angle
					mo.charmy_stinger_angle = mo.angle
				end
			end
			
			-- for homing
			mo.charmy_stinger_homingspeed = FixedHypot(mo.momz, FixedHypot(mo.momx, mo.momy))
			
			-- set tics depending on speed
			local tics = charmy.StingerTicConversion[1]
			local fracs = charmy.StingerTicConversion[2]
			
			mo.charmy_stinger_tics = ((abs(thrust) * tics) / fracs)
			mo.charmy_stinger_ghosttics = mo.charmy_stinger_tics/2
			if (mo.eflags & MFE_UNDERWATER) then mo.charmy_stinger_tics = $*2 end
			
			-- store homing target and tics!
			if (closestmo and closestmo.valid)
				mo.charmy_stinger_target = closestmo
				mo.charmy_stinger_homing = charmy.PollenHomingTics
			end
			
			return
		end
		
		-- spawn homing reticule!
		if (closestmo and closestmo.valid)
			P_SpawnLockOn(p, closestmo, mobjinfo[MT_LOCKON].spawnstate)
		end
		
		-- charge up!
		p.pflags = $|(PF_STARTDASH|PF_SPINNING)
		if (state >= 2)	-- air stall only
		and not (CBW_Battle and p.exhaustmeter == 0)	-- not exhausted (battlemod)
			mo.momx = FixedMul(charmy.StingerChargeXYMultiplier, $)
			mo.momy = FixedMul(charmy.StingerChargeXYMultiplier, $)
			mo.momz = FixedMul(charmy.StingerChargeZMultiplier, $)
		end
		
		mo.charmy_stinger_charge = min(charmy.StingerChargeTics, $+1)
		if (mo.charmy_stinger_charge == charmy.StingerChargeTics)
		and not (mo.charmy_stinger_maxed)	-- havent already done it
			mo.charmy_stinger_maxed = true
			S_StartSound(mo, sfx_s3ka2)
		end
		
		-- frame handling
		if (state != 2) then p.drawangle = mo.angle end	-- only dont do this on aerial downwards
		
		local sprite2 = SPR2_MLEE
		if (state >= 2) then sprite2 = (state == 2) and SPR2_MLEL or SPR2_TWIN end
		
		local charge = mo.charmy_stinger_charge
		local bounds = charmy.StingerChargeVibrationStartup
		local range = bounds[1] - bounds[2]
		local maxtics = bounds[1] - ((charge * range) / charmy.StingerChargeTics)
		
		if (mo.charmy_stinger_vibration)
			-- use the vibration frame!
			chaotix.SetFrame(mo, 1, sprite2)
			mo.charmy_stinger_vibration = $-1
			
			-- go back to the first frame
			if not (mo.charmy_stinger_vibration)
				chaotix.SetFrame(mo, 0, sprite2)
			end
		else
			chaotix.SetFrame(mo, 0, sprite2)
			mo.charmy_stinger_frametics = min(maxtics, $ or maxtics)	-- make sure this exists
			
			-- decrease frametics
			mo.charmy_stinger_frametics = $-1
			if not (mo.charmy_stinger_frametics)
				chaotix.SetFrame(mo, 1, sprite2)
				mo.charmy_stinger_vibration = charmy.StingerChargeVibrationTics
			end
		end
	elseif (mo.state == S_CHARMY_STINGER)
		-- adjust how we're drawn
		if (mo.charmy_stinger_angle != nil)
			if (FixedHypot(mo.momx, mo.momy))
				mo.charmy_stinger_angle = R_PointToAngle2(0, 0, mo.momx, mo.momy)
			end
			
			-- drawangle should be the momentum angle!
			p.drawangle = mo.charmy_stinger_angle
		end
		
		local state = mo.charmy_stinger_state
		local sprite2 = SPR2_MLEE
		if (state >= 2) then sprite2 = (state == 2) and SPR2_MLEL or SPR2_TWIN end
		chaotix.SetFrame(mo, 0, sprite2)
		
		-- ghosts for good effect
		if (mo.charmy_stinger_ghosttics)
		and (mo.charmy_stinger_ghosttics % 2)
			chaotix.SpawnSpeedTrail(mo, true, {scale = mo.scale/2})
			local g = P_SpawnGhostMobj(mo)
			g.tics = 3
		end
		mo.charmy_stinger_ghosttics = $ and $-1
		
		-- homing functionality!
		local target = mo.charmy_stinger_target
		if (target and target.valid)
			local angle = R_PointToAngle2(0, 0, mo.momx, mo.momy) - R_PointToAngle2(mo.x, mo.y, target.x, target.y)
			
			if (target.health)
			and not (target.player and P_PlayerInPain(target.player))
			and not (abs(angle) > ANGLE_90)	-- dont snap!
				-- redirect momentum!
				local speed = mo.charmy_stinger_homingspeed
				local angle = R_PointToAngle2(mo.x, mo.y, target.x, target.y)
				p.drawangle = angle
								
				if (state == 1)
					P_InstaThrust(mo, angle, speed)
				elseif (state == 3)
					local dist = R_PointToDist2(mo.x, mo.y, target.x, target.y)
					local zdiff = target.z - mo.z    
					local zangle = atan(FixedDiv(zdiff, dist))
						
					P_InstaThrust(mo, angle, FixedMul(speed, cos(zangle)))
					P_SetObjectMomZ(mo, FixedMul(speed, sin(zangle)))
				end
				
				P_SpawnLockOn(p, target, mobjinfo[MT_LOCKON].spawnstate)
				
				-- end condition(s)
				mo.charmy_stinger_homing = $ and $-1 or 0
				if not (mo.charmy_stinger_homing)	-- stop homing, now!
					mo.charmy_stinger_target = nil
					mo.charmy_stinger_homing = nil
					mo.charmy_stinger_homingspeed = nil
				end
			else
				-- reset stuff for safety
				mo.charmy_stinger_target = nil
				mo.charmy_stinger_homing = nil
				mo.charmy_stinger_homingspeed = nil
			end
		else
			-- reset stuff for safety
			mo.charmy_stinger_target = nil
			mo.charmy_stinger_homing = nil
			mo.charmy_stinger_homingspeed = nil
		end
				
		-- natural end
		mo.charmy_stinger_tics = $ and $-1
		if not (mo.charmy_stinger_tics)
			p.pflags = $ & ~(PF_SPINNING)
			mo.state = (P_IsObjectOnGround(mo)) and S_PLAY_RUN or S_PLAY_FALL
			charmy.ResetStinger(mo)
			
			-- get jump flags properly when not grounded
			if not (P_IsObjectOnGround(mo))
				local jumpflags = P_GetJumpFlags(p)
				p.pflags = $|jumpflags
			end
		end
	end
end

-- rebound off of things we damage properly!
charmy.StingerDiveDamage = function(mobj, mo)
	if not (mobj and mobj.valid) then return end
	if not ((mo and mo.valid) and mo.health) then return end
	local p = mo.player
	
	-- are we doing the stinger dive?
	if not (mo.charmy_stinger) then return end	-- no stinger activity
	if not (mo.charmy_stinger_state >= 2) then return end	-- not a dive
	if not (mo.charmy_stinger_tics) then return end	-- we're not thrusting
	
	p.pflags = $ & ~(PF_SPINNING)
	mo.state = S_PLAY_JUMP
	mo.charmy_ascent_count = min(charmy.QuickAscentCount - 1, $)	-- get at least one quick ascent back
	charmy.ResetStinger(mo)
end
chaotix.AddHook("MobjDamage", charmy.StingerDiveDamage)
chaotix.AddHook("MobjDeath", charmy.StingerDiveDamage)

-- get stuck to walls
local function charmy_checkLine(p, line)
	local mo = p.mo
	
	-- sector checks
	if not (line.backsector) then return true end	-- apparently in these situations source always returns true?
	local sec = (line.backsector and not P_PointOnLineSide(mo.x, mo.y, line)) and line.backsector or line.frontsector
	
	-- shouldnt be too hard to check this
	local checkz = mo.z
	if (mo.eflags & MFE_VERTICALFLIP) then checkz = mo.z + mo.height end
	
	-- make sure we cant climb thokbarriers
	local thokbarrier = (sec.ceilingpic:upper():find("F_SKY"))
	if (thokbarrier and (checkz >= sec.ceilingheight))
		return false
	end
	
	-- should be good?
	return true
end

chaotix.AddHook("MobjMoveBlocked", function(mo, thing, line)
	if not ((mo and mo.valid) and mo.health) then return end
	if not (line and line.valid) then return end
	
	-- stinger attack?
	if not (mo.charmy_stinger) then return end	-- no stinger activity
	if not (mo.charmy_stinger_state != 2) then return end	-- not straight down
	if not (mo.charmy_stinger_tics) then return end	-- we're not thrusting
	
	local p = mo.player
	if not (charmy_checkLine(p, line)) then return end
	
	-- stick into the wall!
	local whichside = P_PointOnLineSide(mo.x, mo.y, line)
	local stuckangle = R_PointToAngle2(line.v1.x, line.v1.y, line.v2.x, line.v2.y)
	stuckangle = $ + (ANGLE_90 * (whichside and -1 or 1))
		
	p.drawangle = stuckangle
	p.pflags = ($|(PF_JUMPDOWN)) & ~(PF_SPINNING)
	mo.state = S_CHARMY_STINGER_STUCK
	mo.sprite2 = SPR2_CLNG
	mo.charmy_stinger_stuck = charmy.StingerStuckTics
	mo.charmy_stinger_stuckline = line
	charmy.ResetStinger(mo)
end, MT_PLAYER)