-- everything for vector in battle mod
local vector = chaotix.vector

-- loading vector's stuff
local loaded = false
vector.BattleSetup = function()
	if not (CBW_Battle) then return end	-- battle isnt loaded
	if (loaded) then return end	-- already setup for battle? dont run
	local B = CBW_Battle
	local S = B.SkinVars
	local CV = B.Console
	
	-- vector's priority
	local function vector_battlePriority(p)
		local mo = p.mo
		
		-- failsafe for encore
		if (p.actionstate)
			B.SetPriority(p, 0, 0, "can_damage", 0, 0, "attack")
			return
		end
		
		-- beat dash
		if (mo.vector_beatbash)
			local satk = (mo.vector_beatbash_timing) and 2 or 1
			B.SetPriority(p, 1, 1, "knuckles_glide", satk, 1, "beat bash")
			return
		end
		
		-- fix up vector's priority when alt spindashing
		if (mo.vector_riptire_spindash)
			B.SetPriority(p, 1, 1, "can_damage", 1, 1, "spin attack")
			return
		end
	end
	
	-- vector's battle special
	local function angtoint(a)
		if not a return 0 end
		a = $/ANG1
		if a < 0
			a = $+360
		end
		return a
	end
	
	local st_startup = 1
	local st_charging = 2
	local st_scream = 3

	local function vector_stopChargeSound(mo)
		for level = 0, vector.EncoreMaxLevel
			local charge_sound = sfx_vecch1 + level
			S_StopSoundByID(mo, charge_sound)
		end
	end 

	local function vector_startScream(mo)
		local p = mo.player
		
		-- set actionstate + actiontime
		p.actionstate = st_scream
		p.actiontime = 0
		mo.state = P_IsObjectOnGround(mo) and S_VECTOR_ENCORE_SCREAM or S_VECTOR_ENCORE_SCREAM_AIR
		chaotix.SetFrame(mo, 3)
		
		-- store angle
		p.drawangle = mo.angle
		mo.vector_battleangle = mo.angle	 -- preserve this

		--oh waiter play my sound please
		local screech_sound = sfx_vecl0 + mo.vector_battlelevel
		S_StartSound(mo, screech_sound)
		
		-- this'll be good for fixed-point stuff
		local fracModifier = (mo.vector_battlelevel * FRACUNIT) / vector.EncoreMaxLevel
		
		-- recoil
		local recoil_range = vector.EncoreScreamMaxRecoil - vector.EncoreScreamMinRecoil
		local recoil = vector.EncoreScreamMinRecoil + FixedMul(fracModifier, recoil_range)
		P_Thrust(mo, mo.angle, FixedMul(mo.scale, -(recoil)))
		
		-- spawn the object that'll blast players away!
		vector.SpawnEncoreScream(mo, fracModifier)
		mo.vector_battlescream = true	-- this'll be useful for resetting stuff
						
		-- apply cooldown based on charge level
		local cooldown_range = vector.EncoreMaxCooldown - vector.EncoreMinCooldown
		local cooldown = vector.EncoreMinCooldown + ((mo.vector_battlelevel * cooldown_range) / vector.EncoreMaxLevel)
		B.ApplyCooldown(p, cooldown)
	end
	
	local function vector_battleSpecial(mo, doaction)
		local p = mo.player
		
		-- init vector's charge
		mo.vector_battlecharge = $ or 0
		mo.vector_battlelevel = $ or 0
		
		-- some stuff depends on not being in this state
		if (p.actionstate != st_scream)
			if (mo.vector_battlescream)
				-- make SURE this is reset
				mo.vector_battlecharge = 0
				mo.vector_battlelevel = 0
				mo.vector_battlescream = false
			end
			
			-- trigger the visual flash
			if (mo.vector_battlelevel)
				local level, maxlevel = mo.vector_battlelevel - 1, vector.EncoreMaxLevel - 1
				local frequency_range = vector.EncoreFlashMaxFrequency - vector.EncoreFlashMinFrequency
				local frequency = vector.EncoreFlashMinFrequency + ((level * frequency_range) / maxlevel)
				
				mo.vector_battleflashtics = $ and min(frequency, $) - 1 or 0
				if not (mo.vector_battleflashtics)
					mo.vector_battleflash = vector.EncoreFlashTics
					
					-- reset this
					mo.vector_battleflashtics = frequency
				end
			else
				-- reset this
				mo.vector_battleflashtics = 0
			end
		end
		
		-- show the visual flash
		if (mo.vector_battleflash)
			local maxtrans = vector.EncoreFlashMaxTrans
			local trans = 10 - ((mo.vector_battleflash * maxtrans) / vector.EncoreFlashTics)
				
			if (trans >= 0 and trans <= 9)
				local gmo = mo
				while (gmo and gmo.valid)
					local g = P_SpawnGhostMobj(gmo)
					g.frame = ($ & ~FF_TRANSMASK) | (trans << FF_TRANSSHIFT)
					g.color, g.colorized = SKINCOLOR_SUPERGOLD5, true
					g.blendmode = AST_ADD
					g.tics = 1
						
					-- attempt to spawn for a followmobj as well?
					gmo = g.tracer
				end
			end
				
			mo.vector_battleflash = $-1
		end

		-- battle hud handling
		if (p.actionstate < st_charging)
			p.actiontext = "Charge"
			p.actionrings = 0
				
			-- it'd be released instantly in this case
			if (mo.vector_battlelevel == 3)
				p.actiontext = "Release"
				p.actionrings = vector.EncoreCost
			end
		else
			-- show the new action
			p.actiontext = "Release"
			p.actionrings = vector.EncoreCost
			p.actiontextflags = (p.rings < p.actionrings) and 1 or 0
				
			-- show the percent from leveling up!
			if not (mo.vector_battlelevel == vector.EncoreMaxLevel)
				local level_tics = (mo.vector_battlelevel * vector.EncoreMaxLevelTics) / vector.EncoreMaxLevel
				local nextLevel_tics = (((mo.vector_battlelevel + 1) * vector.EncoreMaxLevelTics) / vector.EncoreMaxLevel) - level_tics
				local charge_tics = mo.vector_battlecharge - level_tics
				
				local levelUp_tics = min(100, (charge_tics * 100) / nextLevel_tics)
				p.action2text = "Level Up "..levelUp_tics.."%"
				p.action2textflags = 0
			else
				p.action2text = "Level Maxed!"
				p.action2textflags = (leveltime&4 <= 2) and 2 or 0
			end
		end
		p.actiontext = $ +  " (Lv "..mo.vector_battlelevel..")"	-- append the level to the actiontext!
		
		-- action states
		if not (p.actionstate)
			if (doaction == 1)
				-- go to the charging state!
				p.actionstate = st_startup
				p.actiontime = 0
				B.PayRings(p)
				
				-- set player stufff
				mo.state = P_IsObjectOnGround(mo) and S_VECTOR_ENCORE_STARTUP or S_VECTOR_ENCORE_STARTUP_AIR
				chaotix.SetFrame(mo, 0)
				p.pflags = $ & ~(PF_JUMPED|PF_SPINNING|PF_STARTDASH)
			end
		else
			if (P_PlayerInPain(p))
				-- punish vector players for getting hit while in the actionstate
				p.actionstate = 0
				mo.vector_battlecharge = 0
				mo.vector_battlelevel = 0
				mo.vector_battlescream = false	-- also reset this here
			else
				-- affect all states
				p.pflags = $|(PF_FULLSTASIS)
				
				local speed = FixedDiv(FixedHypot(mo.momx, mo.momy), mo.scale)
				if (speed >= vector.EncoreDustThreshold)
				and (P_IsObjectOnGround(mo))
					local speed = 3
					local dust = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_SPINDUST)
					dust.scale = $/2
					dust.destscale = FRACUNIT*2
					dust.scalespeed = dust.scale/12
					dust.momx = mo.momx + (P_RandomRange(-(speed), speed) << FRACBITS)
					dust.momy = mo.momy + (P_RandomRange(-(speed), speed) << FRACBITS)
					P_SetObjectMomZ(dust, P_RandomRange(1, 6)*FRACUNIT)
					if (mo.eflags & (MFE_TOUCHWATER|MFE_UNDERWATER))
						dust.state = S_SPINDUST_BUBBLE1
					end
					if not (leveltime%6) then S_StartSound(mo, sfx_s3k7e) end
				end
				
				-- state specific thinkers
				if (p.actionstate == st_startup)
				or (p.actionstate == st_charging)
					-- turn to camera angle
					local diff = mo.angle - p.drawangle
					local factor = 8
					if (diff)
						diff = (angtoint(diff) > 180) and InvAngle(InvAngle(diff)/factor) or $/factor
						p.drawangle = $ + diff
					end
					
					-- if we're airborne, allow movement!
					if not (P_IsObjectOnGround(mo)) then p.pflags = $ & ~(PF_STASIS) end
					
					-- state specific thinkers (2)
					if (p.actionstate == st_startup)
						local state = P_IsObjectOnGround(mo) and S_VECTOR_ENCORE_STARTUP or S_VECTOR_ENCORE_STARTUP_AIR
						if (mo.state != state)
							mo.state = state
							chaotix.SetFrame(mo, 0)
						end
						p.actiontime = $+1
						
						-- play the whole animation out if we have max charge!
						local startup_tics = vector.EncoreStartupTics
						
						if (mo.vector_battlelevel == vector.EncoreMaxLevel)
							local frame = 0
							
							if (p.actiontime > (startup_tics*3/4))
								frame = 3
							elseif (p.actiontime > (startup_tics/2))
								frame = 2
							elseif (p.actiontime > (startup_tics/4))
								frame = 1
							end
							chaotix.SetFrame(mo, frame)
						end
						
						if (p.actiontime >= vector.EncoreStartupTics)
							-- go to a state depending on charge
							if (mo.vector_battlelevel != vector.EncoreMaxLevel)	-- not maxed out
								-- go to charging state
								p.actionstate = st_charging
								mo.state = S_VECTOR_ENCORE_CHARGE
								chaotix.SetFrame(mo, 1)
							else
								-- go to scream state
								vector_startScream(mo)
							end
						end
					elseif (p.actionstate == st_charging)
						-- set state
						local state = P_IsObjectOnGround(mo) and S_VECTOR_ENCORE_CHARGE or S_VECTOR_ENCORE_CHARGE_AIR
						if (mo.state != state)
							mo.state = state
						end
						
						-- set frame
						local frame = max(1, (mo.vector_battlelevel * 3) / vector.EncoreMaxLevel)
						chaotix.SetFrame(mo, frame)
						
						-- cancel out of the charge state!
						if (doaction == 1)	-- activate!
							-- go to scream state
							B.PayRings(p)	-- pay additional rings!
							vector_startScream(mo)
						elseif (B.PlayerButtonPressed(p, p.battleconfig_guard, false))	-- guard cancel!
							-- funnily enough, the jump cancel has a 1 tic window where the jump won't come out...
							p.actionstate = 0
							p.canguard = false
							B.ApplyCooldown(p, vector.EncoreCancelCooldown)
							
							local grounded = P_IsObjectOnGround(mo)
							local jumped = not grounded
							local thokked = nil
							if not (grounded) then thokked = false end
						
							B.ResetPlayerProperties(p, jumped, thokked)
							if (grounded) 
								local state = S_PLAY_STND
								if (p.speed) then state = (p.speed > p.runspeed) and S_PLAY_RUN or S_PLAY_WALK end
								mo.state = state
							end
						else
							-- no cancel conditions met, charge up!
							mo.vector_battlecharge = min(vector.EncoreMaxLevelTics, $+1)
							for level = 0, vector.EncoreMaxLevel
								if (mo.vector_battlelevel >= level) then continue end	-- remove unnecessary maths
								
								-- see if we've reached the level!
								local levelTics = (level * vector.EncoreMaxLevelTics) / vector.EncoreMaxLevel
								if not (mo.vector_battlecharge >= levelTics) then break end
								mo.vector_battlelevel = level	-- set level!
								if (mo.vector_battlelevel == vector.EncoreMaxLevel)
									S_StartSound(mo, sfx_vecchf)
								end
							end
							for level = 0, vector.EncoreMaxLevel - 1
								local charge_sound = sfx_vecch1 + level 
								if (mo.vector_battlelevel == level) 
									if not (S_SoundPlaying(mo, charge_sound))
										S_StartSound(mo, charge_sound)
									end
								else
									if (S_SoundPlaying(mo, charge_sound))
										S_StopSoundByID(mo, charge_sound)
									end
								end
							end
						end
					end
				elseif (p.actionstate == st_scream)
					local state = P_IsObjectOnGround(mo) and S_VECTOR_ENCORE_SCREAM or S_VECTOR_ENCORE_SCREAM_AIR
					if (mo.state != state)
						mo.state = state
					end
					
					-- increase action time to end scream state
					p.actiontime = $+1
					
					-- set frame based on actiontime
					local frame = (p.actiontime > vector.EncoreSmearTics) and 5 or 4
					chaotix.SetFrame(mo, frame)
					
					local screamTics_range = vector.EncoreScreamEndlagMaxTics - vector.EncoreScreamEndlagMinTics
					local screamTics = vector.EncoreScreamEndlagMinTics + ((mo.vector_battlelevel * screamTics_range) / vector.EncoreMaxLevel)
					if (p.actiontime >= screamTics)
						-- reset charge + level
						mo.vector_battlecharge = 0
						mo.vector_battlelevel = 0
						mo.vector_battlescream = false	-- won't need this anymore
						
						-- reset state
						local grounded = P_IsObjectOnGround(mo)
						local jumped = not grounded
						local thokked = nil
						if not (grounded) then thokked = false end
						
						p.actionstate = 0
						B.ResetPlayerProperties(p, jumped, thokked)
						if (grounded) 
							local state = S_PLAY_STND
							if (p.speed) then state = (p.speed > p.runspeed) and S_PLAY_RUN or S_PLAY_WALK end
							mo.state = state
						end
					end
				end
			end
		end
		if (p.actionstate != st_charging)
			vector_stopChargeSound(mo)
		end
	end
	
	-- collision functions
	local function vector_battlePreCollide(s, t, p, mo, atk, def, weight, hurt, pain, ground, angle, thrust, thrust2, collisiontype)
		if not (p[s] and mo[s]) then return end
		
		-- laziness check for hurt value (so we don't have to use 2 different checks...)
		local newhurt = hurt
		if (abs(newhurt) == 1)
		and (s == 2)	-- s is actually t
			newhurt = $ * -1
		end
		
		-- make beat bash recoil off of damage correctly
		if (mo[s].vector_beatbash and newhurt == 1)	-- make sure it's only THEM getting damaged
		and not (p[t].guard > 0)	-- bruh... why do i even need to check this
			mo[s].vector_beatbash_battlerebound = true
		end
	end
	
	local function vector_battlePostCollide(s, t, p, mo, atk, def, weight, hurt, pain, ground, angle, thrust, thrust2, collisiontype)
		if not (p[s] and mo[s]) then return end
		if not (p[t] and mo[t]) then return end
		
		if (mo[s].vector_beatbash_battlerebound)
			-- set stuff based on timing
			local multiplier = FRACUNIT
			local state = S_PLAY_JUMP	-- default is rolling state
				
			if (mo[s].vector_beatbash_timing)	-- we had good timing
				local timing = mo[s].vector_beatbash_timing
				local maxtiming = vector.BeatBashTimingTics
				local range = vector.BeatBashTimingMaxMultiplier - FRACUNIT
				
				multiplier = FRACUNIT + ((timing * range) / maxtiming)
				state = S_VECTOR_BEATBASH_POSE
				p[s].pflags = $ & ~(PF_THOKKED)	-- get thokability back, always
				S_StartSound(mo[s], sfx_vcclap)
			else	-- lose a rebound!
				mo[s].vector_beatbash_rebounds = max(0, $-1)
				mo[s].vector_beatbash_pose = 0	-- reset pose
				
				if (mo[s].vector_beatbash_rebounds)
					-- we still have rebound, get thokability back!
					p[s].pflags = $ & ~(PF_THOKKED)
				else
					-- lose jumpability on our last messup
					state = S_PLAY_SPRING
					p[s].pflags = $ & ~(PF_JUMPED)
				end
			end
				
			-- rebound the player (consider the player a wall
			local playerangle = R_PointToAngle2(mo[s].x, mo[s].y, mo[t].x, mo[t].y)
			local hitangle = playerangle - mo[s].vector_beatbash_angle
			local reflectangle = playerangle + hitangle
				
			local anglefactor = abs(cos(hitangle))
			local range = FRACUNIT - vector.BeatBashReboundWallModifier
			local modifier = FRACUNIT - FixedMul(anglefactor, range)
				
			mo[s].state = state
			P_InstaThrust(mo[s], reflectangle, FixedMul(modifier, -(mo[s].vector_beatbash_speed)))
			P_SetObjectMomZ(mo[s], FixedMul(multiplier, chaotix.WaterMom(mo[s], vector.BeatBashReboundMomZ)))
			S_StartSound(mo[s], sfx_s3k7b)
			vector.SetBeatPose(mo[s], true)
			vector.ResetBeatBash(mo[s])
			
			-- remove this var
			mo[s].vector_beatbash_battlerebound = nil
		end
	end
	
	-- setup vector's skinvars
	S[vector.SKIN] = {
		weight = 130,
		special = vector_battleSpecial,
		func_priority_ext = vector_battlePriority,
		func_precollide = vector_battlePreCollide,
		func_postcollide = vector_battlePostCollide,
	}
	
	-- little bit of a hack to fix tumble getting rid of credit!
	B.UpdateCollisionHistory = function(pmo)
		//Needs comments
		//Initialize
		if pmo.pushtics == nil then pmo.pushtics = 0 end
		//If we've recently been pushed
		if pmo.pushed or pmo.pushed_last then
			//Update the timer
			if pmo.pushtics == 0 then
				pmo.pushtics = 1
			else
				pmo.pushtics = $+1
				//After a certain time, clean the slate
				if pmo.pushtics > CV.CollisionTimer.value then
					pmo.pushed_last = nil
					pmo.pushed = nil
					pmo.pushtics = 0
				end
			end
		else
			//No pusher, no need for pushtics
			if pmo.pushtics then
				pmo.pushtics = 0 //!! If this is reset just off of pushed being nil, and pushed is reset per frame, then when is this not 1 or 0??
			end
		end
		//Update other player object's pushed stats
		if pmo.pushed and pmo.pushed.valid and pmo.pushed.pushed == pmo and not(pmo.pushtics) then
			pmo.pushed.flags = $|MF_SPECIAL
			pmo.pushed.pushed_last = pmo
			pmo.pushed.pushed = nil
		end
		pmo.flags = $|MF_SPECIAL //Reallow collisions
		if pmo.pushed then
			pmo.pushed_last = pmo.pushed //History
			pmo.pushed = nil //Done with this
		end
		local player = pmo.player
		if player.pushed_creditplr != nil
			if player.pushed_creditplr.valid and P_IsObjectOnGround(pmo) and not(player.powers[pw_nocontrol] or player.powers[pw_flashing] or player.tumble) then	-- snu: also check for tumble!
				player.pushed_creditplr = nil
			end
		end
	end

	
	loaded = true
end