local vector = chaotix.vector

-- command for when the text shows up
vector.BeatBashCommandVar = CV_RegisterVar({
	name = vector.BeatBashCommand,
	defaultvalue = "Local",
	flags = CV_CALL,
	PossibleValue = {Off = 0, Local = 1, Global = 2},
	func = function(cv)
		if consoleplayer and consoleplayer.valid
			chaotix.WriteConfig(consoleplayer)	-- update config
		end
	end
})

-- start the beat bash!
vector.AbilitySpecial = function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	if not (vector.AbilityCheck(p, 1)) then return end
	
	if (p.pflags & PF_THOKKED) then return end	-- thokked state
	if (p.gotcrystal) then return end	-- battlemod support
	if (CBW_Battle and p.exhaustmeter == 0) then return end	-- we're exhausted (battlemod)
	
	p.pflags = ($|PF_THOKKED) & ~(PF_STARTJUMP|PF_NOJUMPDAMAGE|PF_SPINNING)
	mo.state = S_VECTOR_BEATBASH	-- set state
	mo.vector_beatbash = true
	mo.vector_beatbash_tics = vector.BeatBashTics
	mo.vector_beatbash_angle = mo.angle
	//mo.vector_beatbash_timing = vector.BeatBashTimingTics -- this'll be set in the thinker
	S_StartSound(mo, sfx_kc5b)
	
	-- transfer speed (affected by angle so we dont get fast 180 turns)
	local angle = R_PointToAngle2(0, 0, mo.momx, mo.momy) - mo.angle
	local anglefactor = abs(cos((chaotix.AngToInt(angle)/2) * ANG1))
	local anglespeed = FixedMul(anglefactor, FixedHypot(mo.momx, mo.momy))
	mo.vector_beatbash_speed = max(FixedMul(mo.scale, chaotix.WaterMom(mo, p.actionspd)), anglespeed)
	P_InstaThrust(mo, mo.angle, mo.vector_beatbash_speed)
end

-- reset stuff related to the beat bash
vector.ResetBeatBash = function(mo)
	mo.vector_beatbash = nil
	mo.vector_beatbash_tics = nil
	mo.vector_beatbash_angle = nil
	mo.vector_beatbash_wallangle = nil
	mo.vector_beatbash_line = nil
	mo.vector_beatbash_speed = nil
	mo.vector_beatbash_timing = nil
	mo.vector_beatbash_visual = nil
end

-- function to build poses for each character
vector.NumSkinPoses = {}
vector.BuildBeatBashPoses = function()
	for skin in skins.iterate
		if (vector.NumSkinPoses[skin.name]) then continue end	-- already done
		
		-- set up the dummy state
		local state = S_UNKNOWN
		local stateinfo = states[state]
		local statedef = {
			stateinfo.sprite,
			stateinfo.frame,
			stateinfo.tics,
			stateinfo.action,
			stateinfo.var1,
			stateinfo.var2,
			stateinfo.nextstate
		}
		stateinfo.sprite, stateinfo.frame = SPR_PLAY, SPR2_TAL0
		stateinfo.action = nil
		
		-- set up our dummy
		local numposes = -1
		local thok = P_SpawnMobj(0, 0, 0, MT_THOK)
		thok.skin = skin.name
		thok.state = state
		
		-- count poses
		while ((thok.frame & FF_FRAMEMASK) > numposes)
			numposes = (thok.frame & FF_FRAMEMASK)
			thok.state = state
		end
		
		-- register pose count
		numposes = $+1	-- turn 0 into 1
		vector.NumSkinPoses[skin.name] = numposes
		states[state] = statedef	-- restore this
		P_RemoveMobj(thok)
	end
end


-- beat bash function
vector.BeatBash = function(p)
	local mo = p.mo
	if not ((mo and mo.valid) and mo.health) then return end
	
	-- ability check
	if not (vector.AbilityCheck(p, 1))
		vector.ResetBeatBash(mo)
		return
	end
	
	-- if we're grounded, reset our status for getting a guaranteed perfect rebound!
	if (mo.vector_beatbash_rebounds == nil)
	or (P_IsObjectOnGround(mo) or p.climbing)
		mo.vector_beatbash_rebounds = 2
		mo.vector_beatbash_pose = 0
	end
	
	-- special handling for the pose
	if (mo.state == S_VECTOR_BEATBASH_POSE)
		mo.vector_beatbash_posetics = $ and $-1
		local tics = mo.vector_beatbash_posetics
		
		if not (tics)
			mo.state = S_PLAY_JUMP
			p.pflags = $ &~ PF_NOJUMPDAMAGE
		else
			-- spawn ghosts!
			local ghosttics = vector.BeatBashPoseTics*2/3
			if (tics > ghosttics)
			and not (tics%2)
				local g = P_SpawnGhostMobj(mo)
				g.tics = 3
			end
			
			-- spawn speed trail!
			local trailtics = vector.BeatBashPoseTics/2
			chaotix.SpawnSpeedTrail(mo, (tics > trailtics) and not (tics%2))
		end
		
		-- don't do anything else
		return
	end
	
	-- not doing to beat bash?? dont go forward
	if not (mo.vector_beatbash) then return end
	
	-- stop conditions for the beat bash
	if (mo.state != S_VECTOR_BEATBASH)	-- we got knocked out of it
	or (p.gotflag or p.gotcrystal)	-- we got the flag (or the crystal in battlemod)
	or ((p.cmd.buttons & BT_SPIN) and not (p.lastbuttons & BT_SPIN))	-- we cancelled it
		-- reset to rolling frames
		if (mo.state == S_VECTOR_BEATBASH)
			mo.state = S_PLAY_JUMP
		end
		
		vector.ResetBeatBash(mo)
		return
	end
	
	-- handle the timing for getting a boost!
	if (mo.vector_beatbash_timing == nil)	-- not set yet
	and not (p.cmd.buttons & BT_JUMP)	-- not holding jump anymore, start the timing!
		mo.vector_beatbash_timing = vector.BeatBashTimingTics
			
		-- start visuals
		mo.vector_beatbash_visual = leveltime
		S_StartSound(mo, sfx_s3k42)
	else
		mo.vector_beatbash_timing = $ and $-1
	end
		
	-- while we have tics, thrust forward with no z momentum!
	mo.vector_beatbash_tics = $ and $-1
	if (mo.vector_beatbash_tics)
		-- still got tics left, thrust forward!
		local angle = mo.vector_beatbash_angle
		p.drawangle = angle
		P_InstaThrust(mo, angle, mo.vector_beatbash_speed)
		P_SetObjectMomZ(mo, 0)
		
		-- nice ghost effect
		if not (mo.vector_beatbash_tics%2)
			chaotix.SpawnSpeedTrail(mo)
			local g = P_SpawnGhostMobj(mo)
			g.tics = 3
		end
	elseif not mo.vector_beatbash_line
		-- update these for the sake of the rebound
		mo.vector_beatbash_speed = FixedHypot(mo.momx, mo.momy)
		mo.vector_beatbash_angle = R_PointToAngle2(0, 0, mo.momx, mo.momy)
	end
	
	-- visual flash!
	if (mo.vector_beatbash_visual != nil)
		local visual = (leveltime - mo.vector_beatbash_visual)
		
		if (visual < vector.BeatBashTimingTics)
			-- use ghosts to make it a tint
			local visualtics = vector.BeatBashTimingTics - visual
			local maxtrans = 5
			local trans = 10 - ((visualtics * maxtrans) / vector.BeatBashTimingTics)
			
			if (trans >= 1 and trans <= 9)
				local gmo = mo
				while (gmo and gmo.valid)
					local g = chaotix.SpawnGhostMobj(gmo)
					g.frame = ($ & ~FF_TRANSMASK) | (trans << FF_TRANSSHIFT)
					g.color, g.colorized = SKINCOLOR_FULLBRIGHT, true
					g.blendmode = AST_ADD
					g.tics = 1
					
					-- attempt to spawn for a followmobj as well?
					gmo = g.tracer
				end
			end
		end
	end
	
	-- did we hit a wall? if so, continue and bounce off
	local line = mo.vector_beatbash_line
	if not (line and line.valid) then return end
	
	-- set stuff based on timing
	local multiplier = FRACUNIT
	local state = S_PLAY_JUMP	-- default is rolling state
		
	if (mo.vector_beatbash_timing)	-- we had good timing
		local timing = mo.vector_beatbash_timing
		local maxtiming = vector.BeatBashTimingTics
		local range = vector.BeatBashTimingMaxMultiplier - FRACUNIT
		
		multiplier = FRACUNIT + ((timing * range) / maxtiming)
		state = S_VECTOR_BEATBASH_POSE
		p.pflags = $ | PF_NOJUMPDAMAGE & ~(PF_THOKKED)	-- get thokability back, always
	else	-- lose a rebound!
		mo.vector_beatbash_rebounds = max(0, $-1)
		
		if (mo.vector_beatbash_rebounds)
			-- we still have rebound, get thokability back!
			p.pflags = $ & ~(PF_THOKKED)
		else
			-- lose jumpability on our last messup
			state = S_PLAY_SPRING
			p.pflags = $ & ~(PF_JUMPED)
		end
	end
		
	-- rebound the player
	local wallangle = R_PointToAngle2(line.v1.x, line.v1.y, line.v2.x, line.v2.y) - ANGLE_90
	mo.vector_beatbash_wallangle = wallangle
	local hitangle = wallangle - mo.vector_beatbash_angle
	local reflectangle = wallangle + hitangle
		
	local anglefactor = abs(cos(hitangle))
	local range = FRACUNIT - vector.BeatBashReboundWallModifier
	local modifier = FRACUNIT - FixedMul(anglefactor, range)
		
	mo.state = state
	P_InstaThrust(mo, reflectangle, FixedMul(modifier, -(mo.vector_beatbash_speed)))
	P_SetObjectMomZ(mo, FixedMul(multiplier, chaotix.WaterMom(mo, vector.BeatBashReboundMomZ)))
	vector.SetBeatPose(mo)
	vector.ResetBeatBash(mo)
		
	-- battlemod exhaust
	if (CBW_Battle)
		local maxrebounds = vector.BeatBashBattleMaxRebounds
		local exhaustremove = (FRACUNIT / maxrebounds) + 1	-- depending on the constant, it sometimes doesnt add up to FRACUNIT... bandaid time!
		p.exhaustmeter = max(0, $ - exhaustremove)
	end
end

-- Spawn Beat Bash Notes
vector.SpawnNotes = function(mo, fail, radial)
	local p = mo.player
	local angle = mo.angle
	if mo.vector_beatbash_wallangle != nil
		angle = mo.vector_beatbash_wallangle
	elseif mo.vector_beatbash_angle != nil
		angle = mo.vector_beatbash_angle
	elseif p
		angle = p.drawangle
	end
	angle = $ - ANGLE_90
	
	local speed = FRACUNIT*12
	local colors = fail and vector.FailColors or vector.SuccessColors
	local coloroff = P_RandomKey(#colors)
	local reversed = P_RandomChance(FRACUNIT/2)
	local amount = #colors
	local diff = 360/amount
	local angleoff = P_RandomKey(diff)
	
	for i = 1, amount
		local mobj = P_SpawnMobjFromMobj(mo, 0,0,
		FixedDiv(mo.height*2/3, mo.scale), MT_THOK)
		
		-- color
		local color = ((i + coloroff) % #colors) + 1
		if reversed then color = #colors - $ + 1 end
		mobj.color = colors[color]
			
		-- state
		if fail then mobj.cusval = 1 end
		mobj.state = S_VECTOR_NOTE
			
		if (radial)
			local horzangle = (i*diff + angleoff)*ANG1
			
			-- horizontal
			P_InstaThrust(mobj, horzangle, speed)
			mobj.angle = angle
			
			-- vertical
			local thrust = FixedMul(speed, sin(ANGLE_45))
			P_SetObjectMomZ(mobj, thrust)
		else
			local vertangle = (i*diff + angleoff)*ANG1
			
			-- horizontal
			local thrust = FixedMul(speed, sin(vertangle))
			P_InstaThrust(mobj, angle, FixedMul(thrust, mo.scale))
			mobj.angle = angle
			
			-- vertical
			thrust = FixedMul(speed, cos(vertangle))
			P_SetObjectMomZ(mobj, thrust)
		end
	end
end

-- Set the Beat Bash Pose
vector.SetBeatPose = function(mo, enemy)
	-- go back to the first pose
	local maxpose = vector.NumSkinPoses[mo.skin]
	if (mo.vector_beatbash_pose == nil)
	or (abs(mo.vector_beatbash_pose) >= maxpose)
		mo.vector_beatbash_pose = 0
	end
	
	-- increase pose count
	if mo.vector_beatbash_pose >= 0
		mo.vector_beatbash_pose = $ + 1
	else
		mo.vector_beatbash_pose = $ - 1
	end
	local note = abs(mo.vector_beatbash_pose)
	
	-- show the text based off the command
	local p = mo.player
	local command = vector.BeatBashCommandVar.value
	local hidetext = not command or (command < 2 and (p != consoleplayer and p != secondaryconsoleplayer))
	
	-- during pose
	if (mo.state == S_VECTOR_BEATBASH_POSE)
		-- reset pose if we were failing
		if mo.vector_beatbash_pose < 0
			mo.vector_beatbash_pose = 1
			note = 1
		end
		
		note = _G['sfx_vchrs' + mo.vector_beatbash_pose]
		if note then S_StartSound(mo, note) end
		S_StartSound(mo, sfx_vcclap)
		
		chaotix.SetFrame(mo, mo.vector_beatbash_pose - 1)
		mo.vector_beatbash_posetics = vector.BeatBashPoseTics
		
		-- show flavor text
		local thok = P_SpawnMobjFromMobj(mo, 0,0,0, MT_THOK)
		P_SetObjectMomZ(thok, 16*FRACUNIT)
		
		thok.threshold = G + mo.vector_beatbash_pose - 1
		if mo.vector_beatbash_pose >= maxpose or thok.threshold > L
			thok.threshold = L
		end
		thok.target = mo
		thok.state = S_VECTOR_TEXT
		if hidetext then thok.flags2 = $ | MF2_DONTDRAW end
	
	-- failed pose
	else
		mo.vector_beatbash_pose = -abs($) -- failure means negative pose
		
		note = _G['sfx_vchrf' + note]
		if note then S_StartSound(mo, note) end
		S_StartSound(mo, sfx_s3k7b)
		
		-- show if you were early or late
		local thok = P_SpawnMobjFromMobj(mo, 0,0,0, MT_THOK)
		P_SetObjectMomZ(thok, 16*FRACUNIT)
		
		if mo.vector_beatbash_timing == nil
			thok.threshold = R
		else
			thok.threshold = Q
		end
		thok.target = mo
		thok.state = S_VECTOR_TEXT
		if hidetext then thok.flags2 = $ | MF2_DONTDRAW end
	end
	vector.SpawnNotes(mo, mo.state != S_VECTOR_BEATBASH_POSE, enemy)
end

-- rebound off of things we damage!
vector.BeatBashDamage = function(mobj, mo)
	if not (mobj and mobj.valid) then return end
	if not ((mo and mo.valid) and mo.health) then return end
	local p = mo.player
	
	-- are we doing the beat bash?
	if not (mo.vector_beatbash) then return end
	
	-- don't do anything in battle mod! (handled in collision function)
	if (mobj.player)
	and (CBW_Battle)
		return
	end
		
	-- set stuff based on timing
	local multiplier = FRACUNIT
	local state = S_PLAY_JUMP	-- default is rolling state
		
	if (mo.vector_beatbash_timing)	-- we had good timing
		local timing = mo.vector_beatbash_timing
		local maxtiming = vector.BeatBashTimingTics
		local range = vector.BeatBashTimingMaxMultiplier - FRACUNIT
		
		multiplier = FRACUNIT + ((timing * range) / maxtiming)
		state = S_VECTOR_BEATBASH_POSE
		p.pflags = $ | PF_NOJUMPDAMAGE & ~(PF_THOKKED)	-- get thokability back, always
	else	-- lose a rebound!
		mo.vector_beatbash_rebounds = max(0, $-1)
		
		if (mo.vector_beatbash_rebounds)
			-- we still have rebound, get thokability back!
			p.pflags = $ & ~(PF_THOKKED)
		else
			-- lose jumpability on our last messup
			state = S_PLAY_SPRING
			p.pflags = $ & ~(PF_JUMPED)
		end
	end
	
	-- get a stronger bounce!
	local reboundmomz = FixedMul(multiplier, chaotix.WaterMom(mo, vector.BeatBashReboundMomZ))
	local bouncemomz = FixedDiv(FixedMul(multiplier, mo.momz), mo.scale)
	mo.state = state
	P_SetObjectMomZ(mo, max(reboundmomz, bouncemomz))
	vector.SetBeatPose(mo, true)
	vector.ResetBeatBash(mo)
end
chaotix.AddHook("MobjDamage", vector.BeatBashDamage)
chaotix.AddHook("MobjDeath", vector.BeatBashDamage)

-- rebound off walls!
local function vector_checkLine(p, line)
	local mo = p.mo
	
	-- horizon line checks
	if (line.special == 41) then return end
	
	-- sector checks
	if not (line.backsector) then return true end	-- apparently in these situations source always returns true?
	local sec = (line.backsector and not P_PointOnLineSide(mo.x, mo.y, line)) and line.backsector or line.frontsector
	
	-- shouldnt be too hard to check this
	local checkz = mo.z
	if (mo.eflags & MFE_VERTICALFLIP) then checkz = mo.z + mo.height end
	
	-- make sure we cant climb thokbarriers
	local thokbarrier = (sec.ceilingpic:upper():find("F_SKY"))
	if (thokbarrier and (checkz >= sec.ceilingheight))
		return false
	end
	
	-- should be good?
	return true
end

chaotix.AddHook("MobjMoveBlocked", function(mo, thing, line)
	if not ((mo and mo.valid) and mo.health) then return end
	if not (line and line.valid) then return end
	
	-- are we doing the beat bash?
	if not (mo.vector_beatbash and not mo.vector_beatbash_line) then return end
	local p = mo.player
	if not (vector_checkLine(p, line)) then return end
	
	-- set the line here so we can use it in the main beat bash thinker later
	mo.vector_beatbash_line = line
end, MT_PLAYER)

-- additional hook for busting walls (MobjMoveBlocked gets the wrong line 99% of the time...)
chaotix.AddHook("MobjLineCollide", function(mo, line)
	if not ((mo and mo.valid) and mo.health) then return end
	if not (line and line.valid) then return end
	
	-- are we doing the beat bash?
	if not (mo.vector_beatbash) then return end
	local p = mo.player
	if not (vector_checkLine(p, line)) then return end
	
	-- break bustable walls!
	if (line.frontsector == nil or line.backsector == nil) then return end	-- one-sided, dont bother
	local sector = (P_PointOnLineSide(mo.x, mo.y, line) == 1) and line.frontsector or line.backsector
	
	for rover in sector.ffloors()
		if not (rover.flags & FF_EXISTS) then continue end
		if not ((rover.flags & FF_STRONGBUST) or (rover.flags & FF_BUSTUP)) then continue end
			
		-- make sure we timed it right for strong walls!
		if (rover.flags & FF_STRONGBUST) 
		and not (mo.vector_beatbash_timing)
			continue
		end

		-- height checks
		local topheight = rover.topheight
		local bottomheight = rover.bottomheight
			
		if (mo.z > topheight)
		or ((mo.z + mo.height) < bottomheight)
			continue
		end
			
		EV_CrumbleChain(rover)
		if (rover.master.flags & ML_EFFECT5)
			P_LinedefExecute(P_AproxDistance(rover.master.dx, rover.master.dy)>>FRACBITS, mo, mo.subsector.sector)
		end
		
		-- rebound off if we get a strong bustable wall!
		if (rover.flags & FF_STRONGBUST)
			local timing = mo.vector_beatbash_timing
			local maxtiming = vector.BeatBashTimingTics
			local range = vector.BeatBashTimingMaxMultiplier - FRACUNIT
			local multiplier = FRACUNIT + ((timing * range) / maxtiming)
				
			local wallangle = R_PointToAngle2(line.v1.x, line.v1.y, line.v2.x, line.v2.y) - ANGLE_90
			mo.vector_beatbash_wallangle = wallangle
			local hitangle = wallangle - mo.vector_beatbash_angle
			local reflectangle = wallangle + hitangle
				
			local anglefactor = abs(cos(hitangle))
			local range = FRACUNIT - vector.BeatBashReboundWallModifier
			local modifier = FRACUNIT - FixedMul(anglefactor, range)
			
			p.pflags = $ | PF_NOJUMPDAMAGE & ~(PF_THOKKED)	-- get thokability back, always
			mo.state = S_VECTOR_BEATBASH_POSE
			P_InstaThrust(mo, reflectangle, FixedMul(modifier, -(mo.vector_beatbash_speed)))
			P_SetObjectMomZ(mo, FixedMul(multiplier, chaotix.WaterMom(mo, vector.BeatBashReboundMomZ)))
			vector.SetBeatPose(mo)
			vector.ResetBeatBash(mo)
			
			chaotix.SetFrame(mo, mo.vector_beatbash_pose - 1)
			mo.vector_beatbash_posetics = vector.BeatBashPoseTics
		end
	end
end, MT_PLAYER)

-- collision hook to fix springs
vector.BeatBashCollide = function(mo, mo2)
	-- object checks
	if not ((mo and mo.valid) and mo.health) then return end
	if not ((mo2 and mo2.valid) and mo2.health) then return end
	
	-- beat bash check
	if not (mo.vector_beatbash) then return end
	
	-- the rest of these use regular height checks
	if not ((mo.z <= mo2.z + mo2.height)
	and (mo2.z <= mo.z + mo.height))
		return
	end
	
	-- spring collision
	if (mo2.flags & MF_SPRING)
	and not (mobjinfo[mo2.type].mass)	-- vertical
	and (mobjinfo[mo2.type].damage)	-- horizontal
		mo.player.pflags = $ & ~(PF_JUMPED)
		mo.state = S_PLAY_FALL
		vector.ResetBeatBash(mo)
		return
	end
end