-- vector's defs, makes things easier
chaotix.vector = {}
local vector = chaotix.vector
vector.SKIN = "vector"
chaotix.AddCharacter(vector)

-- vector's sounds
local sounds = {
	{sound = "sfx_vcjump", caption = "Jump"},
	{sound = "sfx_vcdrft", caption = "Drifting"},
	{sound = "sfx_vcclap", caption = "Clap"},
	{sound = "sfx_vecl0", caption = "Level 0 Scream"},
	{sound = "sfx_vecl1", caption = "Level 1 Scream"},
	{sound = "sfx_vecl2", caption = "Level 2 Scream"},
	{sound = "sfx_vecl3", caption = "LEVEL 3 Scream"},
	{sound = "sfx_vecch1", caption = "Charging"},
	{sound = "sfx_vecch2", caption = "More Charging"},
	{sound = "sfx_vecch3", caption = "Even More Charging"},
	{sound = "sfx_vecchf", caption = "Fully Charged"},
}
for i = 1, 6
	table.insert(sounds, {sound = "sfx_vchrs" + i, caption = "/"})
	table.insert(sounds, {sound = "sfx_vchrf" + i, caption = "/"})
end
for k, v in pairs(sounds)
	chaotix.SafeFreeslot(v.sound)
	sfxinfo[_G[v.sound]].caption = v.caption
end

-- vector's abilities
local abilityslot = 97
rawset(_G, "CA_JESUSVECTOR", abilityslot)
rawset(_G, "CA2_ACCELSPINDASH", abilityslot)	-- will do more on this later
vector.AbilityCheck = function(p, ability)
	-- since they're both the same number, we can do some fancy stuff!
	local checkability = (ability == 2) and "2" or ""
	
	if (ability == 1)
		-- same as mighty, need a CA_GLIDEANDCLIMB buffer
		local abilitycheck = (p.charability != CA_JESUSVECTOR)
		if (abilitycheck and not p.vector_beatbash_climbing)
			return false
		end
		
		-- if we're here, we got the ability
		return true
	elseif (ability == 2)
		-- ability2 is easy to check (though this might not be needed...)
		return (p.charability2 == CA2_ACCELSPINDASH)
	end
end

-- jukebox constants
vector.JukeboxFadeVolume = 25
vector.JukeboxFadeOutTime = TICRATE/4
vector.JukeboxFadeInTime = TICRATE/2
vector.JukeboxDefaultFlags = SF_X8AWAYSOUND

-- headset death constants
vector.HeadsetDeathAngleOffset = 33
vector.HeadsetDeathThrustBounds = {FRACUNIT*4/3, FRACUNIT*11/4}
vector.HeadsetDeathLaunch = 2*FRACUNIT
vector.HeadsetDeathSpeedMultiplier = FRACUNIT*3/2
vector.HeadsetDeathUnderwaterMultiplier = FRACUNIT/3

-- thumbs up constants
vector.ThumbsUpBaseChance = 224	-- base chance
vector.ThumbsUpChanceSpeedReduce = vector.ThumbsUpBaseChance*2/3	-- how much speed can increase the chance of the pose
vector.ThumbsUpTicBounds = {TICRATE*5/4, TICRATE*2}	-- how long we thumbs up for
vector.ThumbsUpStretch = FRACUNIT*23/20
vector.ThumbsUpStretchTics = 6
vector.ThumbsUpCooldownTics = TICRATE*5/3	-- how long we don't roll for rng again after a pose

-- record rush constants
vector.RiptireStabilizer = FRACUNIT/3	-- for fixing vector's angle
vector.RiptireStabilizerModifier = FRACUNIT*3/4	-- for making up how much speed we lose from stabilising!
vector.RiptireSpeedBoost = FRACUNIT*4/3	-- based off of normalspeed
vector.RiptireSneakerSpeedBoost = FRACUNIT*4/3
vector.RiptireSneakerAccelBoost = FRACUNIT*7/3
vector.RiptireAccelValue = FRACUNIT*3/4
vector.RiptireAccelBoost = FRACUNIT*17/4	-- for turning around
vector.RiptireMaxRollAngle = 64
vector.RiptireRollAngleChange = 9	-- how much rollangle changes by to meet targetrollangle
vector.RiptireDustRollAngle = vector.RiptireMaxRollAngle	-- when dust starts to spawn while turning
vector.RiptireDriftHeight = 32*FRACUNIT	-- to fit through tight spaces; ...
vector.RiptireSpindashDustThreshold = FRACUNIT*8
vector.RiptireSpindashSpeedMultiplier = FRACUNIT*2/3
vector.RiptireXStretch = FRACUNIT*4/5
vector.RiptireYStretch = FRACUNIT*4/3
vector.RiptireStretchTics = 8
vector.RiptireStretchTics = 8
vector.RiptireEngineAdjust = 6
vector.RiptireEngineSounds = 8	-- how many sounds there are in total
vector.RiptireEngineVolume = 144
vector.RiptireBobThreshold = FRACUNIT*4/10
vector.RiptireBobTics = TICRATE/5
vector.RiptireBobHeight = FRACUNIT*8
vector.RiptireSpinModifier = FRACUNIT*4/5	-- how fast we need to go to go into spin frames
//vector.RiptireMinGhostThreshold = FRACUNIT*2/3	-- consistent with vanilla spintrail speed
vector.RiptireMaxGhostThreshold = FRACUNIT*64
vector.RiptireMaxGhosts = 3
vector.RiptireMinSpeedlines, vector.RiptireMaxSpeedlines = 8, 1
vector.RiptireSpeedlineModifier = FRACUNIT/4

-- record rush sounds (easier to freeslot here)
for i = 1, vector.RiptireEngineSounds
	local sound = "sfx_vceng"..i
	chaotix.SafeFreeslot(sound)
	sfxinfo[_G[sound]].caption = "Engine"
end

-- beat bash constants
vector.BeatBashCommand = "vector_text"
vector.BeatBashTics = TICRATE/2
vector.BeatBashReboundMomZ = FRACUNIT*11
vector.BeatBashTimingTics = 5
vector.BeatBashTimingMaxMultiplier = FRACUNIT*6/5	-- will scale up to this from FRACUNIT for good timing
vector.BeatBashPoseTics = 16
vector.BeatBashBattleMaxRebounds = 3
vector.BeatBashReboundWallModifier = FRACUNIT/3	-- will scale up to this depending on your angle!

-- battlemod constants
vector.BeatBashRebounds = 2
vector.EncoreCost = 10
vector.EncoreMaxLevel, vector.EncoreMaxLevelTics = 3, TICRATE*5/2	-- the max level and how long it takes to get to it
vector.EncoreMinCooldown, vector.EncoreMaxCooldown = TICRATE*2, TICRATE*17/4
vector.EncoreCancelCooldown = TICRATE	-- shorter cooldown for cancelling the charge state
vector.EncoreStartupTics = TICRATE/3
vector.EncoreScreamMinRecoil, vector.EncoreScreamMaxRecoil = FRACUNIT*6, FRACUNIT*18
vector.EncoreScreamMinTics, vector.EncoreScreamMaxTics = TICRATE/4, TICRATE/2
vector.EncoreScreamMinHeight, vector.EncoreScreamMaxHeight = FRACUNIT*64, FRACUNIT*96	-- this gets added on TOP of mo.height
vector.EncoreScreamAngle = 32
vector.EncoreScreamMinRange, vector.EncoreScreamMaxRange = FRACUNIT*224, FRACUNIT*352
vector.EncoreScreamCloseRange = FRACUNIT*8	-- used for the frontal hitbox
vector.EncoreScreamMinModifier = FRACUNIT*3/5	-- max is always FRACUNIT
vector.EncoreScreamThrust = FRACUNIT*40	-- used ONLY for level 0
vector.EncoreScreamMinTumbleThrust, vector.EncoreScreamMaxTumbleThrust = FRACUNIT*18, FRACUNIT*28	-- keep in mind it's supposed to be really strong from the base
vector.EncoreScreamTumbleHeightModifier = FRACUNIT*7/9	-- based on the thrust
vector.EncoreScreamTumbleTicDivider = FRACUNIT*16	-- ^ ditto
vector.EncoreScreamEndlagMinTics, vector.EncoreScreamEndlagMaxTics = TICRATE/3, TICRATE*4/5
vector.EncoreDustThreshold = 1
vector.EncoreFlashMinFrequency, vector.EncoreFlashMaxFrequency = TICRATE/3, TICRATE/7
vector.EncoreFlashTics = TICRATE/5
vector.EncoreFlashMaxTrans = 6
vector.EncoreSmearTics = 1
vector.EncoreScreamWaveScalar = FRACUNIT*80
vector.EncoreScreamWaveModifier = FRACUNIT/5
vector.EncoreMinShakeTics, vector.EncoreMaxShakeTics = 2, 8
vector.EncoreMinShakeIntensity, vector.EncoreMaxShakeIntensity = FRACUNIT*16, FRACUNIT*60

-- record rush spindash
chaotix.SafeFreeslot("S_VECTOR_SPINDASH")
states[S_VECTOR_SPINDASH] = {SPR_PLAY, SPR2_MLEE, 2, nil, 0, 0, S_VECTOR_SPINDASH}

-- beat bash states
chaotix.SafeFreeslot("S_VECTOR_BEATBASH", "S_VECTOR_BEATBASH_POSE")
states[S_VECTOR_BEATBASH] = {SPR_PLAY, SPR2_GLID, 2, nil, 0, 0, S_VECTOR_BEATBASH}
states[S_VECTOR_BEATBASH_POSE] = {SPR_PLAY, SPR2_TAL0, -1, nil, 0, 0, S_VECTOR_BEATBASH_POSE}

-- encore stats
chaotix.SafeFreeslot("S_VECTOR_ENCORE_STARTUP", "S_VECTOR_ENCORE_CHARGE", "S_VECTOR_ENCORE_SCREAM")
states[S_VECTOR_ENCORE_STARTUP] = {SPR_PLAY, SPR2_FIRE, -1, nil, 0, 0, S_VECTOR_ENCORE_STARTUP}
states[S_VECTOR_ENCORE_CHARGE] = {SPR_PLAY, SPR2_FIRE, -1, nil, 0, 0, S_VECTOR_ENCORE_CHARGE}
states[S_VECTOR_ENCORE_SCREAM] = {SPR_PLAY, SPR2_FIRE, -1, nil, 0, 0, S_VECTOR_ENCORE_SCREAM}
chaotix.SafeFreeslot("S_VECTOR_ENCORE_STARTUP_AIR", "S_VECTOR_ENCORE_CHARGE_AIR", "S_VECTOR_ENCORE_SCREAM_AIR")
states[S_VECTOR_ENCORE_STARTUP_AIR] = {SPR_PLAY, SPR2_TWIN, -1, nil, 0, 0, S_VECTOR_ENCORE_STARTUP_AIR}
states[S_VECTOR_ENCORE_CHARGE_AIR] = {SPR_PLAY, SPR2_TWIN, -1, nil, 0, 0, S_VECTOR_ENCORE_CHARGE_AIR}
states[S_VECTOR_ENCORE_SCREAM_AIR] = {SPR_PLAY, SPR2_TWIN, -1, nil, 0, 0, S_VECTOR_ENCORE_SCREAM_AIR}

-- vector's headset (death)
chaotix.SafeFreeslot("SPR_VHST", "S_VECTOR_HEADSET")
A_VectorHeadset = function(mo)
	if not ((mo and mo.valid) and mo.health) then return end
	if not (mo.vector_isheadset) then return end	-- lol
	
	-- target validity checks
	local target = mo.target
	
	if not (target and target.valid)	-- target doesn't exist
	or (target.health)	-- they revived??
	or (target.skin != vector.SKIN)	-- we are no longer mighty
		P_RemoveMobj(mo)
		return
	end
	
	-- fall with the player
	local multiplier = (mo.vector_drowned) and vector.HeadsetDeathUnderwaterMultiplier or FRACUNIT
	mo.momz = $ - FixedMul(multiplier, (43690 * P_MobjFlip(mo)))	-- magic number for now until I find out the actual value for death falling
	
	-- add some rotation based on momentum!
	if not (mo.vector_drowned)	-- but only if we didnt drown
		local speed = FixedDiv(FixedHypot(mo.momz, FixedHypot(mo.momx, mo.momy)), mo.scale)
		local rollangle_add = FixedAngle(FixedMul(vector.HeadsetDeathSpeedMultiplier, speed))
		mo.rollangle = $ + (rollangle_add * mo.vector_launchangle)
		
		-- play a sound effect for the apex of the fall!
		if not (mo.vector_soundplayed)
		and ((mo.momz * P_MobjFlip(mo)) < 0)
			S_StartSound(mo, sfx_s3k51)
			mo.vector_soundplayed = true
		end
	end
end
states[S_VECTOR_HEADSET] = {SPR_VHST, A, 1, A_VectorHeadset, 0, 0, S_VECTOR_HEADSET}

-- vector's notes
chaotix.SafeFreeslot("SPR_VNOT", "S_VECTOR_NOTE", "S_VECTOR_TEXT")
vector.SuccessColors = {
	SKINCOLOR_RED,
	SKINCOLOR_ORANGE,
	SKINCOLOR_YELLOW,
	SKINCOLOR_EMERALD,
	SKINCOLOR_SKY,
	SKINCOLOR_BLUE,
	SKINCOLOR_PURPLE,
	SKINCOLOR_MAGENTA,
}
vector.FailColors = {
	SKINCOLOR_PINK,
	SKINCOLOR_YOGURT,
	SKINCOLOR_TAN,
	SKINCOLOR_MOSS,
	SKINCOLOR_TEAL,
	SKINCOLOR_AZURE,
	SKINCOLOR_LAVENDER,
	SKINCOLOR_ROSY,
}

local fuse = TICRATE
local half = fuse/2
A_VectorNote = function(mo)
	if not mo.fuse
		mo.threshold = P_RandomKey(C+1)
		if mo.cusval
			mo.threshold = $ + C+1
			mo.flags = $ &~ MF_NOGRAVITY
		end
		
		mo.fuse = fuse
		mo.renderflags = $ | RF_SEMIBRIGHT
	end
	mo.frame = ($ &~ FF_FRAMEMASK) | mo.threshold
	
	-- fade out when nearing the end of fuse...
	if ((mo.frame & FF_TRANSMASK) >> FF_TRANSSHIFT) < (NUMTRANSMAPS-1) - mo.fuse / 2
		mo.frame = (mo.frame & ~FF_TRANSMASK) | (((NUMTRANSMAPS-1) - mo.fuse / 2) << FF_TRANSSHIFT)
	end
	
	-- slow down
	local grav = P_GetMobjGravity(mo)*4
	mo.momz = $ - grav
	for i, variable in ipairs({'momx', 'momy', 'momz'})
		mo[variable] = $*(half - 1)/half
	end
	mo.momz = $ + grav
	
	-- set silly rollangle
	if mo.cusval
		mo.rollangle = $ - UINT8_MAX*FixedHypot(FixedHypot(mo.momx, mo.momy), mo.momz)
	else
		local angle = FixedMul(cos(FixedAngle(mo.fuse*12*FRACUNIT)), ANGLE_45*3/2)
		mo.rollangle = angle
	end
	
	-- set silly offseting
	mo.spriteyoffset = 32*abs(sin(FixedAngle(mo.fuse*16*FRACUNIT)))
end
states[S_VECTOR_NOTE] = {SPR_VNOT, A|FF_PAPERSPRITE, 1, A_VectorNote, 0, 0, S_VECTOR_NOTE}

local ending = fuse/4
A_VectorText = function(mo)
	-- slow down
	mo.momz = $*(ending - 1)/ending
	
	local frame = mo.threshold
	if frame == L then frame = $ + 4 - ((mo.fuse/3) % 5) end
	mo.frame = ($ &~ FF_FRAMEMASK) | frame
	
	-- follow vector around
	mo.movefactor = $ + mo.momz*P_MobjFlip(mo)
	if mo.target and mo.target.health
		if mo.target.eflags & MFE_VERTICALFLIP
		and mo.target.player and not (mo.target.player.pflags & PF_FLIPCAM)
			mo.frame = $ | FF_VERTICALFLIP
		end
		A_CapeChase(mo, mo.movefactor/mo.scale*FRACUNIT, FRACUNIT)
	end
	
	-- fuse
	if not mo.fuse
		mo.fuse = fuse
		mo.renderflags = $ | RF_FULLBRIGHT | RF_NOCOLORMAPS
		return
		
	-- intro squish
	elseif mo.fuse >= ending
		local squish = (mo.fuse - (fuse - ending))*FRACUNIT/ending
		if squish >= 0
			squish = ease.outsine($, 0, FRACUNIT)
			mo.spritexscale = FRACUNIT - squish
			mo.spriteyscale = FRACUNIT + squish
		end
		
		return
	end
	
	-- outro squish
	local squish = (ending - mo.fuse)*FRACUNIT/ending
	squish = ease.outsine($, 0, FRACUNIT)
	mo.spritexscale = FRACUNIT + squish
	mo.spriteyscale = FRACUNIT - squish
end
states[S_VECTOR_TEXT] = {SPR_VNOT, A, 1, A_VectorText, 0, 0, S_VECTOR_TEXT}

-- super apple
chaotix.SafeFreeslot("SKINCOLOR_SUPERAPPLE1")
skincolors[SKINCOLOR_SUPERAPPLE1] = {
	name = "Super Apple 1",
	ramp = {80, 80, 82, 82, 82, 72, 72, 188, 100, 100, 102, 102, 104, 104, 104, 104},
	accessible = false
}
chaotix.SafeFreeslot("SKINCOLOR_SUPERAPPLE2")
skincolors[SKINCOLOR_SUPERAPPLE2] = {
	name = "Super Apple 2",
	ramp = {82, 82, 83, 72, 72, 188, 188, 100, 102, 102, 104, 104, 105, 105, 105, 105},
	accessible = false
}
chaotix.SafeFreeslot("SKINCOLOR_SUPERAPPLE3")
skincolors[SKINCOLOR_SUPERAPPLE3] = {
	name = "Super Apple 3",
	ramp = {83, 83, 72, 73, 73, 188, 188, 101, 103, 103, 105, 105, 106, 106, 106, 106},
	accessible = false
}
chaotix.SafeFreeslot("SKINCOLOR_SUPERAPPLE4")
skincolors[SKINCOLOR_SUPERAPPLE4] = {
	name = "Super Apple 4",
	ramp = {72, 72, 73, 188, 188, 188, 188, 102, 104, 104, 106, 106, 107, 107, 107, 107},
	accessible = false
}
chaotix.SafeFreeslot("SKINCOLOR_SUPERAPPLE5")
skincolors[SKINCOLOR_SUPERAPPLE5] = {
	name = "Super Apple 5",
	ramp = {73, 73, 188, 189, 189, 189, 189, 103, 105, 105, 107, 107, 108, 108, 108, 108},
	accessible = false
}

-- encore visuals
local wave_frames = 3
local wave_tics = wave_frames * 2

function A_VectorWaveDeath(actor, var1, var2)
	-- store the starting threshold so we can go from it
	if (actor.saved_threshold == nil)
		actor.saved_threshold = actor.threshold or 0
	end
	
	-- increase threshold
	actor.threshold = $ and $+1 or 1
	local threshold = actor.threshold - actor.saved_threshold
	local wave_maxtics = wave_tics - actor.saved_threshold
	
	-- get transparency
	local trans = (threshold * 10) / wave_maxtics
	local trans_flag = (trans << FF_TRANSSHIFT)
	
	-- get frame
	local frame = 0
	for i = (wave_frames - 1), 1, -1
		local frame_threshold = (wave_maxtics / wave_frames) * i
		if not (threshold >= frame_threshold) then continue end
		
		frame = i
		break
	end
	
	-- set frame + transparency
	actor.frame = frame|trans_flag 
	
	-- remove mobj
	if (threshold >= wave_maxtics)
		P_RemoveMobj(actor)
		return
	end
end
chaotix.SafeFreeslot("S_VECTOR_WAVE", "S_VECTOR_WAVE_DEATH", "SPR_VSCM")
states[S_VECTOR_WAVE] = {SPR_VSCM, 0|FF_PAPERSPRITE|FF_ANIMATE, -1, nil, 1, 2, S_VECTOR_WAVE}
states[S_VECTOR_WAVE_DEATH] = {SPR_VSCM, 0|FF_PAPERSPRITE, 1, A_VectorWaveDeath, 0, 0, S_VECTOR_WAVE_DEATH}

-- the hitbox for vector's encore
chaotix.SafeFreeslot("S_VECTOR_SCREAM", "MT_VECTOR_SCREAM")
states[S_VECTOR_SCREAM] = {SPR_NULL, 0, -1, nil, 0, 0, S_NULL}	-- tics NEEDS to be set properly in vector.SpawnScream
mobjinfo[MT_VECTOR_SCREAM] = {
	spawnstate = S_VECTOR_SCREAM,
	flags = MF_NOBLOCKMAP|MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT,
}

vector.cv_showrange = CV_RegisterVar({
	name = "vector_showrange",
	defaultvalue = "No",
	flags = CV_NETVAR,
	PossibleValue = CV_YesNo,
})

vector.EncoreScreamHitbox = function(scream)
	-- not valid anymore?
	local screaminfo = scream.vector_screaminfo
	if not (screaminfo)
		P_RemoveMobj(scream)
		return 
	end
	
	-- setup hitbox table
	scream.vector_hitmobjs = $ or {}
	
	-- search with blockmap
	local source = scream.target
	local tics = screaminfo["maxtics"] - scream.tics
	
	local range_modifier = ease.outquart((tics * FRACUNIT) / screaminfo["maxtics"], FRACUNIT)
	local range2_modifier
	if (tics)
		-- using this to make sure the hitbox doesnt just zip past targets
		range2_modifier = ease.outquart(((tics - 1) * FRACUNIT) / screaminfo["maxtics"], FRACUNIT)
	end
	
	-- distance
	local range = FixedMul(range_modifier, screaminfo["range"])
	local xoff = FixedMul(range, cos(scream.angle))
	local yoff = FixedMul(range, sin(scream.angle))
	
	local wave = scream.tracer
	if not (wave and wave.valid)
		wave = P_SpawnMobjFromMobj(scream, xoff, yoff, 0, MT_THOK)
		wave.angle = scream.angle + ANGLE_90
		wave.state = S_VECTOR_WAVE
		wave.color = scream.color
		scream.tracer = wave	-- store it
	end
	
	local xpos = scream.x + xoff
	local ypos = scream.y + yoff
	P_MoveOrigin(wave, xpos, ypos, scream.z)
		
	-- go to death state (if not done so already)
	local death_tics = tics - (screaminfo["maxtics"] - wave_tics + 1)
		
	if (wave.state != S_VECTOR_WAVE_DEATH)
	and (death_tics >= 0)
		wave.threshold = death_tics
		wave.state = S_VECTOR_WAVE_DEATH
	end
	
	-- scale
	local maxScale = FixedDiv(screaminfo["range"], vector.EncoreScreamWaveScalar)
	local minScale = FixedMul(vector.EncoreScreamWaveModifier, maxScale)
	local scale = minScale + FixedMul(range_modifier, maxScale - minScale)
	wave.scale = scale
		
	if (vector.cv_showrange.value)
		-- height
		local height_thoks = 5
		for hthoks = -(height_thoks / 2), height_thoks / 2
			local zoffset = (hthoks * FRACUNIT) / (height_thoks / 2)
				
			local angle_thoks = 8
			for athoks = 0, (angle_thoks - 1)
				-- spawn angle
				local trigAngle = vector.EncoreScreamAngle - ((athoks * (vector.EncoreScreamAngle * 2)) / (angle_thoks - 1))
				trigAngle = scream.angle + ($ * ANG1)
				
				local modifiers = {range_modifier, range2_modifier}
				for mthoks = 1, #modifiers
					local modifier = modifiers[mthoks]
					if (modifier == nil) then continue end
					
					-- distance
					local range = FixedMul(modifier, screaminfo["range"])
					local xpos = scream.x + FixedMul(range, cos(trigAngle))
					local ypos = scream.y + FixedMul(FixedMul(modifier, screaminfo["range"]), sin(trigAngle))
								
					-- height
					local startz = scream.z
					if (scream.eflags & MFE_VERTICALFLIP) then startz = scream.z + screaminfo["startheight"] end
					local heightOffset = (screaminfo["startheight"] + FixedMul(modifier, screaminfo["height"])) * P_MobjFlip(scream)
					local zpos = startz + FixedMul(heightOffset, zoffset)
					
					-- height
					local heightOffset = FixedMul(modifier, screaminfo["height"])
					local bottomz = scream.z - heightOffset
					local topz = scream.z + screaminfo["startheight"] + heightOffset
					local zpos = scream.z + FixedMul(zoffset, (topz - bottomz) / 2)
								
					-- spawn the thok
					local thok = P_SpawnMobj(xpos, ypos, zpos, MT_THOK)
					thok.frame = ($ & ~(FF_TRANSMASK)) | (FF_FULLBRIGHT|FF_TRANS70)
					thok.scale = FixedMul(FRACUNIT/2, $)
					thok.color = (mthoks == 1) and SKINCOLOR_RED or SKINCOLOR_ORANGE
					thok.tics = 2
				end
			end
		end
	end
	
	searchBlockmap("objects", function(ref, found)
		if not ((found and found.valid) or found.health) then return end
		if (ref == found) then return end	-- don't damage the inflictor
		if (found == source) then return end	-- don't damage the source!
		
		-- already hit!
		if (ref.vector_hitmobjs[found]) then return end
			
		-- only player support for now
		if not (found.player) then return end
		if (found.player and found.player.spectator) then return end	-- NUH UH
		local B = CBW_Battle
		if (B.MyTeam(source.player, found.player)) then return end -- same team
			
		-- positional checks
		if not (P_CheckSight(ref, found)) then return end
		
		local dist = R_PointToDist2(ref.x, ref.y, found.x, found.y)
		if (dist > screaminfo["range"]) then return end	-- max range
		local distModifier = FRACUNIT - FixedDiv(dist, screaminfo["range"])
		
		local angle = R_PointToAngle2(ref.x, ref.y, found.x, found.y)
		local angle_diff = ref.angle - angle
		if (chaotix.AngToInt(angle_diff) > 180) then angle_diff = InvAngle(angle_diff) end
		if (chaotix.AngToInt(angle_diff) > vector.EncoreScreamAngle) then return end
		
		local heightOffset = FixedMul(distModifier, screaminfo["height"])
		local bottomz = ref.z - heightOffset
		local topz = ref.z + screaminfo["startheight"] + heightOffset
		if not ((topz >= found.z) and (bottomz <= found.z + found.height)) then return end	-- z checks
		
		local range = FixedMul(range_modifier, screaminfo["range"])
		local range2
		if (range2_modifier != nil) then range2 = FixedMul(range2_modifier, screaminfo["range"]) end
		if not (abs(dist - range) <= vector.EncoreScreamCloseRange)	-- not in range of the front range check
		and not (range2 != nil and (dist >= range2 and dist <= range))	-- not in between the 2 ranges
			return
		end
			
		-- airdodge/guard protection
		if (found.player.intangible)
			return
		elseif (found.player.guard > 0)	-- ...unless they're guarding!
			-- using a thok to get around getting vector tumbled!
			if (B.SkinVars[found.player.skinvars].func_guard_trigger)
				B.SkinVars[found.player.skinvars].func_guard_trigger(found, ref, source)
			elseif (B.SkinVars[-1].func_guard_trigger)
				B.SkinVars[-1].func_guard_trigger(found, ref, source)
			end
				
			return
		end
			
		-- affect them!
		local launchModifier_range = FRACUNIT - vector.EncoreScreamMinModifier
		local launchModifier = vector.EncoreScreamMinModifier + FixedMul(distModifier, launchModifier_range)
		if not (source.vector_battlelevel)
			-- only a simple thrust
			local thrust = FixedMul(launchModifier, vector.EncoreScreamThrust)
			P_Thrust(found, angle, FixedMul(source.scale, thrust))
		else
			-- tumble them!
			local level, maxlevel = source.vector_battlelevel - 1, vector.EncoreMaxLevel - 1
				
			local thrust_range = vector.EncoreScreamMaxTumbleThrust - vector.EncoreScreamMinTumbleThrust
			local thrust = vector.EncoreScreamMinTumbleThrust + (level * thrust_range) / maxlevel
			local momz = FixedMul(vector.EncoreScreamTumbleHeightModifier, thrust)
			local tumble = (FixedDiv(thrust, vector.EncoreScreamTumbleTicDivider) * TICRATE) / FRACUNIT
			
			momz = FixedMul(launchModifier, $)
			tumble = FixedMul(launchModifier, $)
			thrust = FixedMul(launchModifier, $)
				
			found.momz = FixedMul(ref.scale, momz) * P_MobjFlip(ref)
			B.DoPlayerTumble(found.player, tumble, angle, FixedMul(ref.scale, thrust), true)
		end
		B.PlayerCreditPusher(found.player, source)
		ref.vector_hitmobjs[found] = true
	end, scream, scream.x - screaminfo["range"], scream.x + screaminfo["range"], scream.y - screaminfo["range"], scream.y + screaminfo["range"])
end

vector.SpawnEncoreScream = function(mo, strength)	-- strength is usually from 0 to FRACUNIT
	if not (mo and mo.valid) then return end	-- need a source
	if (strength == nil) then return end	-- needs a strength!
	
	-- calculate values based on strength
	local range_range = vector.EncoreScreamMaxRange - vector.EncoreScreamMinRange
	local range = vector.EncoreScreamMinRange + FixedMul(strength, range_range)
	local height_range = vector.EncoreScreamMaxHeight - vector.EncoreScreamMinHeight
	local height = vector.EncoreScreamMinHeight + FixedMul(strength, height_range)
	local tics_range = vector.EncoreScreamMaxTics - vector.EncoreScreamMinTics
	local tics = vector.EncoreScreamMinTics + FixedMul(strength, tics_range)	-- damn i should use FixedMul for non fixed-point values more often
	
	-- the actual damage range
	local scream = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_VECTOR_SCREAM)
	scream.color = SKINCOLOR_AETHER
	if (mo.player)
		P_ColorTeamMissile(scream, mo.player)
	end
	scream.angle = mo.angle
	scream.target = mo
	scream.tics = tics
	scream.vector_screaminfo = {
		range = range,
		startheight = mo.height,
		height = height,	-- added on TOP of startheight
		maxtics = tics,
	}
	vector.EncoreScreamHitbox(scream)	-- make sure the hitbox exists on the first tic
	
	if mo.player == displayplayer
		local quakeIntensity_range = vector.EncoreMaxShakeIntensity - vector.EncoreMinShakeIntensity
		local quakeIntensity = vector.EncoreMinShakeIntensity + FixedMul(strength, quakeIntensity_range)
		local quakeTics_range = vector.EncoreMaxShakeTics - vector.EncoreMinShakeTics
		local quakeTics = vector.EncoreMinShakeTics + FixedMul(strength, quakeTics_range)
		P_StartQuake(quakeIntensity, quakeTics)
	end
	
	-- show range (debug)
	if (vector.cv_showrange.value)
		local trig_dist = (range / FRACUNIT)	-- needs to not be fixed point
			
		-- height
		local height_thoks = 5
		for hthoks = -(height_thoks / 2), height_thoks / 2
			local zoffset = (hthoks * FRACUNIT) / (height_thoks / 2)
				
			local angle_thoks = 8
			for athoks = 0, (angle_thoks - 1)
				-- spawn angle
				local trigAngle = vector.EncoreScreamAngle - ((athoks * (vector.EncoreScreamAngle * 2)) / (angle_thoks - 1))
				trigAngle = mo.angle + ($ * ANG1)
						
				local length_thoks = 8
				for lthoks = 0, length_thoks - 1
					-- define the lining
					if (athoks and not lthoks) then continue end	-- don't stack on the start point
						
					if not (athoks == (angle_thoks - 1) or not athoks)
					and not (lthoks == (length_thoks - 1) or not lthoks)
						continue
					end
						
					if (hthoks % 2)
					and not ((lthoks == (length_thoks - 1) or not lthoks) and (athoks == (angle_thoks - 1) or not  athoks))
						continue
					end
						
					-- distance
					local thok_dist = (lthoks * trig_dist) / (length_thoks - 1)
					local xpos = mo.x + thok_dist*cos(trigAngle)
					local ypos = mo.y + thok_dist*sin(trigAngle)
						
					-- height
					local heightOffset = FixedMul((lthoks * FRACUNIT) / (length_thoks - 1), height)
					local bottomz = mo.z - heightOffset
					local topz = mo.z + mo.height + heightOffset
					local zpos = mo.z + FixedMul(zoffset, (topz - bottomz) / 2)
						
					-- spawn the thok
					local thok = P_SpawnMobj(xpos, ypos, zpos, MT_THOK)
					thok.frame = ($ & ~(FF_TRANSMASK)) | (FF_FULLBRIGHT|FF_TRANS70)
					thok.scale = FixedMul(FRACUNIT/2, $)
					thok.color = SKINCOLOR_WHITE
					thok.tics = tics
				end
			end
		end
	end
end

chaotix.AddHook("MobjThinker", function(scream)
	-- make sure it's a valid scream
	local source = scream.target
	local screaminfo = scream.vector_screaminfo
	
	if not (source and source.valid)	-- no source???
	or not (screaminfo)	-- no scream info?
		P_RemoveMobj(scream)
		return
	end
	
	-- scream at stuff
	vector.EncoreScreamHitbox(scream)	-- make sure the hitbox exists on the first tic
end, MT_VECTOR_SCREAM)