-- vector's record rush
local vector = chaotix.vector

-- functions to init and reset vars
local function vector_resetRiptire(mo)
	-- accleration related vars
	if (mo.vector_riptire_rollangle)
		mo.rollangle = 0
		mo.vector_riptire_rollangle = nil
	end
	mo.vector_riptire_lockout = nil
	
	-- spindash related vars
	mo.vector_riptire_spindash = nil
	mo.vector_riptire_dashspeed = nil
	mo.vector_riptire_frame = nil
	mo.vector_riptire_frametics = nil
	mo.vector_riptire_wasjumping = nil
	mo.vector_riptire_engine = nil
	mo.vector_riptire_prevsound = nil
	mo.vector_riptire_leveltime = nil
	mo.vector_riptire_speedup = nil
	
	-- bobbing resetting
	if (mo.vector_riptire_bobtics)
		mo.vector_riptire_bobtics = nil
		mo.spriteyoffset = 0
	end
	
	-- stretch resetting
	if (mo.vector_riptire_stretch)
		mo.vector_riptire_stretch = nil
		mo.spritexscale = FRACUNIT
		mo.spriteyscale = FRACUNIT
	end
end

-- the main function
local function angtoint(a)
	if not a return 0 end
	a = $/ANG1
	if a < 0
		a = $+360
	end
	return a
end

vector.AccelSpindash = function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	
	-- always have this running
	if (mo.vector_riptire_spin == nil) then mo.vector_riptire_spin = 0 end
	mo.vector_riptire_spin = (p.cmd.buttons & BT_SPIN) and $+1 or 0
	
	-- checks
	if not (mo.health)
	or not (mo.skin == vector.SKIN)
		vector_resetRiptire(mo)
		return
	end
	
	if not (mo.vector_riptire_spindash)	-- attempt to accelerate!
		-- make sure we're rolling!
		if not ((p.pflags & PF_SPINNING) and not (p.pflags & PF_STARTDASH)
		and p.powers[pw_super] != 1)
			vector_resetRiptire(mo)
			return
		end
		
		-- attempt to start the spindash!
		if (P_IsObjectOnGround(mo))	-- need to be grounded
		and (mo.vector_riptire_spin == 1)	-- need a spin input (and make sure it's a press!)
		and (p.cmd.forwardmove < -16)	-- backwards input
		and not (mo.vector_riptire_lockout)	-- dont kill ourself!
		and not (CBW_Battle and (p.gotflag or p.gotcrystal))	-- we dont have the flag
			local old_engine = mo.vector_riptire_engine
			vector_resetRiptire(mo)
			mo.vector_riptire_engine = old_engine	-- little hack to keep the engine
			
			-- setup spindash vars
			mo.vector_riptire_spindash = true
			mo.vector_riptire_frame = 0
			mo.vector_riptire_frametics = 3
			
			-- transfer current speed into dashspeed!
			local multiplier = vector.RiptireSpindashSpeedMultiplier	-- adjust the speed based on a modifier
			local speed = FixedMul(multiplier, FixedDiv(FixedHypot(mo.momx, mo.momy), mo.scale))
			mo.vector_riptire_dashspeed = max(p.mindash, min(p.maxdash, speed))
			S_StartSound(mo, sfx_spndsh)
			
			-- set player state
			p.pflags = $|(PF_FULLSTASIS)
			mo.state = S_VECTOR_SPINDASH
			return
		end
		
		-- reduce stretch, if we have it!
		if (mo.vector_riptire_stretch)
			mo.vector_riptire_stretch = $-1
			if (mo.vector_riptire_stretch)
				-- x stretch (LESS than FRACUNIT)
				local xrange = FRACUNIT - vector.RiptireXStretch
				local xscale = FRACUNIT - ((mo.vector_riptire_stretch * xrange) / vector.RiptireStretchTics)
				mo.spritexscale = xscale
				
				-- y stretch (MORE than FRACUNIT)
				local yrange = vector.RiptireYStretch - FRACUNIT
				local yscale = FRACUNIT + ((mo.vector_riptire_stretch * yrange) / vector.RiptireStretchTics)
				mo.spriteyscale = yscale
			else
				-- reset stuff
				mo.spritexscale = FRACUNIT
				mo.spriteyscale = FRACUNIT
			end
		end
		
		-- change jump sound while spinning!
		if (S_SoundPlaying(mo, sfx_jump))
		and not (mo.vector_riptire_wasjumping)
			S_StartSound(mo, sfx_vcjump)
			S_StopSoundByID(mo, sfx_jump)
			
			-- handle jump stretch for rolling
			mo.vector_riptire_stretch = vector.RiptireStretchTics
			mo.spritexscale = vector.RiptireXStretch
			mo.spriteyscale = vector.RiptireYStretch
		end
		mo.vector_riptire_wasjumping = (p.pflags & PF_JUMPED)
		
		-- make sure this var exists
		if (mo.vector_riptire_rollangle == nil) then mo.vector_riptire_rollangle = 0 end
		
		-- lower lockout tics if they exist
		mo.vector_riptire_lockout = $ and $-1
		
		local accelerating = 0	-- just to make my life easier
		if not (mo.vector_riptire_lockout)
		and (p.cmd.buttons & BT_SPIN)
			-- laziness check for later?
			accelerating = 1
			
			if (P_IsObjectOnGround(mo))	-- only accelerate on the ground
				-- laziness check for later?
				accelerating = 2
				
				-- get which angle most of our momentum is currently going!
				local momentumangle = R_PointToAngle2(0, 0, mo.momx, mo.momy) - mo.angle
				local speed = FixedHypot(mo.momx, mo.momy)
					
				-- first stabilize angle
				local stableangle = momentumangle + ANGLE_90
				local stablefactor = cos(stableangle)
				local stablespeed = FixedMul(stablefactor, speed)
				local stabilize = 0	-- trust me on this
					
				if (stablespeed)	-- we have momentum on our 90 degree angles
					stabilize = min(abs(stablespeed), FixedMul(mo.scale, vector.RiptireStabilizer))	-- we want to end up going straight
					if (stablefactor < 0) then stabilize = $ * -1 end
					P_Thrust(mo, mo.angle + ANGLE_90, chaotix.WaterMom(mo, stabilize))
				end
					
				-- now accel to max speed!
				local speedup = (p.powers[pw_sneakers] or p.powers[pw_super])
				local anglefactor = cos(momentumangle)
				local anglespeed = FixedMul(anglefactor, speed)	-- this isnt local so we can save the maxspeed afterwards!
				
				local maxspeedboost = (CBW_Battle and (p.gotflag or p.gotcrystal)) and FRACUNIT or vector.RiptireSpeedBoost
				local maxspeed = FixedMul(mo.scale, FixedMul(maxspeedboost, p.normalspeed))
				if (speedup) then maxspeed = FixedMul(vector.RiptireSneakerSpeedBoost, maxspeed) end
					
				if (anglespeed < maxspeed)
					-- bob if we're accelerating from a slow speed
					local bobspeed = FixedMul(vector.RiptireBobThreshold, maxspeed)
					if (anglespeed < bobspeed)
					and not (mo.vector_riptire_bobtics)
						mo.vector_riptire_bobtics = vector.RiptireBobTics
					end
					
					-- get acceleration
					local accel = FixedMul(mo.scale, vector.RiptireAccelValue)
					accel = $ + FixedMul(vector.RiptireStabilizerModifier, abs(stabilize))	-- make up a bit of momentum from stabilizing
					accel = min(maxspeed - anglespeed, $)	-- dont go above max speed
					if (anglespeed < 0)
						-- turn around quicker
						accel = FixedMul(vector.RiptireAccelBoost, $)
					end
					if (speedup) then accel = FixedMul(vector.RiptireSneakerAccelBoost, $) end
					
					-- thrust forward
					P_Thrust(mo, mo.angle, chaotix.WaterMom(mo, accel, FRACUNIT/3))	-- gain less underwater
				end
			end
			
			-- change the engine based on speed
			local engine_speed = FixedHypot(mo.momx, mo.momy)
			local engine_maxspeedboost = (p.gotflag or p.gotcrystal) and FRACUNIT or vector.RiptireSpeedBoost
			local engine_maxspeed = FixedMul(mo.scale, FixedMul(FixedMul(engine_maxspeedboost, p.normalspeed), vector.RiptireSneakerSpeedBoost))
			
			local targ_engine = max(1, (engine_speed * 100) / engine_maxspeed)
			mo.vector_riptire_engine = $ or 1	-- init this value
			
			if (mo.vector_riptire_engine < targ_engine)
				mo.vector_riptire_engine = $ and min(targ_engine, $+vector.RiptireEngineAdjust) or vector.RiptireEngineAdjust
			elseif (mo.vector_riptire_engine > targ_engine)
				mo.vector_riptire_engine = $ and max(targ_engine, $-vector.RiptireEngineAdjust) or 1
			end
			
			-- go into spin frames when going fast enough!
			local speed = FixedHypot(mo.momx, mo.momy)
			local maxspeed = FixedMul(mo.scale, FixedMul(vector.RiptireSpeedBoost, p.normalspeed))
			
			if (speed > FixedMul(vector.RiptireSpinModifier, maxspeed))
				chaotix.SetFrame(mo, leveltime%4, SPR2_SPIN)
				if not (mo.vector_riptire_speedup)
					mo.vector_riptire_speedup = true
					S_StartSoundAtVolume(mo, sfx_cdfm40, 144)
				end
			else
				mo.vector_riptire_speedup = false
			end
		else
			-- reduce "engine"
			mo.vector_riptire_engine = $ and max(0, $-vector.RiptireEngineAdjust) or 0
		end
			
		-- roll angle handling
		if (P_IsObjectOnGround(mo))	-- only update angle when we're on the ground (could make for some sick effects!)
		and not (mo.vector_riptire_lockout)
			local targetrollangle = 0
				
			if (P_IsObjectOnGround(mo))
			and (p.cmd.buttons & BT_SPIN)
				local maxrollangle = vector.RiptireMaxRollAngle
				targetrollangle = (mo.angle - R_PointToAngle2(0, 0, mo.momx, mo.momy)) / ANG1
				targetrollangle = max(-(maxrollangle), min(maxrollangle, targetrollangle)) -- min/max it
			end
			
			if (mo.vector_riptire_rollangle > targetrollangle)
				mo.vector_riptire_rollangle = max(targetrollangle, $ - vector.RiptireRollAngleChange)
			elseif (mo.vector_riptire_rollangle < targetrollangle)
				mo.vector_riptire_rollangle = min(targetrollangle, $ + vector.RiptireRollAngleChange)
			end
		end
		mo.rollangle = mo.vector_riptire_rollangle * ANG1	-- leave this outside of the block so we don't spontaneously die under spinheight gaps
			
		if (abs(mo.vector_riptire_rollangle) >= vector.RiptireDustRollAngle)	-- spawn dust for hard angles
		and (P_IsObjectOnGround(mo))	-- we need to be on the ground for it
			local radius = mo.radius >> FRACBITS
			local xoff = P_RandomRange(-(radius), radius) << FRACBITS
			local yoff = P_RandomRange(-(radius), radius) << FRACBITS
			local speed = 3
				
			local dust = P_SpawnMobjFromMobj(mo, xoff, yoff, 0, MT_SPINDUST)
			dust.scale = $*2/3
			dust.destscale = FRACUNIT*2
			dust.scalespeed = dust.scale/12
			dust.momx = mo.momx + (P_RandomRange(-(speed), speed) << FRACBITS)
			dust.momy = mo.momy + (P_RandomRange(-(speed), speed) << FRACBITS)
			P_SetObjectMomZ(dust, P_RandomRange(1, 6)*FRACUNIT)
			if (mo.eflags & (MFE_TOUCHWATER|MFE_UNDERWATER))
				dust.state = S_SPINDUST_BUBBLE1
			end
			
			-- play sound
			if not (mo.eflags & (MFE_TOUCHWATER|MFE_UNDERWATER))
				if not (leveltime%6) then S_StartSound(mo, sfx_vcdrft) end
			else
				if not (leveltime%4) then S_StartSound(mo, sfx_bubbl1 + P_RandomRange(0, 4)) end
			end
			
			-- add bobbing if not done so already
			if (mo.vector_riptire_lockout)
				-- nevermind, REMOVE bobbing
				mo.vector_riptire_bobtics = 0
				mo.spriteyoffset = 0
			else
				if not (mo.vector_riptire_bobtics)
					mo.vector_riptire_bobtics = vector.RiptireBobTics
				end
			end
		end
		
		-- animate the bobbing
		if (mo.vector_riptire_bobtics)
			mo.vector_riptire_bobtics = $-1
			if (mo.vector_riptire_bobtics)
				local bobtics = vector.RiptireBobTics - mo.vector_riptire_bobtics
				local multiplier = abs(cos(ease.outquart((bobtics * FRACUNIT) / vector.RiptireBobTics, 90, 270) * ANG1))
				mo.spriteyoffset = chaotix.WaterMom(mo, FixedMul(multiplier, vector.RiptireBobHeight))
			else
				mo.spriteyoffset = 0
			end
		end
		
		-- vector's spin trail (after images)
		local min_trailspeed = FixedMul(15<<FRACBITS, mo.scale)
		local speed = P_AproxDistance(p.speed, mo.momz)
		
		if (speed >= min_trailspeed)
		and not ((p.pflags & PF_JUMPED) and not accelerating)
			local maxspeed = FixedMul(mo.scale, FixedMul(vector.RiptireSpeedBoost, p.normalspeed))
			speed = $ - min_trailspeed	-- easier for calcs
			local max_trailspeed = FixedMul(mo.scale, vector.RiptireMaxGhostThreshold) - min_trailspeed
			
			if (leveltime%2)
				local g = P_SpawnGhostMobj(mo)
				g.tics = max(1, min(vector.RiptireMaxGhosts, (speed * vector.RiptireMaxGhosts) / max_trailspeed)) + 1
			end
			
			-- spawn speedlines!
			if (accelerating)
				local frequency_range = vector.RiptireMinSpeedlines - vector.RiptireMaxSpeedlines
				local frequency = max(vector.RiptireMaxSpeedlines, vector.RiptireMinSpeedlines - ((speed * frequency_range) / max_trailspeed))
				chaotix.SpawnSpeedTrail(mo, not (leveltime % frequency),
				{radius = mo.radius*3/2})
			end
		end
	else	-- spindash handling!
		-- cancel it early!
		if (P_PlayerInPain(p))	-- we got hurt
		or (mo.eflags & MFE_SPRUNG)	-- hit a spring
		or (p.powers[pw_carry])	-- we got carried
		or (p.panim == PA_PAIN)	-- pain pain pain pain pain
		or (p.powers[pw_nocontrol])	-- no control
		or (p.actionstate)	-- battle support
		or (p.airdodge)	-- battle support (2)
			vector_resetRiptire(mo)
			return
		end
		
		-- release the spindash!
		if not (p.cmd.buttons & BT_SPIN)
			p.pflags = ($|PF_SPINNING) & ~(PF_FULLSTASIS)
			mo.state = S_PLAY_ROLL
			P_InstaThrust(mo, mo.angle, FixedMul(mo.scale, mo.vector_riptire_dashspeed))
			S_StartSound(mo, sfx_zoom)
			
			local old_engine = mo.vector_riptire_engine
			vector_resetRiptire(mo)
			mo.vector_riptire_engine = old_engine	-- little hack to keep the engine
			return
		end
		
		-- set state
		if (mo.state != S_VECTOR_SPINDASH) then mo.state = S_VECTOR_SPINDASH end
		
		-- get max frametics
		local step = (p.maxdash - p.mindash) / 4
		local speed = (mo.vector_riptire_dashspeed - p.mindash)
		local maxtics = 3	-- default is 3 with no charge
			
		if (speed > (3 * step))
			maxtics = 1
		elseif (speed > step)
			maxtics = 2
		end
		mo.vector_riptire_frametics = $ and min(maxtics, $) or maxtics
		
		-- reduce frame tics and get the frame we want
		local numframes = 5
			
		mo.vector_riptire_frametics = $-1
		if not (mo.vector_riptire_frametics)
			mo.vector_riptire_frame = $+1
				
			-- reset frame after max frame is reached
			if (mo.vector_riptire_frame > (numframes - 1))
				mo.vector_riptire_frame = 0
			end
		end
			
		-- set frame
		chaotix.SetFrame(mo, mo.vector_riptire_frame, SPR2_MLEE)
		
		-- increase our dashspeed
		if (mo.vector_riptire_dashspeed < p.maxdash)
			local soundcalculation = (6 * (mo.vector_riptire_dashspeed - p.mindash)) / (p.maxdash - p.mindash)
			mo.vector_riptire_dashspeed = $ + FRACUNIT
				
			-- play the rev sound based on local vars
			local chargecalculation = (6 * (mo.vector_riptire_dashspeed - p.mindash)) / (p.maxdash - p.mindash)
			if (soundcalculation != chargecalculation)
				S_StartSound(mo, sfx_spndsh)
			end
		end
		
		-- handle stasis
		p.pflags = $|(PF_FULLSTASIS)
		
		-- spawn dust from skidding!
		local speed = FixedDiv(FixedHypot(mo.momx, mo.momy), mo.scale)
			
		if (P_IsObjectOnGround(mo))
		and (speed >= vector.RiptireSpindashDustThreshold)
			local speed = 3
			local dust = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_SPINDUST)
			dust.scale = $/2
			dust.destscale = FRACUNIT*2
			dust.scalespeed = dust.scale/12
			dust.momx = mo.momx + (P_RandomRange(-(speed), speed) << FRACBITS)
			dust.momy = mo.momy + (P_RandomRange(-(speed), speed) << FRACBITS)
			P_SetObjectMomZ(dust, P_RandomRange(1, 6)*FRACUNIT)
			if (mo.eflags & (MFE_TOUCHWATER|MFE_UNDERWATER))
				dust.state = S_SPINDUST_BUBBLE1
			end
			if not (leveltime%6) then S_StartSound(mo, sfx_vcdrft) end
		end
		
		-- face the direction we'll launch
		local diff = mo.angle - p.drawangle
		local factor = 8
		if (diff)
			diff = (angtoint(diff) > 180) and InvAngle(InvAngle(diff)/factor) or $/factor
			p.drawangle = $ + diff
		end
		
		-- reduce "engine"
		mo.vector_riptire_engine = $ and max(0, $-vector.RiptireEngineAdjust) or 0
	end
	
	-- play engine
	if (mo.vector_riptire_spindash)
	or ((p.pflags & PF_SPINNING) and not (p.pflags & PF_STARTDASH))
		-- play engine sound
		local engine = mo.vector_riptire_engine
		mo.vector_riptire_leveltime = $ or leveltime
		local engine_leveltime = leveltime - mo.vector_riptire_leveltime
		
		if (engine)
		and not (engine_leveltime%5)
			local sound = max(1, min(vector.RiptireEngineSounds, (engine * vector.RiptireEngineSounds) / 99))	-- slightly lower threshold for the sound
			
			-- stop the old sound
			if (mo.vector_riptire_prevsound)
				S_StopSoundByID(mo, _G["sfx_vceng"..mo.vector_riptire_prevsound])
			end
			
			-- play sound, then store it
			S_StartSoundAtVolume(mo, _G["sfx_vceng"..sound], vector.RiptireEngineVolume)
			mo.vector_riptire_prevsound = sound
		end
	end
end

-- make it break spikes
vector.AccelSpindashCollide = function(mo, mo2)
	-- object checks
	if not ((mo and mo.valid) and mo.health) then return end
	if not ((mo2 and mo2.valid) and mo2.health) then return end
	
	-- checks
	local p = mo.player
	if not (mo.skin == vector.SKIN) then return end
	
	-- we need to be rolling!
	if not (p.pflags & PF_SPINNING)
	and not (mo.vector_riptire_spindash)
	and not (p.climbing and (mo.vector_beatbash_scalespeed != nil))
		return
	end
	
	-- roll straight through spikes!
	if (mo2.type == MT_SPIKE or mo2.type == MT_WALLSPIKE)
		-- adjust this based on momz
		local topheight = (mo2.z + mo2.height) - mo.momz
		local bottomheight = mo2.z - mo.momz
		
		if ((mo.z <= topheight)
		and (bottomheight <= mo.z + mo.height))
			P_KillMobj(mo2, mo, mo)
			return false	-- they're getting destroyed, we WOULDNT collide
		end
	end
end
addHook("MobjCollide", vector.AccelSpindashCollide, MT_PLAYER)
addHook("MobjMoveCollide", vector.AccelSpindashCollide, MT_PLAYER)

addHook("PlayerHeight", function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	if not (mo.skin == vector.SKIN) then return end
	
	-- while doing the special spindash, make sure we have spinheight!
	if (mo.vector_riptire_spindash)
		return p.spinheight
	end
	
	-- when drifting, lower spin height to enter spin gaps!
	if (p.pflags & PF_SPINNING)
	and (mo.vector_riptire_rollangle)	-- validation
	and (abs(mo.vector_riptire_rollangle) >= vector.RiptireDustRollAngle)	-- consider this state "drifting"
		return vector.RiptireDriftHeight
	end
end)
addHook("PlayerCanEnterSpinGaps", function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	if not (mo.skin == vector.SKIN) then return end
	
	if (p.pflags & PF_SPINNING)
	and (mo.vector_riptire_rollangle)	-- validation
	and (abs(mo.vector_riptire_rollangle) >= vector.RiptireDustRollAngle)	-- consider this state "drifting"
		p.pflags = $|(PF_JUMPSTASIS)	-- prevent jumping for instant death (might move this to the main function)
		mo.vector_riptire_lockout = 2
		return true
	end
end)

addHook("PlayerCanDamage", function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	if not (mo.skin == vector.SKIN) then return end
	
	if (mo.vector_riptire_spindash)
		return true
	end
end)