local CV_HardHeavy = CV_RegisterVar({
	name = 'heavy_hardmode',
	defaultvalue = "Off",
	flags = CV_NETVAR|CV_SHOWMODIF,
	PossibleValue = CV_OnOff,
})

local Heavy = chaotix.Heavy

Heavy.GroundPound = 1 << 0
Heavy.AirCharge = 1 << 1

Heavy.GPStart = 15

local sfx_chargejump = sfx_steam2
local sfx_chargejump2 = sfx_s3kd1l
local sfx_heavydrop = sfx_s3k51
local sfx_heavydrop2 = sfx_hevfal
local sfx_heavywalk = sfx_s3k98
local sfx_heavyrun = sfx_s3k98
-- local sfx_heavyrun = sfx_s3k96
local sfx_heavylanding = sfx_s3k98
-- local sfx_heavylanding = sfx_s3k96
-- local sfx_heavylanding = sfx_pstop
local sfx_heavycrash = sfx_pstop
local sfx_heavycrash2 = sfx_s3k9b
local sfx_heavytink = sfx_cdfm46
-- local sfx_wallbump = sfx_s3k49
local sfx_wallbump = sfx_s3k5d
-- local sfx_enemybump = sfx_s3k5d
local sfx_phantomjump = sfx_hevpj
local sfx_heavydeathpit = sfx_hevdie

local getDust = function(mo, mt)
	return mo.eflags & MFE_UNDERWATER and MT_MEDIUMBUBBLE or mt or MT_DUST
end

local initTable = function(mo)
	if type(mo.heavyvars) != 'table'
		mo.heavyvars = {}
	end
	local t = mo.heavyvars
	t.fastcharge = false
	t.groundpound = 0
	t.chargejump = false
	t.dashspeed = 0
	t.buttonslast = mo.player.cmd.buttons
	t.tics = 0
	t.time = 0
	t.spritexscale = mo.spritexscale
	t.spriteyscale = mo.spriteyscale
	t.spritexoffset = mo.spritexoffset
	t.spriteyoffset = mo.spriteyoffset
	t.lasthitfloor = mo.eflags & MFE_JUSTHITFLOOR
	t.lastrolling = S_PLAY_ROLL
	t.visualstomp = false
	t.hitstun = 0
	t.wallcollide = 0
	t.lastmomz = mo.momz
	t.lavatimer = 0
	t.pow = 0
	t.powangle = 0
end

local resetVisual = function(mo)
	local t = mo.heavyvars
	t.visualstomp = false
	mo.spritexscale = t.spritexscale
	mo.spriteyscale = t.spriteyscale
end
local resetHeavy = function(mo)
	local t = mo.heavyvars
	if mo.player.pflags & PF_SPINNING
		mo.player.pflags = $ &~ PF_SPINNING
		if mo.state == S_PLAY_STND
			mo.state = S_PLAY_WALK
		end
	end
	mo.spritexscale = t.spritexscale
	mo.spriteyscale = t.spriteyscale
	t.fastcharge = false
	t.groundpound = 0
	t.chargejump = false
	t.dashspeed = 0
	t.tics = 0
	t.time = 0
	t.lasthitfloor = mo.eflags & MFE_JUSTHITFLOOR
	t.lastrolling = S_PLAY_ROLL
	t.visualstomp = false
	t.wallcollide = 0
	t.lavatimer = 0
	t.pow = 0
end

local heavyCanAct = function(player, stasis)
	return not(
		P_PlayerInPain(player) or player.playerstate != PST_LIVE
		or player.powers[pw_nocontrol]
		or (stasis and player.pflags & stasis)
		or player.exiting
		or (player.powers[pw_carry] != player.heavy_lastcarry)
		or (player.powers[pw_carry] and player.powers[pw_carry] != CR_ROLLOUT
		and player.powers[pw_carry] != CR_ROPEHANG
		and player.powers[pw_carry] != CR_PLAYER)
		or player.pflags & PF_SHIELDABILITY
		or player.mo and player.mo.flags & MF_SCENERY
	)
end

local startChargeJump = function(mo)
	local t = mo.heavyvars
	if t.chargejump == false
		t.chargejump = true
		if t.visualstomp
			resetVisual(mo)
		else
			t.spritexscale = mo.spritexscale
			t.spriteyscale = mo.spriteyscale
		end
		mo.time = 0
		S_StartSound(mo, sfx_chargejump)
		local thrust = mo.scale * 6
		local mt = getDust(mo)
		for n = 1, 12 do
			local angle = FixedAngle(FRACUNIT * n * 30)
			local x = P_ReturnThrustX(nil, angle, thrust)
			local y = P_ReturnThrustY(nil, angle, thrust)
			local fx = P_SpawnMobjFromMobj(mo, x, y, 0, mt)
			if fx and fx.valid
				fx.momx = x
				fx.momy = y
				P_SetObjectMomZ(fx, FRACUNIT*2)
			end
		end
	end
end

local endChargeJump = function(mo)
	local t = mo.heavyvars
	if t.chargejump == true
		t.chargejump = false
		mo.spritexscale = t.spritexscale
		mo.spriteyscale = t.spriteyscale
		mo.time = 0
		S_StopSoundByID(mo, sfx_chargejump)
		S_StopSoundByID(mo, sfx_chargejump2)
	end
end

local doNearQuake = function(mo, dist, intensity, time)
	dist, intensity, time =
		$ or mo.scale * 640, 
		$ or mo.scale * 2,
		$ or 5
		
	if R_PointToDist(mo.x, mo.y) < dist
		P_StartQuake(intensity, time)
	end
end

local doCoil = function(player)
	local mo = player.mo
	local amt = FixedDiv(mo.heavyvars.dashspeed, player.maxdash)
	local jumpfactor1 = player.jumpfactor
	local jumpfactor2 = player.actionspd/10
	local thrust = FixedMul(jumpfactor1, FRACUNIT-amt) + FixedMul(jumpfactor2, amt)
	thrust = FixedMul($, mo.scale * 39/4)
	if mo.eflags & (MFE_UNDERWATER|MFE_GOOWATER)
		thrust = $*11/20
	end
	if not P_IsObjectOnGround(mo) and not player.powers[pw_carry]
		S_StartSound(mo, sfx_hevpj)
		player.pflags = $ &~ PF_JUMPED
		P_DoJump(player, false)
		player.pflags = $ | P_GetJumpFlags(player)
		mo.state = player.charflags & SF_NOJUMPSPIN and S_PLAY_JUMP or S_PLAY_ROLL
	else
		P_DoJump(player, true)
		mo.momz = P_MobjFlip(mo) * thrust
	end
	if not(player.cmd.buttons & BT_JUMP)
		player.pflags = $ &~ PF_STARTJUMP
	end
	mo.heavyvars.groundpound = 0
end

local doShockwave = function(player, range)
	local mo = player.mo
	-- Do aesthetic
	for n = 1, 24 do
		-- Do radius dust (Telegraphs the max shockwave distance)
		local angle = FixedAngle(FRACUNIT * n * 15)
		local dust = P_SpawnMobjFromMobj(mo, 
			P_ReturnThrustX(nil, angle, range), 
			P_ReturnThrustY(nil, angle, range),
			0, MT_DUST)
		dust.scale = $<<1
		dust.z = min(max($, dust.floorz), dust.ceilingz - dust.height)
		P_SetObjectMomZ(dust, FRACUNIT)
	
		-- Do more aesthetic dust
		for n = 1, 2 do
			angle = $+ANG15/n
			local dust = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_SPINDUST)
			dust.momx = P_ReturnThrustX(nil, angle, range>>n)
			dust.momy = P_ReturnThrustY(nil, angle, range>>n)
			P_SetObjectMomZ(dust, FRACUNIT*n)
		end
	end
	-- Do object interactions, stealing Mighty's behavior!
	searchBlockmap("objects", function(pmo, found)
		return chaotix.mighty.doHammerDropSearch(pmo, found, range, true)
	end, mo, mo.x - range, mo.x + range, mo.y - range, mo.y + range)
end

-- Main thinker
chaotix.AddHook("MobjThinker", function(mo)
	local player = mo.player
	if not(player)
		return
	end
	
	if not Heavy.CheckStatus(player)
		if mo.heavyvars
			resetVisual(mo)
			mo.heavyvars = nil
			if (mo.state >= S_HEAVY_POW_START and mo.state <= S_HEAVY_POW_END)
				mo.state = S_PLAY_STND
			end
		end
		return
	elseif mo.heavyvars == nil
		initTable(mo)
	end
	
	-- POW haha!
	if (mo.state >= S_HEAVY_POW_START and mo.state <= S_HEAVY_POW_END)
		-- remove skidding
		player.pflags = $|(PF_FULLSTASIS)
		player.charflags = $|(SF_NOSKID)
		player.heavy_noskid = true
		
		-- lock angle
		player.drawangle = mo.heavyvars.powangle
		
		-- spawn dust
		if (P_IsObjectOnGround(mo))
		and (FixedHypot(mo.momx, mo.momy))
			local speed = 3
			
			local dust = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_SPINDUST)
			dust.scale = $/2
			dust.destscale = FRACUNIT*2
			dust.scalespeed = dust.scale/12
			dust.momx = mo.momx + (P_RandomRange(-(speed), speed) << FRACBITS)
			dust.momy = mo.momy + (P_RandomRange(-(speed), speed) << FRACBITS)
			P_SetObjectMomZ(dust, P_RandomRange(1, 6)*FRACUNIT)
			if (mo.eflags & (MFE_TOUCHWATER|MFE_UNDERWATER))
				dust.state = S_SPINDUST_BUBBLE1
			end
			if not (leveltime%6) then S_StartSound(mo, sfx_s3k7e) end
		end
		
		-- state handling
		if (mo.state == S_HEAVY_POW_START)	-- start the shot!
			mo.heavyvars.pow = $+1
			if (mo.heavyvars.pow >= TICRATE)
				mo.state = S_HEAVY_POW_WINDUP
			end
		elseif (mo.state == S_HEAVY_POW_FIRE)
			-- frame handling
			local frame = 0
			if (mo.tics < (states[S_HEAVY_POW_FIRE].tics - 3))
				frame = 1
			end
				
			local frameflags = (mo.frame & ~FF_FRAMEMASK)
			mo.frame = frame|frameflags
		end
		
		-- don't do anything else
		return
	elseif (player.heavy_noskid)
		-- reset skidding
		player.charflags = $ & ~(SF_NOSKID)
		player.heavy_noskid = false
	end
	
	if heavyCanAct(player) == false
		if not(mo.flags & MF_SCENERY)
			resetHeavy(mo)
		end
		return
	end
	
	local heavylanding = false
	local skin = skins[mo.skin]
	local grounded = P_IsObjectOnGround(mo) or player.powers[pw_carry] == CR_ROLLOUT
	
	-- Skidding dust
	if grounded and mo.state == S_PLAY_CHARGE
	and FixedDiv(player.speed, mo.scale) > player.runspeed
	and abs(R_PointToAngle2(0, 0, player.rmomx, player.rmomy) - mo.angle) > ANGLE_45
		local dust = getDust(mo, MT_SPINDUST)
		local x = P_RandomRange(-mo.radius/FRACUNIT, mo.radius/FRACUNIT) * FRACUNIT
		local y = P_RandomRange(-mo.radius/FRACUNIT, mo.radius/FRACUNIT) * FRACUNIT
		P_SpawnMobjFromMobj(mo, x, y, 0, dust)
		S_StartSound(mo, sfx_skid)
	end
	
	-- Runspeed regulation
	if player.powers[pw_sneakers] or player.powers[pw_super]
		-- Sneakers / Super
		player.normalspeed = skin.normalspeed * 3
	elseif grounded and not (mo.flags2 & MF2_TWOD or twodlevel) -- Screw 2D
		-- Grounded movement
		local momentum = FixedDiv(R_PointToDist2(0,0, player.rmomx,player.rmomy), mo.scale)
		momentum = FixedDiv($, mo.friction or FRACUNIT) -- Compensate for friction
		if mo.eflags & MFE_UNDERWATER
			momentum = $<<1 -- Adjust for underwater physics
		end
		local decel = mo.friction / 2
		player.normalspeed = max(skin.normalspeed, momentum - decel) -- Maintain current momentum, with degradation
	elseif mo.state == S_PLAY_ROLL
		-- Air movement (rolling)
		player.normalspeed = FRACUNIT*36
	else
		-- Air movement (not rolling)
		player.normalspeed = skin.normalspeed
	end
	
	-- Heavyweight
	if mo.heavyvars.groundpound
		if grounded
			mo.state = S_PLAY_LAND
			player.panim = PA_ABILITY
			
			-- Do shockwave
			if mo.heavyvars.groundpound >= 2
				S_StartSound(mo, sfx_heavycrash2)
				S_StopSoundByID(mo, sfx_heavydrop2)
				doShockwave(player, mo.scale * 512)
				doNearQuake(mo, mo.scale * 5120, mo.scale * 20, 10)
			else
				S_StartSound(mo, sfx_heavycrash)
				S_StopSoundByID(mo, sfx_heavydrop)
				doShockwave(player, mo.scale * 256)
				doNearQuake(mo, mo.scale * 1280, mo.scale * 10, 10)
			end
			
			//player.charflags = $ &~ nativebustwalls
			mo.heavyvars.groundpound = 0
			heavylanding = true
			if mo.heavyvars.lastrolling
				-- Do rebound
				mo.momz = mo.scale * 12 -- lastmomz resets before i can use it here. should investigate, not high priority though
				player.pflags = ($|PF_JUMPED) &~ PF_NOJUMPDAMAGE
				mo.state = S_PLAY_ROLL
			end
		elseif not (player.pflags & PF_THOKKED) -- Something has forced us out of thokked state, so we should exit this move
			mo.heavyvars.groundpound = 0
		else -- Falling FX and state regulation
			if mo.state == S_PLAY_BOUNCE_LANDING and not mo.heavyvars.tics
				mo.sprite2 = SPR2_FALL
				if mo.tics == 1
					mo.state = S_PLAY_FALL
					mo.tics = -1
				end
			elseif player.panim >= PA_IDLE and player.panim <= PA_RUN
				mo.state = S_PLAY_FALL
				mo.tics = -1
			elseif mo.state != S_PLAY_ROLL
			and not (mo.frame & FF_SPR2ENDSTATE)
				-- Pause sprite animation
				mo.tics = -1
			end
			
			-- Strong
			player.panim = PA_ABILITY
			if STR_ANIM
				player.powers[pw_strong] = STR_BOUNCE|STR_HEAVY|STR_SPRING|STR_SPIKE 
			end
			
			-- Stretch effect
			mo.heavyvars.visualstomp = true
			local stretch = (mo.momz*P_MobjFlip(mo) + Heavy.GPStart*FRACUNIT)/64
			stretch = min(max(FixedDiv($, mo.scale), INT16_MIN), INT16_MAX)
			mo.spritexscale = FRACUNIT + stretch
			mo.spriteyscale = FRACUNIT - stretch
		end
	end
	
	if mo.state == S_PLAY_BOUNCE_LANDING and mo.heavyvars.tics
		mo.sprite2 = SPR2_TAL2
		if mo.tics == 1
			mo.state = S_PLAY_CHARGE
		end
	elseif mo.state == S_PLAY_CHARGE_START
		-- Train rush start
		mo.heavyvars.tics = 0
		if not (player.cmd.buttons & BT_SPIN)
		and player.pflags & PF_DIRECTIONCHAR
			player.drawangle = $ + (mo.angle - $) / 8
		end
	elseif mo.state == S_PLAY_CHARGE
		if grounded and mo.heavyvars.fastcharge
			mo.heavyvars.fastcharge = false
		end
		
		-- Train rush
		if not (player.cmd.buttons & BT_SPIN) and not(mo.heavyvars.tics)
		or mo.heavyvars.tics == 1
			-- Train rush cancel
			if grounded
				mo.state = (FixedDiv(player.speed, mo.scale) >= player.runspeed) and S_PLAY_RUN or S_PLAY_WALK
				player.skidtime = 20
			elseif mo.momz*P_MobjFlip(mo) > 0
				mo.state = S_PLAY_JUMP
			else
				mo.state = S_PLAY_FALL
			end
			player.pflags = $ &~ (PF_SPINNING|PF_STARTDASH)
			//player.charflags = $ &~ nativebustwalls
			mo.heavyvars.tics = 0
		else
			-- Train rush state regulation
			local angle = mo.angle
			player.panim = PA_ABILITY
			if STR_ANIM
				player.powers[pw_strong] = STR_MELEE &~ STR_FLOOR
			end
			player.pflags = $|PF_SPINNING|PF_STARTDASH
			if not(grounded) and not(player.pflags & PF_JUMPED)
				player.pflags =  $ | PF_JUMPED | PF_THOKKED
			end
			if player.powers[pw_carry] == CR_ROLLOUT
				angle = $ + ANGLE_180
			end
			if player.pflags & PF_DIRECTIONCHAR
				player.drawangle = $ + (angle - $) / 8
			end
			
			-- Movement and animation
			if grounded
				-- Accelerate				
				local movescale = FixedMul(mo.eflags & (MFE_UNDERWATER|MFE_GOOWATER) and FRACUNIT*11/20 or FRACUNIT, FRACUNIT)
				angle = mo.angle + FixedAngle(-FRACUNIT*player.cmd.sidemove/2)
				local speed = FixedHypot(player.rmomx, player.rmomy)
				local friction = P_IsObjectOnGround(mo) and speed - FixedMul(FixedMul(speed, mo.friction), FRACUNIT) or 0
				local fwd = player.cmd.forwardmove
				if (mo.flags2 & MF2_TWOD or twodlevel) then fwd = abs(player.cmd.sidemove) end
				fwd = FRACUNIT * ($ + 50) / 100
				local mindash = FixedMul(player.mindash, mo.scale)
				local maxdash = max(FixedMul(player.maxdash, mo.scale), speed + friction)
				local accel, controldash
				if player.powers[pw_sneakers] or player.powers[pw_super]
					accel = FixedMul(movescale, mo.scale)*3/2 + friction
					controldash = FixedMul(movescale, FixedMul(maxdash*3/2, fwd) + FixedMul(mindash, FRACUNIT - fwd))
				else
					accel = FixedMul(movescale, mo.scale) + friction
					controldash = FixedMul(movescale, FixedMul(maxdash, fwd) + FixedMul(mindash, FRACUNIT - fwd))
				end
				if mo.heavyvars.chargejump
					local heavy_dashspeed = mo.heavyvars.dashspeed >> FRACBITS
					local heavy_maxdash = player.maxdash >> FRACBITS
					local modifier = FRACUNIT - ((heavy_dashspeed * (FRACUNIT/2)) / heavy_maxdash)
					controldash = FixedMul(modifier, $)
				end

				P_Thrust(player.powers[pw_carry] and mo.tracer or mo, angle, accel)
							
				-- Speed limit
				player.rmomx = mo.momx - player.cmomx
				player.rmomy = mo.momy - player.cmomy
				speed = min( FixedHypot(player.rmomx, player.rmomy), controldash)
				angle = R_PointToAngle2(0, 0, player.rmomx, player.rmomy)
				player.rmomx = P_ReturnThrustX(nil, angle, speed)
				player.rmomy = P_ReturnThrustY(nil, angle, speed)
				mo.momx = player.rmomx + player.cmomx
				mo.momy = player.rmomy + player.cmomy
				
				-- Control animation speed
				if speed > maxdash - accel
				or mo.heavyvars.tics
					mo.tics = min($, 1)
					mo.heavyvars.fastcharge = true
				elseif speed > mindash - accel
					mo.tics = min($, 2)
				end
			else -- Aerial movement
				-- Stop on a keyframe for a jump-like pose
				if mo.frame % 4 == 1
					mo.tics = $+1
				end
			end
			
			-- Do visual dust
			if mo.tics == 1
				local r = mo.radius>>FRACBITS
				local h = mo.height>>FRACBITS
				local x = P_RandomRange(-r, r)*FRACUNIT
				local y = P_RandomRange(-r, r)*FRACUNIT
				local z = P_RandomRange(0, h)*FRACUNIT
				P_SpawnMobjFromMobj(mo, x, y, z, getDust(mo))
			end
		end
	elseif not mo.heavyvars.groundpound
		-- Neutral state regulation
		if player.panim == PA_ABILITY
			player.panim = 0
		end
		if player.pflags & (PF_SPINNING|PF_STARTDASH) == PF_SPINNING|PF_STARTDASH
			player.pflags = $ &~ (PF_SPINNING|PF_STARTDASH)
		end
	end
	
	-- Charge jump regulation
	if mo.heavyvars.chargejump
		local squished = (mo.ceilingz - mo.floorz) < P_GetPlayerHeight(player)
		if ((player.cmd.buttons & BT_JUMP or squished)
		and (grounded or player.powers[pw_carry]))
			local maxcharge = player.maxdash
			if (player.gotflagdebuff)
				maxcharge = FixedMul(FRACUNIT/3, $)
			end
			
			-- Store energy for jump
			mo.heavyvars.dashspeed = min($ + FRACUNIT * 9/8, maxcharge)
			-- Do glow
			if mo.heavyvars.time % 12 == 0
				local ghost = P_SpawnGhostMobj(mo)
				ghost.blendmode = AST_ADD
				ghost.spritexscale = mo.spritexscale
				ghost.spriteyscale = mo.spriteyscale
			end
			-- Do "kettle" steam
			if mo.heavyvars.time % 6 == 0
			and mo.heavyvars.dashspeed >= maxcharge
				local r = mo.radius>>FRACBITS
				local h = mo.height>>FRACBITS
				local fx = P_SpawnMobjFromMobj(mo,
					P_RandomRange(-r,r)*FRACUNIT,
					P_RandomRange(-r,r)*FRACUNIT,
					P_RandomRange(0,h)*FRACUNIT,
					getDust(mo))
				P_SetObjectMomZ(fx, FRACUNIT*2)
				fx.scale = $>>P_RandomRange(0,2)
			end
			-- Disallow idle state
			if mo.state == S_PLAY_STND
				mo.tics = -1
			elseif mo.state == S_PLAY_WAIT
				mo.state = S_PLAY_STND
				mo.tics = -1
			end
		elseif not (player.cmd.buttons & BT_JUMP or squished)
			if not (mo.state == S_PLAY_LAND)	-- so we can CHARGE it, but not release it early
				local state = mo.state
				-- Perform jump
				doCoil(player)
				if state == S_PLAY_CHARGE
					mo.state = state
				end
			end
			
			endChargeJump(mo)
		end
	end
	
	-- Heavyweight state regulation
	if -P_MobjFlip(mo)*mo.momz >= mo.scale * Heavy.GPStart
		-- Falling
		if not(mo.heavyvars.groundpound)
			-- Set falling groundpound state and sfx
			S_StartSound(mo, sfx_heavydrop)
			mo.heavyvars.groundpound = 1
			player.pflags = $ | PF_THOKKED
		elseif mo.heavyvars.groundpound == 1
		and -P_MobjFlip(mo)*mo.momz >= mo.scale * Heavy.GPStart * 3
			-- Set falling groundpound state and sfx
			S_StartSound(mo, sfx_heavydrop2)
			mo.heavyvars.groundpound = 2
			player.pflags = $ | PF_THOKKED
		end
	elseif mo.heavyvars.groundpound and not player.powers[pw_justsprung] and not(mo.eflags & MFE_GOOWATER)
		-- Falling groundpound (apply minimum fall speed)
		mo.momz = -P_MobjFlip(mo) * mo.scale * Heavy.GPStart
	else
		mo.heavyvars.groundpound = 0
	end
	
	-- Wallcollide regulation
	if grounded
		mo.heavyvars.wallcollide = 0
	end
	
	-- Footstep/landing FX
	if grounded
		if sfx_heavylanding and mo.eflags & MFE_JUSTHITFLOOR and not(mo.heavyvars.lasthitfloor) and not(heavylanding)
			S_StartSound(mo, sfx_heavylanding)
			doNearQuake(mo)
		elseif mo.state == S_PLAY_CHARGE and sfx_heavyrun
			if mo.frame % 4 == 0 and mo.tics == 1 and grounded
				S_StartSound(mo, sfx_heavyrun)
				doNearQuake(mo)
			end
		elseif mo.state == S_PLAY_WALK and sfx_heavywalk
			if mo.frame % 4 == 0 and mo.tics == 1 and grounded
				S_StartSound(mo, sfx_heavywalk)
			end
		elseif mo.state == S_PLAY_RUN and sfx_heavyrun
			if mo.frame % 2 == 0 and mo.tics == 1 and grounded
				S_StartSound(mo, sfx_heavyrun)
			end
		end
	end
	
	-- Visual dust (falling/rising velocity)
	if abs(mo.momz) >= mo.scale * 10
		local r = mo.radius>>FRACBITS
		local h = mo.height>>FRACBITS
		local x = P_RandomRange(-r, r)*FRACUNIT
		local y = P_RandomRange(-r, r)*FRACUNIT
		local z = P_RandomRange(0, h)*FRACUNIT
		P_SpawnMobjFromMobj(mo, x, y, z, MT_DUST)
	end
	
	-- Timer regulation
	mo.heavyvars.tics = max(0, $-1)
	mo.heavyvars.time = $+1

	-- History
	mo.heavyvars.lasthitfloor = mo.eflags & MFE_JUSTHITFLOOR
	mo.heavyvars.lastrolling = (mo.state == S_PLAY_ROLL)
end, MT_PLAYER)

-- I guess since vanilla does this for drowning, we can do it for the other damage types too?
chaotix.AddHook("MobjDeath", function(mo, inf, src, dmgtype)
	if (mo and mo.heavyvars) then mo.movedir = dmgtype end
end, MT_PLAYER)

-- where's my fucking floor landing hook
chaotix.AddHook("PostThinkFrame", do
	for player in players.iterate do
		local mo = player.mo
		if not (mo and mo.heavyvars) continue end
		
		-- Death pit death
		if not mo.health
			if mo.movedir == DMG_DEATHPIT
				mo.momz = mo.heavyvars.lastmomz
				mo.state = S_PLAY_FALL
				for i = 1, 4
					S_StopSoundByID(mo, _G["sfx_altdi" + i])
				end
				S_StartSound(mo, sfx_heavydeathpit)
			end
			
		-- Heavyweight
		elseif mo.heavyvars.groundpound and P_IsObjectOnGround(mo)
			local floor = P_MobjFlip(mo) == 1 and mo.floorrover and mo.floorrover.sector
				or P_MobjFlip(mo) == -1 and mo.ceilingrover and mo.ceilingrover.sector
				or mo.subsector.sector
			if GetSecSpecial(floor.special, 1) == 15 -- Bouncy floor
				mo.heavyvars.groundpound = 0
				mo.momz = abs(mo.heavyvars.lastmomz) * P_MobjFlip(mo)
				mo.state = S_PLAY_FALL
				player.panim = 0
			end
		end
	end
end)

-- Post-gametic thinker
chaotix.AddHook("PostThinkFrame", do
	for player in players.iterate do
		if (player.mo and player.mo.heavyvars)
			local mo = player.mo
			local vars = mo.heavyvars
			if vars.visualstomp and mo.state != S_PLAY_LAND
			and not vars.groundpound
				resetVisual(mo)
			end
			if vars.chargejump and (P_IsObjectOnGround(mo) or player.powers[pw_carry])
				if mo.heavyvars.dashspeed == player.maxdash and not S_SoundPlaying(mo, sfx_chargejump2)
					S_StartSound(mo, sfx_chargejump2)
				end
				-- Squash and stretch while charging jump
				local lerpamt = FixedDiv(mo.heavyvars.dashspeed, player.maxdash)
				local lerpamt2 = FRACUNIT/2 + abs(lerpamt - FRACUNIT/2)
				mo.spriteyscale = chaotix.FixedLerp(vars.spriteyscale, FRACUNIT*3/4, FixedSqrt(lerpamt))
				local xstretch = FixedDiv(FRACUNIT, mo.spriteyscale)
				mo.spritexscale = chaotix.FixedLerp(FRACUNIT*3/8, xstretch, FixedSqrt(lerpamt2))
				mo.spritexscale = $ + P_RandomRange(-2, 2) * FRACUNIT / 100
			elseif mo.state == S_PLAY_LAND
				vars.visualstomp = true
				local maxtime = states[S_PLAY_LAND].tics
				local time = mo.tics
				local lerpamt = (cos(FixedAngle(FRACUNIT * time / maxtime * 360)) + FRACUNIT) / 2
				local lerpamt2 = abs(lerpamt - FRACUNIT)
				lerpamt = $
				mo.spriteyscale = chaotix.FixedLerp(FRACUNIT*3/4, vars.spriteyscale, lerpamt)
				local xstretch = FixedDiv(FRACUNIT, FixedSqrt(mo.spriteyscale))
				mo.spritexscale = chaotix.FixedLerp(FRACUNIT, xstretch, lerpamt2)
				
				-- prevent actionability during the landing stomp
				player.pflags = $|(PF_JUMPDOWN|PF_SPINDOWN)
			end

			-- Update history
			vars.buttonslast = player.cmd.buttons
			vars.lastmomz = mo.momz
			if (mo.state == S_PLAY_LAND) then vars.buttonslast = $|(BT_SPIN) end	-- a bit of a hack
		end
		
		-- Update lastcarry always
		player.heavy_lastcarry = player.powers[pw_carry]
	end
end)


-- Ground spin move trigger
local chargeFunc = function(player, airspecial)
	local mo = player.mo
	if not Heavy.CheckStatus(player) or not mo.heavyvars
	or not heavyCanAct(player, PF_STASIS)
		return
	end
	local onground = P_IsObjectOnGround(mo) or player.powers[pw_carry]
	local holding = mo.heavyvars.buttonslast & BT_SPIN
	local active = not onground or (mo.state == S_PLAY_CHARGE_START
	or mo.state == S_PLAY_CHARGE or mo.state == S_PLAY_LAND
	or (mo.state == S_PLAY_BOUNCE_LANDING and mo.heavyvars.tics))
	if holding and active
	or not (airspecial or onground)
		return true
	end
	if not(mo.state == S_PLAY_CHARGE_START or mo.state == S_PLAY_CHARGE)
		mo.state = S_PLAY_CHARGE_START
	elseif mo.state == S_PLAY_CHARGE_START
		S_StartSound(mo, sfx_zoom)
		mo.state = S_PLAY_BOUNCE_LANDING
		mo.sprite2 = SPR2_TAL2
		mo.tics = 3
		mo.heavyvars.tics = 15 + mo.tics
		P_InstaThrust(player.powers[pw_carry] and mo.tracer or mo,
		mo.angle, FixedMul(mo.scale, player.mindash))
		player.pflags = $|PF_SPINNING|PF_STARTDASH
		player.panim = PA_ABILITY
		if STR_ANIM
			player.powers[pw_strong] = STR_MELEE &~ STR_FLOOR
		end
	end
	return true
end
chaotix.AddHook("SpinSpecial", chargeFunc)
-- Air spin move trigger
chaotix.AddHook("JumpSpinSpecial", function(player)
--	if not CBW_Battle
---		return chargeFunc(player, true)
--	end
	local mo = player.mo
	if not Heavy.CheckStatus(player) or not mo.heavyvars
	or not heavyCanAct(player, PF_STASIS)
		return
	end
	if mo.heavyvars.buttonslast & BT_SPIN or player.heavy_lastcarry
	or (mo.state >= S_HEAVY_POW_START and mo.state <= S_HEAVY_POW_END)
		return
	end
	if not(mo.state == S_PLAY_CHARGE) and player.pflags & PF_JUMPED
	and not (player.glidetime & Heavy.AirCharge)
		S_StartSound(mo, sfx_zoom)
		mo.state = S_PLAY_BOUNCE_LANDING
		mo.sprite2 = SPR2_TAL2
		mo.tics = 3
		mo.heavyvars.tics = 15 + mo.tics
		mo.momx = $/2
		mo.momy = $/2
		mo.momz = $/2
		P_Thrust(player.powers[pw_carry] and mo.tracer or mo,
		mo.angle, FixedMul(mo.scale, player.mindash))
		player.pflags = $|PF_SPINNING|PF_STARTDASH
		player.panim = PA_ABILITY
		if STR_ANIM
			player.powers[pw_strong] = STR_BOUNCE|STR_HEAVY|STR_SPRING|STR_SPIKE 
		end
		player.glidetime = $ | Heavy.AirCharge
	end
	return true
end)

-- Double jump trigger
chaotix.AddHook("AbilitySpecial", function(player)
	local mo = player.mo
	if not (Heavy.CheckStatus(player) and mo.heavyvars)
		return
	end
	if mo.heavyvars.buttonslast & BT_JUMP
	or (mo.state >= S_HEAVY_POW_START and mo.state <= S_HEAVY_POW_END)
		return
	end
	local speed = -Heavy.GPStart*4/3
	if player.pflags & PF_JUMPED
	and not (player.glidetime & Heavy.GroundPound)
	and not (mo.eflags & MFE_GOOWATER)
	and not (player.pflags & PF_SHIELDABILITY)
	and not (player.gotflag or player.gotcrystal)
		player.pflags = ($ | PF_THOKKED | PF_SPINNING) &~ PF_NOJUMPDAMAGE
		player.glidetime = $ | Heavy.GroundPound
		mo.heavyvars.groundpound = 1
		mo.heavyvars.tics = 0
		mo.state = S_PLAY_BOUNCE_LANDING
		mo.sprite2 = SPR2_FALL
		mo.tics = 3
		mo.momx = 0
		mo.momy = 0
		P_SetObjectMomZ(mo, FRACUNIT * speed, mo.momz*P_MobjFlip(mo) < 0)
		S_StartSound(mo, sfx_zoom)
		player.panim = PA_ABILITY
	end
	return true
end)

-- Jump trigger
chaotix.AddHook("JumpSpecial", function(player)
	local mo = player.mo
	if not (Heavy.CheckStatus(player) and mo.heavyvars)
		return
	end
	if mo.heavyvars.buttonslast & BT_JUMP
	or (mo.state >= S_HEAVY_POW_START and mo.state <= S_HEAVY_POW_END)
	or (not P_IsObjectOnGround(mo) and player.powers[pw_carry] != CR_ROLLOUT
	and player.powers[pw_carry] != CR_ROPEHANG
	and player.powers[pw_carry] != CR_PLAYER)
		return
	end
	if mo.heavyvars.chargejump then return true end
	
	startChargeJump(mo)
	mo.heavyvars.dashspeed = 0
	return true
end)

-- Destroy spikes on touch
chaotix.AddHook("MobjMoveCollide", function(mo, other)
	if not(mo.valid and Heavy.CheckStatus(mo.player) and mo.heavyvars)
	or not(mo.z <= other.z + other.height and other.z <= mo.z + mo.height)
	or not (other.valid and (other.type == MT_SPIKE or other.type == MT_WALLSPIKE))
		return
	end
	if (other.type == MT_SPIKE or other.type == MT_WALLSPIKE)
		P_KillMobj(other, mo, mo)
	end
end, MT_PLAYER)

-- Monitor collision
chaotix.AddHook("MobjMoveCollide", function(mo, other)
	if not(mo.valid and Heavy.CheckStatus(mo.player) and mo.heavyvars)
	or not(other.valid and other.flags & MF_MONITOR)
	or not(mo.z <= other.z + other.height and other.z <= mo.z + mo.height)
	or not (mo.player.pflags & (PF_SPINNING|PF_STARTDASH) or mo.heavyvars.groundpound)
		return
	end
	P_DamageMobj(other, mo, mo)
	return false
end, MT_PLAYER)

-- Wall-bash recoil
chaotix.AddHook("MobjMoveBlocked", function(mo)
	if not(mo and mo.valid and mo.player and Heavy.CheckStatus(mo.player) and mo.heavyvars)
		return
	end
	if mo.state == S_PLAY_CHARGE
		mo.state = S_PLAY_SPRING
		local spd = FixedHypot(mo.momx, mo.momy)
		if spd > FixedMul(mo.player.mindash, mo.scale)
			mo.momz = $ + spd/(6+mo.heavyvars.wallcollide) * P_MobjFlip(mo)
			mo.player.pflags = ($ | PF_JUMPED) &~ (PF_SPINNING|PF_STARTDASH)
			mo.player.panim = 0
			S_StartSound(mo, sfx_heavycrash)
			doNearQuake(mo, mo.scale * 1280, mo.scale * 10, 10)
			mo.heavyvars.wallcollide = $+1
		else
			S_StartSound(mo, sfx_wallbump)
		end
		mo.momx = $/3
		mo.momy = $/3

		P_BounceMove(mo)
	end
end, MT_PLAYER)

-- Damage Invulnerability
local yesflags = MF_MISSILE|MF_ENEMY|MF_BOSS
chaotix.AddHook("ShouldDamage", function(mo, inflictor, source, damage, damagetype)
	-- Get damaged anyways under certain conditions
	local player = mo.player
	if not mo.valid
	or not mo.heavyvars
	or damagetype & DMG_INSTAKILL
	or CV_HardHeavy.value
	or (inflictor and inflictor.valid and inflictor.player and inflictor != mo)
	or (source and source.valid and source.player and source != mo)
	or not Heavy.CheckStatus(player)
	or (CBW_Battle and CBW_Battle.BattleGametype())
		return
	end
	
	if inflictor and inflictor.valid
		if inflictor.flags & yesflags
			local missile = inflictor.flags & MF_MISSILE
			P_InstaThrust(mo, R_PointToAngle2(inflictor.x, inflictor.y, mo.x, mo.y), mo.scale * 4)
			if missile and not S_SoundPlaying(mo, sfx_heavytink)
				S_StartSound(mo, sfx_heavytink)
				P_PlayerFlagBurst(player,true)
			--elseif not missile and not S_SoundPlaying(mo, sfx_enemybump)
				--S_StartSound(mo, sfx_enemybump)
			end
		end
	elseif not (source and source.valid) and damagetype == DMG_FIRE
	and not (player.powers[pw_shield] & SH_PROTECTFIRE)
	and not player.powers[pw_invulnerability] and not player.powers[pw_flashing]
	and not player.powers[pw_super] and not (player.pflags & PF_GODMODE)
		if not (mo.heavyvars.lavatimer % TICRATE) or mo.eflags & MFE_JUSTHITFLOOR
			mo.heavyvars.lavatimer = 1
			if player.rings > 1
				S_StartSound(mo, sfx_antiri)
				P_GivePlayerRings(player, -1)
				
				local ring = P_SpawnMobjFromMobj(mo, 0,0,8*FRACUNIT, MT_FLINGRING)
				ring.flags = $ | MF_NOCLIPTHING | MF_NOCLIPHEIGHT
				ring.fuse = TICRATE/2
				ring.shadowscale = 0
				
				P_SpawnMobjFromMobj(ring, 0,0,0, MT_SPARK)
				P_SetObjectMomZ(ring, FRACUNIT*4)
				P_InstaThrust(ring, P_RandomRange(0, 359)*ANG1, mo.scale*4)
			else
				return
			end
		else
			mo.heavyvars.lavatimer = $ + 1
		end
	end
	
	return false
end, MT_PLAYER)

-- Height Control
chaotix.AddHook("PlayerHeight", function(player)
	local mo = player.mo
	if mo and mo.heavyvars
		local height = P_PlayerShouldUseSpinHeight(player) and P_GetPlayerSpinHeight(player) or P_GetPlayerHeight(player)
		local changed = false
		
		-- The game will try to use spinheight due to these relying on PF_SPINNING
		if player.pflags & PF_SPINNING and (mo.heavyvars.groundpound
		or mo.state == S_PLAY_CHARGE_START or mo.state == S_PLAY_CHARGE
		or (mo.state == S_PLAY_BOUNCE_LANDING and mo.heavyvars.tics))
			height = P_GetPlayerHeight(player)
			changed = true
		end
		
		-- Shorten yourself during charge jump
		-- Makes you a little shorter than visually indicated
		-- This is to allow going into spin gaps easier
		if mo.heavyvars.chargejump
			height = FixedMul($, FixedMul(mo.spriteyscale - FRACUNIT, FRACUNIT + INT16_MAX) + FRACUNIT)
			changed = true
		end
		
		if changed
			return height
		end
	end
end)

chaotix.AddHook("PlayerCanEnterSpinGaps", function(player)
	local mo = player.mo
	if mo and mo.heavyvars and mo.heavyvars.chargejump
		return true
	end
end)

-- Damage while moving
chaotix.AddHook("PlayerCanDamage", function(player)
	local mo = player.mo
	if mo and mo.heavyvars
		local runspeed = FixedMul(player.runspeed, mo.scale)
		if mo.heavyvars.groundpound
		or player.speed > runspeed or abs(mo.momz) > runspeed
			return true
		end
	end
end)

-- Charge effect
chaotix.AddHook("FollowMobj", function(player, mobj)
	local mo = player.mo
	local active = mo and mo.heavyvars and (mo.heavyvars.groundpound or
	(mo.state == S_PLAY_CHARGE
	and (mo.heavyvars.tics or mo.heavyvars.fastcharge)))
	if active
		mobj.reactiontime = $ + 3
	elseif mobj.reactiontime
		mobj.reactiontime = max(min($ - 3, 10), 0)
	end
		
	if mobj.reactiontime
		//Visual copying
		mobj.skin = mo.skin
		mobj.scale = mo.scale
		mobj.radius = mo.radius
		mobj.height = mo.height
		
		mobj.rollangle = mo.rollangle
		mobj.mirrored = mo.mirrored
		mobj.spritexscale = mo.spritexscale
		mobj.spriteyscale = mo.spriteyscale
		mobj.spritexoffset = mo.spritexoffset
		mobj.spriteyoffset = mo.spriteyoffset
		mobj.renderflags = mo.renderflags | RF_FULLBRIGHT //| RF_NOCOLORMAPS
		mobj.blendmode = mo.blendmode
		mobj.flags2 = mo.flags2 & (MF2_DONTDRAW|MF2_SHADOW|MF2_SPLAT) | MF2_LINKDRAW
		
		//Set frame and state
		if active
			local state = mobj.info.seestate
			+ ((mo.heavyvars.groundpound and mo.state != S_PLAY_CHARGE) and 1 or 0)
			mobj.threshold = $ + 1
			if mobj.state != state
				mobj.state = state
			end
			mobj.frame = (mo.frame &~ FF_FRAMEMASK) | ($ & FF_FRAMEMASK)
		end
		if mobj.tics > 0 then mobj.tics = min($, 2) end
		
		//Set transparency
		local trans = min(max((10 - mobj.reactiontime) << FF_TRANSSHIFT, 0), FF_TRANS90)
		if (mobj.frame & FF_TRANSMASK) < trans
			mobj.frame = ($ &~ FF_TRANSMASK) | trans
		end
		if not (mobj.threshold % 3) and (mobj.frame & FF_TRANSMASK) < FF_TRANS50
			mobj.frame = ($ &~ FF_TRANSMASK) | FF_TRANS50
		end
		
		//Set angle
		if mobj.state == mobj.info.seestate + 1
			A_CapeChase(mobj, 1, 0)
		elseif player.powers[pw_carry] == CR_ROLLOUT
		and player.pflags & PF_DIRECTIONCHAR
			A_CapeChase(mobj, 1, -FRACUNIT)
			mobj.angle = $ + ANGLE_180
			mobj.mirrored = not $
		else
			A_CapeChase(mobj, 1, FRACUNIT)
		end
		
		//Spawn speed trail
		chaotix.SpawnSpeedTrail(mo, active and (leveltime & 1)
		and (not mo.heavyvars.groundpound or not P_IsObjectOnGround(mo)))
		
		//Spawn Ghosts
		if not active then return true end
		local slow = (mo.heavyvars.fastcharge or mo.heavyvars.groundpound == 1)
		if not (not slow or (slow and leveltime & 1)) then return true end
		
		local ghost = P_SpawnGhostMobj(mo)
		if ghost.tracer then P_RemoveMobj(ghost.tracer) end
		
		local timer = (mo.followmobj and mo.followmobj.reactiontime)
		or mo.heavyvars.tics or leveltime
		ghost.color = (timer % 6)*5 + SKINCOLOR_SUPERRED5
		ghost.colorized = true
		ghost.blendmode = AST_ADD
		ghost.renderflags = $ | RF_FULLBRIGHT | RF_NOCOLORMAPS
		return true
	elseif mobj.state != mobj.info.spawnstate
		mobj.state = mobj.info.spawnstate
	end
end, MT_HEAVY_CHARGE)

-- Hitstun control
-- chaotix.AddHook("MobjDamage", function(mo, inflictor)
-- 	if inflictor and inflictor.valid and inflictor.heavyvars and inflictor.state == S_PLAY_CHARGE
-- 		local t = inflictor.heavyvars
-- 		if not(mo.flags & MF_SCENERY)
-- 			t.spritexoffset = inflictor.spritexoffset
-- 			t.spriteyoffset = inflictor.spriteyoffset
-- 		end
-- 		t.hitstun = $+2
-- 	end
-- end)
chaotix.AddHook("MobjThinker", function(mo)
	if mo.valid and Heavy.CheckStatus(mo.player)
		local t = mo.heavyvars
		if t.hitstun
			mo.flags = $ | MF_SCENERY
			t.hitstun = $-1
			mo.spritexoffset = P_RandomRange(-2, 2)*FRACUNIT
			return true
		elseif mo.flags & MF_SCENERY
			mo.flags = $ &~ MF_SCENERY
			mo.spritexoffset = t.spritexoffset
			mo.spriteyoffset = t.spriteyoffset
		end
	end
end, MT_PLAYER)

-- POW, haha!
chaotix.AddHook("PlayerMsg", function(p, msgtype, target, msg)
	if (msgtype) then return end	-- needs to be a regular message
	if not (gamestate == GS_LEVEL or titlemapinaction) then return end -- need to be in a level
	
	-- heavy checks
	if not (Heavy.CheckStatus(p)) then return end
	local mo = p.mo
	if not (P_IsObjectOnGround(mo)) then return end	-- gotta be grounded
	if (p.isjettysyn) then return end	-- NO
	if (mo.state >= S_HEAVY_POW_START and mo.state <= S_HEAVY_POW_END) then return end	-- already doing it!
	
	-- check message
	msg = $:lower()
	if (msg == "pow") -- start the pow state
		p.panim = PA_IDLE	-- this should make us interupt it
		mo.state = S_HEAVY_POW_START
		mo.heavyvars.pow = 0
		mo.heavyvars.powangle = mo.angle
		return true
	end
end)

-- teeter animation speed adjustment
chaotix.AddHook("ThinkFrame", do
	for p in players.iterate
		local mo = p.mo
		if not (mo and mo.valid) then continue end
		if not (Heavy.CheckStatus(p)) then mo.heavy_teeterleveltime = nil continue end
		
		if (mo.sprite2 != SPR2_EDGE) then mo.heavy_teeterleveltime = nil continue end
		
		-- init vars
		if (mo.heavy_teeterleveltime == nil) then mo.heavy_teeterleveltime = leveltime end	-- so the animation always starts from the first frame
		local teetertime = leveltime - mo.heavy_teeterleveltime
		
		-- set frame
		local frames = 4	-- frame count of the animation
		local tics = 5	-- tics per frame
		chaotix.SetFrame(mo, (teetertime % (frames * tics)) / tics)
	end
end)