local Heavy = chaotix.Heavy
Heavy.tumbleTics = 25

local B = CBW_Battle
local BattleFuncSet = false
chaotix.AddHook("ThinkFrame",function()
	if (BattleFuncSet) or not(CBW_Battle and CBW_Battle.SkinVars) then return end
	BattleFuncSet = true
	
local cooldown = TICRATE*17/4	-- 4.25
local moveduration = TICRATE*4

local HeavyPriority = function(player)
	-- only make train rush damage when we charge INTO someone
	if player.mo.state == S_PLAY_CHARGE then
		local satk = 1	-- base atk
		
		-- when going above a speed, get 2 atk!
		local speed = FixedHypot(player.rmomx, player.rmomy)
		local movescale = FixedMul(player.mo.eflags & (MFE_UNDERWATER|MFE_GOOWATER) and FRACUNIT*11/20 or FRACUNIT, FRACUNIT)
		local friction = P_IsObjectOnGround(player.mo) and speed - FixedMul(FixedMul(speed, player.mo.friction), FRACUNIT) or 0
		local maxdash = max(FixedMul(player.maxdash, player.mo.scale), speed + friction)
		local accel = FixedMul(movescale, player.mo.scale) + friction
		if player.powers[pw_sneakers] or player.powers[pw_super]
			accel = FixedMul(movescale, player.mo.scale)*3/2 + friction
		end
				
		if speed > maxdash - accel
			satk = 2
		end
		
		-- set priority
		CBW_Battle.SetPriority(player, 0, 0, "knuckles_glide", satk, 1, "train rush")
	end
	if player.mo.state == S_PLAY_CHARGE and player.actionstate == 1 then
		-- titanium gives more defence, so increase the non-sweetspot version too
		CBW_Battle.SetPriority(player, 0, 3, "knuckles_glide", 3, 3, "titanium train rush")
	end
	
	-- make heavy have a bit of priority, but a lot of defence below him
	if player.mo.heavyvars and player.mo.heavyvars.groundpound
	and player.mo.state == S_PLAY_FALL then 
		CBW_Battle.SetPriority(player, 1, 2, "fang_tailbounce", 2, 3, "anvil drop")
	end
	
	if player.tumble then
		--circumventing Tumble stunlock bullshit before it happens
		CBW_Battle.SetPriority(player,0,0,nil,0,0,nil)
	end
end

local Titanium = function(mo, doaction)
	local player = mo.player
	player.actiontext = "Titanium"
	player.actionrings = 15
	player.actionsuper = true
	
	-- no guard for pow
	if mo.state >= S_HEAVY_POW_START and mo.state <= S_HEAVY_POW_END
		player.canguard = false
	end
	
	player.action2text = nil
	if player.actionstate == 1
		player.action2text = "Duration "..G_TicsToSeconds(player.actiontime).."."..G_TicsToCentiseconds(player.actiontime)
	end
	if player.rings < player.actionrings and not player.actionstate
		player.actiontext = "Not enough Rings!"
	end
	-- Activate special
	if doaction == 1 and player.actionstate == 0
		if player.rings < player.actionrings
			S_StartSound(nil, sfx_s3k8c, player)
			return
		end
		
		CBW_Battle.PayRings(player)
		player.actionstate = 1
		player.actiontime = moveduration
		S_StartSound(mo,sfx_bedeen)
		S_StartSound(mo,sfx_kc62)
		player.powers[pw_flashing] = max($,moveduration)
		-- Do effects
		for n = 1, 20 do
			local fx = P_SpawnMobjFromMobj(mo, 0, 0, mo.height/2, MT_BOXSPARKLE)
			fx.momx = P_RandomRange(-10, 10)*fx.scale
			fx.momy = P_RandomRange(-10, 10)*fx.scale
			fx.momz = P_RandomRange(-10, 10)*fx.scale
			fx.colorized = true
			fx.color = CBW_Battle.Choose(SKINCOLOR_WHITE, SKINCOLOR_CYAN, SKINCOLOR_PURPLE, SKINCOLOR_JET)
		end
	end
	
	--Titanium mode
	if player.actionstate == 1
	if player.actiontime <= 0 or player.mo.eflags&MFE_SPRUNG or player.tumble
			-- Reset to neutral
			CBW_Battle.ApplyCooldown(player, cooldown)
			player.powers[pw_flashing] = 0
			player.actionstate = 0
			return
		end
		
		-- Regulate state and timer
		player.actiontime = $-1		
		-- Do effects
		if player.actiontime % 3 == 0
			local ghost = P_SpawnGhostMobj(mo)
			ghost.color = P_RandomRange(1,10)
			ghost.fuse = 3
			ghost.destscale = $<<1
			ghost.blendmode = AST_ADD
			local fx = P_SpawnMobjFromMobj(mo, 0, 0, mo.height/2, MT_BOXSPARKLE)
			fx.momx = P_RandomRange(-10, 10)*fx.scale
			fx.momy = P_RandomRange(-10, 10)*fx.scale
			fx.momz = P_RandomRange(-10, 10)*fx.scale
			fx.colorized = true
			fx.color = CBW_Battle.Choose(SKINCOLOR_WHITE, SKINCOLOR_CYAN, SKINCOLOR_PURPLE, SKINCOLOR_JET)
			local fx = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_IVSP)
			fx.color = CBW_Battle.Choose(SKINCOLOR_WHITE, SKINCOLOR_CYAN, SKINCOLOR_PURPLE, SKINCOLOR_JET)
		end
		
	end
end


local function isTitanium(player)
	return (player and player.valid and player.playerstate == PST_LIVE)
		and player.mo
		and player.actionstate == 1
		and player.mo.health
end

local Titanium_PreCollide = function(n1,n2,plr,mo,atk,def,weight,hurt,pain,ground,angle,thrust,thrust2,collisiontype)
	if isTitanium(plr[n1])
		plr[n1].heavymarker = true
	end
end

local Titanium_PostCollide = function(n1,n2,plr,mo,atk,def,weight,hurt,pain,ground,angle,thrust,thrust2,collisiontype)
	if plr[n1] and plr[n1].heavymarker
		plr[n1].heavymarker = nil
	end
end

local Titanium_Collide = function(n1,n2,plr,mo,atk,def,weight,hurt,pain,ground,angle,thrust,thrust2,collisiontype)
	if not (plr[n1] and plr[n1].heavymarker and atk[n1] <= def[n2])
		return false
	end
	if (hurt != 1 and n1 == 1) or (hurt != -1 and n1 == 2)
		if not (plr[n2] and plr[n2].heavymarker)
			--Anything special we want to do here?
		end
		if plr[n2]
			CBW_Battle.DoPlayerTumble(plr[n2], 35, angle[n1], mo[n1].scale*3, true, true)
		end
		P_InstaThrust(mo[n2], angle[n2], mo[n1].scale * 5)
		CBW_Battle.ZLaunch(mo[n2], 7 * mo[n2].scale, false)
		return true
	end
end
CBW_Battle.SkinVars[Heavy.SKIN] = {
		flags = SKINVARS_GUARD|SKINVARS_NOSPINSHIELD,
		weight = 200,
		shields = 1, 
		special = Titanium,
		guard_frame = 1,
		func_priority_ext = HeavyPriority,
		func_precollide = Titanium_PreCollide,
		func_collide = Titanium_Collide,
		func_postcollide = Titanium_PostCollide
}
end)