// ------
// states
// ------

chaotix.SafeFreeslot('sfx_hevdie', 'sfx_hevfal','sfx_hevpj')
sfxinfo[sfx_s3kd1l].caption = 'Steaming'
sfxinfo[sfx_hevdie].caption = '/'
sfxinfo[sfx_hevfal].caption = 'Falling'
sfxinfo[sfx_hevpj].caption = 'Phantom Jump'
sfxinfo[sfx_kc62].caption = 'Hardening'

chaotix.SafeFreeslot('MT_HEAVY_CHARGE')
mobjinfo[MT_HEAVY_CHARGE] = {
	spawnstate = S_INVISIBLE,
	seestate = S_TAILSOVERLAY_PLUS60DEGREES,
	spawnhealth = 1000,
	radius = mobjinfo[MT_PLAYER].radius,
	height = mobjinfo[MT_PLAYER].height,
	dispoffset = 1,
	flags = MF_SCENERY|MF_NOGRAVITY|MF_NOBLOCKMAP|MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOCLIPTHING
}

chaotix.SafeFreeslot('S_PLAY_CHARGE_START', 'S_PLAY_CHARGE',
	'S_PLAY_LAND')

states[S_PLAY_CHARGE_START] = {
	sprite = SPR_PLAY, 
	frame = SPR2_TAL2,
	tics = 8,
	nextstate = S_PLAY_CHARGE
}

states[S_PLAY_CHARGE] = {
	sprite = SPR_PLAY, 
	frame = SPR2_DASH,
	tics = 3,
	nextstate = S_PLAY_CHARGE
}

states[S_PLAY_LAND] = {
	sprite = SPR_PLAY, 
	frame = SPR2_LAND,
	tics = 12,
	nextstate = S_PLAY_STND
}

local pow_range = FRACUNIT*480
local pow_angle = 5

local function angtoint(a)
	if not a return 0 end
	a = $/ANG1
	if a < 0
		a = $+360
	end
	return a
end
	
function A_HeavyPOW(actor, var1, var2)
	if not (actor.player) then return end
	
	-- kill the closest object we can find!
	local range = FixedMul(actor.scale, pow_range)
	local pow_targets = {}
	
	local debug_range = false
	if (debug_range)
		local trig_dist = (range / FRACUNIT)	-- needs to not be fixed point
			
		-- height
		local height_thoks = 3
		for hthoks = 0, height_thoks - 1
			local zoffset = (hthoks * FRACUNIT) / (height_thoks - 1)
				
			local angle_thoks = 8
			for athoks = 0, (angle_thoks - 1)
				-- spawn angle
				local trigAngle = pow_angle - ((athoks * (pow_angle * 2)) / (angle_thoks - 1))
				trigAngle = actor.heavyvars.powangle + ($ * ANG1)
						
				local length_thoks = 8
				for lthoks = 0, length_thoks - 1
					-- define the lining
					if (athoks and not lthoks) then continue end	-- don't stack on the start point
						
					if not (athoks == (angle_thoks - 1) or not athoks)
					and not (lthoks == (length_thoks - 1) or not lthoks)
						continue
					end
						
					if (hthoks % 2)
					and not ((lthoks == (length_thoks - 1) or not lthoks) and (athoks == (angle_thoks - 1) or not  athoks))
						continue
					end
						
					-- distance
					local thok_dist = (lthoks * trig_dist) / (length_thoks - 1)
					local xpos = actor.x + thok_dist*cos(trigAngle)
					local ypos = actor.y + thok_dist*sin(trigAngle)
						
					-- height
					local startz = actor.z
					if (actor.eflags & MFE_VERTICALFLIP) then startz = actor.z + actor.height end
					local heightOffset = actor.height * P_MobjFlip(actor)
					local zpos = startz + FixedMul(heightOffset, zoffset)
						
					-- spawn the thok
					local thok = P_SpawnMobj(xpos, ypos, zpos, MT_THOK)
					thok.frame = ($ & ~(FF_TRANSMASK)) | (FF_FULLBRIGHT|FF_TRANS70)
					thok.scale = FixedMul(FRACUNIT/4, $)
					thok.color = SKINCOLOR_WHITE
					thok.tics = TICRATE*3/4
				end
			end
		end
	end
	
	searchBlockmap("objects", function(ref, found)
		if not ((found and found.valid) or found.health) then return end
		if (ref == found) then return end
			
		-- get the object type
		if not ((found.flags & (MF_SHOOTABLE|MF_ENEMY|MF_MONITOR)) or found.player) then return end
			
		-- positional checks
		local dist = R_PointToDist2(ref.x, ref.y, found.x, found.y)
		if not (dist <= range) then return end	-- making it radial
			
		local heightOffset = ref.height * P_MobjFlip(ref)
		local topz, bottomz
		if not (ref.eflags & MFE_VERTICALFLIP)
			bottomz = ref.z
			topz = bottomz + heightOffset
		else
			topz = ref.z + ref.height
			bottomz = topz + heightOffset
		end
		if not ((topz >= found.z) and (bottomz <= found.z + found.height)) then return end	-- z checks
			
		if not (P_CheckSight(ref, found)) then return end
		
		-- angle check
		local angle = R_PointToAngle2(ref.x, ref.y, found.x, found.y)
		local angle_diff = ref.heavyvars.powangle - angle
		if (angtoint(angle_diff) > 180) then angle_diff = InvAngle(angle_diff) end
		if (angtoint(angle_diff) > pow_angle) then return end
		
		-- add them to the table!
		table.insert(pow_targets, found)
	end, actor, actor.x - range, actor.x + range, actor.y - range, actor.y + range)
	
	table.sort(pow_targets, function(a, b) 
		local a_dist = R_PointToDist2(actor.x, actor.y, a.x, a.y)
		local b_dist = R_PointToDist2(actor.x, actor.y, b.x, b.y)
		return a_dist < b_dist 
	end)
	
	for _, target in ipairs(pow_targets)
		if (P_DamageMobj(target, actor, actor, 1, DMG_INSTAKILL))
			if (target.valid)
				local dust = P_SpawnMobjFromMobj(target, 0, 0, 0, MT_THOK)
				dust.state = S_XPLD1
				S_StartSound(dust, sfx_rail2)
			end
			
			break
		end
	end
	S_StartSound(actor, sfx_rail1)
	
	-- dust effect
	local dusts = 8
	for i = 1, dusts
		local angle = (360/dusts*i)*ANG1

		local dust = P_SpawnMobj(actor.x + 32*cos(angle), actor.y + 32*sin(angle), actor.z, MT_SPINDUST)
		dust.scale = $/2
		P_Thrust(dust, angle, FixedMul(actor.scale, FRACUNIT*8))
		P_SetObjectMomZ(dust, FRACUNIT*2)
		if (actor.eflags & (MFE_TOUCHWATER|MFE_UNDERWATER))
			dust.state = S_SPINDUST_BUBBLE1
		end
	end
	
	-- message and sound
	COM_BufInsertText(actor.player, "say POW, haha!")
	S_StartSound(actor, sfx_hevpow)
end
function A_EndPOWState(actor, var1, var2)
	-- go to proper state after ending
	local state = S_PLAY_FALL
	if (P_IsObjectOnGround(actor))
		state = S_PLAY_STND
		local speed = FixedDiv(FixedHypot(actor.momx, actor.momy), actor.scale)
		if (speed)
			state = (speed > FixedMul(actor.scale, actor.player.runspeed)) and S_PLAY_RUN or S_PLAY_WALK
		end
	end
	
	actor.state = state
end
chaotix.SafeFreeslot("S_HEAVY_POW_START", "S_HEAVY_POW_WINDUP", "S_HEAVY_POW_FIRE", "S_HEAVY_POW_END", "sfx_hevpow")
states[S_HEAVY_POW_START] = {SPR_PLAY, SPR2_MLEE, 4, nil, 0, 0, S_HEAVY_POW_START}
states[S_HEAVY_POW_WINDUP] = {SPR_PLAY, SPR2_FIRE|FF_SPR2ENDSTATE, 2, nil, S_HEAVY_POW_FIRE, 0, S_HEAVY_POW_WINDUP}
states[S_HEAVY_POW_FIRE] = {SPR_PLAY, SPR2_MLEL, TICRATE*4/3, A_HeavyPOW, 0, 0, S_HEAVY_POW_END}
states[S_HEAVY_POW_END] = {SPR_PLAY, SPR2_MLEL, 1, A_EndPOWState, 0, 0, S_PLAY_STND}	-- just to set the state back properly
sfxinfo[sfx_hevpow].caption = "\"POW, haha!\""