-- everything for bomb's battle support
local bomb = chaotix.bomb

local loaded = false
bomb.BattleSetup = function()
	if not (CBW_Battle) then return end	-- battle isnt loaded
	if (loaded) then return end	-- already setup for battle? dont run
	local B = CBW_Battle
	local S = B.SkinVars
	
	-- bomb's priority
	local function bomb_battleVolatile(mo, other)
		-- get the sweetspot hitbox when volatile!
		return (mo.bomb_volatile) and true or false
	end
	B.AddPriorityFunction("bomb_volatile", bomb_battleVolatile)
	
	local function bomb_battlePriority(p)
		local mo = p.mo
		
		-- bomb doesn't spin... he topspins!
		local spinjump = ((p.pflags & PF_JUMPED) and not (p.pflags & PF_NOJUMPDAMAGE))
		local spinning = (p.pflags & PF_SPINNING)
			
		if (spinjump or spinning)
		or (mo.state >= S_BOMB_TOPSPIN and mo.state <= S_BOMB_TOPSPIN_FAST)
			local atk, def = 1, 1
			local satk, sdef = atk, def
			local text = (spinning) and "topspin attack" or "jumping topspin attack"
			if (mo.state >= S_BOMB_TOPSPIN and mo.state <= S_BOMB_TOPSPIN_FAST)
				text = (p.pflags & PF_JUMPED) and "jumping topspin attack" or "topspin attack"
			end
			
			B.SetPriority(p, atk, def, "bomb_volatile", satk, sdef, text)
		end
		
		-- fuse shot
		if (mo.bomb_fuseshot and mo.bomb_fuseshot_shot)
			local atk, def = 1, 1
			local satk, sdef = atk, def
			local text = "fuse shot"
			
			B.SetPriority(p, atk, def, "bomb_volatile", satk, sdef, text)
		end
		
		-- volatile state is just added on top of existing priority
		if (mo.bomb_volatile)
			-- count how volatile we are
			local count = (P_PlayerFullbright(p)) and 1 or 0
			for _, flag in pairs(bomb.VolatileFlags)
				if not (mo.bomb_volatile & flag) then continue end
				count = $+1
			end
			
			local satk = p.battle_satk + count
			local sdef = p.battle_sdef + (count - 1)
			local text = "volatile "..p.battle_hurttxt
			
			B.SetPriority(p, nil, nil, "bomb_volatile", satk, sdef, text)	-- volatile state amplifies sweetspot priority!
		end
	end
	
	-- bomb's special
	local st_meltdown = 1
	
	local function bomb_battleSpecial(mo, doaction)
		local p = mo.player
		
		-- ability names
		local forceShow = p.actionstate or mo.cream_refresh_tics
		p.actionrings = bomb.MeltdownCost
		p.actiontext = ((p.rings < p.actionrings) and not forceShow) and "Not enough Rings!" or "Overcharge"	-- no ring debt for this
		p.action2text = nil
		if (p.actionstate == 1)
			-- add action2text to show fuse time!
			p.action2text = "Explode "..G_TicsToSeconds(p.actiontime).."."..G_TicsToCentiseconds(p.actiontime)
		end
		
		-- no guard for cupcake
		if p.bomb_cupcake
			p.canguard = false
		end
		
		-- attempt to meltdown!
		if not (p.actionstate)
			if (doaction == 1)	-- just pressed it
				if (p.rings < p.actionrings)	-- no debt!
					S_StartSound(nil, sfx_s3k8c, p)
					return
				end
					
				-- go to the meltdown state!
				p.actionstate = st_meltdown
				p.actiontime = bomb.MeltdownFuse
				B.PayRings(p)
				bomb.TriggerVolatile(p, bomb.VolatileFlags["MELTDOWN"])
				S_StartSound(mo, sfx_s25c)
			end
		elseif (p.actionstate == st_meltdown)
			-- reset this early!
			if not (mo.bomb_volatile & bomb.VolatileFlags["MELTDOWN"])
				p.actionstate = 0
				B.ApplyCooldown(p, bomb.MeltdownCooldown)
				
				-- reset speed
				B.ResetPlayerProperties(p)
				mo.bomb_fixnormalspeed = nil
				return
			end
			
			-- reduce fuse then do stuff
			local sec = p.actiontime / TICRATE
			p.actiontime = $ and $-1
		
			if not (p.actiontime)
				-- set battlemod stuff
				p.actionstate = 0
				S_StopSound(mo)
				B.ApplyCooldown(p, bomb.MeltdownCooldown)
				
				-- reset volatile stuff
				bomb.ResetVolatile(mo)
				B.ResetPlayerProperties(p)
				mo.bomb_fixnormalspeed = nil
				
				-- damage us
				P_DamageMobj(mo)
				mo.bomb_explode = true
				mo.bomb_explode_tics = ((mo.health) and bomb.ExplodeTics or bomb.ExplodeDeathTics) - 1	-- set tics to the moment we're supposed to explode
				mo.bomb_explode_exploded = true	-- hack
				
				-- affect things in a range
				local range = FixedMul(mo.scale, bomb.MeltdownSelfDestructRange)
				local underwater_modifier = (mo.eflags & MFE_UNDERWATER) and bomb.VolatileUnderwaterNerf or FRACUNIT
				bomb.SpawnExplosion(mo, FixedMul(underwater_modifier, range))
				S_StartSound(mo, sfx_kc3b)
				
				-- modify momentum if we're still alive
				if (mo.health)
					mo.momx, mo.momy = 0, 0
					P_SetObjectMomZ(mo, chaotix.WaterMom(mo, bomb.MeltdownSelfDestructLaunch))
				end
			else
				-- gain a speed boost!
				local multiplier_tics = bomb.MeltdownFuse - p.actiontime
				local multiplier_range = bomb.MeltdownSpeedMultiplier - FRACUNIT
				local multiplier = FRACUNIT + FixedMul((multiplier_tics * FRACUNIT) / bomb.MeltdownFuse, multiplier_range)
				p.normalspeed = FixedMul(multiplier, skins[mo.skin].normalspeed)
				mo.bomb_fixnormalspeed = true	-- we'll need this
				
				-- Do ghost and alert
				if not (9000 % p.actiontime)
					local ghost = P_SpawnGhostMobj(mo)
					ghost.colorized = true
					ghost.color = SKINCOLOR_FLAME
					ghost.destscale = $<<(bomb.MeltdownFuse/TICRATE - sec)
					ghost.blendmode = AST_ADD
					S_StartSoundAtVolume(mo, sfx_s3k5c, 196)
				end
				bomb.SpawnFuseSparks(p, 1, FixedDiv(mo.height, mo.scale), bomb.FuseShotSparkHMom, bomb.FuseShotSparkZMom, SKINCOLOR_NONE, true, false, 2)	-- force smoke
			end
		end
		
		-- make sure this'll be reset properly!
		if not (p.actionstate)
		and (mo.bomb_fixnormalspeed)
			B.ResetPlayerProperties(p)
			mo.bomb_fixnormalspeed = nil
		end
	end
	
	-- collision functions
	local function bomb_battlePreCollide(s, t, p, mo, atk, def, weight, hurt, pain, ground, angle, thrust, thrust2, collisiontype)
		if not (p[s] and mo[s]) then return end
		
		-- laziness check for hurt value (so we don't have to use 2 different checks...)
		local newhurt = hurt
		if (abs(newhurt) == 1)
		and (s == 2)	-- s is actually t
			newhurt = $ * -1
		end
		
		-- make beat bash recoil off of damage correctly
		if (mo[s].bomb_volatile and newhurt == 1)	-- make sure it's only THEM getting damaged
		and not (p[t].guard > 0)	-- bruh... why do i even need to check this
			mo[s].bomb_volatile_battle = true
		end
	end
	
	local function bomb_battlePostCollide(s, t, p, mo, atk, def, weight, hurt, pain, ground, angle, thrust, thrust2, collisiontype)
		if not (p[s] and mo[s]) then return end
		if not (p[t] and mo[t]) then return end
		
		if (mo[s].bomb_volatile_battle)
			local oldinfo = mo[s].bomb_volatile_oldinfo
			local momx = (oldinfo) and oldinfo.momx or mo[s].momx
			local momy = (oldinfo) and oldinfo.momy or mo[s].momy
			local momz = (oldinfo) and oldinfo.momz or mo[s].momz
			
			mo[t].bomb_volatile_launcher = true	-- hack
			local launch = max(FixedHypot(FixedHypot(momx, momy), momz) >> 1, abs(momz))
			local height = max(bomb.VolatileDamageMinHeight, FixedMul(bomb.VolatileDamageHeightMultiplier, launch))
			bomb.VolatileLaunch(p[s], height)
			mo[t].bomb_volatile_launcher = false	-- hack (2)
			bomb.ResetThokked(p[s])	-- sounds fun
			
			-- RESET THIS
			mo[s].bomb_volatile_battle = nil
		end
	end
	
	-- setup bomb's skinvars
	S[bomb.SKIN] = {
		weight = 70,
		special = bomb_battleSpecial,
		func_priority_ext = bomb_battlePriority,
		func_precollide = bomb_battlePreCollide,
		func_postcollide = bomb_battlePostCollide,
	}
	
	loaded = true
end