-- all of bomb's definitions
chaotix.bomb = {}
local bomb = chaotix.bomb
bomb.SKIN = "bomb"
chaotix.AddCharacter(bomb)

-- bomb's sounds
local sounds = {
	{sound = "sfx_cdfm16", caption = "Activate"},
	{sound = "sfx_kc2a", caption = "Launch"},
	{sound = "sfx_kc68", caption = "Ignite"},
	{sound = "sfx_kc3b", caption = "Explosion"},
	{sound = "sfx_s25c", caption = "Ignite"},
}
for k, v in pairs(sounds)
	-- these are now espio's sounds
	sfxinfo[_G[v.sound]].caption = v.caption
end

-- bomb's abilities
local abilityslot = 28	-- idk 28 is an explosive number
rawset(_G, "CA_FUSESHOT", abilityslot)
rawset(_G, "CA2_FUSEDASH", abilityslot)

bomb.AbilityCheck = function(p, ability)
	if not (p and p.valid) then return end
	
	-- since they're both the same number, we can do some fancy stuff!
	local checkability = (ability == 2) and "2" or ""
	return (p["charability"..checkability] == abilityslot)
end

-- thokked constants
bomb.ThokkedFlags = {
	["FUSESHOT"] = 2 ^ 0,	-- midair ability
	["FUSEDASH"] = 2 ^ 1,	-- secondary ability
}

-- explode constants
bomb.ExplodeTics = TICRATE*2/3
bomb.ExplodeDeathTics = TICRATE*11/10
bomb.ExplodeStartup = TICRATE/6
bomb.ExplodeFrameTics = {3, 2}
bomb.ExplodeRange = FRACUNIT*128
bomb.ExplodeRecoilModifier = FRACUNIT/2
bomb.ExplodeRecoilHeight = FRACUNIT*8
bomb.ExplodeLosstimeAdd = TICRATE*8

-- fuse shot constants
bomb.FuseShotStartupModifier = FRACUNIT*2/3	-- how much momentum gets cut when we start it
bomb.FuseShotStartupHeight = FRACUNIT*6	-- how much of a little bounce we get
bomb.FuseShotChargeMinTics, bomb.FuseShotChargeMaxTics = TICRATE/2, TICRATE*6/5
bomb.FuseShotChargeXYModifier = FRACUNIT*94/100
bomb.FuseShotChargeSlowfall = FRACUNIT*2/5
bomb.FuseShotMaxModifier = FRACUNIT*7/5	-- how much we can charge up our actionspd
bomb.FuseShotHeight = FRACUNIT*8
bomb.FuseShotOverchargeTics = bomb.FuseShotChargeMaxTics*7/5
bomb.FuseShotSparkColours = {SKINCOLOR_PLUM, SKINCOLOR_RASPBERRY, SKINCOLOR_SUNSET, SKINCOLOR_COPPER, SKINCOLOR_KETCHUP}	-- will end at SKINCOLOR_ROSE no matter what
bomb.FuseShotSparkHMom = FRACUNIT*2
bomb.FuseShotSparkZMom = FRACUNIT*3
bomb.FuseShotRicochetSparkHMom = FRACUNIT*6
bomb.FuseShotRicochetSparkZMom = FRACUNIT*5
bomb.FuseShotBounceHeight = FRACUNIT*9
bomb.FuseShotBounceSparkHMom = FRACUNIT*6
bomb.FuseShotBounceSparkZMom = FRACUNIT*8

-- fuse dash constants
bomb.FuseDashChargeTics = TICRATE*6/5	-- how long it takes to fully charge
bomb.FuseDashSparkColours = {SKINCOLOR_NONE, SKINCOLOR_PLUM, SKINCOLOR_RASPBERRY, SKINCOLOR_SUNSET, SKINCOLOR_COPPER, SKINCOLOR_KETCHUP}	-- the same as bomb.FuseShotSparkColours, but with SKINCOLOR_NONE

-- top spin constants
bomb.TopSpinThreshold = FRACUNIT*12
bomb.TopSpinFastThreshold = FRACUNIT*18

-- volatile constants
bomb.VolatileFlags = {
	["FUSESHOT"] = 2 ^ 0,	-- midair ability
	["FUSEDASH"] = 2 ^ 1,	-- secondary ability
	["MELTDOWN"] = 2 ^ 2,	-- battle ability
}
bomb.VolatileHitlagDivider = FRACUNIT*2
bomb.VolatileManualHeight = FRACUNIT*11
bomb.VolatileBounceMinHeight = FRACUNIT*14
bomb.VolatileBounceHeightMultiplier = FRACUNIT*4/5
bomb.VolatileWallFacingMultiplier = FRACUNIT/4
bomb.VolatileWallMinHeight = FRACUNIT*13
bomb.VolatileWallHeightMultiplier = FRACUNIT/3
bomb.VolatileDamageMinHeight = FRACUNIT*15
bomb.VolatileDamageHeightMultiplier = FRACUNIT*7/9
bomb.VolatileOverchargeHeight = FRACUNIT*26
bomb.VolatileEffectTable = {
	[1] = {
		multiplier = FRACUNIT, 
		sound = sfx_kc3b,
		colorized_func = function(p) return ((leveltime%6) <= 2) end,
		ghost_func = function(p) return not (leveltime%3) and 3 or 0 end,
	},
	[2] = {
		multiplier = FRACUNIT*6/5, 
		sound = sfx_bkpoof,
		colorized_func = function(p) return ((leveltime%4) <= 1) end,
		ghost_func = function(p) return not (leveltime%2) and 4 or 0 end,
	},
	[3] = {
		multiplier = FRACUNIT*7/5, 
		sound = sfx_bgxpld,
		colorized_func = function(p) return (leveltime%2) and true or false end,
		ghost_func = function(p) return not (leveltime%2) and 5 or 0 end,
	},
}
bomb.VolatileDamageRangeMultiplier = FRACUNIT*6
bomb.VolatileDamageMinRange = FRACUNIT*96
bomb.VolatileDamageHeightModifier = FRACUNIT
bomb.VolatileDamageTics = TICRATE*2/3
bomb.VolatileBurstScalar = FRACUNIT*104	-- try keep this equal to the sprite's width
bomb.VolatileExplodeThrustScalar = FRACUNIT*12
bomb.VolatileExplodeScaleScalar = FRACUNIT*92
bomb.VolatileDustThrustScalar = FRACUNIT*17
bomb.VolatileDustScaleScalar = FRACUNIT*172
bomb.VolatileQuakeIntensityScalar = FRACUNIT*8
bomb.VolatileQuakeTicsScalar = FRACUNIT*228
bomb.VolatileUnderwaterNerf = FRACUNIT*4/5
bomb.VolatileUnderwaterModifier = FRACUNIT*2/3	-- used for the explosion vfx
bomb.VolatileScaleSpeed = TICRATE*4/9

-- meltdown constants (battlemod)
bomb.MeltdownCost = 15
bomb.MeltdownCooldown = TICRATE*5
bomb.MeltdownFuse = TICRATE*5
bomb.MeltdownSpeedMultiplier = FRACUNIT*2
bomb.MeltdownExplosionXYMultiplier = FRACUNIT/10	-- based on explosion radius
bomb.MeltdownExplosionZMultiplier = FRACUNIT*10/9	-- based on result of bomb.MeltdownExplosionXYMultiplier
bomb.MeltdownExplosionTumbleDivider = FRACUNIT*10	-- based on result of bomb.MeltdownExplosionXYMultiplier
bomb.MeltdownSelfDestructLaunch = FRACUNIT*20
bomb.MeltdownSelfDestructRange = FRACUNIT*192

-- fuse shot state
chaotix.SafeFreeslot("S_BOMB_FUSESHOT", "S_BOMB_FUSESHOT_THOK")
states[S_BOMB_FUSESHOT] = {SPR_PLAY, SPR2_FALL, 4, nil, 0, 0, S_BOMB_FUSESHOT}
states[S_BOMB_FUSESHOT_THOK] = {SPR_PLAY, SPR2_ROLL, 2, nil, 0, 0, S_BOMB_FUSESHOT_THOK}

-- top spin state
chaotix.SafeFreeslot("S_BOMB_TOPSPIN", "S_BOMB_TOPSPIN_FAST")
states[S_BOMB_TOPSPIN] = {SPR_PLAY, SPR2_TWIN, 2, nil, 0, 0, S_BOMB_TOPSPIN}
states[S_BOMB_TOPSPIN_FAST] = {SPR_PLAY, SPR2_SPIN, 2, nil, 0, 0, S_BOMB_TOPSPIN_FAST}

-- volatile launch state
chaotix.SafeFreeslot("S_BOMB_LAUNCH")
states[S_BOMB_LAUNCH] = {SPR_PLAY, SPR2_GASP, 2, nil, 0, 0, S_BOMB_LAUNCH}

-- function to spawn sparks
chaotix.SafeFreeslot("MT_BOMB_SPARK")
mobjinfo[MT_BOMB_SPARK] = {
	spawnstate = S_BOXSPARKLE2, // for whatever reason, this state is invisible???
	height = 8*FRACUNIT,
	radius = 4*FRACUNIT,
	flags = MF_NOBLOCKMAP|MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOCLIPTHING|MF_SCENERY
}

bomb.SpawnFuseSparks = function(player, num, zOffset, hMom, zMom, color, addMom, variateZMom, smoke)
	local mo = player.mo
	local colorized
		
	if color == nil
		color = SKINCOLOR_NONE
	end
		
	colorized = color ~= SKINCOLOR_NONE
		
	for i = 1, num
		local spark = P_SpawnMobjFromMobj(mo, 0, 0, zOffset, MT_BOMB_SPARK)
		
		if (smoke
		and spark.z <= mo.watertop
		and spark.z + spark.height >= mo.waterbottom)
		or (smoke == 2)	-- force
			spark.state = S_SMOKE1
		end
		P_InstaThrust(spark, FixedAngle(360*P_RandomFixed()), FixedMul(FixedMul(P_RandomFixed(), hMom), spark.scale))
		if variateZMom
			P_SetObjectMomZ(spark, FixedMul(P_RandomFixed(), zMom))
		else
			P_SetObjectMomZ(spark, zMom)
		end
		if addMom
			spark.momz = $ + mo.momz
			spark.momx = $ + mo.momx
			spark.momy = $ + mo.momy
		end
		spark.color = color
		spark.colorized = colorized
	end
end

-- visual pop effect
local popcolors = {
	SKINCOLOR_NONE,
	SKINCOLOR_BUBBLEGUM, 
	SKINCOLOR_YELLOW, 
	SKINCOLOR_FLAME, 
	SKINCOLOR_MINT, 
	SKINCOLOR_WHITE, 
	SKINCOLOR_JET, 
	SKINCOLOR_PASTEL,
	SKINCOLOR_PURPLE, 
	SKINCOLOR_APRICOT,
}
bomb.VisualPop = function(mo, speed, mobj)
	speed = $ or 10
	mobj = $ or MT_BOXSPARKLE
	
	local fx = P_SpawnMobjFromMobj(mo, 0, 0, mo.height/2, mobj or MT_BOXSPARKLE)
	fx.momx = FixedMul(mo.scale, P_RandomRange(-speed, speed) * FRACUNIT)
	fx.momy = FixedMul(mo.scale, P_RandomRange(-speed, speed) * FRACUNIT)
	fx.momz = FixedMul(mo.scale, P_RandomRange(-speed, speed) * FRACUNIT)
	fx.color = chaotix.RandomElement(popcolors)
	fx.colorized = true
	
	return fx
end

-- cupcake
chaotix.SafeFreeslot("S_BOMB_CUPCAKE", "sfx_bomcup")
states[S_BOMB_CUPCAKE] = {SPR_PLAY, SPR2_MLEE|FF_SPR2ENDSTATE, 4, nil, S_PLAY_MELEE_FINISH, 0, S_BOMB_CUPCAKE}
sfxinfo[sfx_bomcup].caption = "\"C'mere, cupcake!\""

-- bomb's explosion
chaotix.SafeFreeslot("S_BOMB_EXPLOSION", "MT_BOMB_EXPLOSION")
states[S_BOMB_EXPLOSION] = {SPR_NULL, 0, bomb.VolatileDamageTics, nil, 0, 0, S_NULL}
mobjinfo[MT_BOMB_EXPLOSION] = {
	spawnstate = S_BOMB_EXPLOSION,
	flags = MF_NOBLOCKMAP|MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT,
}

bomb.cv_showrange = CV_RegisterVar({
	name = "bomb_showrange",
	defaultvalue = "No",
	flags = CV_NETVAR,
	PossibleValue = CV_YesNo,
})

bomb.ExplosionRadialDamage = function(explosion)
	local range = explosion.bomb_range
	local height = explosion.bomb_height
	if not (range and height) then return end
	
	-- setup hitbox table
	explosion.bomb_hitmobjs = $ or {}
	
	-- search with blockmap
	local source = explosion.target
	searchBlockmap("objects", function(ref, found)
		if not ((found and found.valid) and found.health) then return end
		if (ref == found) then return end	-- don't damage the inflictor
		if (found == source) then return end	-- don't damage the source!
		
		-- already hit!
		if (ref.bomb_hitmobjs[found]) then return end
		
		-- valid type?
		if not ((found.flags & (MF_SHOOTABLE|MF_ENEMY|MF_MONITOR)) or found.player)
		and not (found.type == MT_SPIKE or found.type == MT_WALLSPIKE)
			return
		end
		if (found.player and found.player.spectator) then return end	-- NUH UH
		
		-- positional checks
		local dist = R_PointToDist2(ref.x, ref.y, found.x, found.y)
		if not (dist <= range) then return end	-- making it radial
		
		local height_modifier = FRACUNIT - FixedDiv(dist, range)
		local height_offset = ease.outcubic(height_modifier, height)
		
		local bottomz = ref.z - height_offset
		local topz = ref.z + height_offset
		if not ((topz >= found.z) and (bottomz <= found.z + found.height)) then return end	-- z checks
		
		if not (P_CheckSight(ref, found)) then return end
		
		-- damage them!
		if (source.player and source.player.bomb_cupcake == 2 and found.player)
			P_DamageMobj(found, source, source, 1000, DMG_INSTAKILL) -- cupcake instakill
		elseif (found.player and CBW_Battle)
			local B = CBW_Battle	-- laziness
			if (B.MyTeam(source.player, found.player)) then return end -- same team
			if (found.player.airdodge) then return end	-- air dodging, intangible!
			
			if (found.player.guard > 0)	-- guarding
			and not (ref.bomb_noparry)	-- it's transcendent
				if (B.SkinVars[found.player.skinvars].func_guard_trigger)
					B.SkinVars[found.player.skinvars].func_guard_trigger(found, parrymo, source)
				elseif (B.SkinVars[-1].func_guard_trigger)
					B.SkinVars[-1].func_guard_trigger(found, parrymo, source)
				end
			else	-- uh oh!
				local tumble_thrust = FixedMul(range, bomb.MeltdownExplosionXYMultiplier)
				local tumble_height = FixedMul(bomb.MeltdownExplosionZMultiplier, tumble_thrust)
				local tumble_tics = (FixedDiv(FixedDiv(tumble_thrust, ref.scale), bomb.MeltdownExplosionTumbleDivider) * TICRATE) / FRACUNIT	-- surely there's no chance of overflow!!!
				
				found.z = $ + P_MobjFlip(found)	-- GOD i hate SRB2
				found.momz = tumble_height * P_MobjFlip(ref)
				local angle = R_PointToAngle2(source.x, source.y, found.x, found.y)
				B.DoPlayerTumble(found.player, tumble_tics, angle, tumble_thrust, true)
				if (source.player) then B.PlayerCreditPusher(found.player, source) end
			end
		else
			if (found.bomb_volatile_launcher) then return end	-- C stack overflow!
			local func = (found.type == MT_SPIKE or found.type == MT_WALLSPIKE
			or (source.player and source.player.bomb_cupcake == 2)) and P_KillMobj or P_DamageMobj
			func(found, source, source)	-- maybe the explosion could be the inflictor instead of bomb?
		end
		ref.bomb_hitmobjs[found] = true
	end, explosion, explosion.x - range, explosion.x + range, explosion.y - range, explosion.y + range)
end

bomb.SpawnExplosion = function(mo, range, noparry)
	if not (mo and mo.valid) then return end	-- need a source
	if not (range) then return end	-- need a range and height
	
	local height = FixedMul(bomb.VolatileDamageHeightModifier, range)
	local tics = bomb.VolatileDamageTics
	local underwater = (mo.eflags & MFE_UNDERWATER)
	
	-- the burst that scales with range
	local burst = P_SpawnMobjFromMobj(mo, 0, 0, mo.height/2, MT_THOK)
	burst.state = S_TNTBARREL_EXPL4
	if (G_GametypeHasTeams())
		burst.color = mo.color
		burst.colorized = true
	end
	burst.scale = FixedDiv(range, bomb.VolatileBurstScalar)	-- scaled with the sprite's width
	if (underwater) then S_StartSound(mo, sfx_splash) end
	
	-- burst out particles
	local explode_thrust = FixedDiv(range, bomb.VolatileExplodeThrustScalar)
	for n = 1, 359 do
		if n % 45 continue end
		local xyangle = FixedAngle(FRACUNIT*n)
		local x = FixedMul(cos(xyangle), explode_thrust)
		local y = FixedMul(sin(xyangle), explode_thrust)
		for m = 1, 359 do
			if m % 36 continue end
			local zangle = FixedAngle(FRACUNIT*m)
			local x = FixedMul(cos(zangle), x)
			local y = FixedMul(cos(zangle), y)
			local z = FixedMul(sin(zangle), explode_thrust)
			local fx = P_SpawnMobjFromMobj(mo, x, y, (mo.height/2) + z, MT_DUST)
			if P_RandomChance(FRACUNIT/2)
				fx.state = (underwater) and S_WPLD1 or S_BOSSEXPLODE
			end
			if (G_GametypeHasTeams())
				fx.color = mo.color
				fx.colorized = true
			end
			fx.flags = $|MF_NOCLIPHEIGHT|MF_NOCLIP
			fx.scale = FixedDiv(range, bomb.VolatileExplodeScaleScalar)
			fx.destscale = 1
			fx.scalespeed = mo.scale / bomb.VolatileScaleSpeed
			fx.momx = x
			fx.momy = y
			fx.momz = z
			if (underwater) then fx.scale = FixedMul(bomb.VolatileUnderwaterModifier, $) end
		end
	end
	
	local dust_thrust = FixedDiv(range, bomb.VolatileDustThrustScalar)
	for n = 1, 16 do
		local xyangle = FixedAngle(FRACUNIT*P_RandomRange(1,359))
		local zangle = FixedAngle(FRACUNIT*P_RandomRange(1, 359))
		local x = FixedMul(cos(zangle), FixedMul(cos(xyangle) , dust_thrust))
		local y = FixedMul(cos(zangle), FixedMul(sin(xyangle) , dust_thrust))
		local z = FixedMul(sin(zangle), FixedMul(sin(xyangle) , dust_thrust))
		local fx = P_SpawnMobjFromMobj(mo, x, y, (mo.height/2) + z, MT_TNTDUST)
		fx.flags = $|MF_NOCLIPHEIGHT|MF_NOCLIP
		fx.scale = FixedDiv(range, bomb.VolatileDustScaleScalar)
		fx.destscale = 1
		fx.scalespeed = mo.scale / bomb.VolatileScaleSpeed
		fx.momx = x
		fx.momy = y
		fx.momz = z
	end
	
	-- quake the screen depending on strength
	local quake_intensity = FixedDiv(range, bomb.VolatileQuakeIntensityScalar)
	local quake_tics = (FixedDiv(range, bomb.VolatileQuakeTicsScalar) * TICRATE) / FRACUNIT	-- surely there's no chance of overflow!!!
	
	if (mo.player and mo.player.valid)
	and (mo.player == displayplayer or mo.player == secondarydisplayplayer)
		-- make sure it's local
		P_StartQuake(quake_intensity, quake_tics)
	end
	
	-- the actual damage range
	local explosion = P_SpawnMobjFromMobj(mo, 0, 0, mo.height/2, MT_BOMB_EXPLOSION)
	explosion.target = mo
	explosion.tics = tics
	explosion.bomb_range = range
	explosion.bomb_height = height
	explosion.bomb_noparry = noparry
	bomb.ExplosionRadialDamage(explosion)
	
	-- show range (debug)
	if (bomb.cv_showrange.value)
		explosion.sprite = SPR_THOK
		explosion.color = SKINCOLOR_ORANGE
		
		local angle_thoks = 30
		for angle_i = 1, angle_thoks
			local angle = ((360 / angle_thoks) * angle_i) * ANG1
				
			local length_thoks = 8
			for length_i = 0, length_thoks - 1
				local dist_modifier = (FRACUNIT / (length_thoks - 1)) * length_i
				local xoff = FixedMul(FixedMul(dist_modifier, range), cos(angle))
				local yoff = FixedMul(FixedMul(dist_modifier, range), sin(angle))
					
				for height_i = -1, 1, 2
					local height_modifier = FRACUNIT - dist_modifier
					local zoff = ease.outcubic(height_modifier, height) * height_i
						
					local thok = P_SpawnMobjFromMobj(explosion, xoff, yoff, (mo.height/2) + zoff, MT_THOK)
					thok.frame = ($ & ~(FF_TRANSMASK)) | (FF_FULLBRIGHT|FF_TRANS70)
					thok.scale = FixedMul(FRACUNIT/3, $)
					thok.color = SKINCOLOR_ORANGE
					thok.tics = tics
				end
			end
		end
	end
end

chaotix.AddHook("MobjThinker", function(explosion)
	-- make sure it's a valid explosion
	local source = explosion.target
	local range = explosion.bomb_range
	local height = explosion.bomb_height
	
	if not (source and source.valid)	-- no source???
	or not (range and height)	-- no range or height?
		P_RemoveMobj(explosion)
		return
	end
	
	-- damage stuff
	bomb.ExplosionRadialDamage(explosion)
end, MT_BOMB_EXPLOSION)

-- bomb giblets
A_BombGiblet = function(gib)
	if not ((gib and gib.valid) and gib.health) then return end
	
	-- target validity checks
	local target = gib.target
	
	if not (target and target.valid)	-- target doesn't exist
	or (target.health)	-- they revived??
	or (target.skin != bomb.SKIN)	-- we are no longer bomb
		P_RemoveMobj(gib)
		return
	end
	
	-- little hack to prevent movement of the player
	target.momz = 0
	
	-- set frame + colour
	local frameflags = (gib.frame & ~FF_FRAMEMASK)
	gib.frame = gib.bomb_frame|frameflags
	gib.color = target.color
	
	-- spin and fall
	gib.rollangle = $ + gib.bomb_rotate
	gib.momz = $ - (43690 * P_MobjFlip(gib))	-- magic number for now until I find out the actual value for death falling
end

chaotix.SafeFreeslot("S_BOMB_GIBLET")
states[S_BOMB_GIBLET] = {SPR_PLAY, SPR2_TAL2, 1, A_BombGiblet, 0, 0, S_BOMB_GIBLET}

bomb.SpawnGiblet = function(p, xoff, yoff, zoff, frame)
	local mo = p.mo
	
	local gib = P_SpawnMobjFromMobj(mo, xoff, yoff, zoff, MT_GHOST)
	gib.target = mo
	P_SetMobjStateNF(gib, S_BOMB_GIBLET)
	gib.color = mo.color
	gib.flags = MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT
	gib.bomb_giblet = true
	
	-- set frame
	local frameflags = (mo.frame & ~FF_FRAMEMASK)
	gib.skin = mo.skin
	gib.sprite, gib.sprite2 = SPR_PLAY, SPR2_TAL2
	gib.frame = frame|frameflags
	gib.bomb_frame = frame
	
	-- return gib to caller
	return gib
end

-- super fire
chaotix.SafeFreeslot("SKINCOLOR_SUPERFIRE1")
skincolors[SKINCOLOR_SUPERFIRE1] = {
	name = "Super Fire 1",
	ramp = {0, 0, 48, 48, 48, 48, 48, 49, 49, 50, 50, 32, 32, 33, 33, 33},
	accessible = false
}
chaotix.SafeFreeslot("SKINCOLOR_SUPERFIRE2")
skincolors[SKINCOLOR_SUPERFIRE2] = {
	name = "Super Fire 2",
	ramp = {48, 48, 48, 49, 50, 32, 32, 33, 33, 33, 33, 34, 34, 34, 36, 36},
	accessible = false
}
chaotix.SafeFreeslot("SKINCOLOR_SUPERFIRE3")
skincolors[SKINCOLOR_SUPERFIRE3] = {
	name = "Super Fire 3",
	ramp = {48, 48, 49, 50, 32, 33, 33, 34, 34, 36, 36, 38, 38, 40, 40, 42},
	accessible = false
}
chaotix.SafeFreeslot("SKINCOLOR_SUPERFIRE4")
skincolors[SKINCOLOR_SUPERFIRE4] = {
	name = "Super Fire 4",
	ramp = {48, 49, 50, 32, 33, 33, 34, 34, 36, 38, 38, 40, 40, 42, 42, 44},
	accessible = false
}
chaotix.SafeFreeslot("SKINCOLOR_SUPERFIRE5")
skincolors[SKINCOLOR_SUPERFIRE5] = {
	name = "Super Fire 5",
	ramp = {49, 50, 32, 33, 34, 34, 36, 36, 38, 40, 40, 42, 44, 46, 46, 46},
	accessible = false
}