-- make bomb explode when he takes damage
local bomb = chaotix.bomb

bomb.ExplodeDamage = function(mo)
	if (mo.skin != bomb.SKIN) then return end
	
	mo.player.losstime = $ + bomb.ExplodeLosstimeAdd
	mo.bomb_explode = true
	mo.bomb_explode_tics = 1
	return true
end

bomb.ResetExplode = function(mo)
	mo.bomb_explode = nil
	mo.bomb_explode_tics = nil
	mo.bomb_explode_exploded = nil
	
	local p = mo.player
	if (p and p.valid)
		if (p.bomb_fixmachineflag)
			if (skins[mo.skin].flags & SF_MACHINE)
				p.charflags = $|(SF_MACHINE)
			end
			p.bomb_fixmachineflag = nil
		end
	end
end

bomb.Explode = function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	
	-- not exploding
	if not (mo.bomb_explode) then bomb.ResetExplode(mo) return end
	
	if (mo.skin != bomb.SKIN)	-- not bomb!
	or (mo.health and not P_PlayerInPain(p))	-- not hurt anymore!
	or (p.isjettysyn)	-- NO
		bomb.ResetExplode(mo)
		return
	end
	
	-- increase timer
	mo.bomb_explode_tics = $+1
	local tics = mo.bomb_explode_tics
	
	-- explode on cue!
	local explode_tics = (mo.health) and bomb.ExplodeTics or bomb.ExplodeDeathTics
	if (tics >= (explode_tics - bomb.ExplodeStartup))
		if (tics < explode_tics)
			tics = $ - (explode_tics - bomb.ExplodeStartup)
			
			-- startup for the explosion
			local frames = 3	-- this'll be useful to expose
			chaotix.SetFrame(mo, (tics * frames) / bomb.ExplodeStartup, SPR2_FIRE)
		else
			tics = $ - explode_tics
			
			-- frames
			local frame = 2
			local totaltics = 0
			
			for _, frametics in ipairs(bomb.ExplodeFrameTics)
				totaltics = $+frametics
				if not (tics >= totaltics) then break end
				frame = $+1
			end
			chaotix.SetFrame(mo, frame, SPR2_FIRE)
			
			-- actually explode
			if not (tics)
				if not (mo.bomb_explode_exploded)	-- little hack
					-- adjust momentum
					mo.momx = FixedMul(bomb.ExplodeRecoilModifier, $)
					mo.momy = FixedMul(bomb.ExplodeRecoilModifier, $)
					P_SetObjectMomZ(mo, chaotix.WaterMom(mo, bomb.ExplodeRecoilHeight))
					
					-- do the explosion
					local underwater_modifier = (mo.eflags & MFE_UNDERWATER) and bomb.VolatileUnderwaterNerf or FRACUNIT
					local range = FixedMul(mo.scale, FixedMul(underwater_modifier, bomb.ExplodeRange))
					bomb.SpawnExplosion(mo, range, true)
					S_StartSound(mo, sfx_kc3b)
					
					-- set hacky var
					mo.bomb_explode_exploded = true
				end
				
				-- uh oh we died!
				if not (mo.health)
					-- go invisible
					P_SetMobjStateNF(mo, S_INVISIBLE)
					mo.momz = 0
					mo.flags = ($|MF_NOGRAVITY) & ~(MF_NOCLIPHEIGHT)
					p.charflags = $ & ~(SF_MACHINE)
					p.bomb_fixmachineflag = true
					
					-- spawn giblets
					local spawnangle = mo.angle - ANGLE_90
					local function bomb_launchGiblet(gib, thrust, height)
						local speed = FixedHypot(thrust, height)
						if (thrust > 0) then speed = $ * -1 end
						
						-- launch them
						gib.momx, gib.momy = mo.momx, mo.momy
						P_Thrust(gib, spawnangle, FixedMul(gib.scale, chaotix.WaterMom(mo, thrust)))
						P_SetObjectMomZ(gib, chaotix.WaterMom(mo, height))
						
						-- set rotation speed
						local modifier = FRACUNIT*4/3
						gib.bomb_rotate = FixedAngle(FixedMul(modifier, FixedDiv(speed, gib.scale)))
					end
					
					local body = bomb.SpawnGiblet(p, 0, 0, 0, A)
					bomb_launchGiblet(body, P_RandomRange(2, 5)*(not P_RandomRange(0, 1) and -1 or 1)*FRACUNIT, P_RandomRange(11, 15)*FRACUNIT)
					 
					local larm_distance = -(FRACUNIT*26)
					local larm = bomb.SpawnGiblet(p, FixedMul(cos(spawnangle), larm_distance), FixedMul(sin(spawnangle), larm_distance), FRACUNIT*11, B)
					bomb_launchGiblet(larm, P_RandomRange(-16, -9)*FRACUNIT, P_RandomRange(12, 18)*FRACUNIT)
					
					local rarm_distance = FRACUNIT*24
					local rarm = bomb.SpawnGiblet(p, FixedMul(cos(spawnangle), rarm_distance), FixedMul(sin(spawnangle), rarm_distance), FRACUNIT*26, C)
					bomb_launchGiblet(rarm, P_RandomRange(9, 16)*FRACUNIT, P_RandomRange(12, 18)*FRACUNIT)
					
					local lleg_distance = -(FRACUNIT*18)
					local lleg = bomb.SpawnGiblet(p, FixedMul(cos(spawnangle), lleg_distance), FixedMul(sin(spawnangle), lleg_distance), -(FRACUNIT*10), D)
					bomb_launchGiblet(lleg, P_RandomRange(-16, -9)*FRACUNIT, P_RandomRange(5, 7)*FRACUNIT)
					
					local rleg_distance = FRACUNIT*28
					local rleg = bomb.SpawnGiblet(p, FixedMul(cos(spawnangle), rleg_distance), FixedMul(sin(spawnangle), rleg_distance), -(FRACUNIT*4), E)
					bomb_launchGiblet(rleg, P_RandomRange(9, 16)*FRACUNIT, P_RandomRange(5, 7)*FRACUNIT)
				end
			end
		end
	end
end