-- bomb's secondary ability, fuse dash
local bomb = chaotix.bomb

-- start the fuse dash (even in the air!)
bomb.SpinSpecial = function(p)
	if (p.pflags & PF_SPINDOWN) then return end	-- make sure it's not continuous
	
	-- ability checks
	local mo = p.mo
	if not (mo and mo.valid) then return end
	if not (bomb.AbilityCheck(p, 2)) then return end
	
	-- state checks
	if (p.pflags & PF_SPINNING) then return end	-- easy spinning check
	if (p.powers[pw_carry]) then return end	-- getting carried
	if (p.exiting) then return end -- exiting the level
	if (P_SuperReady(p)) then return end -- turning super
	
	if not (P_IsObjectOnGround(mo))	-- aerial checks
		-- need a bit more checks for non-grounded activations...
		if not ((p.pflags & PF_JUMPED) and not (mo.bomb_thokked & bomb.ThokkedFlags["FUSEDASH"]))	-- havent already done the dash
		or (P_PlayerInPain(p))	-- not while in pain
		or (CBW_Battle and (p.gotflag or p.gotcrystal))	-- not with the flag
			return
		end
	end
	
	-- bomb checks
	if (mo.bomb_fuseshot and mo.state == S_BOMB_FUSESHOT) then return end	-- pre-fire
	if (mo.bomb_volatile & bomb.VolatileFlags["FUSEDASH"]) then return end
	if (mo.bomb_volatile_hitlag) then return end
	
	-- start the charge!
	p.pflags = ($|PF_SPINNING|PF_STARTDASH)
	mo.state = S_PLAY_SPINDASH
	mo.bomb_fusedash = true
	mo.bomb_fusedash_charge = 1
	S_StartSound(mo, sfx_spndsh)
	//bomb.AddThokked(p, bomb.ThokkedFlags["FUSEDASH"])
	
	-- frame stuff
	mo.bomb_fusedash_frame = 0
	mo.bomb_fusedash_frametics = 3
end

-- resetting the fuse dash
bomb.ResetFuseDash = function(mo)
	mo.bomb_fusedash = nil
	mo.bomb_fusedash_charge = nil
	mo.bomb_fusedash_maxed = nil
	mo.bomb_fusedash_frame = nil
	mo.bomb_fusedash_frametics = nil
end

-- main thinker for charging the fuse dash
bomb.FuseDash = function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	
	-- not doing the fuse dash?
	if not (mo.bomb_fusedash) then return end
	
	-- cancel it early!
	if not (mo.health)	-- dead
	or not (bomb.AbilityCheck(p, 2))	-- not using fuse dash
	or (P_PlayerInPain(p))	-- we got hurt
	or (mo.eflags & MFE_SPRUNG)	-- hit a spring
	or (p.powers[pw_carry])	-- we got carried
	or (p.panim == PA_PAIN)	-- pain pain pain pain pain
	or (p.powers[pw_nocontrol])	-- no control
	or (p.airdodge)	-- battle support (2)
		bomb.ResetFuseDash(mo)
		return
	end
	
	-- make sure we're spinning
	p.pflags = $|(PF_SPINNING|PF_STARTDASH)
	if (CBW_Battle and (p.gotflag or p.gotcrystal))
		-- no jumping with the flag
		p.pflags = $|(PF_JUMPSTASIS)
	end
	
	-- release the spindash!
	if not (p.cmd.buttons & BT_SPIN)
		-- launch off
		local charge = mo.bomb_fusedash_charge
		local range = p.maxdash - p.mindash
		local dashspeed = p.mindash + ((charge * range) / bomb.FuseDashChargeTics)
		
		p.pflags = $ & ~(PF_STARTDASH)
		mo.state = S_BOMB_TOPSPIN
		P_InstaThrust(mo, mo.angle, FixedMul(mo.scale, dashspeed))
		S_StartSound(mo, sfx_zoom)
		if (charge == bomb.FuseDashChargeTics)
			-- if we're at max charge, become volatile!
			bomb.TriggerVolatile(p, bomb.VolatileFlags["FUSEDASH"])
		end
		bomb.AddThokked(p, bomb.ThokkedFlags["FUSEDASH"])
		
		-- visual pop
		for i = 1, 20 do
			bomb.VisualPop(mo)
		end
		
		-- reset stuff
		bomb.ResetFuseDash(mo)
		return
	end
	
	-- increase the charge, become volatile at max charge
	mo.bomb_fusedash_charge = min(bomb.FuseDashChargeTics, $+1)
	local charge = mo.bomb_fusedash_charge
	local range = p.maxdash - p.mindash
	local dashspeed = p.mindash + ((charge * range) / bomb.FuseDashChargeTics)
	
	-- max charge sound
	if (charge == bomb.FuseDashChargeTics)
	and not (mo.bomb_fusedash_maxed)
		mo.bomb_fusedash_maxed = true
		S_StartSound(mo, sfx_cdfm50)
	end
	
	-- set state
	if (mo.state != S_PLAY_SPINDASH) then mo.state = S_PLAY_SPINDASH end
	
	-- get max frametics
	local step = (p.maxdash - p.mindash) / 4
	local speed = (dashspeed - p.mindash)
	local maxtics = 3	-- default is 3 with no charge
			
	if (speed > (3 * step))
		maxtics = 1
	elseif (speed > step)
		maxtics = 2
	end
	mo.bomb_fusedash_frametics = $ and min(maxtics, $) or maxtics
	//print(maxtics)
		
	-- reduce frame tics and get the frame we want
	local numframes = 4
			
	mo.bomb_fusedash_frametics = $-1
	if not (mo.bomb_fusedash_frametics)
		mo.bomb_fusedash_frame = $+1
				
		-- reset frame after max frame is reached
		if (mo.bomb_fusedash_frame > (numframes - 1))
			mo.bomb_fusedash_frame = 0
		end
	end
			
	-- set frame
	chaotix.SetFrame(mo, mo.bomb_fusedash_frame, SPR2_SPIN)
	
	-- light the fuse on bomb's head (reused from Fuse Shot)
	local colour = SKINCOLOR_ROSE	-- max charge colour
	if (charge < bomb.FuseDashChargeTics)
		local colours = bomb.FuseDashSparkColours
		local index = (charge * #colours) / bomb.FuseDashChargeTics
		colour = colours[index + 1]
	end
	bomb.SpawnFuseSparks(p, 1, FixedDiv(mo.height, mo.scale), bomb.FuseShotSparkHMom, bomb.FuseShotSparkZMom, colour, true, false, true)
	
	-- dust particles for the ground
	if (P_IsObjectOnGround(mo))
	and (chaotix.PlayerSpeedLimit(p, FixedMul(mo.scale, FRACUNIT*10), FRACUNIT*19/20))	-- regulate speed
	and (leveltime%2)
		local angle = mo.angle + FixedAngle(P_RandomRange(140, 200)*FRACUNIT)
		local dust = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_SPINDUST)
		
		P_InstaThrust(dust, angle, FixedMul(dust.scale, FRACUNIT*20))
		P_SetObjectMomZ(dust, P_RandomRange(2,4) * FRACUNIT/2)
	end
end