-- bomb's jump ability, fuse shot
local bomb = chaotix.bomb

bomb.ActivateFuseShot = function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	if not (bomb.AbilityCheck(p, 1)) then return end
	
	-- state checks
	if not (p.pflags & PF_JUMPED) then return end	-- gotta be jumping
	if (mo.bomb_thokked & bomb.ThokkedFlags["FUSESHOT"]) then return end	-- thokked state
	if (p.gotcrystal) then return end	-- battlemod support
	if (CBW_Battle and p.exhaustmeter == 0) then return end	-- we're exhausted (battlemod)
	
	-- bomb checks
	if (mo.bomb_fuseshot) then return end	-- already doing it!
	//if (mo.bomb_fusedash) then return end	-- cant be charging a spindash
	if (mo.bomb_volatile_hitlag) then return end
	
	-- start the fuse!
	p.pflags = ($|PF_JUMPED|PF_NOJUMPDAMAGE) & ~(PF_STARTJUMP|PF_SPINNING)
	mo.state = S_BOMB_FUSESHOT
	mo.bomb_fuseshot = true	-- we're gonna use this as a stand-in PF_THOKKED
	mo.bomb_fuseshot_tics = 1
	S_StartSound(mo, sfx_cdfm16)
	bomb.AddThokked(p, bomb.ThokkedFlags["FUSESHOT"])
	
	-- cap our speed a bit
	local modifier = bomb.FuseShotStartupModifier
	mo.momx = FixedMul(modifier, $)
	mo.momy = FixedMul(modifier, $)
	P_SetObjectMomZ(mo, chaotix.WaterMom(mo, bomb.FuseShotStartupHeight))
end

-- reset fuse shot vars
bomb.ResetFuseShot = function(mo)
	mo.bomb_fuseshot = nil
	mo.bomb_fuseshot_tics = nil
	mo.bomb_fuseshot_maxed = nil
	mo.bomb_fuseshot_target = nil
	mo.bomb_fuseshot_shot = nil
	if (mo.bomb_fusethok_bounce)
		mo.flags = $ & ~(MF_BOUNCE)
		mo.bomb_fusethok_bounce = nil
	end
	mo.bomb_fusethok_bounced = nil
end

bomb.FindFustShotTarget = function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	
	-- search for the target
	local target
	local dist = FixedMul(mo.scale, RING_DIST)
	
	searchBlockmap("objects", function(ref, found)
		if not (found.flags & (MF_ENEMY|MF_BOSS|MF_MONITOR))
		or not (found.flags & MF_SHOOTABLE)
		or found.health <= 0
			return
		end
		
		local scale = ref.scale
		local vDist = P_MobjFlip(ref) * ((found.z + (found.height >> 1)) - (ref.z + (ref.height >> 1)) )
		
		// object has to be close enough vertically
		if vDist > 0
		or vDist < -256*scale
			return
		end
		
		// pick the closest object to the player
		local hDist = R_PointToDist2(ref.x, ref.y, found.x, found.y)
		
		if hDist > dist
			return
		end
		
		// object has to line up with the angle the player is facing
		local angle = ref.angle
		local x, y = ref.x + FixedMul(hDist, cos(angle)), ref.y + FixedMul(hDist, sin(angle))
		
		if x - ref.radius > found.x + found.radius
		or x + ref.radius < found.x - found.radius
		or y - ref.radius > found.y + found.radius
		or y + ref.radius < found.y - found.radius
			return
		end
		
		dist = hDist	-- so we can work to find the closest enemy
		target = found
	end, mo, mo.x - dist, mo.x + dist, mo.y - dist, mo.y + dist)
	
	-- return target
	return target
end

bomb.FuseShot = function(p)
	local mo = p.mo
	if not ((mo and mo.valid) and mo.health) then return end
	if not (bomb.AbilityCheck(p, 1)) then bomb.ResetFuseShot(mo) return end
	
	-- dont do anything
	if not (mo.bomb_fuseshot) then return end
	
	-- reset fuse shot stuff
	if not (mo.state >= S_BOMB_FUSESHOT and mo.state <= S_BOMB_FUSESHOT_THOK)	-- we got knocked out of it
	or (p.gotflag or p.gotcrystal)	-- we got the flag (or the crystal in battlemod)
	or (CBW_Battle and p.exhaustmeter == 0)	-- we're exhausted (battlemod)
		-- gotta revert state manually
		if (mo.state >= S_BOMB_FUSESHOT and mo.state <= S_BOMB_FUSESHOT_THOK)
			-- reset state properly
			mo.state = S_PLAY_FALL 
		end
		
		-- ground hit, are we volatile?
		if (mo.eflags & MFE_JUSTHITFLOOR)
		and not (mo.bomb_volatile)	-- we aren't, bounce!
		and (mo.bomb_fuseshot_shot)	-- make sure we actually shot the ability
			p.pflags = ($|PF_JUMPED|PF_STARTJUMP) & ~(PF_SPINNING|PF_NOJUMPDAMAGE)
			mo.state = S_PLAY_ROLL
			P_SetObjectMomZ(mo, chaotix.WaterMom(mo, bomb.FuseShotBounceHeight))
			S_StartSound(mo, sfx_s3k5d)
			bomb.SpawnFuseSparks(p, 8, 0, bomb.FuseShotBounceSparkHMom, bomb.FuseShotBounceSparkZMom, SKINCOLOR_KETCHUP, false)
			
			-- setup thokked stuff
			bomb.AddThokked(p, bomb.ThokkedFlags["FUSESHOT"])
		end
		
		bomb.ResetFuseShot(mo)
		return
	end
	
	if (mo.state == S_BOMB_FUSESHOT)	-- pre-fire
		mo.bomb_fuseshot_tics = $+1	-- increase charge timer
		local timer = mo.bomb_fuseshot_tics
		
		-- max charge sound
		if (timer == bomb.FuseShotChargeMaxTics)
		and not (mo.bomb_fuseshot_maxed)
			mo.bomb_fuseshot_maxed = true
			S_StartSound(mo, sfx_cdfm50)
		end
			
		-- force volatile release if we charge too long
		if (timer >= bomb.FuseShotOverchargeTics)
			-- kill momentum
			mo.momx = 0
			mo.momy = 0
				
			-- set volatile flag directly, then trigger the launch
			mo.bomb_volatile = $ and $|(bomb.VolatileFlags["FUSESHOT"]) or bomb.VolatileFlags["FUSESHOT"]
			bomb.VolatileLaunch(p, bomb.VolatileOverchargeHeight)
			return
		end
			
		-- try to get a target
		local target
		if (p.cmd.buttons & BT_SPIN) then target = bomb.FindFustShotTarget(p) end
		
		-- play lock-on sounds
		local old_target = mo.bomb_fuseshot_target
			
		if (target != old_target)	-- new target?
			S_StartSound(mo, (target and target.valid) and sfx_kc39 or sfx_kc43, p)
		end
		mo.bomb_fuseshot_target = target	-- store target so we can check next tick
		
		-- spawn an overlay for the target
		if (target and target.valid)
			local ret = P_SpawnMobjFromMobj(target, 0, 0, 0, MT_OVERLAY)
			ret.tracer = P_SpawnMobjFromMobj(ret, 0, 0, 0, MT_OVERLAY)
			ret.tracer.target = ret
				
			for i = 1, -1, -2
				if not splitscreen
				and p ~= displayplayer
					ret.flags2 = $ | MF2_DONTDRAW
				end
				ret.rollangle = i * ANGLE_90
				ret.state = S_LOCKON1
				ret.tics = 1
				ret.color = SKINCOLOR_KETCHUP
				ret.colorized = true
				ret.spritexscale = $ * 3 / 2
				ret.spriteyscale = $ * 3 / 2
				ret = $.tracer or $.target
			end
			
			local scale = FixedMul(ret.spritexscale, target.scale)
			local radius = FixedDiv(target.radius, scale) + 35*FRACUNIT/2
			local height = FixedDiv(target.height >> 1, scale) - 19*FRACUNIT/2
				
			for i = 1, -1, -2
				ret.spritexoffset = i*radius
				ret.spriteyoffset = height
				ret = $.tracer
			end
		end
		
		-- release the charge manually!
		if not (p.cmd.buttons & BT_JUMP)
		and (timer >= bomb.FuseShotChargeMinTics)
		or (timer >= bomb.FuseShotOverchargeTics)	-- uh... fallback??
			local max_tics = bomb.FuseShotChargeMaxTics - bomb.FuseShotChargeMinTics
			timer = min(max_tics, $ - bomb.FuseShotChargeMinTics)	-- easier calculations
			
			-- set momentum
			local modifier_range = bomb.FuseShotMaxModifier - FRACUNIT
			local modifier = FRACUNIT + ((timer * modifier_range) / max_tics)
				
			local thrust = FixedMul(modifier, FixedMul(mo.scale, p.actionspd))
				
			if (target and target.valid)
				local vDist = P_MobjFlip(mo) * ((target.z + (target.height >> 1)) - (mo.z + (mo.height >> 1)) )
				local hDist = R_PointToDist2(mo.x, mo.y, target.x, target.y)
				
				local grav = abs(P_GetMobjGravity(mo))
				local vCap = -thrust/2
				local angle = R_PointToAngle2(0, 0, hDist, vDist)
				local tics
				
				local hThrust = FixedMul(thrust, cos(angle))
				local vThrust = FixedMul(thrust, sin(angle))
				
				if vThrust < vCap
					hThrust = FixedMul($, FixedDiv(vCap, vThrust))
					vThrust = vCap
				end
				
				tics = FixedDiv(hDist, mo.scale) / hThrust
				
				while tics > 0
					tics = $ - 1
					vThrust = $ + grav
				end
				
				mo.bomb_fuseshot_homing = target	-- for spin buffering
				P_InstaThrust(mo, mo.angle, hThrust)
				P_SetObjectMomZ(mo, vThrust)
			else
				P_InstaThrust(mo, mo.angle, chaotix.WaterMom(mo, thrust))
				P_SetObjectMomZ(mo, chaotix.WaterMom(mo, bomb.FuseShotHeight))
				//print("speed: "..tostring(FixedDiv(thrust, mo.scale) / FRACUNIT))
			end
			
			-- set state
			p.pflags = ($|PF_JUMPED) & ~(PF_NOJUMPDAMAGE)
			p.drawangle = mo.angle
			mo.state = S_BOMB_FUSESHOT_THOK
			mo.bomb_fuseshot_shot = true
			S_StartSound(mo, sfx_kc2a)
			
			-- trigger volatile state
			bomb.TriggerVolatile(p, bomb.VolatileFlags["FUSESHOT"])
			//p.lastbuttons = $|(BT_JUMP)	-- a little bit of an annoying hack...
			if not (mo.bomb_volatile)
				mo.flags = $|(MF_BOUNCE)
				mo.bomb_fusethok_bounce = true	-- just for resetting it
			end
			
			return
		end
			
		-- reduce momentum while charging
		local moms = {"momx", "momy"}
		for i = 1, 2
			local mom = moms[i]
			mo[mom] = FixedMul(bomb.FuseShotChargeXYModifier, $)
		end
			
		-- slowfall (which is actually just lower gravity)
		if ((mo.momz * P_MobjFlip(mo)) < 0)	-- only when descending!
			mo.momz = $ - (FixedMul(P_GetMobjGravity(mo), bomb.FuseShotChargeSlowfall))
		end
		
		-- preserve drawangle
		local diff = mo.angle - p.drawangle
		local factor = 8
		if (diff)
			diff = (chaotix.AngToInt(diff) > 180) and InvAngle(InvAngle(diff)/factor) or $/factor
			p.drawangle = $ + diff
		end
		
		-- light the fuse on bomb's head
		local colour
		
		if (timer >= bomb.FuseShotChargeMinTics)
			local max_tics = bomb.FuseShotChargeMaxTics - bomb.FuseShotChargeMinTics
			timer = min(max_tics, $ - bomb.FuseShotChargeMinTics)	-- easier calculations
			
			colour = SKINCOLOR_ROSE
			if (timer < max_tics)
				local colours = bomb.FuseShotSparkColours
				local index = (timer * #colours) / max_tics
				colour = colours[index + 1]
			end
		end
		
		bomb.SpawnFuseSparks(p, 1, FixedDiv(mo.height, mo.scale), bomb.FuseShotSparkHMom, bomb.FuseShotSparkZMom, colour, true, false, true)
	elseif (mo.state == S_BOMB_FUSESHOT_THOK)
		-- become bouncy if we're not volatile!
		if (mo.bomb_volatile)
			mo.flags = $ & ~(MF_BOUNCE)
			mo.bomb_fusethok_bounce = nil
		else
			mo.flags = $|(MF_BOUNCE)
			mo.bomb_fusethok_bounce = true	-- just for resetting it
		end
		
		-- change drawangle
		if (mo.bomb_fuseshot_bounced)
			p.drawangle = R_PointToAngle2(0, 0, mo.momx, mo.momy)
			mo.bomb_fuseshot_bounced = nil
		end
		
		-- roll faster with speed!
		local speed = FixedDiv(p.speed, FixedMul(mo.scale, mo.movefactor))
		if (speed > 16<<FRACBITS)
			mo.tics = min(1, $)
		else
			mo.tics = min(2, $)
		end
	end
end

-- bounce around when we're not volatile!
bomb.FuseShotMoveBlocked = function(mo)
	if not (mo.bomb_fuseshot and mo.state == S_BOMB_FUSESHOT_THOK)
	or not (mo.flags & MF_BOUNCE)
		return
	end
	
	-- bounce sparks + sound
	local p = mo.player
	mo.bomb_fuseshot_bounced = true	-- used to change our drawangle
	bomb.SpawnFuseSparks(p, 8, FixedDiv(mo.height, mo.scale) >> 1, bomb.FuseShotRicochetSparkHMom, bomb.FuseShotRicochetSparkZMom, SKINCOLOR_KETCHUP, false)
	S_StartSound(mo, sfx_s3k96)
	return true
end