-- bomb's volatile state he gets from his abilities
local bomb = chaotix.bomb

-- hitlag for volatile activations
bomb.StartHitlag = function(p, tics, jumpbuffer, spinbuffer)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	
	mo.bomb_volatile_hitlag = max(0, tics)
	mo.bomb_volatile_angle = p.drawangle	-- preserve angle
	
	-- hacky shit
	if (jumpbuffer == nil) then jumpbuffer = (p.cmd.buttons & BT_JUMP) end
	if (spinbuffer == nil) then spinbuffer = true end
	
	mo.bomb_volatile_jumpbuffer = (jumpbuffer) and true or false
	mo.bomb_volatile_spinbuffer = (spinbuffer) and true or false
end

bomb.VolatileHitlag = function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return false end
	
	-- no hitlag?
	if not (mo.bomb_volatile_hitlag) then return false end
	
	-- reset stuff on death! (if we somehow die...
	if not (mo.health)
		mo.flags = $ & ~(MF_NOTHINK)
		mo.spritexoffset = 0
		mo.bomb_volatile_hitlag = 0
		return false
	end
	
	-- reduce hitlag
	mo.bomb_volatile_hitlag = max(0, $-1)
	local hitlag = mo.bomb_volatile_hitlag
	
	if (hitlag)	-- we still got hitlag
		local offset = (2 * hitlag) * FRACUNIT
		if (hitlag%2) then offset = $ * -1 end
		
		p.drawangle = mo.bomb_volatile_angle
		mo.flags = $|(MF_NOTHINK)
		mo.spritexoffset = offset
	else	-- endlag finished
		mo.flags = $ & ~(MF_NOTHINK)
		mo.spritexoffset = 0
		
		-- some hacky (but convenient) stuff
		local jumpbuffer = mo.bomb_volatile_jumpbuffer
		local spinbuffer = mo.bomb_volatile_spinbuffer
		
		if (jumpbuffer) then p.pflags = $|(PF_STARTJUMP) end
		if (spinbuffer) then p.pflags = $ & ~(PF_SPINDOWN) end
	end
	
	-- return true to end other thinkers
	return true
end

-- reset volatile state
bomb.ResetVolatile = function(mo, sound)
	local p = mo.player
	
	if (sound and (bomb.AbilityCheck(p, 1) or bomb.AbilityCheck(p, 2)))	-- we need to be bomb to make a sound!
	and (mo.bomb_volatile)	-- we WERE volatile
		S_StartSound(mo, sfx_kc65)
	end
	mo.bomb_volatile = 0
	mo.bomb_volatile_oldinfo = nil
	if (mo.bomb_volatile_colorized)	
		mo.colorized = false
		mo.bomb_volatile_colorized = nil
	end
end

-- trigger volatile states
bomb.TriggerVolatile = function(p, flag)
	local mo = p.mo
	
	if not (bomb.AbilityCheck(p, 1) or bomb.AbilityCheck(p, 2))	-- no bomb abilities
	or (p.gotflag or p.gotcrystal)	-- carrying something
		return
	end
	
	-- try to trigger it
	mo.bomb_volatile = $ or 0
	
	if (mo.bomb_volatile & flag) then return end
	mo.bomb_volatile = $ and $|(flag) or flag
	S_StartSound(mo, sfx_kc68)
	
	-- store old values
	mo.bomb_volatile_oldinfo = {	
		-- old momentum values
		momx = mo.momx,
		momy = mo.momy,
		momz = mo.momz,
		
		-- angle for angled rebounds
		angle = R_PointToAngle2(0, 0, mo.momx, mo.momy),
		
		-- bomb thokked state
		thokked = mo.bomb_thokked,
	}
end

-- the vertical launch we get from volatile activations
bomb.VolatileLaunch = function(p, momz, jumpbuffer, spinbuffer)
	local mo = p.mo
	local underwater_modifier = (mo.eflags & MFE_UNDERWATER) and bomb.VolatileUnderwaterNerf or FRACUNIT
	
	-- might as well
	mo.bomb_volatile = $ or 1
	local volatile = mo.bomb_volatile
	
	-- get how volatile we are to amplify effects
	local multiplier = bomb.VolatileEffectTable[1].multiplier
	local sound = bomb.VolatileEffectTable[1].sound
	
	local count = (P_PlayerFullbright(p)) and 1 or 0	-- start out more volatile if we're super!
	for _, flag in pairs(bomb.VolatileFlags)
		if not (volatile & flag) then continue end
		count = $+1
		
		-- get new info?
		if not (count > 1) then continue end
		
		local effect_table = bomb.VolatileEffectTable[count]
		if not (effect_table) then continue end	-- paranoia
		
		multiplier = bomb.VolatileEffectTable[count].multiplier or $
		sound = bomb.VolatileEffectTable[count].sound or $
	end
	
	-- launch and set hitlag!
	mo.bomb_volatile_launch = true
	P_SetObjectMomZ(mo, chaotix.WaterMom(mo, FixedMul(underwater_modifier, FixedMul(multiplier, momz))))
	bomb.StartHitlag(p, FixedDiv(abs(mo.momz), mo.scale) / bomb.VolatileHitlagDivider, jumpbuffer, spinbuffer)
	//print("height: "..tostring(FixedDiv(abs(mo.momz), mo.scale) / FRACUNIT))
	
	-- set thokked state
	if (mo.bomb_volatile & bomb.VolatileFlags["FUSESHOT"])
		bomb.AddThokked(p, bomb.ThokkedFlags["FUSESHOT"])
	end
	if (mo.bomb_volatile & bomb.VolatileFlags["FUSEDASH"])
		bomb.AddThokked(p, bomb.ThokkedFlags["FUSEDASH"])
	end
	
	-- reset volatile
	bomb.ResetVolatile(mo)
	
	-- spawn explosion to damage stuff!
	local range = FixedMul(FixedDiv(mo.momz, mo.scale), bomb.VolatileDamageRangeMultiplier)
	local minrange = FixedMul(mo.scale, FixedMul(underwater_modifier, bomb.VolatileDamageMinRange))
	range = max(minrange, $)
	bomb.SpawnExplosion(mo, range)
	
	-- set player state
	p.pflags = ($|PF_JUMPED|PF_NOJUMPDAMAGE) & ~(PF_SPINNING)
	mo.state = S_BOMB_LAUNCH
	S_StartSound(mo, sound)
	bomb.ResetFuseShot(mo)	-- also reset this
	bomb.ResetFuseDash(mo)	-- make sure to reset
end

-- the active volatile state
bomb.Volatile = function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	
	-- NEVER have this as nil
	mo.bomb_volatile = $ or 0
	
	-- volatile launch state
	if (mo.state == S_BOMB_LAUNCH)
		-- spin in the air!
		mo.bomb_volatile_angle = $ + FixedAngle(FixedDiv(7*mo.momz, 2*mo.scale))
		p.drawangle = mo.bomb_volatile_angle
		
		-- when our momz declines completely, go to falling
		-- (just like S_PLAY_SPRING)
		if ((mo.momz * P_MobjFlip(mo)) <= 0)
			mo.state = S_PLAY_FALL
			mo.bomb_volatile_angle = nil
		else
			// spawns smoke that trails @player's hitbox
			local scale = mo.scale
			
			local momx = FixedDiv(mo.momx, scale)
			local momy = FixedDiv(mo.momy, scale)
			local momz = FixedDiv(mo.momz, scale) * P_MobjFlip(mo)
			local radius = FixedDiv(mo.radius, scale) * 2 / 3
			local height = FixedDiv(mo.height, scale) * 2 / 3
			
			local xoff = FixedMul(P_RandomFixed(), radius << 1) - radius - momx
			local yoff = FixedMul(P_RandomFixed(), radius << 1) - radius - momy
			local zoff = FixedMul(P_RandomFixed(), height) - momz
			P_SpawnMobjFromMobj(mo, xoff, yoff, zoff, MT_SMOKE)
		end
	end
	
	-- remove volatile flags
	local volatile = mo.bomb_volatile
	
	if (volatile)
		for flag, value in pairs(bomb.VolatileFlags)
			if not (volatile & value) then continue end
			
			if not (mo.health)	-- dead
			or (P_PlayerInPain(p))	-- got hurt
			or (p.gotflag or p.gotcrystal)	-- carrying something
				volatile = $ & ~(value)
				continue
			end
			
			if (flag == "MELTDOWN")	-- special contitions
				if not (p.actionstate)
					volatile = $ & ~(value)
					continue
				end
			else
				if not (bomb.AbilityCheck(p, 1) or bomb.AbilityCheck(p, 2))	-- no bomb abilities
				or (p.powers[pw_carry])	-- got carried
				or (not (mo.eflags & MFE_JUSTHITFLOOR) and not ((p.pflags & PF_JUMPED) or (p.pflags & PF_SPINNING)))
					volatile = $ & ~(value)
					continue
				end
			end
		end
	end
	if not (volatile)
		-- not volatile anymore!
		bomb.ResetVolatile(mo, mo.health)
		return
	end
	mo.bomb_volatile = volatile	-- make sure this is updated
	
	-- manual activations
	local shot_activate = false
	if (volatile & bomb.VolatileFlags["FUSESHOT"])
	and ((p.cmd.buttons & BT_JUMP) and not (p.lastbuttons & BT_JUMP))
		shot_activate = true
	end
	
	local dash_activate = false
	if (volatile & bomb.VolatileFlags["FUSEDASH"])
	and ((p.cmd.buttons & BT_SPIN) and not (p.lastbuttons & BT_SPIN))
	and not (mo.bomb_fuseshot and mo.state == S_BOMB_FUSESHOT)
		dash_activate = true
	end
	
	if (shot_activate or dash_activate)
		bomb.VolatileLaunch(p, bomb.VolatileManualHeight)
		return
	end
	
	-- ground bounce activation
	if (mo.eflags & MFE_JUSTHITFLOOR)
	and not (volatile == bomb.VolatileFlags["MELTDOWN"])
		local momz = abs(mo.momz)
		if (mo.bomb_volatile_oldinfo["momz"]) then momz = abs(mo.bomb_volatile_oldinfo["momz"]) end
		momz = max(bomb.VolatileBounceMinHeight, FixedMul(bomb.VolatileBounceHeightMultiplier, $))
		
		local old_thokked = mo.bomb_volatile_oldinfo["thokked"]	-- save this
		bomb.VolatileLaunch(p, momz)
		bomb.AddThokked(p, old_thokked)	-- add back old thokked stuff
		
		-- break floors! (we can return here since it's the last thing in the block!)
		local rover = mo.floorrover
		
		-- rover checks
		if not (rover and rover.valid) then return end
		if not (rover.flags & FF_EXISTS) then return end
		if not ((rover.flags & FF_STRONGBUST) or (rover.flags & FF_BUSTUP)) then return end
		
		-- height checks
		/*local topheight = rover.topheight - momz
		local bottomheight = rover.bottomheight - momz
			
		if (mo.z >= topheight) then return end
		if (mo.z + mo.height < bottomheight) then return end*/
		
		-- break it if we're here!
		EV_CrumbleChain(rover)
		if (rover.master.flags & ML_EFFECT5)
			P_LinedefExecute(P_AproxDistance(rover.master.dx, rover.master.dy)>>FRACBITS, mo, mo.subsector.sector)
		end
		
		return
	end
	
	-- get how volatile we are to set some visuals
	local colorized_func = bomb.VolatileEffectTable[1].colorized_func
	local ghost_func = bomb.VolatileEffectTable[1].ghost_func
	
	local count = (P_PlayerFullbright(p)) and 1 or 0	-- start out more volatile if we're super!
	for _, flag in pairs(bomb.VolatileFlags)
		if not (volatile & flag) then continue end
		count = $+1
		
		-- get new info?
		if not (count > 1) then continue end
		
		local effect_table = bomb.VolatileEffectTable[count]
		if not (effect_table) then continue end	-- paranoia
		
		colorized_func = bomb.VolatileEffectTable[count].colorized_func or $
		ghost_func = bomb.VolatileEffectTable[count].ghost_func or $
	end
	
	mo.colorized = colorized_func(p)
	local ghost_tics = ghost_func(p)
	if (ghost_tics)
		local ghost = P_SpawnGhostMobj(mo)
		ghost.colorized = true
		ghost.tics = ghost_tics
	end
	mo.bomb_volatile_colorized = mo.colorized
	if (leveltime%2) then S_StartSoundAtVolume(mo, sfx_s3k48, 128) end	-- sound effect to show ignition
	
	-- store old values
	mo.bomb_volatile_oldinfo = {	
		-- old momentum values
		momx = mo.momx,
		momy = mo.momy,
		momz = mo.momz,
		
		-- angle for angled rebounds
		angle = R_PointToAngle2(0, 0, mo.momx, mo.momy),
		
		-- bomb thokked state
		thokked = mo.bomb_thokked,
	}
end

-- explode off of walls!
local function bomb_checkLine(p, line)
	local mo = p.mo
	
	-- horizon line checks
	if (line.special == 41) then return end
	
	-- sector checks
	if not (line.backsector) then return true end	-- apparently in these situations source always returns true?
	local sec = (line.backsector and not P_PointOnLineSide(mo.x, mo.y, line)) and line.backsector or line.frontsector
	
	-- shouldnt be too hard to check this
	local checkz = mo.z
	if (mo.eflags & MFE_VERTICALFLIP) then checkz = mo.z + mo.height end
	
	-- make sure we cant climb thokbarriers
	local thokbarrier = (sec.ceilingpic:upper():find("F_SKY"))
	if (thokbarrier and (checkz >= sec.ceilingheight))
		return false
	end
	
	-- should be good?
	return true
end

bomb.VolatileMoveBlocked = function(mo, thing, line)
	if not (line and line.valid) then return end
	
	-- are we volatile?
	if not (mo.bomb_volatile)
	or (mo.bomb_volatile == bomb.VolatileFlags["MELTDOWN"])
		return
	end
	
	-- check the line
	local p = mo.player
	if not (bomb_checkLine(p, line)) then return end
	
	-- oldinfo (laziness)
	local oldinfo = mo.bomb_volatile_oldinfo
	
	-- rebound the player
	local wallangle = R_PointToAngle2(line.v1.x, line.v1.y, line.v2.x, line.v2.y) - ANGLE_90
	local hitangle = wallangle - oldinfo["angle"]
	local reflectangle = wallangle + hitangle
		
	local anglefactor = abs(cos(hitangle))
	local range = FRACUNIT - bomb.VolatileWallFacingMultiplier	-- dont get launched as far away from walls we're facing!
	local modifier = FRACUNIT - FixedMul(anglefactor, range)
	
	local speed = FixedHypot(oldinfo["momx"], oldinfo["momy"])
	P_InstaThrust(mo, reflectangle, FixedMul(modifier, -(speed)))
	bomb.VolatileLaunch(p, max(bomb.VolatileWallMinHeight, FixedMul(bomb.VolatileWallHeightMultiplier, speed)))
	return true
end

-- additional hook for busting walls (MobjMoveBlocked gets the wrong line 99% of the time...)
chaotix.AddHook("MobjLineCollide", function(mo, line)
	if not ((mo and mo.valid) and mo.health) then return end
	if not (line and line.valid) then return end
	
	-- are we volatile?
	if not (mo.bomb_volatile)
	or (mo.bomb_volatile == bomb.VolatileFlags["MELTDOWN"])
		return
	end
	
	-- check the line
	local p = mo.player
	if not (bomb_checkLine(p, line)) then return end
	
	-- break bustable walls!
	if (line.frontsector == nil or line.backsector == nil) then return end	-- one-sided, dont bother
	local sector = (P_PointOnLineSide(mo.x, mo.y, line) == 1) and line.frontsector or line.backsector
	
	for rover in sector.ffloors()
		if not (rover.flags & FF_EXISTS) then continue end
		if not ((rover.flags & FF_STRONGBUST) or (rover.flags & FF_BUSTUP)) then continue end
			
		-- make sure we're volatile enough for strong walls!
		local count = (P_PlayerFullbright(p)) and 1 or 0
		for _, flag in pairs(bomb.VolatileFlags)
			if not (mo.bomb_volatile & flag) then continue end
			count = $+1
		end
		
		if (rover.flags & FF_STRONGBUST) 
		and not (count >= 2)
			continue
		end

		-- height checks
		local topheight = rover.topheight
		local bottomheight = rover.bottomheight
			
		if (mo.z > topheight)
		or ((mo.z + mo.height) < bottomheight)
			continue
		end
			
		EV_CrumbleChain(rover)
		if (rover.master.flags & ML_EFFECT5)
			P_LinedefExecute(P_AproxDistance(rover.master.dx, rover.master.dy)>>FRACBITS, mo, mo.subsector.sector)
		end
		
		-- oldinfo (laziness)
		local oldinfo = mo.bomb_volatile_oldinfo
		
		-- rebound the player
		local wallangle = R_PointToAngle2(line.v1.x, line.v1.y, line.v2.x, line.v2.y) - ANGLE_90
		local hitangle = wallangle - oldinfo["angle"]
		local reflectangle = wallangle + hitangle
			
		local anglefactor = abs(cos(hitangle))
		local range = FRACUNIT - bomb.VolatileWallFacingMultiplier	-- dont get launched as far away from walls we're facing!
		local modifier = FRACUNIT - FixedMul(anglefactor, range)
		
		local speed = FixedHypot(oldinfo["momx"], oldinfo["momy"])
		P_InstaThrust(mo, reflectangle, FixedMul(modifier, -(speed)))
		bomb.VolatileLaunch(p, max(bomb.VolatileWallMinHeight, FixedMul(bomb.VolatileWallHeightMultiplier, speed)))
	end
end, MT_PLAYER)

-- blow up on things we damage while volatile!
bomb.VolatileDamage = function(mobj, mo)
	if not ((mo and mo.valid) and mo.health) then return end
	
	-- are we volatile?
	if not (mo.bomb_volatile) then return end
	local p = mo.player
	
	-- don't do anything in battle mod! (handled in collision function)
	if (mobj.player)
	and (CBW_Battle)
		return
	end
	
	-- deal MORE damage!
	local count = (P_PlayerFullbright(p)) and 1 or 0
	for _, flag in pairs(bomb.VolatileFlags)
		if not (mo.bomb_volatile & flag) then continue end
		count = $+1
	end
	mobj.health = max(1, $ - count)	-- don't go below 1, let mobjdamage kill them!
	mobj.bomb_volatile_launcher = true	-- hack
	
	-- spinbuffer handling
	local spinbuffer = true
	if (mo.bomb_fuseshot_target and mo.bomb_fuseshot_target.valid)
	and (mo.bomb_fuseshot_target == mobj)
		spinbuffer = false
	end
	
	local launch = max(FixedHypot(FixedHypot(mo.momx, mo.momy), mo.momz) >> 1, abs(mo.momz))
	local height = max(bomb.VolatileDamageMinHeight, FixedMul(bomb.VolatileDamageHeightMultiplier, launch))
	bomb.VolatileLaunch(p, height, nil, spinbuffer)
	mobj.bomb_volatile_launcher = false	-- hack (2)
	bomb.ResetThokked(p)	-- sounds fun
	return true
end