-- chaotixchar's config
chaotix.ConfigPath = "client/chaotixchars/config.cfg"

-- function to write a config (should be called whenever anything is set)
chaotix.WriteConfig = function(p)
	-- consoleplayer check, limit netgame shenanigans
	if (p != consoleplayer) then return end
	
	-- it's espio, mighty, and vector!
	local espio = chaotix.espio
	local mighty = chaotix.mighty
	local vector = chaotix.vector
	
	-- load the config we'll be writing to!
	local config = io.openlocal(chaotix.ConfigPath, "w+")
	
	-- write espio's shift colours
	local shift_colours = p.espio_shiftcolours or espio.ColourShiftDefault
	
	local shift_string = espio.ColourShiftCommand.." "
	for index, colour in ipairs(shift_colours)
		shift_string = $ + R_GetNameByColor(colour)
		if (index != #shift_colours) then shift_string = $ + " " end	-- split each colour up with a space
	end
	config:write(shift_string)
	config:write("\n")	-- line break
	
	-- write espio's invis mode
	local invis_mode = espio.InvisModeDefault
	if (p.espio_invismode != nil) then invis_mode = p.espio_invismode end
	
	local invis_modes = {"Hold", "Toggle"}
	local invis_string = espio.InvisModeCommand.." \""..invis_modes[invis_mode].."\""
	config:write(invis_string)
	config:write("\n")	-- line break
	
	-- write vector's text mode
	local beatbash_modes = {"Off", "Local", "Global"}
	local beatbash_string = vector.BeatBashCommand.." \""..beatbash_modes[vector.BeatBashCommandVar.value + 1].."\""
	config:write(beatbash_string)
	
	-- close the config
	config:close()
end

-- load the config once we are valid
local loadedplayer
chaotix.LoadConfig = function()
	if (consoleplayer and consoleplayer.valid)
	and not (loadedplayer and loadedplayer.valid)	-- this'll let it reload whenever we join a server
		-- load the config we'll be reading from
		local config = io.openlocal(chaotix.ConfigPath, "r")
		if not (config)	-- no config? create one real quick
			chaotix.WriteConfig(consoleplayer)
			config = io.openlocal(chaotix.ConfigPath, "r")	-- now this shouldn't fail
		end
		
		-- load each line of the config (technically you could stuff other commands in here!)
		for line in config:lines()
			CommandQueue_Add(line)
		end
		
		-- close the config, we're done
		config:close()
		
		-- save the consoleplayer so this isnt run constantly
		loadedplayer = consoleplayer
	end
end

-- add hook
chaotix.AddHook("ThinkFrame", chaotix.LoadConfig)