/*
Command Queue
	by Lach

To use this, simply call CommandQueue_Add(myCommand), instead of COM_BufInsertText(consoleplayer, myCommand) or COM_BufAddText(consoleplayer, myCommand).

A big problem with IO config loading at the moment is that if a whole bunch of
mods decide to send commands when a player joins a server, the netxcmd buffer
can overflow. james' textcmd refactor will solve this issue, but this will
probably be merged by 2.2.11 at the earliest.

This script lets you put your commands in a queue. Instead of flooding the
console with every command at once, it will execute the first one in the queue,
wait for it to be received, and then execute the next command in the queue.
This means your command may not execute immediately when the client joins the
game if the queue is large, but that sounds much better to me than not executing
at all :p

Obviously, this won't work if other mods fill the buffer anyway. :(
*/

if CommandQueue_Add
	return
end

local queue = {} // table of commands to execute
local received = true // boolean; whether we have received the last command sent

local function ExecuteNext()
	if not received
	or #queue == 0
		return
	end
	COM_BufAddText(consoleplayer, table.remove(queue, 1))
	COM_BufAddText(consoleplayer, "cq_send")
	received = false
end

local function Add(command)
	table.insert(queue, command)
	ExecuteNext()
end

rawset(_G, "CommandQueue_Add", Add)

COM_AddCommand("cq_send", function(player)
	if player ~= consoleplayer
		return
	end
	received = true
	ExecuteNext()
end)

addHook("GameQuit", function(quit)
	if quit
		return
	end
	queue = {}
end)